Package: pizzarr
Type: Package
Title: Slice into 'Zarr' Arrays
Version: 0.2.0
Authors@R: c(
    person(given = "David",
           family = "Blodgett",
           role = c("cre", "aut"),
           email = "dblodgett@usgs.gov",
           comment = c(ORCID = "0000-0001-9489-1710")),
    person(
      given = "Mark",
      family = "Keller",
      email = "mark_keller@g.harvard.edu",
      role = c("aut"),
      comment = c(ORCID = "0000-0003-3003-874X")
    ),
    person(
      given = "Artür", 
      family = "Manukyan", 
      email = "artur-man@hotmail.com",
      role = c("aut"), 
      comment = c(ORCID = "0000-0002-0441-9517")),
    person(given = "zarr-developers", role = "cph")
  )
Description: An implementation of chunked, compressed,
  N-dimensional arrays for R. 'Zarr' spec V2 (2024)
  <doi:10.5281/zenodo.11320255>.
Language: en-US
License: MIT + file LICENSE
BugReports: https://github.com/zarr-developers/pizzarr/issues
URL: https://zarr.dev/pizzarr/,
        https://github.com/zarr-developers/pizzarr
Depends: R (>= 4.2)
Imports: jsonlite, stats, R6, qs2, stringr, memoise, utils
Encoding: UTF-8
RoxygenNote: 7.3.3
VignetteBuilder: knitr
Suggests: testthat, knitr, covr, bslib, pkgdown, rmarkdown, crul,
        blosc, vcr (>= 0.6.0), bench, bit64, withr
Config/testthat/parallel: false
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-04-20 12:57:51 UTC; dblodgett
Author: David Blodgett [cre, aut] (ORCID:
    <https://orcid.org/0000-0001-9489-1710>),
  Mark Keller [aut] (ORCID: <https://orcid.org/0000-0003-3003-874X>),
  Artür Manukyan [aut] (ORCID: <https://orcid.org/0000-0002-0441-9517>),
  zarr-developers [cph]
Maintainer: David Blodgett <dblodgett@usgs.gov>
Repository: CRAN
Date/Publication: 2026-04-20 16:50:02 UTC
Built: R 4.6.0; ; 2026-04-28 08:27:57 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method("[",ZarrArray)
S3method("[<-",ZarrArray)
S3method(as.array,NestedArray)
S3method(as.array,ZarrArray)
export(Attributes)
export(BloscCodec)
export(Bz2Codec)
export(Codec)
export(DirectoryStore)
export(Dtype)
export(GcsStore)
export(GzipCodec)
export(HttpStore)
export(Lz4Codec)
export(LzmaCodec)
export(MemoryStore)
export(NestedArray)
export(S3Store)
export(Store)
export(VLenUtf8Codec)
export(ZarrArray)
export(ZarrGroup)
export(ZlibCodec)
export(ZstdCodec)
export(as_scalar)
export(int)
export(is_key_error)
export(is_scalar)
export(is_slice)
export(obj_list)
export(pizzarr_compiled_features)
export(pizzarr_config)
export(pizzarr_option_defaults)
export(pizzarr_sample)
export(pizzarr_upgrade)
export(slice)
export(zarr_create)
export(zarr_create_array)
export(zarr_create_empty)
export(zarr_create_group)
export(zarr_create_zeros)
export(zarr_open)
export(zarr_open_array)
export(zarr_open_group)
export(zarr_save_array)
export(zarr_volcano)
export(zarrs_compiled_features)
export(zb_int)
export(zb_slice)
importFrom(R6,R6Class)
importFrom(memoise,memoise)
importFrom(memoise,timeout)
importFrom(qs2,zstd_compress_raw)
importFrom(qs2,zstd_decompress_raw)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# pizzarr <a href="https://zarr.dev/pizzarr/"><img src="man/figures/logo.png" align="right" height="139" alt="pizzarr website" /></a>

[![codecov](https://codecov.io/gh/zarr-developers/pizzarr/graph/badge.svg?token=vhidertN9l)](https://app.codecov.io/gh/zarr-developers/pizzarr)
[![R-CMD-check](https://github.com/zarr-developers/pizzarr/actions/workflows/R-CMD-Check.yml/badge.svg)](https://github.com/zarr-developers/pizzarr/actions/workflows/R-CMD-Check.yml)
[![logs](https://cranlogs.r-pkg.org/badges/pizzarr)](https://cran.r-project.org/package=pizzarr)
[![CRAN
status](https://www.r-pkg.org/badges/version/pizzarr)](https://CRAN.R-project.org/package=pizzarr)
[![r-universe](https://zarr-developers.r-universe.dev/badges/pizzarr)](https://zarr-developers.r-universe.dev/pizzarr)

A Zarr implementation for R.

## Installation

pizzarr ships in two tiers — same version, different builds. The CRAN
build is pure R and needs nothing beyond base R dependencies. The
[r-universe](https://zarr-developers.r-universe.dev/pizzarr) build links
in the [zarrs](https://github.com/zarrs/zarrs) Rust crate via
[extendr](https://cran.r-project.org/package=rextendr), adding parallel
decompression and additional codec support.

### CRAN (pure R)

``` r
install.packages("pizzarr")
```

### r-universe (zarrs backend)

``` r
install.packages("pizzarr",
                 repos = c("https://zarr-developers.r-universe.dev",
                            "https://cloud.r-project.org"))
```

Both sources serve the same release version. The r-universe build
distributes pre-compiled binaries for Windows and macOS — no Rust
toolchain needed.

The r-universe binaries compile the zarrs backend with local
filesystem I/O, synchronous HTTP reads, gzip/zstd/blosc codecs,
sharding, and S3/GCS cloud store support via `object_store`. The
build system enables these features automatically when `NOT_CRAN` is
set (which r-universe does). To compile with a different feature set,
set `PIZZARR_FEATURES` explicitly (see below).

`pizzarr_compiled_features()` lists what the zarrs backend provides,
and `pizzarr_upgrade()` prints the install command when zarrs is not
compiled in. `pizzarr_config()` controls concurrency settings (thread
pool size, codec concurrency, HTTP range request batching). See
`vignette("zarrs-backend")` for details.

### Building from source (zarrs backend)

If you need to compile the zarrs backend yourself — either because no
binary is available for your platform, or because you are developing
the Rust side — you need rustc \>= 1.91. On Windows, add the GNU
target:

``` bash
rustup target add x86_64-pc-windows-gnu
```

The build system uses two environment variables to control compilation:

- **`DEBUG`** — set to any non-empty value for a debug build. This
  enables Cargo's incremental compilation, so subsequent rebuilds are
  fast (seconds instead of minutes). The resulting library is slower at
  runtime because it skips link-time optimization and does not strip
  symbols. Use this during active development on the Rust code.
- **`PIZZARR_FEATURES`** — a comma-separated list of extra Cargo
  features. When `NOT_CRAN` is set and `PIZZARR_FEATURES` is empty,
  the build defaults to `s3,gcs` (S3 and GCS cloud store support via
  `object_store` and `tokio`). Set this explicitly to override — e.g.,
  `PIZZARR_FEATURES=none` for default features only, or
  `PIZZARR_FEATURES=s3` for S3 without GCS.

A release build (no `DEBUG`) uses LTO, single codegen unit, and symbol
stripping — fast at runtime but takes several minutes to compile from
scratch.

``` bash
# Development: fast rebuilds, slower runtime, S3/GCS included
DEBUG=1 R CMD INSTALL .

# Production: slow first build, fast runtime, S3/GCS included
NOT_CRAN=true R CMD INSTALL .

# Production with default features only (no cloud stores)
R CMD INSTALL .
```

`NOT_CRAN` is set automatically when `DEBUG` is present. For builds
without vendored crates (the normal case for local development),
`NOT_CRAN` prevents the build system from attempting offline
compilation.

Development happens on the `develop` branch. See
[CONTRIBUTING.md](https://github.com/zarr-developers/pizzarr/blob/main/CONTRIBUTING.md)
for the branching and release model.

## Usage

``` r
library(pizzarr)

# Open a sample BCSD climate dataset (Zarr V3)
v3_root <- pizzarr_sample("bcsd_v3")
v3 <- zarr_open(v3_root)

# Print the group summary
v3
#> <ZarrGroup> /
#>   Store type  : DirectoryStore
#>   Zarr format : 3
#>   Read-only   : FALSE
#>   No. members : 5

# View the hierarchy
v3$tree()
#> / 
#> ├── latitude (33) <f4
#> ├── longitude (81) <f4
#> ├── pr (12, 33, 81) <f4
#> ├── tas (12, 33, 81) <f4
#> └── time (12) <f8

# Inspect an array
v3$get_item("pr")
#> <ZarrArray> /pr
#>   Shape       : (12, 33, 81)
#>   Chunks      : (12, 33, 81)
#>   Data type   : <f4
#>   Fill value  : 1.00000002004088e+20
#>   Order       : C
#>   Read-only   : FALSE
#>   Compressor  : ZstdCodec
#>   Store type  : DirectoryStore
#>   Zarr format : 3

# Read a slice: first 3 time steps, first 3 latitudes, first longitude
v3$get_item("pr")$get_item(list(slice(1, 3), slice(1, 3), 1))$data
#> , , 1
#> 
#>        [,1]   [,2]   [,3]
#> [1,] 133.97 144.51 149.92
#> [2,]  75.40  72.38  68.62
#> [3,]  93.14  91.24  89.37
```

Create an array from scratch:

``` r
a <- array(data = 1:20, dim = c(2, 10))
z <- zarr_create(shape = dim(a), dtype = "<f4", fill_value = NA)
z$set_item("...", a)
```

``` r
z
#> <ZarrArray> /
#>   Shape       : (2, 10)
#>   Chunks      : (2, 10)
#>   Data type   : <f4
#>   Fill value  : 0
#>   Order       : F
#>   Read-only   : FALSE
#>   Compressor  : ZstdCodec
#>   Store type  : MemoryStore
#>   Zarr format : 2
z$get_item(list(slice(1, 2), slice(1, 5)))$data
#>      [,1] [,2] [,3] [,4] [,5]
#> [1,]    1    3    5    7    9
#> [2,]    2    4    6    8   10
```

## Features

- **Zarr V2 and V3** read and write (format auto-detected on open)
- **Stores:** MemoryStore, DirectoryStore (read/write); HttpStore
  (read-only)
- **Data types:** boolean, int8–int64, uint8–uint64, float16/32/64,
  string, Unicode, VLenUTF8
- **Compression:** zlib/gzip, bzip2, blosc, LZMA, LZ4, Zstd
- **Blosc** requires the optional
  [`blosc`](https://cran.r-project.org/package=blosc) package
  (`install.packages("blosc")`)

## How It Works

pizzarr uses [R6](https://r6.r-lib.org/) classes mirroring the
[zarr-python](https://github.com/zarr-developers/zarr-python) object
model:

- **Store** — backend storage (`DirectoryStore` for local files,
  `MemoryStore` for in-memory, `HttpStore` for remote read-only)
- **ZarrGroup** — hierarchical container holding arrays and sub-groups
  (like a directory)
- **ZarrArray** — chunked, compressed N-dimensional array (like a file)
- **Codec** — compression/decompression (zlib, zstd, blosc, lz4, etc.)
- **Dtype** — data-type mapping between R and Zarr

Data flows through the stack: a **Store** holds raw chunk bytes, a
**Codec** pipeline compresses and decompresses them, and **ZarrArray**
presents typed N-dimensional data to R. Groups and arrays are addressed
by path within a store, just like files in a directory tree.

See `vignette("pizzarr")` for a full walkthrough.

## Ecosystem

pizzarr implements the [Zarr
specification](https://zarr-specs.readthedocs.io/) (V2 and V3) for R.
Related projects:

- [zarr-python](https://github.com/zarr-developers/zarr-python) — the
  reference Python implementation
- [zarr.js](https://github.com/gzuidhof/zarr.js) — JavaScript
  implementation
- [zarr](https://cran.r-project.org/package=zarr) — native R V3
  implementation (CRAN)
- [Rarr](https://bioconductor.org/packages/Rarr/) — Bioconductor package
  for reading and writing individual Zarr arrays (V2, limited write
  support)
- [zarr-conformance-tests](https://github.com/Bisaloo/zarr-conformance-tests)
  — cross-implementation validation

## Validation with zarr-python

A standalone integration test cross-validates that pizzarr and
zarr-python produce equivalent Zarr stores. Both implementations write
the same arrays (V2 and V3 formats, multiple dtypes, codecs, chunk
layouts, and groups with attributes), then each reads the other’s output
and verifies the data matches.

**Prerequisites:** Python 3.10+ with `zarr>=3` and `numpy` installed.

``` bash
Rscript inst/extdata/cross-validate.R
```

The script skips gracefully (exit 0) if Python is not available. On
success all checks pass and exit code is 0; any mismatch is reported and
exits 1.

## Zarr Conformance Tests

pizzarr participates in the
[zarr-conformance-tests](https://github.com/Bisaloo/zarr-conformance-tests)
framework, which validates that Zarr implementations can correctly read
standard test arrays (V2 and V3 formats, multiple dtypes). These tests
run automatically in CI on every push and pull request to `main`.

## Performance

The zarrs backend (r-universe install only) handles chunk I/O and codec execution in Rust, bypassing
R's single-threaded chunk loop entirely. For compressed data — the common
case with real Zarr stores — this translates to 6–8x throughput over pure-R.

The table below compares pizzarr's two tiers against zarr-python/xarray on
a 500×500×100 float64 array (~200 MB uncompressed) with 100×100×50 chunks.
All numbers are best-of-three throughput in MB/s, measured on Windows
(pizzarr 0.2.0-dev, zarr-python 3.1.5, April 2026).

| Scenario | pizzarr zarrs | pizzarr R-native | xarray |
|---|---|---|---|
| read_all (gzip) | 253 MB/s | 39 MB/s | 545 MB/s |
| read_all (none) | 247 MB/s | 39 MB/s | 1811 MB/s |
| read_subset (gzip) | **200 MB/s** | 38 MB/s | 138 MB/s |
| read_subset (none) | 229 MB/s | 38 MB/s | 704 MB/s |
| write_all (gzip) | 97 MB/s | 12 MB/s | 143 MB/s |
| write_all (none) | 286 MB/s | 144 MB/s | 477 MB/s |

xarray leads on bulk operations because numpy can transpose array memory
layout via stride views (zero copy) — R's `array()` always copies. The
gap narrows with compression since decompression dominates, and zarrs wins
on cross-chunk subset reads (200 vs 138 MB/s gzip) because it decodes
only the chunks that overlap the selection.

Reproduce with `Rscript inst/extdata/benchmark-zarrs-vs-xarray.R` (requires
Python with `zarr>=3` and `xarray` for the xarray column).

## Contributing

See
[CONTRIBUTING.md](https://github.com/zarr-developers/pizzarr/blob/main/CONTRIBUTING.md)
for development setup, testing, and documentation build instructions.

Note: these files have been copied from https://github.com/manzt/zarrita.js/tree/4acc746/fixtures


```sh
cd inst/extdata/fixtures
conda env create -f environment.yml
conda activate pizzarr-fixtures

cd v2
python ./generate-v2.py
```

#!/usr/bin/env Rscript

# Minimal CLI for zarr-conformance-tests (Bisaloo/zarr-conformance-tests).
# Reads a Zarr array at the given path and exits 0 on success, 1 on failure.
#
# Usage:
#   Rscript pizzarr-cli.R --array_path=/path/to/array.zarr

args <- commandArgs(trailingOnly = TRUE)

# Simple --key=value parser
opts <- grep("^--", args, value = TRUE)
parsed <- strsplit(opts, "=")
opt_list <- setNames(
  vapply(parsed, function(x) paste(x[-1], collapse = "="), ""),
  vapply(parsed, function(x) sub("^--", "", x[1]), "")
)

if (!"array_path" %in% names(opt_list)) {
  cat("Error: --array_path is required\n", file = stderr())
  quit(status = 1)
}

array_path <- opt_list[["array_path"]]

library(pizzarr)

arr <- zarr_open(array_path, mode = "r")
data <- arr$get_item("...")$data
cat(sprintf("OK: read array at %s (%d elements)\n", array_path, length(data)))

# Convert bcsd_v2.zarr.zip (V2) to bcsd_v3.zarr.zip (V3) using pizzarr.
#
# R equivalent of convert-bcsd-v3.py.
# Unzips the V2 store, reads each array and group,
# and writes an equivalent Zarr V3 store, then zips it.
#
# Usage:
#   Rscript convert-bcsd-v3.R

library(pizzarr)

get_script_dir <- function() {
  # Works with Rscript --file, source(), and commandArgs
  args <- commandArgs(trailingOnly = FALSE)
  file_arg <- grep("^--file=", args, value = TRUE)
  if (length(file_arg) > 0) {
    return(dirname(normalizePath(sub("^--file=", "", file_arg))))
  }
  # Fallback: assume working directory contains inst/extdata
  return(normalizePath("inst/extdata"))
}
script_dir <- get_script_dir()

v2_zip <- file.path(script_dir, "bcsd_v2.zarr.zip")
v3_zip <- file.path(script_dir, "bcsd_v3.zarr.zip")

stopifnot(file.exists(v2_zip))

# Clean up previous output
if (file.exists(v3_zip)) file.remove(v3_zip)

# Work in temp directories
tmpdir <- tempfile("bcsd_convert")
dir.create(tmpdir)
on.exit(unlink(tmpdir, recursive = TRUE), add = TRUE)

v2_extract <- file.path(tmpdir, "v2")
v3_dir <- file.path(tmpdir, "bcsd_v3.zarr")

# Unzip V2 store
utils::unzip(v2_zip, exdir = v2_extract)
v2_path <- file.path(v2_extract, "bcsd_v2.zarr")
cat("V2 store extracted to:", v2_path, "\n")

# Open V2 root group
v2_root <- zarr_open_group(v2_path, mode = "r")

# Create V3 output store
dir.create(v3_dir)
v3_store <- DirectoryStore$new(v3_dir)
v3_root <- zarr_create_group(store = v3_store, zarr_format = 3L)

# Copy root-level attributes
root_attrs <- v2_root$get_attrs()$to_list()
if (length(root_attrs) > 0) {
  v3_attrs <- v3_root$get_attrs()
  for (nm in names(root_attrs)) {
    v3_attrs$set_item(nm, root_attrs[[nm]])
  }
  cat("Root attributes:", length(root_attrs), "keys copied\n")
}

# Known arrays in bcsd store
array_names <- c("latitude", "longitude", "time", "pr", "tas")

for (name in array_names) {
  v2_arr <- v2_root$get_item(name)
  data <- v2_arr$as.array()
  shape <- v2_arr$get_shape()
  chunks <- v2_arr$get_chunks()
  dtype <- v2_arr$get_dtype()
  fill_value <- v2_arr$get_fill_value()
  compressor <- v2_arr$get_compressor()

  cat(sprintf("  Array '%s': shape=(%s), dtype=%s, chunks=(%s)\n",
              name,
              paste(shape, collapse = ", "),
              dtype$dtype,
              paste(chunks, collapse = ", ")))

  # Create V3 array (zarr_format inherited from parent group)
  v3_arr <- v3_root$create_dataset(
    name = name,
    shape = as.integer(shape),
    chunks = as.integer(chunks),
    dtype = dtype$dtype,
    fill_value = fill_value,
    compressor = compressor
  )

  # Write data
  v3_arr$set_item("...", data)

  # Copy array attributes
  arr_attrs <- v2_arr$get_attrs()$to_list()
  if (length(arr_attrs) > 0) {
    v3_a <- v3_arr$get_attrs()
    for (nm in names(arr_attrs)) {
      v3_a$set_item(nm, arr_attrs[[nm]])
    }
    cat("    Attributes:", paste(names(arr_attrs), collapse = ", "), "\n")
  }
}

cat("\nV3 store written to:", v3_dir, "\n")

# Verify by reading back
cat("\nVerifying V3 store...\n")
v3_check_store <- DirectoryStore$new(v3_dir)
v3_check <- zarr_open_group(v3_check_store, mode = "r")

for (name in array_names) {
  arr <- v3_check$get_item(name)
  v2_arr <- v2_root$get_item(name)

  v3_data <- arr$as.array()
  v2_data <- v2_arr$as.array()

  shape_str <- paste(arr$get_shape(), collapse = ", ")
  match <- identical(dim(v3_data), dim(v2_data)) && all(v3_data == v2_data, na.rm = TRUE)
  cat(sprintf("  Verify '%s': shape=(%s), data_match=%s\n", name, shape_str, match))
}

# Zip it up
cat(sprintf("\nCreating %s...\n", v3_zip))
old_wd <- setwd(dirname(v3_dir))
zip(v3_zip, files = basename(v3_dir))
setwd(old_wd)

cat("Done!\n")

## ----include = FALSE----------------------------------------------------------
  knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  out.width = "100%"
)

## -----------------------------------------------------------------------------
library(pizzarr)

# get an empty store to add to later
mem_store <- MemoryStore$new()

class(mem_store)

# create a store in line when creating an empty array
demo_array <- zarr_create(c(1,2,3), store = NA)

demo_array_store <- demo_array$get_store()

class(demo_array_store)

demo_array_store$listdir()

# or create a store when creating a group to contain arrays
# notice that passing a path creates a directory store
store_path <- file.path(tempdir(), "demo.zarr")

demo_group <- zarr_create_group(store = store_path)

demo_group_store <- demo_group$get_store()

basename(demo_group_store$root)

class(demo_group_store)

demo_group_store$listdir()


## -----------------------------------------------------------------------------

demo_mem_store_group <- zarr_create_group(mem_store)

demo_mem_store_group$get_attrs()$key

demo_mem_store_group$get_attrs()$set_item("this is", "an attribute")

demo_mem_store_group$get_attrs()$to_list()

demo_mem_store_group$get_attrs()$del_item("this is")

demo_mem_store_group$get_attrs()$to_list()


## -----------------------------------------------------------------------------

demo_array_store$listdir()

demo_array$get_attrs()$set_item("this is", "array metadata")

demo_array$get_attrs()$to_list()

# notice that when we added the item, our demo array store got a .zattrs
demo_array_store$listdir()


## -----------------------------------------------------------------------------

zarr_volcano <- zarr_create_array(volcano, # the R array classic
                                  shape = dim(volcano),
                                  store = demo_group_store, # the store we want the array in
                                  path = "volcano") # the path we want the array stored in

demo_group_store$listdir()

zarr_volcano$get_shape()

all.equal(zarr_volcano$as.array(), 
          volcano) 


## -----------------------------------------------------------------------------
sub_zarr_volcano <- zarr_volcano$get_item(list(slice(1, 10), slice(1, 20)))

all.equal(sub_zarr_volcano$as.array(), 
          volcano[1:10, 1:20]) 

sub_zarr_volcano <- zarr_volcano[1:10, 1:20]

class(sub_zarr_volcano)

sub_zarr_volcano$shape

all.equal(sub_zarr_volcano$as.array(), 
          volcano[1:10, 1:20]) 


## -----------------------------------------------------------------------------

# this woll work once implemented?
# zarr_volcano[1:10, 1:20] <- zarr_volcano[1:10, 1:20] * 10

zarr_volcano$set_item(list(slice(1, 10), slice(1, 20)), 
                      zarr_volcano[1:10, 1:20]$as.array() * 10)


## -----------------------------------------------------------------------------

sub_zarr_volcano <- zarr_volcano$get_item(list(1, "..."))

sub_zarr_volcano$shape

sub_zarr_volcano <- zarr_volcano$get_item(list(":", 1))

sub_zarr_volcano$shape

sub_zarr_volcano <- zarr_volcano$get_item(list("..."))

sub_zarr_volcano$shape

sub_zarr_volcano <- zarr_volcano$get_item(list(slice(1, 20, 2), slice(1, 10, 1)))

sub_zarr_volcano$shape

# Benchmark pizzarr (R-native and zarrs Rust backend) vs xarray (Python).
#
# Modeled after zarrs/zarr_benchmarks: 3D arrays, multiple codecs,
# read-all / chunk / subset / write scenarios, throughput in MB/s.
#
# Prerequisites: Python 3.10+ with zarr>=3, xarray, numpy, numcodecs.
# Skips xarray benchmarks gracefully if Python is unavailable.
#
# Environment variables:
#   PIZZARR_BENCH_ITERS  — iterations per scenario (default 3)
#   PIZZARR_BENCH_CSV    — path to save full results CSV (optional)
#
# Usage:
#   Rscript inst/extdata/benchmark-zarrs-vs-xarray.R

# ---------------------------------------------------------------------------
# 0. Package loading
# ---------------------------------------------------------------------------
if (file.exists("DESCRIPTION") &&
    any(grepl("^Package: pizzarr", readLines("DESCRIPTION", n = 5)))) {
  devtools::load_all(".")
} else {
  library(pizzarr)
}

has_zarrs <- .pizzarr_env$zarrs_available

# ---------------------------------------------------------------------------
# 1. Python detection
# ---------------------------------------------------------------------------
python_cmd <- Sys.which("python")
if (nchar(python_cmd) == 0) python_cmd <- Sys.which("python3")
has_python <- nchar(python_cmd) > 0

if (has_python) {
  zarr_ver <- tryCatch(
    system2(python_cmd, c("-c", shQuote("import zarr; print(zarr.__version__)")),
            stdout = TRUE, stderr = TRUE),
    error = function(e) NULL
  )
  xr_ver <- tryCatch(
    system2(python_cmd, c("-c", shQuote("import xarray; print(xarray.__version__)")),
            stdout = TRUE, stderr = TRUE),
    error = function(e) NULL
  )
  if (is.null(zarr_ver) || !is.null(attr(zarr_ver, "status")) ||
      is.null(xr_ver) || !is.null(attr(xr_ver, "status"))) {
    message("Python found but zarr/xarray/numpy not installed. ",
            "Skipping xarray benchmarks.")
    message("Install via: pip install 'zarr>=3' xarray numpy numcodecs")
    has_python <- FALSE
  } else {
    cat(sprintf("Python: zarr %s, xarray %s (%s)\n",
                zarr_ver, xr_ver, python_cmd))
  }
} else {
  message("Python not found on PATH. Skipping xarray benchmarks.")
}

# Locate companion Python script
script_dir <- if (file.exists("inst/extdata/benchmark-xarray.py")) {
  normalizePath("inst/extdata", winslash = "/")
} else {
  normalizePath(system.file("extdata", package = "pizzarr"), winslash = "/")
}
py_bench_script <- file.path(script_dir, "benchmark-xarray.py")

if (has_python && !file.exists(py_bench_script)) {
  message("Cannot find benchmark-xarray.py in ", script_dir)
  has_python <- FALSE
}

cat(sprintf("R: pizzarr %s | zarrs backend: %s\n",
            packageVersion("pizzarr"),
            if (has_zarrs) "available" else "not available"))

# ---------------------------------------------------------------------------
# 2. Configuration
# ---------------------------------------------------------------------------
n_iter <- as.integer(Sys.getenv("PIZZARR_BENCH_ITERS", "3"))
cat(sprintf("Iterations per scenario: %d\n", n_iter))

configs <- list(
  list(
    name    = "small_none",
    shape   = c(100L, 100L, 100L),
    chunks  = c(100L, 100L, 100L),
    dtype   = "float64",
    dtype_r = "<f8",
    codec   = "none",
    size_mb = 100 * 100 * 100 * 8 / 1e6
  ),
  list(
    name    = "small_gzip",
    shape   = c(100L, 100L, 100L),
    chunks  = c(100L, 100L, 100L),
    dtype   = "float64",
    dtype_r = "<f8",
    codec   = "gzip",
    size_mb = 100 * 100 * 100 * 8 / 1e6
  ),
  list(
    name    = "medium_none",
    shape   = c(500L, 500L, 100L),
    chunks  = c(100L, 100L, 50L),
    dtype   = "float64",
    dtype_r = "<f8",
    codec   = "none",
    size_mb = 500 * 500 * 100 * 8 / 1e6
  ),
  list(
    name    = "medium_gzip",
    shape   = c(500L, 500L, 100L),
    chunks  = c(100L, 100L, 50L),
    dtype   = "float64",
    dtype_r = "<f8",
    codec   = "gzip",
    size_mb = 500 * 500 * 100 * 8 / 1e6
  )
)

# Per-config selection coordinates (1-based inclusive for R's slice())
chunk_sels <- list(
  small_none   = list(slice(1L, 100L), slice(1L, 100L), slice(1L, 100L)),
  small_gzip   = list(slice(1L, 100L), slice(1L, 100L), slice(1L, 100L)),
  medium_none  = list(slice(1L, 100L), slice(1L, 100L), slice(1L, 50L)),
  medium_gzip  = list(slice(1L, 100L), slice(1L, 100L), slice(1L, 50L))
)

subset_sels <- list(
  small_none   = list(slice(26L, 75L), slice(26L, 75L), slice(26L, 75L)),
  small_gzip   = list(slice(26L, 75L), slice(26L, 75L), slice(26L, 75L)),
  medium_none  = list(slice(51L, 250L), slice(51L, 250L), slice(11L, 60L)),
  medium_gzip  = list(slice(51L, 250L), slice(51L, 250L), slice(11L, 60L))
)

# Uncompressed sizes for subset/chunk throughput calculations
chunk_sizes_mb <- list(
  small_none  = 100 * 100 * 100 * 8 / 1e6,
  small_gzip  = 100 * 100 * 100 * 8 / 1e6,
  medium_none = 100 * 100 * 50 * 8 / 1e6,
  medium_gzip = 100 * 100 * 50 * 8 / 1e6
)

subset_sizes_mb <- list(
  small_none  = 50 * 50 * 50 * 8 / 1e6,
  small_gzip  = 50 * 50 * 50 * 8 / 1e6,
  medium_none = 200 * 200 * 50 * 8 / 1e6,
  medium_gzip = 200 * 200 * 50 * 8 / 1e6
)

make_compressor <- function(codec) {
  switch(codec,
    gzip = GzipCodec$new(),  # default level 6 (R-native only supports level 6)
    none = NA,
    stop("Unknown codec: ", codec)
  )
}

# ---------------------------------------------------------------------------
# 3. Working directory and data generation
# ---------------------------------------------------------------------------
work_dir <- tempfile("pizzarr_bench_")
dir.create(work_dir)
on.exit(unlink(work_dir, recursive = TRUE), add = TRUE)
cat("Working directory:", work_dir, "\n")

configs_json <- jsonlite::toJSON(configs, auto_unbox = TRUE)

if (has_python) {
  cat("\n--- Generating test stores via Python ---\n")
  py_work <- normalizePath(work_dir, winslash = "/")
  gen_out <- system2(python_cmd,
                     c(py_bench_script, "generate",
                       shQuote(py_work), shQuote(configs_json)),
                     stdout = TRUE, stderr = TRUE)
  cat(paste(gen_out, collapse = "\n"), "\n")
  if (!is.null(attr(gen_out, "status")) && attr(gen_out, "status") != 0) {
    stop("Python data generation failed:\n", paste(gen_out, collapse = "\n"))
  }
} else {
  # Generate stores with pizzarr so R benchmarks can still run
  cat("\n--- Generating test stores via pizzarr ---\n")
  for (cfg in configs) {
    store_path <- file.path(work_dir, paste0("data_", cfg$name, ".zarr"))
    dir.create(store_path)
    store <- DirectoryStore$new(store_path)
    root <- zarr_create_group(store = store, zarr_format = 2L)
    compressor <- make_compressor(cfg$codec)
    arr <- root$create_dataset(
      name       = "data",
      shape      = cfg$shape,
      chunks     = cfg$chunks,
      dtype      = cfg$dtype_r,
      fill_value = 0,
      compressor = compressor
    )
    set.seed(42)
    data_vec <- runif(prod(cfg$shape))
    arr$set_item("...", array(data_vec, dim = cfg$shape))
    cat(sprintf("  Generated %s: %s (%.1f MB)\n",
                cfg$name, store_path, cfg$size_mb))
  }
}

# ---------------------------------------------------------------------------
# 4. Results collector
# ---------------------------------------------------------------------------
results <- data.frame(
  scenario       = character(),
  array_config   = character(),
  codec          = character(),
  implementation = character(),
  iteration      = integer(),
  time_seconds   = double(),
  throughput_mb_s = double(),
  stringsAsFactors = FALSE
)

add_result <- function(scenario, config_name, codec, impl, iter, elapsed,
                       size_mb) {
  results <<- rbind(results, data.frame(
    scenario        = scenario,
    array_config    = config_name,
    codec           = codec,
    implementation  = impl,
    iteration       = iter,
    time_seconds    = round(elapsed, 6),
    throughput_mb_s = round(size_mb / elapsed, 2),
    stringsAsFactors = FALSE
  ))
}

# ---------------------------------------------------------------------------
# 5. R benchmark helpers
# ---------------------------------------------------------------------------

clear_store_cache <- function(store_path) {
  if (has_zarrs) {
    try(zarrs_close_store(normalizePath(store_path, winslash = "/")),
        silent = TRUE)
  }
}

bench_read_all <- function(store_path, cfg, impl_name) {
  cat(sprintf("    %s / read_all / %s\n", cfg$name, impl_name))
  for (i in seq_len(n_iter)) {
    clear_store_cache(store_path)
    t0 <- proc.time()
    z <- zarr_open(store_path, mode = "r")
    arr <- z$get_item("data")
    val <- arr$get_item("...")$data
    elapsed <- (proc.time() - t0)[["elapsed"]]
    add_result("read_all", cfg$name, cfg$codec, impl_name, i,
               elapsed, cfg$size_mb)
  }
}

bench_read_chunk <- function(store_path, cfg, impl_name) {
  cat(sprintf("    %s / read_chunk / %s\n", cfg$name, impl_name))
  sel <- chunk_sels[[cfg$name]]
  mb <- chunk_sizes_mb[[cfg$name]]
  for (i in seq_len(n_iter)) {
    clear_store_cache(store_path)
    t0 <- proc.time()
    z <- zarr_open(store_path, mode = "r")
    arr <- z$get_item("data")
    val <- arr$get_item(sel)$data
    elapsed <- (proc.time() - t0)[["elapsed"]]
    add_result("read_chunk", cfg$name, cfg$codec, impl_name, i, elapsed, mb)
  }
}

bench_read_subset <- function(store_path, cfg, impl_name) {
  cat(sprintf("    %s / read_subset / %s\n", cfg$name, impl_name))
  sel <- subset_sels[[cfg$name]]
  mb <- subset_sizes_mb[[cfg$name]]
  for (i in seq_len(n_iter)) {
    clear_store_cache(store_path)
    t0 <- proc.time()
    z <- zarr_open(store_path, mode = "r")
    arr <- z$get_item("data")
    val <- arr$get_item(sel)$data
    elapsed <- (proc.time() - t0)[["elapsed"]]
    add_result("read_subset", cfg$name, cfg$codec, impl_name, i, elapsed, mb)
  }
}

bench_write_all <- function(cfg, impl_name) {
  cat(sprintf("    %s / write_all / %s\n", cfg$name, impl_name))
  set.seed(42)
  data_array <- array(runif(prod(cfg$shape)), dim = cfg$shape)
  compressor <- make_compressor(cfg$codec)
  for (i in seq_len(n_iter)) {
    out_path <- file.path(work_dir,
                          paste0("write_r_", impl_name, "_",
                                 cfg$name, "_", i, ".zarr"))
    dir.create(out_path)
    store <- DirectoryStore$new(out_path)
    t0 <- proc.time()
    root <- zarr_create_group(store = store, zarr_format = 2L)
    arr <- root$create_dataset(
      name       = "data",
      shape      = cfg$shape,
      chunks     = cfg$chunks,
      dtype      = cfg$dtype_r,
      fill_value = 0,
      compressor = compressor
    )
    arr$set_item("...", data_array)
    elapsed <- (proc.time() - t0)[["elapsed"]]
    clear_store_cache(out_path)
    add_result("write_all", cfg$name, cfg$codec, impl_name, i,
               elapsed, cfg$size_mb)
  }
}

# ---------------------------------------------------------------------------
# 6. Run R benchmarks
# ---------------------------------------------------------------------------
cat("\n--- R benchmarks ---\n")

for (cfg in configs) {
  store_path <- file.path(work_dir, paste0("data_", cfg$name, ".zarr"))

  if (!dir.exists(store_path)) {
    cat(sprintf("  SKIP %s: store not found at %s\n", cfg$name, store_path))
    next
  }

  # -- pizzarr with zarrs backend --
  if (has_zarrs) {
    cat(sprintf("  pizzarr_zarrs: %s\n", cfg$name))
    bench_read_all(store_path, cfg, "pizzarr_zarrs")
    bench_read_chunk(store_path, cfg, "pizzarr_zarrs")
    bench_read_subset(store_path, cfg, "pizzarr_zarrs")
    bench_write_all(cfg, "pizzarr_zarrs")
    gc()
  }

  # -- pizzarr R-native (zarrs disabled) --
  cat(sprintf("  pizzarr_rnative: %s\n", cfg$name))
  orig_zarrs <- .pizzarr_env$zarrs_available
  .pizzarr_env$zarrs_available <- FALSE
  tryCatch({
    bench_read_all(store_path, cfg, "pizzarr_rnative")
    bench_read_chunk(store_path, cfg, "pizzarr_rnative")
    bench_read_subset(store_path, cfg, "pizzarr_rnative")
    bench_write_all(cfg, "pizzarr_rnative")
  }, finally = {
    .pizzarr_env$zarrs_available <- orig_zarrs
  })
  gc()
}

# ---------------------------------------------------------------------------
# 7. Run xarray benchmarks
# ---------------------------------------------------------------------------
if (has_python) {
  cat("\n--- xarray benchmarks ---\n")
  py_work <- normalizePath(work_dir, winslash = "/")
  py_out <- system2(python_cmd,
                    c(py_bench_script, "bench",
                      shQuote(py_work), shQuote(configs_json),
                      as.character(n_iter)),
                    stdout = TRUE, stderr = TRUE)

  # Last line of stdout is the JSON results; everything else is stderr progress
  # system2 with stdout=TRUE captures both (stderr merged when stderr=TRUE too)
  # Parse each line: JSON is the one that starts with "["
  json_line <- grep("^\\[", py_out, value = TRUE)
  stderr_lines <- grep("^\\[", py_out, value = TRUE, invert = TRUE)
  if (length(stderr_lines) > 0) cat(paste(stderr_lines, collapse = "\n"), "\n")

  if (length(json_line) == 1) {
    py_results <- jsonlite::fromJSON(json_line)
    results <- rbind(results, py_results)
    cat(sprintf("  Parsed %d xarray results\n", nrow(py_results)))
  } else {
    message("WARNING: Could not parse xarray benchmark results.")
    if (length(py_out) > 0) cat(paste(py_out, collapse = "\n"), "\n")
  }
}

# ---------------------------------------------------------------------------
# 8. Summary
# ---------------------------------------------------------------------------
cat("\n")
cat("=============================================================\n")
cat("  pizzarr vs xarray Benchmark Results\n")
cat("=============================================================\n")
cat(sprintf("R: pizzarr %s | zarrs: %s\n",
            packageVersion("pizzarr"),
            if (has_zarrs) "compiled" else "not available"))
if (has_python) {
  cat(sprintf("Python: zarr %s, xarray %s\n", zarr_ver, xr_ver))
}
cat(sprintf("Iterations: %d | Metric: min time, throughput at min time\n\n",
            n_iter))

if (nrow(results) == 0) {
  cat("No results collected.\n")
  quit(status = 1)
}

# Compute summary: min and median per group
agg <- do.call(rbind, lapply(
  split(results, list(results$scenario, results$array_config,
                      results$codec, results$implementation)),
  function(g) {
    if (nrow(g) == 0) return(NULL)
    data.frame(
      scenario       = g$scenario[1],
      array_config   = g$array_config[1],
      codec          = g$codec[1],
      implementation = g$implementation[1],
      min_s          = round(min(g$time_seconds), 4),
      median_s       = round(median(g$time_seconds), 4),
      max_throughput  = round(max(g$throughput_mb_s), 1),
      stringsAsFactors = FALSE
    )
  }
))
rownames(agg) <- NULL

# Sort for readable output
agg <- agg[order(agg$scenario, agg$array_config, agg$codec,
                 agg$implementation), ]

cat(sprintf("%-14s %-14s %-6s %-18s %8s %8s %12s\n",
            "Scenario", "Config", "Codec", "Implementation",
            "Min(s)", "Med(s)", "MB/s"))
cat(paste(rep("-", 86), collapse = ""), "\n")
for (i in seq_len(nrow(agg))) {
  r <- agg[i, ]
  cat(sprintf("%-14s %-14s %-6s %-18s %8.4f %8.4f %12.1f\n",
              r$scenario, r$array_config, r$codec, r$implementation,
              r$min_s, r$median_s, r$max_throughput))
}

# Optional CSV export
csv_path <- Sys.getenv("PIZZARR_BENCH_CSV", "")
if (nchar(csv_path) > 0) {
  write.csv(results, csv_path, row.names = FALSE)
  cat(sprintf("\nFull results saved to: %s\n", csv_path))
}

cat("\nDone.\n")

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  out.width = "100%"
)
sample_dir <- tools::R_user_dir("pizzarr")
clean <- !dir.exists(sample_dir)

## -----------------------------------------------------------------------------
library(pizzarr)

has_zarrs <- pizzarr:::.pizzarr_env$zarrs_available

## -----------------------------------------------------------------------------
pizzarr_compiled_features()

## -----------------------------------------------------------------------------
pizzarr:::.pizzarr_env$zarrs_available

## -----------------------------------------------------------------------------
pizzarr_upgrade()

## ----eval=has_zarrs-----------------------------------------------------------
# v2_root <- pizzarr_sample("fixtures/v2/data.zarr")
# 
# # Root group
# zarrs_node_exists(v2_root, "")

## ----eval=has_zarrs-----------------------------------------------------------
# # An array within the store
# zarrs_node_exists(v2_root, "1d.contiguous.lz4.i2")

## ----eval=has_zarrs-----------------------------------------------------------
# # A path that does not exist
# zarrs_node_exists(v2_root, "does_not_exist")

## ----eval=has_zarrs-----------------------------------------------------------
# v3_root <- pizzarr_sample("fixtures/v3/data.zarr")
# 
# zarrs_node_exists(v3_root, "")

## ----eval=has_zarrs-----------------------------------------------------------
# zarrs_close_store(v2_root)
# zarrs_close_store(v2_root)

## ----eval=has_zarrs-----------------------------------------------------------
# zarrs_close_store(v3_root)

## ----eval=has_zarrs-----------------------------------------------------------
# v2_root <- pizzarr_sample("fixtures/v2/data.zarr")
# zarrs_open_array_metadata(v2_root, "1d.contiguous.raw.i2")

## ----eval=has_zarrs-----------------------------------------------------------
# v3_root <- pizzarr_sample("fixtures/v3/data.zarr")
# zarrs_open_array_metadata(v3_root, "1d.contiguous.gzip.i2")

## ----eval=has_zarrs-----------------------------------------------------------
# zarrs_close_store(v2_root)
# zarrs_close_store(v3_root)

## ----eval=has_zarrs-----------------------------------------------------------
# zarrs_runtime_info()

## ----eval=has_zarrs-----------------------------------------------------------
# # View current settings
# pizzarr_config()
# 
# # Set codec concurrency to 2 parallel operations per read/write
# pizzarr_config(concurrent_target = 2L)
# zarrs_runtime_info()$codec_concurrent_target

## ----eval=has_zarrs-----------------------------------------------------------
# zarrs_set_codec_concurrent_target(2L)
# zarrs_runtime_info()$codec_concurrent_target

## ----eval=has_zarrs-----------------------------------------------------------
# d <- tempfile("zarrs_vignette_")
# z <- zarr_create(store = d, shape = c(100L, 50L), chunks = c(10L, 10L),
#                  dtype = "<f8")
# z$set_item("...", array(as.double(seq_len(5000)), dim = c(100, 50)))
# 
# # Re-open and read a subset --- zarrs handles the chunk I/O
# z2 <- zarr_open(store = d)
# result <- z2$get_item(list(slice(1L, 10L), slice(1L, 5L)))
# dim(result$data)

## ----eval=has_zarrs-----------------------------------------------------------
# result <- zarrs_get_subset(d, "", list(c(0L, 10L), c(0L, 5L)), NULL)
# str(result)

## ----eval=has_zarrs-----------------------------------------------------------
# result <- zarrs_get_subset(d, "", list(c(0L, 10L), c(0L, 5L)), 1L)
# length(result$data)

## ----eval=has_zarrs-----------------------------------------------------------
# zarrs_close_store(d)
# unlink(d, recursive = TRUE)

## ----eval=has_zarrs-----------------------------------------------------------
# # V3 array with gzip compression
# d <- tempfile("zarrs_create_vignette_")
# z <- zarr_create(store = d, shape = c(20L, 10L), chunks = c(10L, 10L),
#                  dtype = "<f8", zarr_format = 3L)
# z

## ----eval=has_zarrs-----------------------------------------------------------
# # Confirm V3 metadata was written
# file.exists(file.path(d, "zarr.json"))

## ----eval=has_zarrs-----------------------------------------------------------
# zarrs_close_store(d)
# unlink(d, recursive = TRUE)

## ----eval=has_zarrs-----------------------------------------------------------
# d <- tempfile("zarrs_create_direct_")
# dir.create(d)
# 
# meta <- zarrs_create_array(
#   store_url = d,
#   array_path = "",
#   shape = c(100L, 50L),
#   chunks = c(10L, 10L),
#   dtype = "float64",
#   codec_preset = "gzip",
#   fill_value = 0.0,
#   attributes_json = "{}",
#   zarr_format = 3L
# )
# str(meta)

## ----eval=has_zarrs-----------------------------------------------------------
# zarrs_set_subset(d, "", list(c(0L, 10L), c(0L, 5L)),
#                  as.double(1:50), NULL)
# result <- zarrs_get_subset(d, "", list(c(0L, 10L), c(0L, 5L)), NULL)
# head(result$data)

## ----eval=has_zarrs-----------------------------------------------------------
# zarrs_close_store(d)
# unlink(d, recursive = TRUE)

## ----eval=has_zarrs-----------------------------------------------------------
# d <- tempfile("zarrs_write_vignette_")
# z <- zarr_create(store = d, shape = c(20L, 10L), chunks = c(10L, 10L),
#                  dtype = "<f8")
# 
# # set_item dispatches to zarrs when eligible
# z$set_item("...", array(as.double(1:200), dim = c(20, 10)))
# 
# # Read back to confirm
# z2 <- zarr_open(store = d)
# result <- z2$get_item(list(slice(1L, 5L), slice(1L, 3L)))
# result$data

## ----eval=has_zarrs-----------------------------------------------------------
# # Overwrite rows 3-7, columns 1-2
# z$set_item(list(slice(3L, 7L), slice(1L, 2L)),
#            array(rep(-1.0, 10), dim = c(5, 2)))
# 
# result <- z2$get_item(list(slice(1L, 10L), slice(1L, 3L)))
# result$data

## ----eval=has_zarrs-----------------------------------------------------------
# # Write 10 values to the first row (0-based range [0, 1) x [0, 10))
# zarrs_set_subset(d, "", list(c(0L, 1L), c(0L, 10L)),
#                    as.double(101:110), NULL)
# 
# result <- zarrs_get_subset(d, "", list(c(0L, 1L), c(0L, 10L)), NULL)
# result$data

## ----eval=has_zarrs-----------------------------------------------------------
# zarrs_close_store(d)
# unlink(d, recursive = TRUE)

## ----eval=has_zarrs && ("http_sync" %in% pizzarr_compiled_features())---------
# url <- "https://raw.githubusercontent.com/DOI-USGS/rnz/main/inst/extdata/bcsd.zarr"
# 
# z <- zarr_open(store = HttpStore$new(url))
# 
# # zarrs handles the HTTP reads + parallel decompression
# pr <- z$get_item("pr")
# pr

## ----eval=has_zarrs && ("http_sync" %in% pizzarr_compiled_features())---------
# # Read a subset --- zarrs fetches only the chunks that overlap
# result <- pr$get_item(list(slice(1L, 3L), slice(1L, 5L), slice(1L, 5L)))
# dim(result$data)

## ----eval=has_zarrs && ("http_sync" %in% pizzarr_compiled_features())---------
# meta <- zarrs_open_array_metadata(url, "pr")
# str(meta[c("shape", "dtype", "zarr_format")])

## ----eval=has_zarrs && ("http_sync" %in% pizzarr_compiled_features())---------
# # Read a single element (first along each dimension)
# ranges <- lapply(seq_along(meta$shape), function(i) c(0L, 1L))
# result <- zarrs_get_subset(url, "pr", ranges, NULL)
# result$data

## ----eval=has_zarrs && ("http_sync" %in% pizzarr_compiled_features())---------
# zarrs_close_store(url)

## ----eval=has_zarrs-----------------------------------------------------------
# "http_sync" %in% pizzarr_compiled_features()

## ----eval=has_zarrs && ("s3" %in% pizzarr_compiled_features())----------------
# # OME-Zarr bonsai dataset on AWS Open Data (V2, zstd, uint8)
# s3_url <- "s3://ome-zarr-scivis/v0.4/64x0/bonsai.ome.zarr"
# 
# # Read array metadata
# meta <- zarrs_open_array_metadata(s3_url, "scale0/bonsai")
# str(meta[c("shape", "dtype", "zarr_format")])

## ----eval=has_zarrs && ("s3" %in% pizzarr_compiled_features())----------------
# # Read a small subset (first 4x4x4 corner)
# result <- zarrs_get_subset(s3_url, "scale0/bonsai",
#                            list(c(0L, 4L), c(0L, 4L), c(0L, 4L)), NULL)
# str(result)

## ----eval=has_zarrs && ("s3" %in% pizzarr_compiled_features())----------------
# zarrs_close_store(s3_url)

## ----eval=has_zarrs && ("blosc" %in% pizzarr_compiled_features())-------------
# # Pangeo ECCO ocean basins (V2, blosc/lz4, float32)
# gcs_url <- "https://storage.googleapis.com/pangeo-data/ECCO_basins.zarr"
# 
# meta <- zarrs_open_array_metadata(gcs_url, "basin_mask")
# cat("Shape:", paste(meta$shape, collapse = " x "), "\n")
# cat("Dtype:", meta$dtype, "\n")

## ----eval=has_zarrs && ("blosc" %in% pizzarr_compiled_features())-------------
# # Read a single basin mask slice
# result <- zarrs_get_subset(gcs_url, "basin_mask",
#                            list(c(0L, 1L), c(0L, 90L), c(0L, 90L)), NULL)
# cat("Slice dimensions:", paste(result$shape, collapse = " x "), "\n")

## ----eval=has_zarrs && ("blosc" %in% pizzarr_compiled_features())-------------
# zarrs_close_store(gcs_url)

## ----include=FALSE------------------------------------------------------------
if (clean) unlink(sample_dir, recursive = TRUE)

# Cross-validate pizzarr (R) and zarr-python produce equivalent Zarr stores.
#
# Writes identical arrays from both R and Python (V2 + V3), then each
# implementation reads the other's output.  Prints a pass/fail summary
# and exits with code 0 (all pass) or 1 (any fail).
#
# Prerequisites: Python 3.10+ with zarr>=3 and numpy installed.
# Skips gracefully if Python is not available.
#
# Usage:
#   Rscript inst/extdata/cross-validate.R

# Use devtools::load_all() when running from the source tree, otherwise
# fall back to the installed package.
if (file.exists("DESCRIPTION") &&
any(grepl("^Package: pizzarr", readLines("DESCRIPTION", n = 5)))) {
  devtools::load_all(".")
} else {
  library(pizzarr)
}

# ---------------------------------------------------------------------------
# 0. Python detection
# ---------------------------------------------------------------------------
python_cmd <- Sys.which("python")
if (nchar(python_cmd) == 0) python_cmd <- Sys.which("python3")
if (nchar(python_cmd) == 0) {
  message("SKIP: Python not found on PATH.")
  quit(status = 0)
}

zarr_ver <- tryCatch(
  system2(python_cmd, c("-c", shQuote("import zarr; print(zarr.__version__)")),
  stdout = TRUE, stderr = TRUE),
  error = function(e) NULL
)
if (is.null(zarr_ver) || !is.null(attr(zarr_ver, "status"))) {
  message("SKIP: zarr-python not installed. Install via: pip install 'zarr>=3' numpy")
  quit(status = 0)
}
cat(sprintf("Python zarr %s detected (%s)\n", zarr_ver, python_cmd))

# Locate companion Python scripts (next to this script in inst/extdata)
script_dir <- if (file.exists("inst/extdata/cross-validate-write.py")) {
  normalizePath("inst/extdata", winslash = "/")
} else {
  normalizePath(system.file("extdata", package = "pizzarr"), winslash = "/")
}
py_write_script <- file.path(script_dir, "cross-validate-write.py")
py_read_script  <- file.path(script_dir, "cross-validate-read.py")

if (!file.exists(py_write_script) || !file.exists(py_read_script)) {
  stop("Cannot find cross-validate-write.py and cross-validate-read.py in ", script_dir)
}

# ---------------------------------------------------------------------------
# 1. Temp working directory
# ---------------------------------------------------------------------------
work_dir <- tempfile("pizzarr_crossval")
dir.create(work_dir)
on.exit(unlink(work_dir, recursive = TRUE), add = TRUE)
cat("Working directory:", work_dir, "\n")

# ---------------------------------------------------------------------------
# 2. Test case definitions (shared contract between R and Python)
# ---------------------------------------------------------------------------
test_cases <- list(
  list(
    name       = "int32_1d_gzip",
    data       = c(10L, 20L, 30L, 40L, 50L, 60L),
    shape      = 6L,
    chunks     = 3L,
    dtype_r    = "<i4",
    codec      = "gzip",
    fill_value = 0L
  ),
  list(
    name       = "float64_1d_zstd",
    data       = c(1.1, 2.2, 3.3, 4.4, 5.5),
    shape      = 5L,
    chunks     = 5L,
    dtype_r    = "<f8",
    codec      = "zstd",
    fill_value = 0
  ),
  list(
    name       = "bool_1d_none",
    data       = c(TRUE, FALSE, TRUE, FALSE),
    shape      = 4L,
    chunks     = 4L,
    dtype_r    = "|b1",
    codec      = "none",
    fill_value = FALSE
  ),
  list(
    name       = "uint8_1d_none",
    data       = as.integer(c(0, 127, 128, 255)),
    shape      = 4L,
    chunks     = 2L,
    dtype_r    = "|u1",
    codec      = "none",
    fill_value = 0L
  ),
  list(
    name       = "float32_1d_gzip",
    data       = c(-1.5, 0.0, 1.5, 3.140000104904175),
    shape      = 4L,
    chunks     = 4L,
    dtype_r    = "<f4",
    codec      = "gzip",
    fill_value = 0
  ),
  list(
    name       = "int16_2d_gzip",
    # C-order: np.arange(12).reshape(4,3) = [[0,1,2],[3,4,5],[6,7,8],[9,10,11]]
    # In R (column-major), same logical array is stored transposed:
    data       = matrix(c(0L,3L,6L,9L, 1L,4L,7L,10L, 2L,5L,8L,11L), nrow = 4L, ncol = 3L),
    shape      = c(4L, 3L),
    chunks     = c(2L, 3L),
    dtype_r    = "<i2",
    codec      = "gzip",
    fill_value = 0L
  ),
  list(
    name       = "float64_3d_zstd",
    # C-order: np.arange(24).reshape(2,3,4)
    # Build the R equivalent via aperm of C-order fill
    data       = aperm(array(as.double(0:23), dim = c(4L, 3L, 2L)), c(3, 2, 1)),
    shape      = c(2L, 3L, 4L),
    chunks     = c(2L, 3L, 4L),
    dtype_r    = "<f8",
    codec      = "zstd",
    fill_value = 0
  ),
  list(
    name       = "fill_custom",
    data       = c(1.0, 2.0, 3.0, -9999.0, -9999.0, -9999.0),
    shape      = 6L,
    chunks     = 3L,
    dtype_r    = "<f8",
    codec      = "none",
    fill_value = -9999.0,
    # write all values (including fill region) so both sides are identical
    partial    = FALSE
  ),
  list(
    name       = "ragged_2d_gzip",
    # C-order: np.arange(35).reshape(5,7) — rows filled left-to-right
    data       = matrix(as.integer(0:34), nrow = 5L, ncol = 7L, byrow = TRUE),
    shape      = c(5L, 7L),
    chunks     = c(3L, 4L),
    dtype_r    = "<i4",
    codec      = "gzip",
    fill_value = 0L
  )
)

# ---------------------------------------------------------------------------
# 3. Helpers
# ---------------------------------------------------------------------------
make_compressor <- function(codec) {
  switch(codec,
    gzip = GzipCodec$new(),
    zstd = ZstdCodec$new(),
    none = NA,
    stop("Unknown codec: ", codec)
  )
}

compare_arrays <- function(actual, expected, dtype_r) {
  # Flatten to vectors for comparison
  a <- as.vector(actual)
  e <- as.vector(expected)
  if (length(a) != length(e)) return(FALSE)
  if (grepl("^[<>|]?f", dtype_r)) {
    isTRUE(all.equal(a, e, tolerance = 1e-5))
  } else if (grepl("b1", dtype_r)) {
    identical(as.logical(a), as.logical(e))
  } else {
    identical(as.integer(a), as.integer(e))
  }
}

# Results collector
results <- data.frame(
  direction = character(),
  format    = character(),
  array     = character(),
  status    = character(),
  stringsAsFactors = FALSE
)

add_result <- function(direction, fmt, arr_name, pass) {
  results <<- rbind(results, data.frame(
    direction = direction,
    format    = fmt,
    array     = arr_name,
    status    = if (pass) "PASS" else "FAIL",
    stringsAsFactors = FALSE
  ))
  if (!pass) cat(sprintf("  ** FAIL: %s / %s / %s\n", direction, fmt, arr_name))
}

# ---------------------------------------------------------------------------
# 4. R write phase
# ---------------------------------------------------------------------------
cat("\n--- R write phase ---\n")
for (fmt in c(2L, 3L)) {
  store_path <- file.path(work_dir, paste0("r_v", fmt, ".zarr"))
  dir.create(store_path)
  store <- DirectoryStore$new(store_path)
  root <- zarr_create_group(store = store, zarr_format = fmt)

  for (tc in test_cases) {
    compressor <- make_compressor(tc$codec)
    arr <- root$create_dataset(
      name       = tc$name,
      shape      = as.integer(tc$shape),
      chunks     = as.integer(tc$chunks),
      dtype      = tc$dtype_r,
      fill_value = tc$fill_value,
      compressor = compressor
    )
    arr$set_item("...", tc$data)
  }

  # Group with attributes
  sub <- root$create_group("meta_group")
  sub$get_attrs()$set_item("description", "test group")
  sub$get_attrs()$set_item("version", 1L)

  temp_arr <- sub$create_dataset(
    name       = "temperature",
    shape      = c(3L, 4L),
    chunks     = c(3L, 4L),
    dtype      = "<f4",
    fill_value = 0,
    compressor = GzipCodec$new()
  )
  temp_arr$set_item("...", matrix(c(
    280.1, 281.2, 282.3, 283.4,
    284.5, 285.6, 286.7, 287.8,
    288.9, 290.0, 291.1, 292.2
  ), nrow = 3L, ncol = 4L, byrow = TRUE))
  temp_arr$get_attrs()$set_item("units", "K")

  cat(sprintf("  V%d: wrote %d arrays + meta_group to %s\n",
  fmt, length(test_cases), store_path))
}

# ---------------------------------------------------------------------------
# 5. Run Python write script
# ---------------------------------------------------------------------------
cat("\n--- Python write phase ---\n")

py_work <- normalizePath(work_dir, winslash = "/")
py_out <- system2(python_cmd, c(py_write_script, shQuote(py_work)),
                  stdout = TRUE, stderr = TRUE)
cat(paste(py_out, collapse = "\n"), "\n")
if (!is.null(attr(py_out, "status")) && attr(py_out, "status") != 0) {
  cat("ERROR: Python write script failed:\n")
  cat(paste(py_out, collapse = "\n"), "\n")
  quit(status = 1)
}

# ---------------------------------------------------------------------------
# 6. R reads Python's stores
# ---------------------------------------------------------------------------
cat("\n--- R reads Python stores ---\n")
for (fmt in c(2L, 3L)) {
  py_store_path <- file.path(work_dir, paste0("python_v", fmt, ".zarr"))
  root <- zarr_open_group(py_store_path, mode = "r")

  for (tc in test_cases) {
    actual <- tryCatch(
      root$get_item(tc$name)$as.array(),
      error = function(e) { cat(sprintf("  ERROR reading %s: %s\n", tc$name, e$message)); NULL }
    )
    pass <- if (is.null(actual)) FALSE else compare_arrays(actual, tc$data, tc$dtype_r)
    add_result("R reads Py", paste0("V", fmt), tc$name, pass)
  }

  # Check meta_group attributes
  sub <- root$get_item("meta_group")
  attrs <- sub$get_attrs()$to_list()
  attr_pass <- identical(attrs$description, "test group") && identical(as.integer(attrs$version), 1L)
  add_result("R reads Py", paste0("V", fmt), "meta_group/attrs", attr_pass)

  # Check temperature array + attributes
  temp <- sub$get_item("temperature")
  temp_data <- temp$as.array()
  temp_expected <- matrix(c(
    280.1, 281.2, 282.3, 283.4,
    284.5, 285.6, 286.7, 287.8,
    288.9, 290.0, 291.1, 292.2
  ), nrow = 3L, ncol = 4L, byrow = TRUE)
  temp_pass <- compare_arrays(temp_data, temp_expected, "<f4")
  add_result("R reads Py", paste0("V", fmt), "meta_group/temperature", temp_pass)

  temp_attrs <- temp$get_attrs()$to_list()
  add_result("R reads Py", paste0("V", fmt), "meta_group/temperature/attrs",
  identical(temp_attrs$units, "K"))
}

# ---------------------------------------------------------------------------
# 7. Run Python read script
# ---------------------------------------------------------------------------
cat("\n--- Python reads R stores ---\n")

arr_names <- paste(vapply(test_cases, function(tc) tc$name, ""), collapse = ",")
py_json <- system2(python_cmd, c(py_read_script, shQuote(py_work), shQuote(arr_names)),
                   stdout = TRUE, stderr = TRUE)
if (!is.null(attr(py_json, "status")) && attr(py_json, "status") != 0) {
  cat("ERROR: Python read script failed:\n")
  cat(paste(py_json, collapse = "\n"), "\n")
  quit(status = 1)
}

py_results <- jsonlite::fromJSON(paste(py_json, collapse = ""))

for (i in seq_len(nrow(py_results))) {
  row <- py_results[i, ]
  fmt_str <- paste0("V", row$format)

  if (!row$ok) {
    add_result("Py reads R", fmt_str, row$name, FALSE)
    next
  }

  # Attribute checks
  if (row$name == "meta_group/attrs") {
    d <- row$data[[1]]
    pass <- identical(d$description, "test group") && identical(as.integer(d$version), 1L)
    add_result("Py reads R", fmt_str, row$name, pass)
    next
  }
  if (row$name == "meta_group/temperature/attrs") {
    d <- row$data[[1]]
    pass <- identical(d$units, "K")
    add_result("Py reads R", fmt_str, row$name, pass)
    next
  }

  # Array data checks
  tc_match <- Filter(function(tc) tc$name == row$name, test_cases)
  if (length(tc_match) == 0 && row$name == "meta_group/temperature") {
    expected_flat <- c(280.1, 281.2, 282.3, 283.4, 284.5, 285.6,
      286.7, 287.8, 288.9, 290.0, 291.1, 292.2)
      actual_flat <- unlist(row$data)
      pass <- isTRUE(all.equal(actual_flat, expected_flat, tolerance = 1e-4))
      add_result("Py reads R", fmt_str, row$name, pass)
      next
    }

    tc <- tc_match[[1]]
    actual_flat <- unlist(row$data)
    # Python flattens C-order; for multi-dim R arrays, we need C-order flat too
    if (is.array(tc$data) || is.matrix(tc$data)) {
      expected_flat <- as.vector(aperm(tc$data))
    } else {
      expected_flat <- as.vector(tc$data)
    }

    if (grepl("^[<>|]?f", tc$dtype_r)) {
      pass <- isTRUE(all.equal(as.double(actual_flat), as.double(expected_flat),
      tolerance = 1e-5))
    } else if (grepl("b1", tc$dtype_r)) {
      pass <- identical(as.logical(actual_flat), as.logical(expected_flat))
    } else {
      pass <- identical(as.integer(actual_flat), as.integer(expected_flat))
    }
    add_result("Py reads R", fmt_str, row$name, pass)
  }

  # ---------------------------------------------------------------------------
  # 8. Summary
  # ---------------------------------------------------------------------------
  cat("\n=== pizzarr Cross-Validation Report ===\n")
  cat(sprintf("Python: zarr %s | R: pizzarr %s\n\n",
  zarr_ver, packageVersion("pizzarr")))

  cat(sprintf("%-14s %-6s %-32s %s\n", "Direction", "Format", "Array", "Status"))
  cat(paste(rep("-", 66), collapse = ""), "\n")
  for (i in seq_len(nrow(results))) {
    r <- results[i, ]
    cat(sprintf("%-14s %-6s %-32s %s\n", r$direction, r$format, r$array, r$status))
  }

  n_pass <- sum(results$status == "PASS")
  n_fail <- sum(results$status == "FAIL")
  cat(sprintf("\nResults: %d/%d PASS, %d FAIL\n", n_pass, nrow(results), n_fail))

  if (n_fail > 0) {
    cat("\nFailed checks:\n")
    fails <- results[results$status == "FAIL", ]
    for (i in seq_len(nrow(fails))) {
      cat(sprintf("  - %s / %s / %s\n", fails$direction[i], fails$format[i], fails$array[i]))
    }
  }

  quit(status = as.integer(n_fail > 0))

#' @keywords internal
rmdir <- function(store, path=NA) {
    # Reference: https://github.com/zarr-developers/zarr-python/blob/5dd4a0e6cdc04c6413e14f57f61d389972ea937c/zarr/storage.py#L130

    # Remove all items under the given path. If `store` provides a `rmdir` method,
    # this will be called, otherwise will fall back to implementation via the
    # `Store` interface.

    path <- normalize_storage_path(path)
    if("rmdir" %in% names(store) && store$is_erasable()) {
        store$rmdir(path)
    }
}

#' @keywords internal
init_array_metadata <- function(
    store,
    shape,
    chunks=NA,
    dtype=NA,
    compressor=NA,
    fill_value=NA,
    order=NA,
    overwrite=FALSE,
    path=NA,
    chunk_store=NA,
    filters=NA,
    object_codec=NA,
    dimension_separator=NA,
    zarr_format=2L,
    dimension_names=NULL
) {
    # Reference: https://github.com/zarr-developers/zarr-python/blob/5dd4a0e6cdc04c6413e14f57f61d389972ea937c/zarr/storage.py#L358
    if(is_na(compressor)) {
        compressor <- "default"
    }
    if(is.na(order)) {
        order <- "F"
    }

    # guard conditions
    if (overwrite) {
        # attempt to delete any pre-existing items in store
        store$rmdir(path)
        if(!is.na(chunk_store)) {
            chunk_store$rmdir(path)
        }
    } else if (contains_array(store, path)) {
        stop("ContainsArrayError(path)")
    } else if (contains_group(store, path)) {
        stop("ContainsGroupError(path)")
    }

    # normalize metadata
    dtype <- normalize_dtype(dtype, object_codec = object_codec)

    object_codec <- normalize_object_codec(dtype, object_codec)

    # shape = normalize_shape(shape) + dtype.shape
    # dtype = dtype.base

    shape <- normalize_shape(shape)

    if (!is.null(dimension_names) && length(dimension_names) != length(shape)) {
        stop("DimensionNamesLengthError(length must match length(shape))")
    }

    dtype_itemsize <- dtype$num_bytes
    chunks <- normalize_chunks(chunks, shape, dtype_itemsize)
    order <- normalize_order(order)
    fill_value <- normalize_fill_value(fill_value, dtype)

    # # optional array metadata
    # if dimension_separator is None:
    #     dimension_separator = getattr(store, "_dimension_separator", None)
    # dimension_separator = normalize_dimension_separator(dimension_separator)

    # # compressor prep
    # if shape == ():
    #     # no point in compressing a 0-dimensional array, only a single value
    #     compressor = None
    # elif compressor == 'none':
    #     # compatibility
    #     compressor = None
    if (is.character(compressor) && compressor == "default") {
        compressor <- get_default_compressor()
    }

    # obtain compressor config
    compressor_config = NA
    if(!is_na(compressor)) {
        compressor_config <- tryCatch(
            compressor$get_config(),
            error = function(e) stop("BadCompressorError(compressor)")
        )
    }

    # obtain filters config
    filters_config <- list()
    if(!is_na(filters)) {
        filters_config <- lapply(filters, \(f) f$get_config())
    }

    # Append object codec config to filters (validation already done above)
    if (dtype$is_object && !is_na(object_codec)) {
        filters_config <- append(filters_config, list(object_codec$get_config()))
    }

    # use null to indicate no filters
    if (length(filters_config) == 0) {
        filters_config <- NA
    }

    # --- zarrs create fast path ---
    if (can_use_zarrs_create(store, dtype)) {
        preset <- compressor_to_preset(compressor, compressor_config)
        if (!is.na(preset)) {
            store_id <- store$get_store_identifier()
            v3_dtype <- v2_dtype_to_v3_dtype(dtype$dtype)$data_type
            result <- tryCatch(
                zarrs_create_array(
                    store_id, normalize_storage_path(path),
                    as.integer(shape), as.integer(chunks),
                    v3_dtype, preset, fill_value,
                    "{}", zarr_format
                ),
                error = function(e) NULL
            )
            if (!is.null(result)) return(invisible(NULL))
            # zarrs failed — fall through to R-native path
        }
    }

    # initialize metadata
    if (zarr_format == 3L) {
        v3_meta <- create_v3_array_meta(
            shape = shape,
            chunks = chunks,
            dtype = dtype$dtype,
            compressor = compressor,
            fill_value = fill_value,
            order = order,
            filters = filters,
            dimension_names = dimension_names
        )
        key <- paste0(path_to_prefix(path), ZARR_JSON)
        meta3 <- Metadata3$new()
        encoded_meta <- meta3$encode_array_metadata(v3_meta)
        store$set_item(key, encoded_meta)
    } else {
        if (!is.null(dimension_names)) {
            stop("DimensionNamesV2Error(dimension_names is only supported for zarr_format = 3L)")
        }
        zarray_meta <- create_zarray_meta(
            shape=shape,
            chunks=chunks,
            dtype=dtype,
            compressor=compressor_config,
            fill_value=fill_value,
            order=order,
            filters=filters_config,
            dimension_separator=dimension_separator
        )
        key <- paste0(path_to_prefix(path), ARRAY_META_KEY)
        encoded_meta <- store$metadata_class$encode_array_metadata(zarray_meta)
        store$set_item(key, encoded_meta)
    }
}

#' @keywords internal
init_group_metadata <- function(
    store,
    overwrite = FALSE,
    path = NA,
    chunk_store = NA,
    zarr_format = 2L
) {
    # Reference: https://github.com/zarr-developers/zarr-python/blob/5dd4a0e6cdc04c6413e14f57f61d389972ea937c/zarr/storage.py#L493
    
    # guard conditions
    if(overwrite) {
        # attempt to delete any pre-existing items in store
        store$rmdir(path)
        if(!is.na(chunk_store)) {
            chunk_store$rmdir(path)
        }
    } else if (contains_array(store, path)) {
        stop("ContainsArrayError(path)")
    } else if (contains_group(store, path)) {
        stop("ContainsGroupError(path)")
    }

    # initialize metadata
    if (zarr_format == 3L) {
        key <- paste0(path_to_prefix(path), ZARR_JSON)
        meta3 <- Metadata3$new()
        encoded_meta <- meta3$encode_group_metadata()
        store$set_item(key, encoded_meta)
    } else {
        # N.B., currently no metadata properties are needed, however there may be in future
        zgroup_meta <- obj_list()
        key <- paste0(path_to_prefix(path), GROUP_META_KEY)
        encoded_meta <- store$metadata_class$encode_group_metadata(zgroup_meta)
        store$set_item(key, encoded_meta)
    }
}

#' @keywords internal
require_parent_group <- function(
    path,
    store,
    chunk_store,
    overwrite,
    zarr_format = 2L
) {
    # Reference: https://github.com/zarr-developers/zarr-python/blob/5dd4a0e6cdc04c6413e14f57f61d389972ea937c/zarr/storage.py#L206
    # assume path is normalized
    if(!is.na(path)) {
        segments <- stringr::str_split(path, "/")[[1]]
        for(i in seq_len(length(segments) - 1)) {
            p <- paste(segments[0:i], collapse="/")
            if(contains_array(store, p)) {
                # Parent contained an array, so overwrite as group.
                init_group_metadata(store, path=p, chunk_store=chunk_store, overwrite=overwrite, zarr_format=zarr_format)
            } else if (!contains_group(store, p)) {
                init_group_metadata(store, path=p, chunk_store=chunk_store, zarr_format=zarr_format)
            }
        }
    }

}



#' Initialize an array store with the given configuration. Note that this is a low-level
#' function and there should be no need to call this directly from user code.
#' @param store : Store
#'     A mapping that supports string keys and bytes-like values.
#' @param shape : int or tuple of ints
#'     Array shape.
#' @param chunks : bool, int or tuple of ints, optional
#'     Chunk shape. If True, will be guessed from `shape` and `dtype`. If
#'     False, will be set to `shape`, i.e., single chunk for the whole array.
#' @param dtype : string or dtype, optional
#'     NumPy dtype.
#' @param compressor : Codec, optional
#'     Primary compressor.
#' @param fill_value : object
#'     Default value to use for uninitialized portions of the array.
#' @param order : \code{'C', 'F'}, optional
#'     Memory layout to be used within each chunk.
#' @param overwrite : bool, optional
#'     If True, erase all data in `store` prior to initialisation.
#' @param path : string, bytes, optional
#'     Path under which array is stored.
#' @param chunk_store : Store, optional
#'     Separate storage for chunks. If not provided, `store` will be used
#'     for storage of both chunks and metadata.
#' @param filters : sequence, optional
#'     Sequence of filters to use to encode chunk data prior to compression.
#' @param object_codec : Codec, optional
#'     A codec to encode object arrays, only needed if dtype=object.
#' @param dimension_separator : \code{'.', '/'}, optional
#'     Separator placed between the dimensions of a chunk.
#' @keywords internal
init_array <- function(
    store,
    shape,
    chunks = TRUE,
    dtype=NA,
    compressor=NA,
    fill_value=NA,
    order = NA,
    overwrite = FALSE,
    path = NA,
    chunk_store = NA,
    filters=NA,
    object_codec=NA,
    dimension_separator=NA,
    zarr_format=2L,
    dimension_names=NULL
) {
    if(is_na(compressor)) {
        compressor <- "default"
    }
    if(is.na(order)) {
        order <- "F"
    }

    # normalize path
    path <- normalize_storage_path(path)

    # ensure parent group initialized
    require_parent_group(path, store=store, chunk_store=chunk_store, overwrite=overwrite, zarr_format=zarr_format)

    init_array_metadata(
        store,
        shape=shape,
        chunks=chunks,
        dtype=dtype,
        compressor=compressor,
        fill_value=fill_value,
        order=order,
        overwrite=overwrite,
        path=path,
        chunk_store=chunk_store,
        filters=filters,
        object_codec=object_codec,
        dimension_separator=dimension_separator,
        zarr_format=zarr_format,
        dimension_names=dimension_names
    )

}

#' Initialize a group store. Note that this is a low-level function and there should be no
#' need to call this directly from user code.
#' @inheritParams init_array
#' @keywords internal
init_group <- function(
    store,
    overwrite = FALSE,
    path = NA,
    chunk_store = NA,
    zarr_format = 2L
) {
    # normalize path
    path <- normalize_storage_path(path)

    # ensure parent group initialized
    require_parent_group(path, store=store, chunk_store=chunk_store, overwrite=overwrite, zarr_format=zarr_format)

    init_group_metadata(
        store,
        overwrite=overwrite,
        path=path,
        chunk_store=chunk_store,
        zarr_format=zarr_format
    )
}


#' Create an empty array
#' @inheritParams init_array
#' @param chunks : int or tuple of ints, optional
#'     Chunk shape. If True, will be guessed from `shape` and `dtype`. If
#'     False, will be set to `shape`, i.e., single chunk for the whole array.
#'     If an int, the chunk size in each dimension will be given by the value
#'     of `chunks`. Default is True.
#' @param synchronizer : object, optional
#'     Array synchronizer.
#' @param cache_metadata : bool, optional
#'     If True, array configuration metadata will be cached for the
#'     lifetime of the object. If False, array metadata will be reloaded
#'     prior to all data access and modification operations (may incur
#'     overhead depending on storage and data access pattern).
#' @param cache_attrs : bool, optional
#'     If True (default), user attributes will be cached for attribute read
#'     operations. If False, user attributes are reloaded from the store prior
#'     to all attribute read operations.
#' @param read_only : bool, optional
#'     True if array should be protected against modification.
#' @param write_empty_chunks : bool, optional
#'     If True (default), all chunks will be stored regardless of their
#'     contents. If False, each chunk is compared to the array's fill value
#'     prior to storing. If a chunk is uniformly equal to the fill value, then
#'     that chunk is not be stored, and the store entry for that chunk's key
#'     is deleted. This setting enables sparser storage, as only chunks with
#'     non-fill-value data are stored, at the expense of overhead associated
#'     with checking the data of each chunk.
#' @param zarr_format : int, optional
#'     Zarr format version. Use \code{2L} (default) for Zarr V2 or \code{3L}
#'     for Zarr V3.
#' @param dimension_names : character vector, optional
#'     Named dimensions for V3 arrays. Length must equal \code{length(shape)}.
#'     V3 only -- passing a non-NULL value with \code{zarr_format = 2L} is an
#'     error. Default \code{NULL}.
#' @returns ZarrArray
#' @export
zarr_create <- function(
    shape,
    chunks=TRUE,
    dtype=NA,
    compressor=NA,
    fill_value=NA,
    order=NA,
    store=NA,
    synchronizer=NA,
    overwrite=FALSE,
    path=NA,
    chunk_store=NA,
    filters=NA,
    cache_metadata=TRUE,
    cache_attrs=TRUE,
    read_only=FALSE,
    object_codec=NA,
    dimension_separator=NA,
    write_empty_chunks=TRUE,
    zarr_format=2L,
    dimension_names=NULL
) {
    # Reference: https://github.com/zarr-developers/zarr-python/blob/5dd4a0e6cdc04c6413e14f57f61d389972ea937c/zarr/creation.py#L18
    if(is_na(compressor)) {
        compressor <- "default"
    }
    if(is.na(fill_value) && !is.nan(fill_value)) {
        fill_value <- 0
    }
    if(is.na(order)) {
        order <- "F"
    }

    # handle polymorphic store arg
    store <- normalize_store_arg(store)

    # initialize array metadata
    init_array(store, shape=shape, chunks=chunks, dtype=dtype, compressor=compressor,
               fill_value=fill_value, order=order, overwrite=overwrite, path=path,
               chunk_store=chunk_store, filters=filters, object_codec=object_codec,
               dimension_separator=dimension_separator, zarr_format=zarr_format,
               dimension_names=dimension_names)

    
    # instantiate array
    z <- ZarrArray$new(
        store,
        path=path,
        chunk_store=chunk_store,
        synchronizer=synchronizer,
        cache_metadata=cache_metadata,
        cache_attrs=cache_attrs,
        read_only=read_only,
        write_empty_chunks=write_empty_chunks
    )
    return(z)
}

#' Create an array filled with NAs.
#' @inheritParams zarr_create
#' @param ... The params of zarr_create()
#' @returns ZarrArray
#' @export
zarr_create_empty <- function(shape, ...) {
    return(zarr_create(shape=shape, fill_value=NA, ...))
}

#' Create an array initialized with data.
#' @param data A base R array() or pizzarr NestedArray instance.
#' @param ... The params of zarr_create()
#' @returns ZarrArray
#' @export
zarr_create_array <- function(data, ...) {
    z <- zarr_create(...)
    z$set_item("...", data)
    return(z)
}

#' Create an array filled with zeros.
#' @inheritParams zarr_create
#' @param ... The params of zarr_create()
#' @returns ZarrArray
#' @export
zarr_create_zeros <- function(shape, ...) {
    return(zarr_create(shape=shape, fill_value=0, ...))
}

#' Create a group.
#' @inheritParams init_array
#' @inheritParams zarr_create
#' @returns ZarrGroup
#' @export
zarr_create_group <- function(
    store = NA,
    overwrite = FALSE,
    chunk_store = NA,
    cache_attrs = TRUE,
    synchronizer = NA,
    path = NA,
    zarr_format = 2L
) {
    # Reference: https://github.com/zarr-developers/zarr-python/blob/5dd4a0e6cdc04c6413e14f57f61d389972ea937c/zarr/hierarchy.py#L1061
    # handle polymorphic store arg
    store <- normalize_store_arg(store)
    path <- normalize_storage_path(path)

    if(overwrite || !contains_group(store)) {
        init_group(
            store,
            overwrite = overwrite,
            chunk_store = chunk_store,
            path = path,
            zarr_format = zarr_format
        )
    }
    return(ZarrGroup$new(
        store,
        read_only = FALSE,
        chunk_store = chunk_store,
        cache_attrs = cache_attrs,
        synchronizer = synchronizer,
        path = path
    ))
}

#' Open a group using file-mode-like semantics.
#' @inheritParams zarr_open
#' @inheritParams zarr_create
#' @param storage_options : dict
#'     If using an fsspec URL to create the store, these will be passed to
#'     the backend implementation. Ignored otherwise.
#' @returns ZarrGroup
#' @export
zarr_open_group <- function(
    store = NA,
    mode = NA,
    cache_attrs = TRUE,
    synchronizer = NA,
    path = NA,
    chunk_store = NA,
    storage_options = NA,
    zarr_format = 2L
) {
    # Reference: https://github.com/zarr-developers/zarr-python/blob/5dd4a0e6cdc04c6413e14f57f61d389972ea937c/zarr/hierarchy.py#L1119

    if(is_na(mode)) {
        mode <- "a"
    }

    # handle polymorphic store arg
    store <- normalize_store_arg(store, storage_options=storage_options, mode=mode)
    if(!is_na(chunk_store)) {
        chunk_store <- normalize_store_arg(chunk_store, storage_options=storage_options, mode=mode)
    }
    path <- normalize_storage_path(path)

    # Auto-detect V3 format from existing store
    v <- detect_zarr_version(store, path)
    if (!is.na(v) && v == 3L) {
      zarr_format <- 3L
    }

    # ensure store is initialized

    if(mode == "r" || mode == "r+") {
        if (!contains_group(store, path)) {
            if(contains_array(store, path)) {
                stop("ContainsArrayError(path)")
            }
            stop("GroupNotFoundError(path)")
        }

    } else if (mode == 'w') {
        init_group(store, overwrite=TRUE, path=path, chunk_store=chunk_store, zarr_format=zarr_format)

    } else if(mode == "a") {
        if (!contains_group(store, path)) {
            if(contains_array(store, path)) {
                stop("ContainsArrayError(path)")
            }
            init_group(store, path=path, chunk_store=chunk_store, zarr_format=zarr_format)
        }
    } else if(mode == "w-" || mode == "x") {
        if (contains_array(store, path)) {
            stop("ContainsArrayError(path)")
        } else if (contains_group(store, path)) {
            stop("ContainsGroupError(path)")
        } else {
            init_group(store, path=path, chunk_store=chunk_store, zarr_format=zarr_format)
        }
    }

    # determine read only status
    read_only <- (mode == "r")

    return(ZarrGroup$new(
        store,
        read_only=read_only,
        cache_attrs=cache_attrs,
        synchronizer=synchronizer,
        path=path,
        chunk_store=chunk_store
    ))
}

#' Open an array using file-mode-like semantics.
#' @inheritParams init_array
#' @inheritParams zarr_create
#' @inheritParams zarr_open
#' @inheritParams zarr_open_group
#' @returns ZarrArray
#' @export
zarr_open_array <- function(
    store = NA,
    storage_options = NA,
    mode = NA,
    shape = NA,
    chunks=TRUE,
    dtype=NA,
    compressor=NA,
    fill_value=NA,
    order=NA,
    synchronizer=NA,
    overwrite=FALSE,
    path=NA,
    chunk_store=NA,
    filters=NA,
    cache_metadata=TRUE,
    cache_attrs=TRUE,
    object_codec=NA,
    dimension_separator=NA,
    write_empty_chunks=TRUE,
    zarr_format=2L
) {
    # Reference: https://github.com/zarr-developers/zarr-python/blob/5dd4a0e6cdc04c6413e14f57f61d389972ea937c/zarr/creation.py#L376
    if(is_na(compressor)) {
        compressor <- "default"
    }
    if(is.na(fill_value) && !is.nan(fill_value)) {
        fill_value <- 0
    }
    if(is.na(order)) {
        order <- "F"
    }

    if(is_na(mode)) {
        mode <- "a"
    }

    # handle polymorphic store arg
    store <- normalize_store_arg(store, storage_options=storage_options, mode=mode)
    if(!is_na(chunk_store)) {
        chunk_store <- normalize_store_arg(chunk_store, storage_options=storage_options, mode=mode)
    }
    path <- normalize_storage_path(path)

    # Auto-detect V3 format from existing store
    v <- detect_zarr_version(store, path)
    if (!is.na(v) && v == 3L) {
      zarr_format <- 3L
    }

    # ensure store is initialized
    if(mode == "r" || mode == "r+") {
        if (!contains_array(store, path)) {
            if(contains_group(store, path)) {
                stop("ContainsGroupError(path)")
            }
            stop("ArrayNotFoundError(path)")
        }

    } else if (mode == 'w') {
        init_array(
            store, shape=shape, chunks=chunks, dtype=dtype,
            compressor=compressor, fill_value=fill_value,
            order=order, filters=filters, overwrite=TRUE, path=path,
            object_codec=object_codec, chunk_store=chunk_store,
            zarr_format=zarr_format
        )

    } else if(mode == "a") {
        if (!contains_array(store, path)) {
            if(contains_group(store, path)) {
                stop("ContainGroupError(path)")
            }
            init_array(
                store, shape=shape, chunks=chunks, dtype=dtype,
                compressor=compressor, fill_value=fill_value,
                order=order, filters=filters, path=path,
                object_codec=object_codec, chunk_store=chunk_store,
                zarr_format=zarr_format
            )
        }
    } else if(mode == "w-" || mode == "x") {
        if (contains_group(store, path)) {
            stop("ContainsGroupError(path)")
        } else if (contains_array(store, path)) {
            stop("ContainsArrayError(path)")
        } else {
            init_array(
                store, shape=shape, chunks=chunks, dtype=dtype,
                compressor=compressor, fill_value=fill_value,
                order=order, filters=filters, path=path,
                object_codec=object_codec, chunk_store=chunk_store,
                zarr_format=zarr_format
            )
        }
    }

    # determine read only status
    read_only <- (mode == "r")

    # instantiate array
    z <- ZarrArray$new(
        store,
        path=path,
        chunk_store=chunk_store,
        synchronizer=synchronizer,
        cache_metadata=cache_metadata,
        cache_attrs=cache_attrs,
        read_only=read_only,
        write_empty_chunks=write_empty_chunks
    )
    return(z)
}

#' Convenience function to save a ZarrArray to the local file system.
#' @inheritParams init_array
#' @param arr : ZarrArray
#'     The array with data to save.
#' @param ... Additional arguments to pass to zarr_create_array().
#' @export
zarr_save_array <- function(store, arr, ...) {
    # Reference: https://github.com/zarr-developers/zarr-python/blob/5dd4a0e6cdc04c6413e14f57f61d389972ea937c/zarr/convenience.py#L112
    store <- normalize_store_arg(store)
    zarr_create_array(data=arr$get_item("..."), shape=arr$get_shape(), store=store, ...)
}

#' Convenience function to open a group or array using file-mode-like semantics.
#' @inheritParams init_array
#' @param mode : \code{'r', 'r+', 'a', 'w', 'w-'}, optional
#'     Persistence mode: 'r' means read only (must exist); 'r+' means
#'     read/write (must exist); 'a' means read/write (create if doesn't
#'     exist); 'w' means create (overwrite if exists); 'w-' means create
#'     (fail if exists).
#' @param ... Additional arguments to pass to zarr_open_array or zarr_open_group.
#' @returns ZarrArray or ZarrGroup
#' @export
zarr_open <- function(store = NA, mode = NA, path = NA, ...) {
    kwargs <- list(...)

    store <- normalize_store_arg(store)
    path <- normalize_storage_path(path)
    
    if(is_na(mode)) {
        mode <- "a"
        if(inherits(store, "HttpStore"))
          mode <- "r"
    }

    if(mode %in% c("w", "w-", "x")) {
        if("shape" %in% names(kwargs)) {
            return(zarr_open_array(store=store, mode=mode, path=path, ...))
        } else {
            return(zarr_open_group(store=store, mode=mode, path=path, ...))
        }
    } else if(mode == "a") {
        if("shape" %in% names(kwargs) || contains_array(store, path)) {
            return(zarr_open_array(store=store, mode=mode, path=path, ...))
        } else {
            return(zarr_open_group(store=store, mode=mode, path=path, ...))
        }
    } else {
        if(contains_array(store, path)) {
            return(zarr_open_array(store=store, mode=mode, path=path, ...))
        } else if(contains_group(store, path)) {
            return(zarr_open_group(store=store, mode=mode, path=path, ...))
        } else {
            stop("PathNotFoundError(path)")
        }
    }
}

# --- zarrs_set_subset: Rust write path tests ---

test_that("zarrs round-trip write+read 1D float64", {
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")
  d <- tempfile("zarrs_wr_1d_")
  z <- zarr_create(store = d, shape = 10L, chunks = 5L, dtype = "<f8")

  # Write via zarrs (through set_item dispatch)
  z$set_item("...", as.double(1:10))

  # Read back via zarrs
  result <- zarrs_get_subset(d, "", list(c(0L, 10L)), NULL)
  expect_equal(result$data, as.double(1:10))

  zarrs_close_store(d)
  unlink(d, recursive = TRUE)
})

test_that("zarrs round-trip write+read 2D int32", {
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")
  d <- tempfile("zarrs_wr_2d_")
  z <- zarr_create(store = d, shape = c(4L, 5L), chunks = c(4L, 5L),
                   dtype = "<i4")
  data <- array(1:20, dim = c(4, 5))
  z$set_item("...", data)

  # Read back via zarrs
  result <- zarrs_get_subset(d, "", list(c(0L, 4L), c(0L, 5L)), NULL)
  expect_equal(result$shape, c(4L, 5L))
  # zarrs returns F-order data; reshape directly
  arr <- array(result$data, dim = c(4, 5))
  expect_equal(arr, data)

  zarrs_close_store(d)
  unlink(d, recursive = TRUE)
})

test_that("zarrs write + R-native read cross-path", {
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")
  d <- tempfile("zarrs_wr_cross_")
  z <- zarr_create(store = d, shape = 10L, chunks = 5L, dtype = "<f8")

  # Write via zarrs (through set_item)
  z$set_item("...", as.double(11:20))

  # Close zarrs cache so R-native reads from disk
  zarrs_close_store(d)

  # Read via R-native by disabling zarrs
  old <- .pizzarr_env$zarrs_available
  .pizzarr_env$zarrs_available <- FALSE
  z2 <- zarr_open(store = d)
  result <- z2$get_item("...")
  .pizzarr_env$zarrs_available <- old

  expect_equal(as.double(result$data), as.double(11:20))

  unlink(d, recursive = TRUE)
})

test_that("zarrs partial subset overwrite", {
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")
  d <- tempfile("zarrs_wr_partial_")
  z <- zarr_create(store = d, shape = 10L, chunks = 5L, dtype = "<f8",
                   fill_value = 0.0)
  # Write full array first
  z$set_item("...", as.double(1:10))

  # Overwrite a slice [3:7] (1-based) = elements 3,4,5,6,7
  z$set_item(list(slice(3L, 7L)), as.double(c(99, 98, 97, 96, 95)))

  # Read back and verify
  result <- z$get_item("...")
  expected <- as.double(c(1, 2, 99, 98, 97, 96, 95, 8, 9, 10))
  expect_equal(as.double(result$data), expected)

  zarrs_close_store(d)
  unlink(d, recursive = TRUE)
})

test_that("zarrs writes int16 with narrowing", {
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")
  d <- tempfile("zarrs_wr_i16_")
  z <- zarr_create(store = d, shape = 5L, chunks = 5L, dtype = "<i2")

  # Write R integers to i16 array
  z$set_item("...", 1:5)

  result <- zarrs_get_subset(d, "", list(c(0L, 5L)), NULL)
  expect_type(result$data, "integer")
  expect_equal(result$data, 1:5)

  zarrs_close_store(d)
  unlink(d, recursive = TRUE)
})

test_that("zarrs 2D F-order to C-order conversion on write", {
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")
  d <- tempfile("zarrs_wr_forder_")
  z <- zarr_create(store = d, shape = c(3L, 4L), chunks = c(3L, 4L),
                   dtype = "<f8")
  # R array in F-order (column-major)
  data <- array(as.double(1:12), dim = c(3, 4))
  z$set_item("...", data)

  # Read back and compare zarrs vs R-native
  z2 <- zarr_open(store = d)
  zarrs_result <- z2$get_item("...")

  old <- .pizzarr_env$zarrs_available
  .pizzarr_env$zarrs_available <- FALSE
  r_result <- z2$get_item("...")
  .pizzarr_env$zarrs_available <- old

  expect_equal(zarrs_result$data, r_result$data)

  zarrs_close_store(d)
  unlink(d, recursive = TRUE)
})

test_that("zarrs write falls through for MemoryStore", {
  ms <- MemoryStore$new()
  z <- zarr_create(store = ms, shape = 5L, chunks = 5L, dtype = "<f8")

  # This should work via R-native path (MemoryStore has no store_id)
  z$set_item("...", as.double(1:5))
  result <- z$get_item("...")
  expect_equal(as.double(result$data), as.double(1:5))
})

test_that("zarrs write concurrent_target option propagated", {
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")
  d <- tempfile("zarrs_wr_ct_")
  z <- zarr_create(store = d, shape = 10L, chunks = 5L, dtype = "<f8")

  # Direct call with concurrent_target
  result <- zarrs_set_subset(d, "", list(c(0L, 10L)),
                               as.double(1:10), 1L)
  expect_true(result)

  # Verify data was written
  read_result <- zarrs_get_subset(d, "", list(c(0L, 10L)), NULL)
  expect_equal(read_result$data, as.double(1:10))

  zarrs_close_store(d)
  unlink(d, recursive = TRUE)
})

# --- zarrs_get_subset: Rust read path tests ---

test_that("zarrs retrieves 1D float64 array", {
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")
  d <- tempfile("zarrs_ret_1d_")
  z <- zarr_create(store = d, shape = 10L, chunks = 5L, dtype = "<f8")
  z$set_item("...", as.double(1:10))

  result <- zarrs_get_subset(d, "", list(c(0L, 10L)), NULL)
  expect_equal(result$data, as.double(1:10))
  expect_equal(result$shape, 10L)

  zarrs_close_store(d)
  unlink(d, recursive = TRUE)
})

test_that("zarrs retrieves 2D integer array slice", {
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")
  d <- tempfile("zarrs_ret_2d_")
  z <- zarr_create(store = d, shape = c(10L, 5L), chunks = c(5L, 5L), dtype = "<i4")
  z$set_item("...", array(1:50, dim = c(10, 5)))

  # Read rows 2-5, all columns (0-based: 2:5 x 0:5)
  result <- zarrs_get_subset(d, "", list(c(2L, 5L), c(0L, 5L)), NULL)
  expect_equal(result$shape, c(3L, 5L))
  expect_type(result$data, "integer")
  expect_equal(length(result$data), 15L)

  zarrs_close_store(d)
  unlink(d, recursive = TRUE)
})

test_that("zarrs retrieves V3 array", {
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")
  d <- tempfile("zarrs_ret_v3_")
  store <- DirectoryStore$new(d)
  g <- zarr_create_group(store = store, zarr_format = 3L)
  a <- g$create_dataset("arr", shape = 6L, dtype = "<f8", compressor = NA)
  a$set_item("...", as.double(11:16))

  result <- zarrs_get_subset(d, "arr", list(c(0L, 6L)), NULL)
  expect_equal(result$data, as.double(11:16))
  expect_equal(result$shape, 6L)

  zarrs_close_store(d)
  unlink(d, recursive = TRUE)
})

test_that("zarrs retrieves int16 array widened to integer", {
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")
  d <- tempfile("zarrs_ret_i16_")
  z <- zarr_create(store = d, shape = 5L, chunks = 5L, dtype = "<i2")
  z$set_item("...", 1:5)

  result <- zarrs_get_subset(d, "", list(c(0L, 5L)), NULL)
  expect_type(result$data, "integer")
  expect_equal(result$data, 1:5)

  zarrs_close_store(d)
  unlink(d, recursive = TRUE)
})

test_that("zarrs retrieves uint8 array widened to integer", {
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")
  d <- tempfile("zarrs_ret_u8_")
  z <- zarr_create(store = d, shape = 5L, chunks = 5L, dtype = "|u1")
  z$set_item("...", 1:5)

  result <- zarrs_get_subset(d, "", list(c(0L, 5L)), NULL)
  expect_type(result$data, "integer")
  expect_equal(result$data, 1:5)

  zarrs_close_store(d)
  unlink(d, recursive = TRUE)
})

test_that("zarrs returns fill values for unwritten chunks", {
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")
  d <- tempfile("zarrs_ret_fill_")
  z <- zarr_create(store = d, shape = 10L, chunks = 5L, dtype = "<f8",
                   fill_value = 42.0)

  result <- zarrs_get_subset(d, "", list(c(0L, 10L)), NULL)
  expect_equal(result$data, rep(42.0, 10))

  zarrs_close_store(d)
  unlink(d, recursive = TRUE)
})

test_that("zarrs errors on unsupported dtype (string)", {
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")
  # zarrs_get_subset should error on string arrays
  # (we don't have a convenient way to create a string zarr array from R,
  #  so test via dtype_family returning None — the error path)
  # Instead, test that the dispatch function correctly falls back.
  # Use a MemoryStore to test can_use_zarrs returning FALSE.
  ms <- MemoryStore$new()
  z <- zarr_create(store = ms, shape = 5L, chunks = 5L, dtype = "<f8")
  indexer <- BasicIndexer$new(list(slice(1L, 5L)), z)
  expect_false(can_use_zarrs(indexer, ms))
})

test_that("concurrent_target option is honored", {
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")
  d <- tempfile("zarrs_ret_ct_")
  z <- zarr_create(store = d, shape = 10L, chunks = 5L, dtype = "<f8")
  z$set_item("...", as.double(1:10))

  result <- zarrs_get_subset(d, "", list(c(0L, 10L)), 1L)
  expect_equal(result$data, as.double(1:10))

  zarrs_close_store(d)
  unlink(d, recursive = TRUE)
})

# --- R-side dispatch tests ---

test_that("can_use_zarrs returns FALSE for MemoryStore", {
  ms <- MemoryStore$new()
  z <- zarr_create(store = ms, shape = 5L, chunks = 5L, dtype = "<f8")
  indexer <- BasicIndexer$new(list(slice(1L, 5L)), z)
  expect_false(can_use_zarrs(indexer, ms))
})

test_that("can_use_zarrs returns FALSE when zarrs unavailable", {
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")
  d <- tempfile("zarrs_disp_")
  z <- zarr_create(store = d, shape = 5L, chunks = 5L, dtype = "<f8")
  ds <- DirectoryStore$new(d)
  indexer <- BasicIndexer$new(list(slice(1L, 5L)), z)

  old <- .pizzarr_env$zarrs_available
  .pizzarr_env$zarrs_available <- FALSE
  on.exit(.pizzarr_env$zarrs_available <- old)
  expect_false(can_use_zarrs(indexer, ds))

  unlink(d, recursive = TRUE)
})

test_that("selection_to_ranges converts indexer dims", {
  ms <- MemoryStore$new()
  z <- zarr_create(store = ms, shape = c(10L, 20L), chunks = c(5L, 10L),
                   dtype = "<f8")
  indexer <- BasicIndexer$new(list(slice(2L, 8L), slice(5L, 15L)), z)
  ranges <- selection_to_ranges(indexer)

  expect_length(ranges, 2)
  # slice(2L, 8L) is 1-based; SliceDimIndexer stores 0-based start=1, stop=8
  expect_equal(ranges[[1]], c(1L, 8L))
  # slice(5L, 15L) is 1-based; SliceDimIndexer stores 0-based start=4, stop=15
  expect_equal(ranges[[2]], c(4L, 15L))
})

test_that("can_use_zarrs accepts IntDimIndexer (manual construction)", {
  # IntDimIndexer is only created when normalize_list_selection uses
  # convert_integer_selection_to_slices=FALSE (not the default get_item path).
  # Test the dispatch function handles it correctly anyway.
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")
  d <- tempfile("zarrs_int_dim_")
  z <- zarr_create(store = d, shape = c(10L, 5L), chunks = c(5L, 5L),
                   dtype = "<f8")
  ds <- DirectoryStore$new(d)
  # Manually build an indexer with IntDimIndexer on dim 1
  int_di <- IntDimIndexer$new(3L, 10L, 5L)
  slice_di <- SliceDimIndexer$new(slice(1L, 5L), 5L, 5L)
  idx <- list()
  idx$dim_indexers <- list(int_di, slice_di)
  idx$shape <- list(5L)
  class(idx) <- "BasicIndexer"
  expect_true(can_use_zarrs(idx, ds))

  unlink(d, recursive = TRUE)
})

test_that("selection_to_ranges handles IntDimIndexer", {
  # IntDimIndexer stores 0-based dim_sel from normalize_integer_selection.
  # 3L -> dim_sel=3 (0-based) -> range [3, 4)
  int_di <- IntDimIndexer$new(3L, 10L, 5L)
  slice_di <- SliceDimIndexer$new(slice(5L, 15L), 20L, 10L)
  idx <- list()
  idx$dim_indexers <- list(int_di, slice_di)
  class(idx) <- "BasicIndexer"
  ranges <- selection_to_ranges(idx)

  expect_length(ranges, 2)
  expect_equal(ranges[[1]], c(3L, 4L))
  # slice(5L, 15L) is 1-based; SliceDimIndexer stores 0-based start=4, stop=15
  expect_equal(ranges[[2]], c(4L, 15L))
})

test_that("zarrs and R-native read produce identical results", {
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")
  d <- tempfile("zarrs_compare_")
  z <- zarr_create(store = d, shape = c(10L, 8L), chunks = c(5L, 4L),
                   dtype = "<f8")
  data <- array(as.double(1:80), dim = c(10, 8))
  z$set_item("...", data)

  # Read via zarrs
  zarrs_result <- zarrs_get_subset(d, "", list(c(2L, 7L), c(1L, 6L)), NULL)

  # Read via R-native
  z2 <- zarr_open(store = d)
  r_result <- z2$get_item(list(slice(3L, 7L), slice(2L, 6L)))

  # zarrs returns flat data; R-native returns NestedArray
  expect_equal(length(zarrs_result$data), length(r_result$data))

  zarrs_close_store(d)
  unlink(d, recursive = TRUE)
})

# --- End-to-end dispatch: ZarrArray$get_item via zarrs ---

test_that("ZarrArray$get_item dispatches to zarrs for eligible selection", {
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")
  d <- tempfile("zarrs_e2e_")
  z <- zarr_create(store = d, shape = 10L, chunks = 5L, dtype = "<f8")
  z$set_item("...", as.double(1:10))

  # Re-open to get a DirectoryStore-backed array
  z2 <- zarr_open(store = d)
  result <- z2$get_item("...")
  expect_equal(as.double(result$data), as.double(1:10))

  zarrs_close_store(d)
  unlink(d, recursive = TRUE)
})

test_that("ZarrArray$get_item dispatches to zarrs for 2D slice", {
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")
  d <- tempfile("zarrs_e2e_2d_")
  z <- zarr_create(store = d, shape = c(10L, 5L), chunks = c(5L, 5L),
                   dtype = "<i4")
  z$set_item("...", array(1:50, dim = c(10, 5)))

  z2 <- zarr_open(store = d)
  # Slice rows 3-7 (1-based), all cols
  result <- z2$get_item(list(slice(3L, 7L), slice(1L, 5L)))
  expect_equal(dim(result$data), c(5, 5))

  zarrs_close_store(d)
  unlink(d, recursive = TRUE)
})

test_that("ZarrArray$get_item dispatches to zarrs for scalar + slice", {
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")
  d <- tempfile("zarrs_e2e_scalar_")
  z <- zarr_create(store = d, shape = c(10L, 5L), chunks = c(5L, 5L),
                   dtype = "<f8")
  data <- array(as.double(1:50), dim = c(10, 5))
  z$set_item("...", data)

  z2 <- zarr_open(store = d)
  # 3L (0-based index 3) selects the 4th row; get_item converts to slice(3,4)
  # so output shape keeps the length-1 dim: c(1, 5)
  zarrs_result <- z2$get_item(list(3L, slice(1L, 5L)))
  expect_equal(dim(zarrs_result$data), c(1L, 5L))
  # Compare with R-native
  old <- .pizzarr_env$zarrs_available
  .pizzarr_env$zarrs_available <- FALSE
  r_result <- z2$get_item(list(3L, slice(1L, 5L)))
  .pizzarr_env$zarrs_available <- old
  expect_equal(zarrs_result$data, r_result$data)

  zarrs_close_store(d)
  unlink(d, recursive = TRUE)
})

# Tests for zarrs_create_array — Rust-side array creation.
# All tests gated on zarrs availability (skip on CRAN / pure-R tier).

skip_if_no_zarrs <- function() {
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")
}

# -- V3 creation tests -------------------------------------------------------

test_that("zarrs_create_array V3 float64 no compression", {
  skip_if_no_zarrs()
  d <- tempfile("zarrs_create_v3_none_")
  dir.create(d)
  on.exit(unlink(d, recursive = TRUE))

  meta <- zarrs_create_array(d, "", c(10L, 20L), c(5L, 10L),
                              "float64", "none", 0.0, "{}", 3L)
  expect_equal(meta$shape, c(10L, 20L))
  expect_equal(meta$chunks, c(5L, 10L))
  expect_equal(meta$dtype, "float64")
  expect_equal(meta$r_type, "double")
  expect_equal(meta$zarr_format, 3L)

  # Metadata file should exist
  expect_true(file.exists(file.path(d, "zarr.json")))
  zarrs_close_store(d)
})

test_that("zarrs_create_array V3 float64 gzip roundtrip", {
  skip_if_no_zarrs()
  d <- tempfile("zarrs_create_v3_gzip_")
  dir.create(d)
  on.exit(unlink(d, recursive = TRUE))

  zarrs_create_array(d, "", 10L, 5L, "float64", "gzip", 0.0, "{}", 3L)

  # Write and read back
  zarrs_set_subset(d, "", list(c(0L, 10L)), as.double(1:10), NULL)
  result <- zarrs_get_subset(d, "", list(c(0L, 10L)), NULL)
  expect_equal(result$data, as.double(1:10))
  zarrs_close_store(d)
})

test_that("zarrs_create_array V3 int32 zstd roundtrip", {
  skip_if_no_zarrs()
  skip_if(!("zstd" %in% pizzarr_compiled_features()),
          "zstd not compiled")
  d <- tempfile("zarrs_create_v3_zstd_")
  dir.create(d)
  on.exit(unlink(d, recursive = TRUE))

  zarrs_create_array(d, "", 10L, 5L, "int32", "zstd", 0L, "{}", 3L)

  zarrs_set_subset(d, "", list(c(0L, 10L)), 1:10, NULL)
  result <- zarrs_get_subset(d, "", list(c(0L, 10L)), NULL)
  expect_equal(result$data, 1:10)
  zarrs_close_store(d)
})

# -- V2 creation tests -------------------------------------------------------

test_that("zarrs_create_array V2 float64 no compression", {
  skip_if_no_zarrs()
  d <- tempfile("zarrs_create_v2_none_")
  dir.create(d)
  on.exit(unlink(d, recursive = TRUE))

  meta <- zarrs_create_array(d, "", c(10L, 20L), c(5L, 10L),
                              "float64", "none", 0.0, "{}", 2L)
  expect_equal(meta$shape, c(10L, 20L))
  expect_equal(meta$chunks, c(5L, 10L))
  expect_equal(meta$dtype, "float64")
  expect_equal(meta$zarr_format, 2L)

  # V2 metadata file should exist
  expect_true(file.exists(file.path(d, ".zarray")))
  zarrs_close_store(d)
})

test_that("zarrs_create_array V2 float64 gzip roundtrip", {
  skip_if_no_zarrs()
  d <- tempfile("zarrs_create_v2_gzip_")
  dir.create(d)
  on.exit(unlink(d, recursive = TRUE))

  zarrs_create_array(d, "", 10L, 5L, "float64", "gzip", 0.0, "{}", 2L)

  zarrs_set_subset(d, "", list(c(0L, 10L)), as.double(1:10), NULL)
  result <- zarrs_get_subset(d, "", list(c(0L, 10L)), NULL)
  expect_equal(result$data, as.double(1:10))
  zarrs_close_store(d)
})

test_that("zarrs_create_array V2 zstd rejected", {
  skip_if_no_zarrs()
  d <- tempfile("zarrs_create_v2_zstd_")
  dir.create(d)
  on.exit(unlink(d, recursive = TRUE))

  expect_error(
    zarrs_create_array(d, "", 10L, 5L, "float64", "zstd", 0.0, "{}", 2L),
    "zstd.*zarr_format"
  )
  zarrs_close_store(d)
})

# -- All dtypes ---------------------------------------------------------------

test_that("zarrs_create_array V3 supports all numeric dtypes", {
  skip_if_no_zarrs()

  dtypes <- c("float64", "float32", "int32", "int16", "int8",
              "uint8", "uint16", "uint32", "int64", "uint64", "bool")

  for (dt in dtypes) {
    d <- tempfile(paste0("zarrs_create_dtype_", dt, "_"))
    dir.create(d)

    meta <- zarrs_create_array(d, "", 10L, 5L, dt, "none", 0L, "{}", 3L)
    expect_equal(meta$dtype, dt, info = paste("dtype:", dt))
    expect_true(meta$r_type %in% c("double", "integer", "logical"),
                info = paste("dtype:", dt))

    zarrs_close_store(d)
    unlink(d, recursive = TRUE)
  }
})

# -- End-to-end roundtrip -----------------------------------------------------

test_that("zarrs create + write + read full roundtrip", {
  skip_if_no_zarrs()
  d <- tempfile("zarrs_roundtrip_")
  dir.create(d)
  on.exit(unlink(d, recursive = TRUE))

  # Create 2D V3 array with gzip
  zarrs_create_array(d, "", c(4L, 6L), c(2L, 3L),
                      "float64", "gzip", 0.0, "{}", 3L)

  # Write a subset
  zarrs_set_subset(d, "", list(c(0L, 4L), c(0L, 6L)),
                   as.double(1:24), NULL)

  # Read back
  result <- zarrs_get_subset(d, "", list(c(0L, 4L), c(0L, 6L)), NULL)
  expect_equal(result$data, as.double(1:24))
  expect_equal(result$shape, c(4L, 6L))

  zarrs_close_store(d)
})

# -- Transparent dispatch via zarr_create() -----------------------------------

test_that("zarr_create transparently uses zarrs backend for DirectoryStore", {
  skip_if_no_zarrs()
  d <- tempfile("zarrs_dispatch_")
  on.exit(unlink(d, recursive = TRUE))

  z <- zarr_create(shape = c(10L, 20L), chunks = c(5L, 10L),
                   dtype = "<f8", store = d, zarr_format = 3L)

  expect_s3_class(z, "ZarrArray")
  # V3 metadata should exist
  expect_true(file.exists(file.path(d, "zarr.json")))
})

test_that("zarr_create V2 works via zarrs dispatch", {
  skip_if_no_zarrs()
  d <- tempfile("zarrs_dispatch_v2_")
  on.exit(unlink(d, recursive = TRUE))

  z <- zarr_create(shape = 10L, chunks = 5L, dtype = "<f8",
                   store = d, zarr_format = 2L)

  expect_s3_class(z, "ZarrArray")
  expect_true(file.exists(file.path(d, ".zarray")))
})

# -- R-native fallback --------------------------------------------------------

test_that("zarr_create falls back for MemoryStore", {
  z <- zarr_create(shape = 10L, chunks = 5L, dtype = "<f8")
  expect_s3_class(z, "ZarrArray")
})

# -- Fill value handling ------------------------------------------------------

test_that("zarrs_create_array fill value 0 roundtrip", {
  skip_if_no_zarrs()
  d <- tempfile("zarrs_fv0_")
  dir.create(d)
  on.exit(unlink(d, recursive = TRUE))

  zarrs_create_array(d, "", 5L, 5L, "float64", "none", 0.0, "{}", 3L)
  # Read uninitialized data — should be fill value
  result <- zarrs_get_subset(d, "", list(c(0L, 5L)), NULL)
  expect_equal(result$data, rep(0.0, 5))
  zarrs_close_store(d)
})

test_that("zarrs_create_array fill value 42 roundtrip", {
  skip_if_no_zarrs()
  d <- tempfile("zarrs_fv42_")
  dir.create(d)
  on.exit(unlink(d, recursive = TRUE))

  zarrs_create_array(d, "", 5L, 5L, "float64", "none", 42.0, "{}", 3L)
  result <- zarrs_get_subset(d, "", list(c(0L, 5L)), NULL)
  expect_equal(result$data, rep(42.0, 5))
  zarrs_close_store(d)
})

test_that("zarrs_create_array fill value NA maps to NaN for floats", {
  skip_if_no_zarrs()
  d <- tempfile("zarrs_fvna_")
  dir.create(d)
  on.exit(unlink(d, recursive = TRUE))

  zarrs_create_array(d, "", 5L, 5L, "float64", "none", NA, "{}", 3L)
  result <- zarrs_get_subset(d, "", list(c(0L, 5L)), NULL)
  expect_true(all(is.nan(result$data)))
  zarrs_close_store(d)
})

# -- Cross-read: zarrs-created, R-native read ---------------------------------

test_that("zarrs-created V3 array readable by R-native path", {
  skip_if_no_zarrs()
  d <- tempfile("zarrs_crossread_")
  dir.create(d)
  on.exit(unlink(d, recursive = TRUE))

  # Create and populate via zarrs
  zarrs_create_array(d, "", 10L, 5L, "float64", "gzip", 0.0, "{}", 3L)
  zarrs_set_subset(d, "", list(c(0L, 10L)), as.double(1:10), NULL)
  zarrs_close_store(d)

  # Read via R-native path (open DirectoryStore, read normally)
  store <- DirectoryStore$new(d)
  z <- ZarrArray$new(store)
  data <- z$get_item("...")$data
  expect_equal(as.vector(data), as.double(1:10))
})

# -- Nested path with parent groups -------------------------------------------

test_that("zarrs_create_array with nested path via zarr_create", {
  skip_if_no_zarrs()
  d <- tempfile("zarrs_nested_")
  on.exit(unlink(d, recursive = TRUE))

  z <- zarr_create(shape = 10L, chunks = 5L, dtype = "<f8",
                   store = d, path = "group1/array1",
                   zarr_format = 3L)
  expect_s3_class(z, "ZarrArray")

  # Parent group metadata should exist
  expect_true(file.exists(file.path(d, "group1", "zarr.json")))
  # Array metadata should exist
  expect_true(file.exists(file.path(d, "group1", "array1", "zarr.json")))
})

# -- Error cases --------------------------------------------------------------

test_that("zarrs_create_array rejects unsupported dtype", {
  skip_if_no_zarrs()
  d <- tempfile("zarrs_bad_dtype_")
  dir.create(d)
  on.exit(unlink(d, recursive = TRUE))

  expect_error(
    zarrs_create_array(d, "", 10L, 5L, "string", "none", "", "{}", 3L),
    "DTypeUnsupported"
  )
  zarrs_close_store(d)
})

test_that("zarrs_create_array rejects shape/chunks length mismatch", {
  skip_if_no_zarrs()
  d <- tempfile("zarrs_mismatch_")
  dir.create(d)
  on.exit(unlink(d, recursive = TRUE))

  expect_error(
    zarrs_create_array(d, "", c(10L, 20L), 5L, "float64", "none", 0.0, "{}", 3L),
    "shape length.*chunks length"
  )
  zarrs_close_store(d)
})

library(pizzarr)

# --- Helper: create a pre-existing array in a store ---
create_existing_array <- function(store, path = NA) {
  zarr_create(
    shape = c(4), chunks = c(4), dtype = "<f8",
    store = store, path = path, fill_value = 0
  )
}

create_existing_group <- function(store, path = NA) {
  zarr_open_group(store, path = path)
}

# =============================================
# zarr_open_array mode tests
# =============================================

# --- Mode "r" ---

test_that("zarr_open_array mode='r' fails if array doesn't exist", {
  store <- MemoryStore$new()
  expect_error(zarr_open_array(store, mode = "r"))
})

test_that("zarr_open_array mode='r' succeeds on existing array", {
  store <- MemoryStore$new()
  create_existing_array(store)
  a <- zarr_open_array(store, mode = "r")
  expect_s3_class(a, "ZarrArray")
})

test_that("zarr_open_array mode='r' returns read-only array", {
  store <- MemoryStore$new()
  create_existing_array(store)
  a <- zarr_open_array(store, mode = "r")
  expect_true(a$get_read_only())
})

# --- Mode "r+" ---

test_that("zarr_open_array mode='r+' fails if array doesn't exist", {
  store <- MemoryStore$new()
  expect_error(zarr_open_array(store, mode = "r+"))
})

test_that("zarr_open_array mode='r+' succeeds on existing array", {
  store <- MemoryStore$new()
  create_existing_array(store)
  a <- zarr_open_array(store, mode = "r+")
  expect_s3_class(a, "ZarrArray")
  expect_false(a$get_read_only())
})

# --- Mode "a" ---

test_that("zarr_open_array mode='a' creates new array when none exists", {
  store <- MemoryStore$new()
  a <- zarr_open_array(store, mode = "a", shape = c(4), chunks = c(4), dtype = "<f8")
  expect_s3_class(a, "ZarrArray")
  expect_equal(a$get_shape(), c(4))
})

test_that("zarr_open_array mode='a' opens existing array", {
  store <- MemoryStore$new()
  create_existing_array(store)
  a <- zarr_open_array(store, mode = "a", shape = c(4))
  expect_s3_class(a, "ZarrArray")
})

# --- Mode "w" ---

test_that("zarr_open_array mode='w' creates array on fresh MemoryStore", {
  store <- MemoryStore$new()
  a <- zarr_open_array(store, mode = "w", shape = c(8), chunks = c(4), dtype = "<i4")
  expect_s3_class(a, "ZarrArray")
  expect_equal(a$get_shape(), c(8))
})

test_that("zarr_open_array mode='w' overwrites existing array", {
  store <- MemoryStore$new()
  create_existing_array(store)
  a <- zarr_open_array(store, mode = "w", shape = c(8), chunks = c(4), dtype = "<i4")
  expect_s3_class(a, "ZarrArray")
  expect_equal(a$get_shape(), c(8))
})

# --- Mode "w-" / "x" ---

test_that("zarr_open_array mode='w-' fails if array exists", {
  store <- MemoryStore$new()
  create_existing_array(store)
  expect_error(zarr_open_array(store, mode = "w-", shape = c(4)))
})

test_that("zarr_open_array mode='w-' creates new array if none exists", {
  store <- MemoryStore$new()
  a <- zarr_open_array(store, mode = "w-", shape = c(4), chunks = c(4), dtype = "<f8")
  expect_s3_class(a, "ZarrArray")
})

test_that("zarr_open_array mode='x' is alias for 'w-'", {
  store <- MemoryStore$new()
  a <- zarr_open_array(store, mode = "x", shape = c(4), chunks = c(4), dtype = "<f8")
  expect_s3_class(a, "ZarrArray")
})

# =============================================
# zarr_open_group mode tests
# =============================================

test_that("zarr_open_group mode='r' fails if group doesn't exist", {
  store <- MemoryStore$new()
  expect_error(zarr_open_group(store, mode = "r"))
})

test_that("zarr_open_group mode='r' succeeds on existing group", {
  store <- MemoryStore$new()
  create_existing_group(store)
  g <- zarr_open_group(store, mode = "r")
  expect_s3_class(g, "ZarrGroup")
})

test_that("zarr_open_group mode='w' creates group on fresh MemoryStore", {
  store <- MemoryStore$new()
  g <- zarr_open_group(store, mode = "w")
  expect_s3_class(g, "ZarrGroup")
})

test_that("zarr_open_group mode='w' overwrites existing group", {
  store <- MemoryStore$new()
  create_existing_group(store)
  g <- zarr_open_group(store, mode = "w")
  expect_s3_class(g, "ZarrGroup")
})

test_that("zarr_open_group mode='a' creates group when none exists", {
  store <- MemoryStore$new()
  g <- zarr_open_group(store, mode = "a")
  expect_s3_class(g, "ZarrGroup")
})

test_that("zarr_open_group mode='w-' fails if group exists", {
  store <- MemoryStore$new()
  create_existing_group(store)
  expect_error(zarr_open_group(store, mode = "w-"))
})

# =============================================
# zarr_open dispatch tests
# =============================================

test_that("zarr_open with shape arg opens as array", {
  store <- MemoryStore$new()
  result <- zarr_open(store, mode = "w", shape = c(4), chunks = c(4), dtype = "<f8")
  expect_s3_class(result, "ZarrArray")
})

test_that("zarr_open without shape, existing array opens as array", {
  store <- MemoryStore$new()
  create_existing_array(store)
  result <- zarr_open(store, mode = "r")
  expect_s3_class(result, "ZarrArray")
})

test_that("zarr_open without shape, existing group opens as group", {
  store <- MemoryStore$new()
  create_existing_group(store)
  result <- zarr_open(store, mode = "r")
  expect_s3_class(result, "ZarrGroup")
})

test_that("zarr_open mode='r' fails when nothing exists", {
  store <- MemoryStore$new()
  expect_error(zarr_open(store, mode = "r"))
})

# =============================================================================
# Gap 6: Additional mode matrix coverage — cross-type errors
# =============================================================================

test_that("zarr_open_group mode='r' errors with ContainsArrayError when only array exists", {
  store <- MemoryStore$new()
  create_existing_array(store)
  expect_error(zarr_open_group(store, mode = "r"), "ContainsArrayError")
})

test_that("zarr_open_group mode='r+' errors with ContainsArrayError when only array exists", {
  store <- MemoryStore$new()
  create_existing_array(store)
  expect_error(zarr_open_group(store, mode = "r+"), "ContainsArrayError")
})

test_that("zarr_open_array mode='r' errors with ContainsGroupError when only group exists", {
  store <- MemoryStore$new()
  create_existing_group(store)
  expect_error(zarr_open_array(store, mode = "r"), "ContainsGroupError")
})

test_that("zarr_open_array mode='r+' errors with ContainsGroupError when only group exists", {
  store <- MemoryStore$new()
  create_existing_group(store)
  expect_error(zarr_open_array(store, mode = "r+"), "ContainsGroupError")
})

test_that("zarr_open_group mode='a' errors with ContainsArrayError when array exists", {
  store <- MemoryStore$new()
  create_existing_array(store)
  expect_error(zarr_open_group(store, mode = "a"), "ContainsArrayError")
})

test_that("zarr_open_group mode='w-' errors with ContainsGroupError when group exists", {
  store <- MemoryStore$new()
  create_existing_group(store)
  expect_error(zarr_open_group(store, mode = "w-"), "ContainsGroupError")
})

test_that("zarr_open_group mode='w-' errors with ContainsArrayError when array exists", {
  store <- MemoryStore$new()
  create_existing_array(store)
  expect_error(zarr_open_group(store, mode = "w-"), "ContainsArrayError")
})

test_that("zarr_open_array mode='w-' errors with ContainsGroupError when group exists", {
  store <- MemoryStore$new()
  create_existing_group(store)
  expect_error(zarr_open_array(store, mode = "w-"), "ContainsGroupError")
})

test_that("zarr_open mode='w' without shape opens as group", {
  store <- MemoryStore$new()
  result <- zarr_open(store, mode = "w")
  expect_s3_class(result, "ZarrGroup")
})

test_that("zarr_open mode='a' without shape and no existing data opens as group", {
  store <- MemoryStore$new()
  result <- zarr_open(store, mode = "a")
  expect_s3_class(result, "ZarrGroup")
})

test_that("zarr_open mode='a' with existing array opens as array", {
  store <- MemoryStore$new()
  create_existing_array(store)
  result <- zarr_open(store, mode = "a")
  expect_s3_class(result, "ZarrArray")
})

test_that("zarr_open_group mode='r' returns read-only group", {
  store <- MemoryStore$new()
  create_existing_group(store)
  g <- zarr_open_group(store, mode = "r")
  expect_true(g$get_read_only())
})

test_that("zarr_open_group mode='x' is alias for 'w-'", {
  store <- MemoryStore$new()
  g <- zarr_open_group(store, mode = "x")
  expect_s3_class(g, "ZarrGroup")
})

test_that("zarr_open_group mode='x' fails if group already exists", {
  store <- MemoryStore$new()
  create_existing_group(store)
  expect_error(zarr_open_group(store, mode = "x"))
})

library(pizzarr)

test_that("zarr_create_empty", {
  z <- zarr_create_empty(100, chunks=10)
  shape <- z$get_shape()
  chunks <- z$get_chunks()
  
  expect_equal(shape, c(100))
  expect_equal(chunks, c(10))
})


test_that("zarr_create_zeros", {
  z <- zarr_create_zeros(100, chunks=10)
  shape <- z$get_shape()
  chunks <- z$get_chunks()
  
  expect_equal(shape, c(100))
  expect_equal(chunks, c(10))
})

test_that("zarr_create zero-dimensional", {
    a <- as_scalar(42)
    z <- zarr_create(shape=dim(a), dtype="<f4", fill_value=NA)
    expect_equal(z$get_shape(), NULL)
})

test_that("zarr_save_array and zarr_open_array", {
    a <- array(data=1:20, dim=c(2, 10))
    z <- zarr_create_array(data=a, shape=dim(a), dtype="<f4", fill_value=NA)
    zarr_save_array(file.path("test_data", "test_zarr_save_array.zarr"), z, overwrite = TRUE)

    z2 <- zarr_open_array(file.path("test_data", "test_zarr_save_array.zarr"))
    selection <- z2$get_item("...")

    expect_equal(z$get_shape(), c(2, 10))
    expect_equal(z2$get_shape(), c(2, 10))

    expect_equal(a, selection$data)

    unlink(file.path("test_data", "test_zarr_save_array.zarr"), recursive = TRUE)
})

test_that("BadCompressorError on invalid compressor", {
  bad_compressor <- list(not_a_codec = TRUE)
  expect_error(
    zarr_create(shape = c(10), chunks = c(10), dtype = "<f4",
                compressor = bad_compressor),
    "BadCompressorError"
  )
})

test_that("normalize_object_codec errors for object dtype without codec", {
  dtype <- Dtype$new("|O")
  expect_error(
    normalize_object_codec(dtype, NA),
    "missing object_codec for object array"
  )
})

test_that("normalize_object_codec warns for non-object dtype with codec", {
  dtype <- Dtype$new("<f8")
  codec <- VLenUtf8Codec$new()
  expect_warning(
    normalize_object_codec(dtype, codec),
    "an object_codec is only needed for object arrays"
  )
})

test_that("get_store_identifier returns path for DirectoryStore", {
  d <- tempfile("zarrs_id_")
  dir.create(d)
  ds <- DirectoryStore$new(d)
  id <- ds$get_store_identifier()
  expect_type(id, "character")
  expect_equal(normalizePath(d, mustWork = FALSE), id)
  unlink(d, recursive = TRUE)
})

test_that("get_store_identifier returns NULL for MemoryStore", {
  ms <- MemoryStore$new()
  expect_null(ms$get_store_identifier())
})

test_that("zarrs_node_exists detects V2 array at root", {
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")
  d <- tempfile("zarrs_v2_arr_")
  z <- zarr_create(store = d, shape = 4L, dtype = "<i4")
  z$set_item("...", 1:4)

  result <- zarrs_node_exists(d, "")
  expect_true(result$exists)
  expect_equal(result$node_type, "array")
  expect_equal(result$zarr_format, 2L)

  zarrs_close_store(d)
  unlink(d, recursive = TRUE)
})

test_that("zarrs_node_exists detects V2 group", {
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")
  d <- tempfile("zarrs_v2_grp_")
  g <- zarr_create_group(store = d)

  result <- zarrs_node_exists(d, "")
  expect_true(result$exists)
  expect_equal(result$node_type, "group")
  expect_equal(result$zarr_format, 2L)

  zarrs_close_store(d)
  unlink(d, recursive = TRUE)
})

test_that("zarrs_node_exists detects V3 array at subpath", {
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")
  d <- tempfile("zarrs_v3_")
  store <- DirectoryStore$new(d)
  g <- zarr_create_group(store = store, zarr_format = 3L)
  a <- g$create_dataset("arr", shape = 4L, dtype = "<i4", compressor = NA)
  a$set_item("...", 1:4)

  result <- zarrs_node_exists(d, "arr")
  expect_true(result$exists)
  expect_equal(result$node_type, "array")
  expect_equal(result$zarr_format, 3L)

  zarrs_close_store(d)
  unlink(d, recursive = TRUE)
})

test_that("zarrs_node_exists returns exists=FALSE for missing path", {
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")
  d <- tempfile("zarrs_empty_")
  dir.create(d)

  result <- zarrs_node_exists(d, "nonexistent")
  expect_false(result$exists)
  expect_equal(result$node_type, "none")

  zarrs_close_store(d)
  unlink(d, recursive = TRUE)
})

test_that("zarrs_close_store returns TRUE then FALSE", {
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")
  d <- tempfile("zarrs_close_")
  dir.create(d)

  # Opening store via zarrs_node_exists caches it.
  zarrs_node_exists(d, "")
  expect_true(zarrs_close_store(d))
  expect_false(zarrs_close_store(d))

  unlink(d, recursive = TRUE)
})

test_that("HttpStore get_store_identifier returns URL", {
  hs <- HttpStore$new("http://example.com/data")
  expect_equal(hs$get_store_identifier(), "http://example.com/data")
})

test_that("base Store get_store_identifier returns NULL", {
  s <- Store$new()
  expect_null(s$get_store_identifier())
})

test_that("zarrs_open_array_metadata reads V2 integer array", {
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")
  d <- tempfile("zarrs_meta_v2i_")
  z <- zarr_create(store = d, shape = c(10L, 5L), chunks = c(5L, 5L), dtype = "<i4")
  z$set_item("...", array(1:50, dim = c(10, 5)))

  meta <- zarrs_open_array_metadata(d, "")
  expect_equal(meta$shape, c(10L, 5L))
  expect_equal(meta$chunks, c(5L, 5L))
  expect_equal(meta$r_type, "integer")
  expect_equal(meta$zarr_format, 2L)
  expect_equal(meta$order, "F")
  expect_true(grepl("int32", meta$dtype))

  zarrs_close_store(d)
  unlink(d, recursive = TRUE)
})

test_that("zarrs_open_array_metadata reads V3 float64 array", {
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")
  d <- tempfile("zarrs_meta_v3f_")
  store <- DirectoryStore$new(d)
  g <- zarr_create_group(store = store, zarr_format = 3L)
  a <- g$create_dataset("arr", shape = 6L, dtype = "<f8", compressor = NA)
  a$set_item("...", as.double(1:6))

  meta <- zarrs_open_array_metadata(d, "arr")
  expect_equal(meta$shape, 6L)
  expect_equal(meta$r_type, "double")
  expect_equal(meta$zarr_format, 3L)
  expect_equal(meta$order, "C")
  expect_true(grepl("float64", meta$dtype))

  zarrs_close_store(d)
  unlink(d, recursive = TRUE)
})

test_that("zarrs_open_array_metadata reads 3D array", {
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")
  d <- tempfile("zarrs_meta_3d_")
  z <- zarr_create(store = d, shape = c(4L, 6L, 8L), chunks = c(2L, 3L, 4L), dtype = "<f8")

  meta <- zarrs_open_array_metadata(d, "")
  expect_equal(meta$shape, c(4L, 6L, 8L))
  expect_equal(meta$chunks, c(2L, 3L, 4L))
  expect_equal(meta$r_type, "double")
  expect_equal(meta$zarr_format, 2L)

  zarrs_close_store(d)
  unlink(d, recursive = TRUE)
})

test_that("zarrs_open_array_metadata errors on nonexistent path", {
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")
  d <- tempfile("zarrs_meta_err_")
  dir.create(d)

  expect_error(zarrs_open_array_metadata(d, "nope"), "ArrayOpen")

  zarrs_close_store(d)
  unlink(d, recursive = TRUE)
})

test_that("zarrs_runtime_info returns expected fields", {
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")
  info <- zarrs_runtime_info()
  expect_true(is.list(info))
  expect_true("codec_concurrent_target" %in% names(info))
  expect_true("nthreads" %in% names(info))
  expect_true("store_cache_entries" %in% names(info))
  expect_true("compiled_features" %in% names(info))
  expect_true(info$codec_concurrent_target > 0L)
  expect_true(info$nthreads > 0L)
})

test_that("zarrs_set_codec_concurrent_target validates input", {
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")
  expect_error(zarrs_set_codec_concurrent_target(0L), "positive")
  expect_error(zarrs_set_codec_concurrent_target(-1L), "positive")
})

test_that("zarrs_set_codec_concurrent_target round-trips", {
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")
  zarrs_set_codec_concurrent_target(2L)
  info <- zarrs_runtime_info()
  expect_equal(info$codec_concurrent_target, 2L)
})

# Reference: https://github.com/jeroen/jsonlite/blob/6a30ac/R/unbox.R

#' Convert a value to a scalar to opt-out of R default vector casting behavior.
#' This uses the `jsonlite::unbox` function to "tag" the value as a scalar.
#' @param obj The value to convert.
#' @return The value wrapped as a scalar.
#' @export
as_scalar <- function(obj) {
  return(jsonlite::unbox(obj))
}

#' Check if a value is a scalar (i.e., a one-element vector that was converted with as_scalar).
#' @param s The value to check.
#' @return TRUE if the value is a scalar, FALSE otherwise.
#' @export
is_scalar <- function(s) {
  if(class(s)[[1]] == "scalar") {
    return(TRUE)
  }
  return(FALSE)
}

#' Check if a value is an integer R vector or scalar.
#' @keywords internal
is_integer <- function(s) {
  if(is.atomic(s) && is.numeric(s) && all(s %% 1 == 0) && length(s) == 1) {
    return(TRUE)
  }
  return(FALSE)
}

#' Check if a value is both a scalar (tagged by as_scalar) and an integer.
#' @keywords internal
is_integer_scalar <- function(s) {
  if(is_scalar(s) && is_integer(s)) {
    return(TRUE)
  }
  return(FALSE)
}

#' Check that a value is a vector of one or more integers and has not been
#' explicitly tagged as a scalar.
#' @keywords internal
is_integer_vec <- function(s) {
  if(!is_scalar(s) && is.vector(s) && !is.list(s) && all(sapply(s,is_integer))) {
    if(length(s) > 1){
      return(TRUE)
    }
  }
  return(FALSE)
}

#' Check that a value is a list of one or more integers.
#' @keywords internal
is_integer_list <- function(s) {
  if(is.list(s) && is_integer_vec(unlist(s))) {
    return(TRUE)
  }
  return(FALSE)
}

#' Ensure that scalars and lists of integers are converted to
#' an R vector of integers.
#' @keywords internal
ensure_integer_vec <- function(selection) {
  if(is_integer_scalar(selection)) {
    return(as.numeric(selection))
  } else if(is_integer_vec(selection)) {
    return(selection)
  } else if(is.list(selection)) {
    return(as.numeric(unlist(selection)))
  }
  return(as.numeric(selection))
}

#' Ensure that scalars, single slices, and R integer vectors are converted
#' to a list containing either R integer vectors or slice instances as values.
#' @keywords internal
ensure_list <- function(selection) {
  if(is_slice(selection)) {
    return(list(selection))
  } else if(is_integer_scalar(selection)) {
    return(as.list(as.numeric(selection)))
  } else if(is.list(selection)) {
    return(selection)
  }
  return(as.list(selection))
}

#' Check that a value is boolean with length 1
#' @keywords internal
is_bool <- function(s){
  if(is.logical(s) && length(s) == 1)
    return(TRUE)
  return(FALSE)
}
#' Check that a value is a vector of one or more boolean
#' @keywords internal
is_bool_vec <- function(s) {
  if(is.vector(s) && !is.list(s) && is.logical(s) && length(s) > 1)
    return(TRUE)
  return(FALSE)
}
#' Check that a value is a list of one or more boolean
#' @keywords internal
is_bool_list <- function(s) {
  if(is.list(s) && is_bool_vec(unlist(s)))
    return(TRUE)
  return(FALSE)
}

library(pizzarr)

MockArray <- R6::R6Class("MockArray",
                         public = list(
                           shape = NULL,
                           chunks = NULL,
                           initialize = function(shape, chunks) {
                             self$shape <- shape
                             self$chunks <- chunks
                           },
                           get_shape = function() {
                             return(self$shape)
                           },
                           get_chunks = function() {
                             return(self$chunks)
                           }
                         )
)

test_that("basic indexer for array that spans multiple chunks", {
  z <- MockArray$new(shape = c(10, 10), chunks = c(5, 5))
  
  expect_equal(z$get_shape(), c(10, 10))
  expect_equal(z$get_chunks(), c(5, 5))
  
  # OrthogonalIndexer supports SliceDimIndexer
  bi <- OrthogonalIndexer$new(list(zb_slice(0, 11), zb_slice(0, 11)), z)
  expect_equal(as.numeric(bi$shape), c(10, 10))
  expect_equal(length(bi$dim_indexers), 2)
  expect_equal("SliceDimIndexer" %in% class(bi$dim_indexers[[1]]), TRUE)
  
  # combination of IntArrayDimIndexer and SliceDimIndexer
  bi <- OrthogonalIndexer$new(list(0:9, zb_slice(0, 11)), z)
  expect_equal("IntArrayDimIndexer" %in% class(bi$dim_indexers[[1]]), TRUE)
  expect_equal("SliceDimIndexer" %in% class(bi$dim_indexers[[2]]), TRUE)
  
  # unordered integer vector is a IntArrayDimIndexer
  bi <- OrthogonalIndexer$new(list(0:9, c(2,3,5,4,1)), z)
  expect_equal("IntArrayDimIndexer" %in% class(bi$dim_indexers[[2]]), TRUE)
  
  # empty dimension is a slice
  bi <- OrthogonalIndexer$new(list(1), z)
  expect_equal("SliceDimIndexer" %in% class(bi$dim_indexers[[2]]), TRUE)
  
  # TODO: add tests for adding BoolArrayDimIndexer
})

test_that("basic indexer for array that spans multiple chunks where shape is not a multiple", {
  z <- MockArray$new(shape = c(10, 10), chunks = c(3, 3))
  
  expect_equal(z$get_shape(), c(10, 10))
  expect_equal(z$get_chunks(), c(3, 3))
  
  # OrthogonalIndexer supports SliceDimIndexer
  bi <- OrthogonalIndexer$new(list(zb_slice(0, 11), zb_slice(0, 11)), z)
  expect_equal(as.numeric(bi$shape), c(10, 10))
  expect_equal(length(bi$dim_indexers), 2)
  expect_equal("SliceDimIndexer" %in% class(bi$dim_indexers[[1]]), TRUE)
  
  # combination of IntArrayDimIndexer and SliceDimIndexer
  bi <- OrthogonalIndexer$new(list(0:9, zb_slice(0, 11)), z)
  expect_equal("IntArrayDimIndexer" %in% class(bi$dim_indexers[[1]]), TRUE)
  expect_equal("SliceDimIndexer" %in% class(bi$dim_indexers[[2]]), TRUE)
  
  # unordered integer vector is a IntArrayDimIndexer
  bi <- OrthogonalIndexer$new(list(0:9, c(2,3,5,4,1)), z)
  expect_equal("IntArrayDimIndexer" %in% class(bi$dim_indexers[[2]]), TRUE)
  
  # empty dimension is a slice
  bi <- OrthogonalIndexer$new(list(1), z)
  expect_equal("SliceDimIndexer" %in% class(bi$dim_indexers[[2]]), TRUE)
  
  # TODO: add tests for adding BoolArrayDimIndexer
})

test_that("int array dimension indexer", {
  
  # ordered int array index
  # iad <- IntArrayDimIndexer$new(1:10, 10, 3)
  iad <- IntArrayDimIndexer$new(0:9, 10, 3)
  expect_equal(iad$dim_sel, 0:9)
  expect_equal(iad$dim_chunk_ixs, c(1,2,3,4))
  expect_equal(iad$dim_len, 10)
  expect_equal(iad$dim_chunk_len, 3)
  expect_equal(iad$num_chunks, 4)
  expect_equal(iad$chunk_nitems, c(3,3,3,1))
  expect_equal(iad$order, 1)
  
  # unordered int array index
  iad <- IntArrayDimIndexer$new(c(2,3,5,1,2), 6, 3)
  expect_equal(iad$dim_sel, c(2,1,2,3,5))
  expect_equal(iad$dim_chunk_ixs, c(1,2))
  expect_equal(iad$dim_len, 6)
  expect_equal(iad$dim_chunk_len, 3)
  expect_equal(iad$num_chunks, 2)
  expect_equal(iad$chunk_nitems, c(3,2))
  expect_equal(iad$order, 3)
  
  # error for wrong dimension length
  expect_error(IntArrayDimIndexer$new(1:10, 6, 3))
  
  # missing chunk size
  expect_error(IntArrayDimIndexer$new(1:10, 10))
  
})

test_that("bool array dimension indexer", {
  
  # boolean checks
  expect_equal(is_bool(TRUE), TRUE)
  expect_equal(is_bool(1), FALSE)
  expect_equal(is_bool(1.2), FALSE)
  expect_equal(is_bool(c(TRUE, FALSE, TRUE)), FALSE)
  expect_equal(is_bool_vec(c(TRUE, FALSE, TRUE)), TRUE)
  expect_equal(is_bool_vec(c(TRUE, FALSE, 1)), FALSE)
  expect_equal(is_bool_vec(c(TRUE, 1.2, 1)), FALSE)
  expect_equal(is_bool_list(list(TRUE, FALSE, FALSE)), TRUE)
  expect_equal(is_bool_list(list(TRUE, 1.2, 1)), FALSE)
  expect_equal(is_bool_list(c(TRUE, FALSE, FALSE)), FALSE)
  
  # ordered int array index
  iad <- BoolArrayDimIndexer$new(c(TRUE, TRUE, FALSE, TRUE, FALSE, TRUE, FALSE, TRUE), 8, 5)
  expect_equal(iad$dim_sel, c(TRUE, TRUE, FALSE, TRUE, FALSE, TRUE, FALSE, TRUE))
  expect_equal(iad$dim_chunk_ixs, c(1,2))
  expect_equal(iad$dim_len, 8)
  expect_equal(iad$dim_chunk_len, 5)
  expect_equal(iad$num_chunks, 2)
  expect_equal(iad$chunk_nitems, c(3,2))
  
  # error for wrong dimension length
  expect_error(BoolArrayDimIndexer$new(c(TRUE, TRUE, FALSE, TRUE, FALSE, TRUE, FALSE, TRUE), 3, 5))
  
  # missing chunk size
  expect_error(BoolArrayDimIndexer$new(c(TRUE, TRUE, FALSE, TRUE, FALSE), 5))
  
})

library(pizzarr)

test_that("int", {

  # is_integer
  expect_equal(is_integer(2), TRUE)
  expect_equal(is_integer(2.2), FALSE)
  expect_equal(is_integer(c(1,2,3)), FALSE)
  expect_equal(is_integer(list(1,2,3)), FALSE)
  
  # is_integer_vec
  expect_equal(is_integer_vec(c(2,3)), TRUE)
  expect_equal(is_integer_vec(c(2,3.3)), FALSE)
  expect_equal(is_integer_vec(2), FALSE)
  expect_equal(is_integer_vec(list(2,3)), FALSE)
  
  # is_integer_list
  expect_equal(is_integer_list(list(2,3)), TRUE)
  expect_equal(is_integer_list(c(2,3)), FALSE)
  expect_equal(is_integer_list(list(2,3.3)), FALSE)
})

#' @keywords internal
.onLoad <- function(libname = NULL, pkgname = NULL) {
  init_options()
  .pizzarr_env$zarrs_available <- is_zarrs_available()
  if (.pizzarr_env$zarrs_available) {
    apply_zarrs_config()
  }
}

#' @keywords internal
.onAttach <- function(libname, pkgname) {
  if (!.pizzarr_env$zarrs_available &&
      getOption("pizzarr.suggest_runiverse", TRUE)) {
    packageStartupMessage(
      "pizzarr: running without zarrs backend (pure R).\n",
      "For parallel I/O and cloud storage, install from r-universe:\n",
      '  install.packages("pizzarr", ',
      'repos = "https://zarr-developers.r-universe.dev")'
    )
  }
}

# Tests for .onAttach startup message (R/onload.R)

test_that(".onAttach shows message when zarrs unavailable", {
  orig <- .pizzarr_env$zarrs_available
  on.exit(.pizzarr_env$zarrs_available <- orig)
  .pizzarr_env$zarrs_available <- FALSE

  saved <- options(pizzarr.suggest_runiverse = TRUE)
  on.exit(do.call(options, saved), add = TRUE)

  expect_message(.onAttach(NULL, "pizzarr"), "without zarrs backend")
})

test_that(".onAttach suppressed when suggest_runiverse is FALSE", {
  orig <- .pizzarr_env$zarrs_available
  on.exit(.pizzarr_env$zarrs_available <- orig)
  .pizzarr_env$zarrs_available <- FALSE

  saved <- options(pizzarr.suggest_runiverse = FALSE)
  on.exit(do.call(options, saved), add = TRUE)

  expect_silent(.onAttach(NULL, "pizzarr"))
})

test_that(".onAttach silent when zarrs is available", {
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")
  expect_silent(.onAttach(NULL, "pizzarr"))
})

# Adapted from https://github.com/IRkernel/IRkernel/blob/master/R/options.r

#' pizzarr_option_defaults
#' @description
#' * pizzarr.http_store_cache_time_seconds how long to cache web requests
#' * pizzarr.nthreads number of threads for parallel codec operations
#'   (NULL = all CPUs). Set-once: takes effect only before the first zarrs
#'   operation. Use env var PIZZARR_NTHREADS for reliable session-level control.
#' * pizzarr.concurrent_target codec concurrency level — how many codec
#'   operations zarrs runs in parallel within a single read/write call
#'   (NULL = zarrs default, typically CPU count). Can be changed at any time.
#' * pizzarr.http_batch_range_requests whether HTTP stores use multipart
#'   range requests (TRUE by default). Set to FALSE for servers with
#'   incomplete multipart range support. Takes effect on next store open.
#' @export
pizzarr_option_defaults <- list(
    pizzarr.http_store_cache_time_seconds = 3600,
    pizzarr.nthreads = NULL,
    pizzarr.concurrent_target = NULL,
    pizzarr.http_batch_range_requests = TRUE
)

#' @keywords internal
from_env <- list(
    PIZZARR_HTTP_STORE_CACHE_TIME_SECONDS = as.integer,
    PIZZARR_NTHREADS = as.integer,
    PIZZARR_CONCURRENT_TARGET = as.integer,
    PIZZARR_HTTP_BATCH_RANGE_REQUESTS = function(x) as.logical(toupper(x))
)

# converts e.g. jupyter.log_level to JUPYTER_LOG_LEVEL
#' @keywords internal
opt_to_env <- function(nms) {
    gsub('.', '_', toupper(nms), fixed = TRUE)
}

# called in .onLoad
#' @keywords internal
init_options <- function() {
    for (opt_name in names(pizzarr_option_defaults)) {
        # skip option if it is already set, e.g. in the Rprofile
        if (is.null(getOption(opt_name))) {
            # prepare `options` call from the default
            call_arg <- pizzarr_option_defaults[opt_name]  # single [] preserve names

            # if an env var is set, get value from it.
            env_name <- opt_to_env(opt_name)
            convert <- from_env[[env_name]]
            env_val <- Sys.getenv(env_name, unset = NA)
            if (!is.null(convert) && !is.na(env_val)) {
                call_arg[[opt_name]] <- convert(env_val)
            }

            do.call(options, call_arg)
        }
    }
}

library(pizzarr)

# --- opt_to_env ---

test_that("opt_to_env converts option names to env var names", {
  expect_equal(opt_to_env("pizzarr.http_store_cache_time_seconds"),
               "PIZZARR_HTTP_STORE_CACHE_TIME_SECONDS")
})

# --- pizzarr_option_defaults ---

test_that("pizzarr_option_defaults has expected keys and values", {
  expect_true("pizzarr.http_store_cache_time_seconds" %in% names(pizzarr_option_defaults))
  expect_true("pizzarr.nthreads" %in% names(pizzarr_option_defaults))
  expect_true("pizzarr.concurrent_target" %in% names(pizzarr_option_defaults))
  expect_true("pizzarr.http_batch_range_requests" %in% names(pizzarr_option_defaults))

  expect_equal(pizzarr_option_defaults$pizzarr.http_store_cache_time_seconds, 3600)
  expect_null(pizzarr_option_defaults$pizzarr.nthreads)
  expect_null(pizzarr_option_defaults$pizzarr.concurrent_target)
  expect_true(pizzarr_option_defaults$pizzarr.http_batch_range_requests)
})

# --- init_options sets defaults ---

test_that("init_options sets default option values", {
  saved <- options(
    pizzarr.http_store_cache_time_seconds = NULL
  )
  on.exit(do.call(options, saved), add = TRUE)

  init_options()

  expect_equal(getOption("pizzarr.http_store_cache_time_seconds"), 3600)
})

test_that("init_options does not overwrite pre-existing options", {
  saved <- options(
    pizzarr.http_store_cache_time_seconds = 999
  )
  on.exit(do.call(options, saved), add = TRUE)

  init_options()

  expect_equal(getOption("pizzarr.http_store_cache_time_seconds"), 999)
})

# --- init_options reads environment variables ---

test_that("init_options picks up env var for http_store_cache_time_seconds", {
  saved <- options(pizzarr.http_store_cache_time_seconds = NULL)
  on.exit(do.call(options, saved), add = TRUE)
  withr::local_envvar(PIZZARR_HTTP_STORE_CACHE_TIME_SECONDS = "120")

  init_options()

  expect_equal(getOption("pizzarr.http_store_cache_time_seconds"), 120L)
})

test_that("init_options env var does not override pre-existing option", {
  saved <- options(pizzarr.http_store_cache_time_seconds = 999)
  on.exit(do.call(options, saved), add = TRUE)
  withr::local_envvar(PIZZARR_HTTP_STORE_CACHE_TIME_SECONDS = "120")

  init_options()

  # pre-existing option wins over env var
  expect_equal(getOption("pizzarr.http_store_cache_time_seconds"), 999)
})

# --- init_options reads env vars for new options ---

test_that("init_options picks up PIZZARR_NTHREADS env var", {
  saved <- options(pizzarr.nthreads = NULL)
  on.exit(do.call(options, saved), add = TRUE)
  withr::local_envvar(PIZZARR_NTHREADS = "4")

  init_options()

  expect_equal(getOption("pizzarr.nthreads"), 4L)
})

test_that("init_options picks up PIZZARR_CONCURRENT_TARGET env var", {
  saved <- options(pizzarr.concurrent_target = NULL)
  on.exit(do.call(options, saved), add = TRUE)
  withr::local_envvar(PIZZARR_CONCURRENT_TARGET = "2")

  init_options()

  expect_equal(getOption("pizzarr.concurrent_target"), 2L)
})

test_that("init_options picks up PIZZARR_HTTP_BATCH_RANGE_REQUESTS env var", {
  saved <- options(pizzarr.http_batch_range_requests = NULL)
  on.exit(do.call(options, saved), add = TRUE)
  withr::local_envvar(PIZZARR_HTTP_BATCH_RANGE_REQUESTS = "FALSE")

  init_options()

  expect_false(getOption("pizzarr.http_batch_range_requests"))
})

test_that("init_options does not overwrite pre-existing nthreads option", {
  saved <- options(pizzarr.nthreads = 8L)
  on.exit(do.call(options, saved), add = TRUE)
  withr::local_envvar(PIZZARR_NTHREADS = "2")

  init_options()

  expect_equal(getOption("pizzarr.nthreads"), 8L)
})

# --- from_env converters ---

test_that("from_env converters produce expected types", {
  expect_equal(from_env$PIZZARR_HTTP_STORE_CACHE_TIME_SECONDS("42"), 42L)
  expect_equal(from_env$PIZZARR_NTHREADS("8"), 8L)
  expect_equal(from_env$PIZZARR_CONCURRENT_TARGET("4"), 4L)
  expect_true(from_env$PIZZARR_HTTP_BATCH_RANGE_REQUESTS("TRUE"))
  expect_false(from_env$PIZZARR_HTTP_BATCH_RANGE_REQUESTS("false"))
})

# zarrs backend availability and upgrade helpers
#
# The zarrs Rust backend is compiled into the shared library on the
# r-universe tier. On the CRAN tier (pure R, no src/), the shared
# library does not exist and these functions report that zarrs is
# unavailable.

# Package-level environment, created at source time.
# .onLoad populates $zarrs_available.
.pizzarr_env <- new.env(parent = emptyenv())

#' Check whether the zarrs Rust backend is available
#'
#' @return Logical scalar. \code{TRUE} when the package was installed with the
#'   compiled zarrs backend (r-universe tier), \code{FALSE} on the CRAN tier
#'   (pure R).
#' @keywords internal
is_zarrs_available <- function() {
  tryCatch(
    {
      feats <- zarrs_compiled_features()
      is.character(feats) && length(feats) > 0
    },
    error = function(e) FALSE
  )
}

#' List compiled zarrs features
#'
#' Returns the feature flags compiled into the zarrs backend, or
#' \code{character(0)} with a message when the backend is absent.
#'
#' @return Character vector of feature names (e.g. \code{"filesystem"},
#'   \code{"gzip"}).
#' @export
pizzarr_compiled_features <- function() {
  if (.pizzarr_env$zarrs_available) {
    zarrs_compiled_features()
  } else {
    message(
      "zarrs backend not available (pure R install).\n",
      "See ?pizzarr_upgrade for the r-universe install."
    )
    character(0)
  }
}

#' Upgrade to the zarrs backend
#'
#' Prints the command to install pizzarr from r-universe with the
#' compiled zarrs backend, or reports that zarrs is already available.
#'
#' @export
pizzarr_upgrade <- function() {
  if (.pizzarr_env$zarrs_available) {
    message("zarrs backend is already available.")
  } else {
    message(
      "Install pizzarr with the zarrs backend from r-universe:\n\n",
      '  install.packages("pizzarr", ',
      'repos = "https://zarr-developers.r-universe.dev")\n'
    )
  }
  invisible(NULL)
}

# Tests for zarrs bridge helpers (R/zarrs-bridge.R)

test_that("is_zarrs_available returns logical", {
  expect_true(is.logical(is_zarrs_available()))
})

test_that("pizzarr_compiled_features returns character when available", {
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")
  feats <- pizzarr_compiled_features()
  expect_true(is.character(feats))
  expect_true(length(feats) > 0)
})

test_that("pizzarr_compiled_features messages when unavailable", {
  orig <- .pizzarr_env$zarrs_available
  on.exit(.pizzarr_env$zarrs_available <- orig)
  .pizzarr_env$zarrs_available <- FALSE

  expect_message(feats <- pizzarr_compiled_features(), "not available")
  expect_equal(feats, character(0))
})

test_that("pizzarr_upgrade messages already available when zarrs present", {
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")
  expect_message(pizzarr_upgrade(), "already available")
})

test_that("pizzarr_upgrade shows install instructions when zarrs absent", {
  orig <- .pizzarr_env$zarrs_available
  on.exit(.pizzarr_env$zarrs_available <- orig)
  .pizzarr_env$zarrs_available <- FALSE

  expect_message(pizzarr_upgrade(), "r-universe")
})

test_that("pizzarr_upgrade returns NULL invisibly", {
  result <- pizzarr_upgrade()
  expect_null(result)
})

#' Abstract compressor for Zarr
#' @title Codec Class
#' @docType class
#' @description
#' Abstract class representing a compressor.
#'
#' @format [R6::R6Class]
#' @family Codec classes
#' @rdname Codec
#' @export
Codec <- R6::R6Class("Codec",
   public = list(
     #' @description
     #' Compress data.
     #' @param buf (`raw()`)\cr
    #'   The un-compressed data.
     #' @param zarr_arr ([ZarrArray])\cr
    #'   The ZarrArray instance.
     #' @return Compressed data.
     encode = function(buf, zarr_arr) {
       return(buf)
     },
     #' @description
     #' Decompress data.
     #' @param buf (`raw()`)\cr
    #'   The compressed data.
     #' @param zarr_arr ([ZarrArray])\cr
    #'   The ZarrArray instance.
     #' @return Un-compressed data.
     decode = function(buf, zarr_arr) {
       return(buf)
     },
    #' @description
    #' Get codec configuration as a list.
    #' @return A named list.
     get_config = function() {
       return(jsonlite::unbox(NA))
     }
   )
)

#' @keywords internal
warn_if_unk_args <- function(unk_args, compressor_name) {
  if(length(unk_args) > 0) {
    unk_args_str <- paste(names(unk_args), collapse = ", ")
    warning(paste(compressor_name, "received unrecognized compressor config parameters:", unk_args_str))
  }
}

#' ZSTD compressor for Zarr
#' @title ZstdCodec Class
#' @docType class
#' @description
#' Class representing a ZSTD compressor

#' @format [R6::R6Class] inheriting from [Codec].
#' @family Codec classes
#' @rdname ZstdCodec
#' @importFrom qs2 zstd_compress_raw zstd_decompress_raw
#' @export
ZstdCodec <- R6::R6Class("ZstdCodec",
  inherit = Codec,
  public = list(
    #' @field level The compression level.
    #' @keywords internal
    level = NULL,
    #' @description
    #' Create a new ZSTD compressor.
    #' @param level The compression level, between 1 and 22.
    #' @param ... Not used.
    #' @return A new `ZstdCodec` object.
    initialize = function(level = 1, ...) {
      self$level <- level
      warn_if_unk_args(list(...), "ZstdCodec")
    },
    #' @description
    #' Compress data.
    #' @param buf (`raw()`)\cr
    #'   The un-compressed data.
    #' @param zarr_arr ([ZarrArray])\cr
    #'   The ZarrArray instance.
    #' @return Compressed data.
    encode = function(buf, zarr_arr) {
      # Reference: https://github.com/traversc/qs/blob/84e30f4/R/RcppExports.R#L16
      result <- zstd_compress_raw(buf, self$level)
      return(result)
    },
    #' @description
    #' Decompress data.
    #' @param buf (`raw()`)\cr
    #'   The compressed data.
    #' @param zarr_arr ([ZarrArray])\cr
    #'   The ZarrArray instance.
    #' @return Un-compressed data.
    decode = function(buf, zarr_arr) {
      result <- zstd_decompress_raw(buf)
      return(result)
    },
    #' @description
    #' Get codec configuration as a list.
    #' @return A named list.
    get_config = function() {
      meta <- list(
        id = jsonlite::unbox("zstd"),
        level = jsonlite::unbox(self$level)
      )
      return(meta)
    }
  )
)

#' LZ4 compressor for Zarr
#' @title Lz4Codec Class
#' @docType class
#' @description
#' Class representing a LZ4 compressor
#'
#' @section Deprecation Notice:
#' This codec is deprecated because the underlying `qs` package was archived on CRAN.
#' For LZ4 compression, use `BloscCodec$new(cname = "lz4")` instead.
#' This class will be removed in a future version.
#'
#' @format [R6::R6Class] inheriting from [Codec].
#' @family Codec classes
#' @rdname Lz4Codec
#' @export
Lz4Codec <- R6::R6Class("Lz4Codec",
   inherit = Codec,
   public = list(
     #' @field acceleration The compression level.
     #' @keywords internal
     acceleration = NULL,
     #' @description
     #' Create a new LZ4 compressor.
     #' @param acceleration The compression level.
     #' @param ... Not used.
     #' @return A new `Lz4Codec` object.
     initialize = function(acceleration = 1, ...) {
       .Deprecated(
         new = "BloscCodec$new(cname = 'lz4')",
         old = "Lz4Codec$new()",
         msg = paste(
           "Lz4Codec is deprecated because the 'qs' package was archived on CRAN.",
           "Use BloscCodec$new(cname = 'lz4') for LZ4 compression instead.",
           "Lz4Codec will be removed in a future version of pizzarr."
         )
       )
       self$acceleration <- acceleration
       warn_if_unk_args(list(...), "Lz4Codec")
     },
     #' @description
     #' Compress data.
     #' @param buf (`raw()`)\cr
    #'   The un-compressed data.
     #' @param zarr_arr ([ZarrArray])\cr
    #'   The ZarrArray instance.
     #' @return Compressed data.
     encode = function(buf, zarr_arr) {
       # Reference: https://github.com/traversc/qs/blob/84e30f4/R/RcppExports.R#L24
       if (!requireNamespace("qs", quietly = TRUE)) {
         stop(
           "Lz4Codec requires the archived 'qs' package. ",
           "Use BloscCodec$new(cname = 'lz4') instead for LZ4 compression."
         )
       }
       body <- qs::lz4_compress_raw(buf, self$acceleration)

       # The compressed output includes a 4-byte header storing the original size
       # of the decompressed data as a little-endian 32-bit integer.
       # Reference: https://numcodecs.readthedocs.io/en/stable/lz4.html#numcodecs.lz4.compress
       orig_size <- length(buf)
       header <- writeBin(orig_size, con = raw(), size = 4, endian = "little")

       result <- c(header, body)

       return(result)
     },
     #' @description
     #' Decompress data.
     #' @param buf (`raw()`)\cr
    #'   The compressed data.
     #' @param zarr_arr ([ZarrArray])\cr
    #'   The ZarrArray instance.
     #' @return Un-compressed data.
     decode = function(buf, zarr_arr) {
       if (!requireNamespace("qs", quietly = TRUE)) {
         stop(
           "Lz4Codec requires the archived 'qs' package. ",
           "Use BloscCodec$new(cname = 'lz4') instead for LZ4 compression."
         )
       }
       body <- buf[5:length(buf)]

       result <- qs::lz4_decompress_raw(body)
       return(result)
     },
     #' @description
     #' Get codec configuration as a list.
    #' @return A named list.
     get_config = function() {
       meta <- list(
         id = jsonlite::unbox("lz4"),
         acceleration = jsonlite::unbox(self$acceleration)
       )
       return(meta)
     }
   )
)

#' Zlib compressor for Zarr
#' @title ZlibCodec Class
#' @docType class
#' @description
#' Class representing a zlib compressor
#'
#' @format [R6::R6Class] inheriting from [Codec].
#' @family Codec classes
#' @rdname ZlibCodec
#' @export
ZlibCodec <- R6::R6Class("ZlibCodec",
   inherit = Codec,
   public = list(
      #' @field level The compression level.
    level = NULL,
     #' @description
     #' Create a new Zlib compressor.
     #' @param level The compression level, between 1 and 22.
     #' @param ... Not used.
     #' @return A new `ZlibCodec` object.
     initialize = function(level = 6, ...) {
      self$level <- level
      # No config options for zlib.
      warn_if_unk_args(list(...), "ZlibCodec")
     },
    #' @description
    #' Compress data.
    #' @param buf (`raw()`)\cr
    #'   The un-compressed data.
    #' @param zarr_arr ([ZarrArray])\cr
    #'   The ZarrArray instance.
    #' @return Compressed data.
     encode = function(buf, zarr_arr) {
      if(self$level != 6) {
        stop("Only system default compression level (normally 6) is enabled for writing.")
      }
       # References:
       # - https://github.com/grimbough/Rarr/blob/684541a86b0313a6f354282b60a08dd0ea0a747d/R/write_data.R#L385
       # - https://stat.ethz.ch/R-manual/R-devel/library/base/html/memCompress.html
      result <- memCompress(buf, type = "gzip")
      return(result)
     },
    #' @description
    #' Decompress data.
    #' @param buf (`raw()`)\cr
    #'   The compressed data.
    #' @param zarr_arr ([ZarrArray])\cr
    #'   The ZarrArray instance.
    #' @return Un-compressed data.
     decode = function(buf, zarr_arr) {
      # References:
       # - https://github.com/grimbough/Rarr/blob/684541a86b0313a6f354282b60a08dd0ea0a747d/R/read_data.R#L356C5-L370C6
       # - https://stat.ethz.ch/R-manual/R-devel/library/base/html/memCompress.html
      result <- memDecompress(buf, type = "gzip", asChar = FALSE)
      return(result)
     },
    #' @description
    #' Get codec configuration as a list.
    #' @return A named list.
     get_config = function() {
       meta <- list(
         id = jsonlite::unbox("zlib"),
         level = jsonlite::unbox(as.integer(self$level))
       )
       return(meta)
     }
   )
)

#' Gzip compressor for Zarr
#' @title GzipCodec Class
#' @docType class
#' @description
#' Class representing a gzip compressor.
#'
#' Gzip encoding uses temporary files because R's `memCompress()` produces
#' zlib framing rather than gzip framing. This makes `GzipCodec` slower than
#' [ZstdCodec] for writes. Prefer `ZstdCodec` when performance matters.
#'
#' @format [R6::R6Class] inheriting from [Codec].
#' @family Codec classes
#' @rdname GzipCodec
#' @export
GzipCodec <- R6::R6Class("GzipCodec",
   inherit = Codec,
   public = list(
      #' @field level The compression level.
      level = NULL,
     #' @description
     #' Create a new Gzip compressor.
     #' @param level The compression level, between 1 and 22.
     #' @param ... Not used.
     #' @return A new `GzipCodec` object.
     initialize = function(level = 6, ...) {
      # No config options for gzip.
      self$level <- level
      warn_if_unk_args(list(...), "GzipCodec")
     },
     #' @description
     #' Compress data.
     #' @param buf (`raw()`)\cr
    #'   The un-compressed data.
     #' @param zarr_arr ([ZarrArray])\cr
    #'   The ZarrArray instance.
     #' @return Compressed data.
     encode = function(buf, zarr_arr) {
      if(self$level != 6) {
        stop("Only system default compression level (normally 6) is enabled for writing.")
      }
      # memCompress(type="gzip") produces zlib (RFC 1950), not gzip (RFC 1952).
      # Use gzfile() to produce real gzip output for interoperability with
      # zarr-python and other implementations that expect gzip framing.
      tmp <- tempfile()
      on.exit(unlink(tmp), add = TRUE)
      con <- gzfile(tmp, open = "wb")
      writeBin(buf, con)
      close(con)
      result <- readBin(tmp, "raw", file.info(tmp)$size)
      return(result)
     },
     #' @description
     #' Decompress data.
     #' @param buf (`raw()`)\cr
    #'   The compressed data.
     #' @param zarr_arr ([ZarrArray])\cr
    #'   The ZarrArray instance.
     #' @return Un-compressed data.
     decode = function(buf, zarr_arr) {
      # memDecompress handles both gzip (RFC 1952) and zlib (RFC 1950),
      # so this remains backward-compatible with stores written before
      # the gzip fix.
      result <- memDecompress(buf, type = "gzip", asChar = FALSE)
      return(result)
     },
     #' @description
     #' Get codec configuration as a list.
    #' @return A named list.
     get_config = function() {
       meta <- list(
          id = jsonlite::unbox("gzip"),
          level = jsonlite::unbox(as.integer(self$level))
       )
       return(meta)
     }
   )
)

#' Bz2 compressor for Zarr
#' @title Bz2Codec Class
#' @docType class
#' @description
#' Class representing a bz2 compressor
#'
#' @format [R6::R6Class] inheriting from [Codec].
#' @family Codec classes
#' @rdname Bz2Codec
#' @export
Bz2Codec <- R6::R6Class("Bz2Codec",
   inherit = Codec,
   public = list(
    #' @field level The compression level.
      level = NULL,
     #' @description
     #' Create a new Bz2 compressor.
     #' @param level The compression level, between 1 and 22.
     #' @param ... Not used.
     #' @return A new `Bz2Codec` object.
     initialize = function(level = 6, ...) {
      # No config options for bz2.
      self$level <- level
      warn_if_unk_args(list(...), "Bz2Codec")
     },
     #' @description
     #' Compress data.
     #' @param buf (`raw()`)\cr
    #'   The un-compressed data.
     #' @param zarr_arr ([ZarrArray])\cr
    #'   The ZarrArray instance.
     #' @return Compressed data.
     encode = function(buf, zarr_arr) {
       # References:
       # - https://github.com/grimbough/Rarr/blob/684541a86b0313a6f354282b60a08dd0ea0a747d/R/read_data.R#L356C5-L370C6
       # - https://stat.ethz.ch/R-manual/R-devel/library/base/html/memCompress.html
      if(self$level != 6) {
        stop("Only bzip2 compression level 6 is enabled for writing.")
      }
      result <- memCompress(buf, type = "bzip2")
      return(result)
     },
     #' @description
     #' Decompress data.
     #' @param buf (`raw()`)\cr
    #'   The compressed data.
     #' @param zarr_arr ([ZarrArray])\cr
    #'   The ZarrArray instance.
     #' @return Un-compressed data.
     decode = function(buf, zarr_arr) {
      result <- memDecompress(buf, type = "bzip2", asChar = FALSE)
      return(result)
     },
     #' @description
     #' Get codec configuration as a list.
    #' @return A named list.
     get_config = function() {
       meta <- list(
         id = jsonlite::unbox("bz2"),
         level = jsonlite::unbox(as.integer(self$level))
       )
       return(meta)
     }
   )
)

#' Lzma compressor for Zarr
#' @title LzmaCodec Class
#' @docType class
#' @description
#' Class representing a lzma compressor
#'
#' @format [R6::R6Class] inheriting from [Codec].
#' @family Codec classes
#' @rdname LzmaCodec
#' @export
LzmaCodec <- R6::R6Class("LzmaCodec",
   inherit = Codec,
   public = list(
      #' @field level The compression level.
      level = NULL,
      #' @field format The compression format.
      format = NULL,
     #' @description
     #' Create a new lzma compressor.
     #' @param level The compression level, between 1 and 22.
     #' @param format (`integer(1)`)\cr
     #'   Only 1 is supported.
     #' @param ... Not used.
     #' @return A new `LzmaCodec` object.
     initialize = function(level = 9, format = 1, ...) {
      # No config options for lzma.
      self$level <- level
      self$format <- format
      if(format != 1) {
        stop("Only format 1 is supported for lzma compression")
      }
      warn_if_unk_args(list(...), "LzmaCodec")
     },
     #' @description
     #' Compress data.
     #' @param buf (`raw()`)\cr
    #'   The un-compressed data.
     #' @param zarr_arr ([ZarrArray])\cr
    #'   The ZarrArray instance.
     #' @return Compressed data.
     encode = function(buf, zarr_arr) {
       # References:
       # - https://github.com/grimbough/Rarr/blob/684541a86b0313a6f354282b60a08dd0ea0a747d/R/read_data.R#L356C5-L370C6
      if(self$level != 9) {
        stop("Only lzma compression level 9 is enabled for writing.")
      }
       # - https://stat.ethz.ch/R-manual/R-devel/library/base/html/memCompress.html
      result <- memCompress(buf, type = "xz")
      return(result)
     },
     #' @description
     #' Decompress data.
     #' @param buf (`raw()`)\cr
    #'   The compressed data.
     #' @param zarr_arr ([ZarrArray])\cr
    #'   The ZarrArray instance.
     #' @return Un-compressed data.
     decode = function(buf, zarr_arr) {
      result <- memDecompress(buf, type = "xz", asChar = FALSE)
      return(result)
     },
     #' @description
     #' Get codec configuration as a list.
    #' @return A named list.
     get_config = function() {
       meta <- list(
         id = jsonlite::unbox("lzma"),
          level = jsonlite::unbox(as.integer(self$level)),
          format = jsonlite::unbox(as.integer(self$format))
       )
       return(meta)
     }
   )
)

#' Blosc compressor for Zarr
#' @title BloscCodec Class
#' @docType class
#' @description
#' Class representing a Blosc compressor
#'
#' @format [R6::R6Class] inheriting from [Codec].
#' @family Codec classes
#' @rdname BloscCodec
#' @export
BloscCodec <- R6::R6Class("BloscCodec",
  inherit = Codec,
  public = list(
    #' @field cname (`character(1)`)\cr
    #'   The compression algorithm to use.
    cname = NULL,
    #' @field clevel (`integer(1)`)\cr
    #'   The compression level.
    clevel = NULL,
    #' @field shuffle (`logical(1)` | `integer(1)`)\cr
    #'   The shuffle filter to use.
    shuffle = NULL,
    #' @field blocksize (`integer(1)` | `NA`)\cr
    #'   The block size.
    blocksize = NULL,
    #' @description
    #' Create a new Blosc compressor.
    #' @param cname (`character(1)`)\cr
    #'   The compression algorithm to use.
    #' @param clevel (`integer(1)`)\cr
    #'   The compression level.
    #' @param shuffle (`logical(1)` | `integer(1)`)\cr
    #'   The shuffle filter to use.
    #' @param blocksize (`integer(1)` | `NA`)\cr
    #'   The block size.
    #' @param ... Not used.
    #' @return A new `BloscCodec` object.
    initialize = function(cname = "lz4", clevel = 5, shuffle = TRUE, blocksize = NA, ...) {
      self$cname <- cname
      self$clevel <- clevel
      self$shuffle <- shuffle
      # R's blosc package does not expose a blocksize parameter, so we cannot
      # pass it through to blosc_compress(). Store it for metadata fidelity.
      self$blocksize <- if (is.na(blocksize)) 0L else as.integer(blocksize)
      # No config options for blosc.
      if (!requireNamespace("blosc", quietly = TRUE)) {
        stop("blosc package must be installed to use the Blosc codec. Install with install.packages('blosc')")
      }
      warn_if_unk_args(list(...), "BloscCodec")
    },
    #' @description
    #' Compress data.
    #' @param buf (`raw()`)\cr
    #'   The un-compressed data.
    #' @param zarr_arr ([ZarrArray])\cr
    #'   The ZarrArray instance.
    #' @return Compressed data.
    encode = function(buf, zarr_arr) {
      dtype_size <- zarr_arr$get_dtype()$num_bytes
      shuffle_str <- if (is.logical(self$shuffle)) {
        if (self$shuffle) "shuffle" else "noshuffle"
      } else {
        c("noshuffle", "shuffle", "bitshuffle")[as.integer(self$shuffle) + 1L]
      }
      result <- blosc::blosc_compress(
        buf,
        compressor = self$cname,
        level = as.integer(self$clevel),
        shuffle = shuffle_str,
        typesize = as.integer(dtype_size)
      )
      return(result)
    },
    #' @description
    #' Decompress data.
    #' @param buf (`raw()`)\cr
    #'   The compressed data.
    #' @param zarr_arr ([ZarrArray])\cr
    #'   The ZarrArray instance.
    #' @return Un-compressed data.
    decode = function(buf, zarr_arr) {
      result <- blosc::blosc_decompress(buf)
      return(result)
    },
    #' @description
    #' Get codec configuration as a list.
    #' @return A named list.
    get_config = function() {
       meta <- list(
         id = jsonlite::unbox("blosc"),
         cname = jsonlite::unbox(as.character(self$cname)),
         clevel = jsonlite::unbox(as.integer(self$clevel)),
         shuffle = jsonlite::unbox(as.integer(self$shuffle)),
         blocksize = jsonlite::unbox(as.integer(self$blocksize))
       )
       return(meta)
    }
  )
)

#' Variable-length UTF-8 codec for Zarr
#' @title VLenUtf8Codec Class
#' @docType class
#' @description
#' Class representing a VLenUtf8 compressor
#'
#' @format [R6::R6Class] inheriting from [Codec].
#' @family Codec classes
#' @rdname VLenUtf8Codec
#' @export
VLenUtf8Codec <- R6::R6Class("VLenUtf8Codec",
  inherit = Codec,
  public = list(
    #' @description
    #' Compress data.
    #' @param buf (`character()`)\cr
    #'   The un-compressed data (character vector).
    #' @param zarr_arr ([ZarrArray])\cr
    #'   The ZarrArray instance.
    #' @return Compressed data.
    encode = function(buf, zarr_arr) {
      # Kind: array to bytes
      # Reference: https://github.com/zarr-developers/numcodecs/blob/cb155432e36536e17a2d054c8c24b7bf6f4a7347/numcodecs/vlen.pyx#L74

      num_strings <- length(buf)

      encoded_values <- list()
      encoded_lengths <- integer(num_strings)

      data_length <- 0

      # first iteration to convert to bytes
      for(i in seq_len(num_strings)) {
        orig_str <- buf[i]
        if(is.na(orig_str) || is.null(orig_str)) {
          # treat these as missing value, normalize
          orig_str <- ""
        }
        encoded_str <- charToRaw(orig_str)
        encoded_str_len <- length(encoded_str)
        encoded_values[[i]] <- encoded_str
        encoded_lengths[i] <- encoded_str_len
        data_length <- data_length + encoded_str_len + 4 # 4 bytes to store item length
      }

      # setup output
      total_length <- 4 + data_length # 4 bytes to store number of items in header
      out <- raw(total_length)

      # write header
      out[1:4] <- writeBin(
        num_strings,
        con = raw(),
        size = 4,
        endian = "little"
      )
      
      # second iteration, store data
      pos <- 4
      for(i in seq_len(num_strings)) {
        l <- encoded_lengths[i]
        out[(pos+1):(pos+4)] <- writeBin(
          l,
          con = raw(),
          size = 4,
          endian = "little"
        )
        pos <- pos + 4
        out[(pos+1):(pos+l)] <- encoded_values[[i]]
        pos <- pos + l
      }

      return(out)
    },
    #' @description
    #' Decompress data.
    #' @param buf (`raw()`)\cr
    #'   The compressed data.
    #' @param zarr_arr ([ZarrArray])\cr
    #'   The ZarrArray instance.
    #' @return Un-compressed data.
    decode = function(buf, zarr_arr) {
      # Kind: bytes to array
      # References:
      # - https://github.com/manzt/zarrita.js/blob/050d128265af14ff3c82e125315f3f527112887d/packages/core/src/codecs/vlen-utf8.ts
      # - https://github.com/zarr-developers/numcodecs/blob/cb155432e36536e17a2d054c8c24b7bf6f4a7347/numcodecs/vlen.pyx#L132

      num_strings <- readBin(
        con = buf,
        what = integer(),
        size = 4,
        n = 1,
        signed = TRUE,
        endian = "little"
      )

      vec_of_strings <- rep(NA, times = num_strings)

      pos <- 4
      for(i in seq_len(num_strings)) {
        num_chars <- readBin(
          con = buf[(pos+1):(pos+4)],
          what = integer(),
          size = 4,
          n = 1,
          signed = TRUE,
          endian = "little"
        )
        pos <- pos + 4
        vec_of_strings[i] <- rawToChar(buf[(pos+1):(pos+num_chars)])
        pos <- pos + num_chars
      }

      return(vec_of_strings)
    },
    #' @description
    #' Get codec configuration as a list.
    #' @return A named list.
    get_config = function() {
       meta <- list(
         id = jsonlite::unbox("vlen-utf8")
       )
       return(meta)
    }
  )
)

#' @keywords internal
CODEC_CONSTRUCTORS <- list(
  lz4 = Lz4Codec,
  zstd = ZstdCodec,
  zlib = ZlibCodec,
  gzip = GzipCodec,
  lzma = LzmaCodec,
  bz2 = Bz2Codec,
  blosc = BloscCodec,
  `vlen-utf8` = VLenUtf8Codec
)

#' Get a codec instance from the registry.
#'
#' @param config A codec config as a named list.
#' @return The instance of the codec.
#' @keywords internal
get_codec <- function(config) {
  if (is_na(config)) return(Codec$new())
  codec_id <- as.character(config$id)
  config$id <- NULL
  constructor <- CODEC_CONSTRUCTORS[[codec_id]]
  if (is.null(constructor)) stop("Unknown codec: ", codec_id)
  do.call(constructor$new, config)
}

#' Get the default compressor.
#' @keywords internal
#' @returns A Zstd compressor instance.
get_default_compressor <- function() {
  return(ZstdCodec$new())
}

# Resolve a V3 codec pipeline into V2-compatible compressor/filters.
# Reference: https://zarr-specs.readthedocs.io/en/latest/v3/core/v3.0.html#codecs
#
# V3 codecs are an ordered list with three roles:
#   - array-to-array: "transpose" (memory layout reordering)
#   - array-to-bytes: "endian"/"bytes" (byte order specification)
#   - bytes-to-bytes: "gzip", "zstd", "blosc", etc. (compression)
#
# For read (decode), codecs are applied in reverse order:
#   1. bytes-to-bytes codecs decompress the data
#   2. The endian codec tells us byte order for interpreting raw data
#   3. The transpose codec tells us memory layout order
#
# @param codecs_list List of codec entries from V3 zarr.json "codecs" field.
#   Each entry has $name and optional $configuration.
# @return Named list with:
#   $compressor - Codec instance for decompression (or NA if none)
#   $filters    - List of additional Codec instances (or NA if none)
#   $endian     - "little" or "big"
#   $order      - "C" or "F"
# @keywords internal
resolve_v3_codecs <- function(codecs_list) {
  endian <- "little"  # V3 default
  order <- "C"        # V3 default (C-contiguous)
  bytes_to_bytes <- list()

  for (codec_entry in codecs_list) {
    codec_name <- codec_entry$name
    config <- codec_entry$configuration
    if (is.null(config)) config <- list()

    if (codec_name %in% c("endian", "bytes")) {
      # array-to-bytes codec: specifies byte order.
      # "endian" is the zarrita name; "bytes" is the modern zarr-python name.
      if (!is.null(config$endian)) {
        endian <- config$endian
      }
    } else if (codec_name == "transpose") {
      # array-to-array codec: specifies memory layout.
      # zarr-python uses permutation arrays (e.g., [1, 0] for 2D F-order).
      # Legacy/zarrita may use string "F" or "C".
      if (!is.null(config$order)) {
        ord <- config$order
        if (is.character(ord) && ord == "F") {
          order <- "F"
        } else if (is.numeric(ord) || is.integer(ord)) {
          # Permutation array: reverse of 0:(n-1) means F-order
          n <- length(ord)
          if (identical(as.integer(ord), rev(seq.int(0L, n - 1L)))) {
            order <- "F"
          } else if (!identical(as.integer(ord), seq.int(0L, n - 1L))) {
            warning("Arbitrary transpose permutations not supported; using C-order")
          }
        }
      }
    } else {
      # bytes-to-bytes codec: compression.
      # Convert V3 codec config to V2 format for get_codec().
      v2_config <- v3_codec_to_v2_config(codec_name, config)
      bytes_to_bytes <- append(bytes_to_bytes, list(get_codec(v2_config)))
    }
  }

  # Map to V2 compressor/filters pattern:
  # - If one bytes-to-bytes codec: it's the compressor, no filters
  # - If multiple: last is compressor, rest are filters
  # - If none: no compression (NA)
  if (length(bytes_to_bytes) == 0) {
    compressor <- NA
    filters <- NA
  } else if (length(bytes_to_bytes) == 1) {
    compressor <- bytes_to_bytes[[1]]
    filters <- NA
  } else {
    compressor <- bytes_to_bytes[[length(bytes_to_bytes)]]
    filters <- bytes_to_bytes[1:(length(bytes_to_bytes) - 1)]
  }

  list(
    compressor = compressor,
    filters = filters,
    endian = endian,
    order = order
  )
}

# Map a V3 codec name + config to a V2-style config list for get_codec().
# Reference: https://zarr-specs.readthedocs.io/en/latest/v3/codecs/
# V3 uses "name" field; V2 uses "id" field.
#
# @param codec_name Character. V3 codec name (e.g., "gzip", "blosc").
# @param config List. V3 codec configuration.
# @return Named list with "id" field suitable for get_codec().
# @keywords internal
v3_codec_to_v2_config <- function(codec_name, config) {
  V3_TO_V2_ID <- list(
    "gzip"  = "gzip",
    "zstd"  = "zstd",
    "blosc" = "blosc",
    "zlib"  = "zlib",
    "bz2"   = "bz2",
    "lzma"  = "lzma"
  )

  v2_id <- V3_TO_V2_ID[[codec_name]]
  if (is.null(v2_id)) {
    stop("Unsupported V3 codec: '", codec_name,
         "'. Supported codecs: ", paste(names(V3_TO_V2_ID), collapse = ", "))
  }

  # Build V2-style config with "id" field
  v2_config <- config
  v2_config$id <- v2_id

  if (v2_id == "blosc") {
    # V3 blosc uses string shuffle values; V2 BloscCodec expects logical/integer.
    # For decode, blosc auto-detects shuffle from the compressed header,
    # but we convert for constructor compatibility.
    if (!is.null(v2_config$shuffle) && is.character(v2_config$shuffle)) {
      v2_config$shuffle <- switch(v2_config$shuffle,
        "noshuffle"  = 0L,
        "shuffle"    = 1L,
        "bitshuffle" = 2L,
        as.integer(v2_config$shuffle)
      )
    }
    # Strip typesize: not a BloscCodec constructor parameter.
    # Blosc reads typesize from the compressed data header during decode.
    v2_config$typesize <- NULL
  }

  if (v2_id == "zstd") {
    # V3 zstd may include "checksum" field not used by ZstdCodec.
    v2_config$checksum <- NULL
  }

  v2_config
}

# Convert a V2 codec config (with $id) to a V3 codec entry (with $name + $configuration).
# Inverse of v3_codec_to_v2_config.
#
# @param config Named list. V2 codec config with "$id" field (as from get_config()).
# @param dtype Character or NULL. V2 dtype string for blosc typesize (e.g., "<f4").
# @return Named list with "$name" and optional "$configuration".
# @keywords internal
v2_config_to_v3_codec <- function(config, dtype = NULL) {
  codec_id <- as.character(config$id)

  V2_TO_V3_NAME <- list(
    "gzip"  = "gzip",
    "zstd"  = "zstd",
    "blosc" = "blosc",
    "zlib"  = "zlib",
    "bz2"   = "bz2",
    "lzma"  = "lzma"
  )

  v3_name <- V2_TO_V3_NAME[[codec_id]]
  if (is.null(v3_name)) {
    stop("Cannot convert V2 codec '", codec_id, "' to V3 codec: unsupported codec")
  }

  # Build config without the "id" field
  v3_cfg <- config[names(config) != "id"]

  if (codec_id == "blosc") {
    # Convert integer shuffle to string ("noshuffle", "shuffle", "bitshuffle")
    if (!is.null(v3_cfg$shuffle)) {
      shuffle_int <- as.integer(v3_cfg$shuffle)
      v3_cfg$shuffle <- jsonlite::unbox(switch(as.character(shuffle_int),
        "0" = "noshuffle",
        "1" = "shuffle",
        "2" = "bitshuffle",
        "shuffle"
      ))
    }
    # Add typesize (required in V3 blosc)
    typesize <- if (!is.null(dtype)) get_dtype_numbytes(dtype) else 0L
    v3_cfg$typesize <- jsonlite::unbox(as.integer(typesize))
    # Ensure blocksize is present
    if (is.null(v3_cfg$blocksize)) {
      v3_cfg$blocksize <- jsonlite::unbox(0L)
    }
  }

  if (codec_id == "zstd") {
    # zarr-python includes checksum in V3 zstd config
    if (is.null(v3_cfg$checksum)) {
      v3_cfg$checksum <- jsonlite::unbox(FALSE)
    }
  }

  result <- list(name = jsonlite::unbox(v3_name))
  if (length(v3_cfg) > 0) {
    result$configuration <- v3_cfg
  }
  result
}

# Build a V3 codec pipeline from V2 compressor, filters, and dtype.
# Pipeline order (write/encode direction):
#   1. ArrayBytesCodec: "bytes" (with endian) or "vlen-utf8" if VLenUtf8 filter present
#   2. BytesBytesCodec: compressor (if any)
#
# @param compressor Codec object or NA.
# @param filters List of Codec objects or NA.
# @param dtype Character. V2 dtype string (e.g., "<f8").
# @return List of V3 codec entry lists.
# @keywords internal
build_v3_codec_pipeline <- function(compressor, filters, dtype) {
  codecs <- list()

  # Check for vlen-utf8 filter (it becomes the ArrayBytesCodec in V3)
  has_vlen <- FALSE
  if (!is_na(filters) && is.list(filters)) {
    has_vlen <- any(sapply(filters, function(f) {
      inherits(f, "VLenUtf8Codec") ||
        (!is.null(f$id) && as.character(f$id) == "vlen-utf8")
    }))
  }

  # 1. ArrayBytesCodec (exactly one, required)
  if (has_vlen) {
    codecs <- append(codecs, list(list(name = jsonlite::unbox("vlen-utf8"))))
  } else {
    dtype_info <- v2_dtype_to_v3_dtype(dtype)
    if (is.na(dtype_info$endian)) {
      # Single-byte type: omit endian field entirely
      codecs <- append(codecs, list(list(name = jsonlite::unbox("bytes"))))
    } else {
      codecs <- append(codecs, list(list(
        name = jsonlite::unbox("bytes"),
        configuration = list(endian = jsonlite::unbox(dtype_info$endian))
      )))
    }
  }

  # 2. BytesBytesCodec (compressor, if any)
  if (!is_na(compressor)) {
    compressor_cfg <- if (inherits(compressor, "Codec")) {
      compressor$get_config()
    } else {
      compressor
    }
    codecs <- append(codecs, list(v2_config_to_v3_codec(compressor_cfg, dtype)))
  }

  codecs
}

library(pizzarr)

# =============================================================================
# Gap 8: v2_config_to_v3_codec / build_v3_codec_pipeline
# =============================================================================

test_that("v2_config_to_v3_codec converts gzip", {
  v2_config <- list(id = jsonlite::unbox("gzip"), level = jsonlite::unbox(5L))
  result <- v2_config_to_v3_codec(v2_config)
  expect_equal(as.character(result$name), "gzip")
  expect_equal(as.integer(result$configuration$level), 5L)
})

test_that("v2_config_to_v3_codec converts zstd", {
  v2_config <- list(id = jsonlite::unbox("zstd"), level = jsonlite::unbox(3L))
  result <- v2_config_to_v3_codec(v2_config)
  expect_equal(as.character(result$name), "zstd")
  expect_equal(as.integer(result$configuration$level), 3L)
  # zstd should get checksum added
  expect_false(is.null(result$configuration$checksum))
})

test_that("v2_config_to_v3_codec converts blosc with shuffle int->string", {
  v2_config <- list(
    id = jsonlite::unbox("blosc"),
    cname = jsonlite::unbox("lz4"),
    clevel = jsonlite::unbox(5L),
    shuffle = jsonlite::unbox(1L)
  )
  result <- v2_config_to_v3_codec(v2_config, dtype = "<f4")
  expect_equal(as.character(result$name), "blosc")
  expect_equal(as.character(result$configuration$shuffle), "shuffle")
  expect_equal(as.integer(result$configuration$typesize), 4L)
  expect_false(is.null(result$configuration$blocksize))
})

test_that("v2_config_to_v3_codec converts blosc shuffle=0 to noshuffle", {
  v2_config <- list(
    id = jsonlite::unbox("blosc"),
    cname = jsonlite::unbox("zstd"),
    clevel = jsonlite::unbox(3L),
    shuffle = jsonlite::unbox(0L)
  )
  result <- v2_config_to_v3_codec(v2_config, dtype = "<i4")
  expect_equal(as.character(result$configuration$shuffle), "noshuffle")
})

test_that("v2_config_to_v3_codec converts blosc shuffle=2 to bitshuffle", {
  v2_config <- list(
    id = jsonlite::unbox("blosc"),
    cname = jsonlite::unbox("lz4"),
    clevel = jsonlite::unbox(5L),
    shuffle = jsonlite::unbox(2L)
  )
  result <- v2_config_to_v3_codec(v2_config, dtype = "<f8")
  expect_equal(as.character(result$configuration$shuffle), "bitshuffle")
  expect_equal(as.integer(result$configuration$typesize), 8L)
})

test_that("v2_config_to_v3_codec errors on unsupported codec", {
  v2_config <- list(id = jsonlite::unbox("unknown_codec"))
  expect_error(v2_config_to_v3_codec(v2_config), "unsupported codec")
})

test_that("v2_config_to_v3_codec converts bz2", {
  v2_config <- list(id = jsonlite::unbox("bz2"), level = jsonlite::unbox(9L))
  result <- v2_config_to_v3_codec(v2_config)
  expect_equal(as.character(result$name), "bz2")
})

test_that("v2_config_to_v3_codec converts lzma", {
  v2_config <- list(id = jsonlite::unbox("lzma"), level = jsonlite::unbox(6L),
                    format = jsonlite::unbox(1L))
  result <- v2_config_to_v3_codec(v2_config)
  expect_equal(as.character(result$name), "lzma")
})

test_that("v2_config_to_v3_codec converts zlib", {
  v2_config <- list(id = jsonlite::unbox("zlib"), level = jsonlite::unbox(1L))
  result <- v2_config_to_v3_codec(v2_config)
  expect_equal(as.character(result$name), "zlib")
})

# --- build_v3_codec_pipeline ---

test_that("build_v3_codec_pipeline with no compressor produces bytes codec only", {
  result <- build_v3_codec_pipeline(NA, NA, "<f8")
  expect_equal(length(result), 1)
  expect_equal(as.character(result[[1]]$name), "bytes")
  expect_equal(as.character(result[[1]]$configuration$endian), "little")
})

test_that("build_v3_codec_pipeline with compressor produces bytes + compressor", {
  compressor <- GzipCodec$new()
  result <- build_v3_codec_pipeline(compressor, NA, "<i4")
  expect_equal(length(result), 2)
  expect_equal(as.character(result[[1]]$name), "bytes")
  expect_equal(as.character(result[[2]]$name), "gzip")
})

test_that("build_v3_codec_pipeline single-byte dtype omits endian", {
  result <- build_v3_codec_pipeline(NA, NA, "|u1")
  expect_equal(length(result), 1)
  expect_equal(as.character(result[[1]]$name), "bytes")
  expect_null(result[[1]]$configuration)
})

test_that("build_v3_codec_pipeline with VLenUtf8 filter uses vlen-utf8 codec", {
  vlen_filter <- VLenUtf8Codec$new()
  result <- build_v3_codec_pipeline(NA, list(vlen_filter), "|O")
  expect_equal(length(result), 1)
  expect_equal(as.character(result[[1]]$name), "vlen-utf8")
})

test_that("build_v3_codec_pipeline with compressor and VLenUtf8", {
  vlen_filter <- VLenUtf8Codec$new()
  compressor <- ZstdCodec$new(level = 1)
  result <- build_v3_codec_pipeline(compressor, list(vlen_filter), "|O")
  expect_equal(length(result), 2)
  expect_equal(as.character(result[[1]]$name), "vlen-utf8")
  expect_equal(as.character(result[[2]]$name), "zstd")
})

# --- v3_codec_to_v2_config ---

test_that("v3_codec_to_v2_config round-trips gzip", {
  v2_config <- v3_codec_to_v2_config("gzip", list(level = 5L))
  expect_equal(as.character(v2_config$id), "gzip")
  expect_equal(as.integer(v2_config$level), 5L)
})

test_that("v3_codec_to_v2_config round-trips zstd (strips checksum)", {
  v2_config <- v3_codec_to_v2_config("zstd", list(level = 3L, checksum = FALSE))
  expect_equal(as.character(v2_config$id), "zstd")
  expect_null(v2_config$checksum)
})

# --- resolve_v3_codecs transpose handling ---

test_that("resolve_v3_codecs handles transpose permutation array [1,0] as F-order", {
  codecs <- list(
    list(name = "transpose", configuration = list(order = c(1L, 0L))),
    list(name = "bytes", configuration = list(endian = "little"))
  )
  result <- resolve_v3_codecs(codecs)
  expect_equal(result$order, "F")
})

test_that("resolve_v3_codecs handles transpose permutation array [0,1] as C-order", {
  codecs <- list(
    list(name = "transpose", configuration = list(order = c(0L, 1L))),
    list(name = "bytes", configuration = list(endian = "little"))
  )
  result <- resolve_v3_codecs(codecs)
  expect_equal(result$order, "C")
})

test_that("resolve_v3_codecs handles transpose string 'F'", {
  codecs <- list(
    list(name = "transpose", configuration = list(order = "F")),
    list(name = "bytes", configuration = list(endian = "little"))
  )
  result <- resolve_v3_codecs(codecs)
  expect_equal(result$order, "F")
})

library(pizzarr)

MockArray <- R6::R6Class("MockArray",
    public = list(
        shape = NULL,
        chunks = NULL,
        initialize = function(shape, chunks) {
            self$shape <- shape
            self$chunks <- chunks
        },
        get_shape = function() {
            return(self$shape)
        },
        get_chunks = function() {
            return(self$chunks)
        }
    )
)

test_that("VLenUTF8 codec can decode - raw to array", {
  codec <- VLenUtf8Codec$new()
  zarr_arr <- MockArray$new(c(4), c(4))
  # ['a', 'b', 'cc', 'd']
  str_arr <- codec$decode(as.raw(c(
    4, 0, 0, 0,
    1, 0, 0, 0,
    97, 1, 0, 0,
    0, 98, 2, 0,
    0, 0, 99, 99,
    1, 0, 0, 0,
    100
  )), zarr_arr)

  expect_equal(str_arr, c("a", "b", "cc", "d"))
})

test_that("VLenUTF8 codec can encode - array to raw", {
  codec <- VLenUtf8Codec$new()
  zarr_arr <- MockArray$new(c(4), c(4))
  # ['a', 'b', 'cc', 'd']
  raw_vec <- codec$encode(c("a", "b", "cc", "d"), zarr_arr)

  expect_equal(raw_vec, as.raw(c(
    4, 0, 0, 0,
    1, 0, 0, 0,
    97, 1, 0, 0,
    0, 98, 2, 0,
    0, 0, 99, 99,
    1, 0, 0, 0,
    100
  )))
})

test_that("VLenUTF8 codec can encode with high level API", {
  store <- MemoryStore$new()
  object_codec <- VLenUtf8Codec$new()

  string <- LETTERS[1:5]
  dims <- length(string)
  data <- array(data = string, dim = dims)

  z <- pizzarr::zarr_create_array(data, store = store, path = "string", dtype = "|O", object_codec = object_codec, shape = dims)
  
  sel <- z$get_item("...")
  expect_equal(as.character(sel$data), string)
})

library(pizzarr)

# --- Codec encode/decode round-trip tests ---

test_that("ZstdCodec encode/decode round-trip (default level)", {
  codec <- ZstdCodec$new()
  data <- as.raw(sample(0:255, 256, replace = TRUE))
  encoded <- codec$encode(data)
  decoded <- codec$decode(encoded)
  expect_equal(decoded, data)
})

test_that("ZstdCodec encode/decode round-trip (level 5)", {
  codec <- ZstdCodec$new(level = 5)
  data <- as.raw(sample(0:255, 256, replace = TRUE))
  encoded <- codec$encode(data)
  decoded <- codec$decode(encoded)
  expect_equal(decoded, data)
})

test_that("ZlibCodec encode/decode round-trip", {
  codec <- ZlibCodec$new()
  data <- as.raw(sample(0:255, 256, replace = TRUE))
  encoded <- codec$encode(data)
  decoded <- codec$decode(encoded)
  expect_equal(decoded, data)
})

test_that("GzipCodec encode/decode round-trip", {
  codec <- GzipCodec$new()
  data <- as.raw(sample(0:255, 256, replace = TRUE))
  encoded <- codec$encode(data)
  decoded <- codec$decode(encoded)
  expect_equal(decoded, data)
})

test_that("Bz2Codec encode/decode round-trip", {
  codec <- Bz2Codec$new()
  data <- as.raw(sample(0:255, 256, replace = TRUE))
  encoded <- codec$encode(data)
  decoded <- codec$decode(encoded)
  expect_equal(decoded, data)
})

test_that("LzmaCodec encode/decode round-trip", {
  codec <- LzmaCodec$new()
  data <- as.raw(sample(0:255, 256, replace = TRUE))
  encoded <- codec$encode(data)
  decoded <- codec$decode(encoded)
  expect_equal(decoded, data)
})

test_that("BloscCodec encode/decode round-trip", {
  skip_if_not_installed("blosc")
  a <- zarr_create(
    shape = c(8),
    chunks = c(8),
    dtype = "<i4",
    store = MemoryStore$new(),
    compressor = BloscCodec$new()
  )
  data <- array(c(1L, 2L, 3L, 4L, 5L, 6L, 7L, 8L), dim = c(8))
  a$set_item("...", data)
  result <- a$get_item("...")
  expect_equal(result$data, data)
})

# =============================================================================
# Gap 7: BloscCodec (V2 and V3) — extensive testing
# =============================================================================

test_that("BloscCodec encode/decode direct raw round-trip", {
  skip_if_not_installed("blosc")
  codec <- BloscCodec$new(cname = "lz4", clevel = 5, shuffle = TRUE)
  # Create a mock zarr array to pass dtype info
  z <- zarr_create(shape = c(8), chunks = c(8), dtype = "<i4",
                   store = MemoryStore$new(), compressor = NA)
  raw_data <- as.raw(rep(c(1:4), each = 4, times = 2))  # 32 bytes = 8 int32s
  encoded <- codec$encode(raw_data, z)
  decoded <- codec$decode(encoded, z)
  expect_equal(decoded, raw_data)
})

test_that("BloscCodec with shuffle=FALSE (noshuffle)", {
  skip_if_not_installed("blosc")
  a <- zarr_create(shape = c(8), chunks = c(8), dtype = "<f8",
                   store = MemoryStore$new(),
                   compressor = BloscCodec$new(cname = "lz4", shuffle = FALSE))
  data <- array(c(1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8), dim = 8)
  a$set_item("...", data)
  result <- a$get_item("...")
  expect_equal(result$data, data)
})

test_that("BloscCodec with integer shuffle=0 (noshuffle)", {
  skip_if_not_installed("blosc")
  a <- zarr_create(shape = c(8), chunks = c(8), dtype = "<i4",
                   store = MemoryStore$new(),
                   compressor = BloscCodec$new(cname = "lz4", shuffle = 0L))
  data <- array(1:8, dim = 8)
  a$set_item("...", data)
  result <- a$get_item("...")
  expect_equal(as.integer(result$data), 1:8)
})

test_that("BloscCodec with integer shuffle=2 (bitshuffle)", {
  skip_if_not_installed("blosc")
  a <- zarr_create(shape = c(8), chunks = c(8), dtype = "<i4",
                   store = MemoryStore$new(),
                   compressor = BloscCodec$new(cname = "lz4", shuffle = 2L))
  data <- array(1:8, dim = 8)
  a$set_item("...", data)
  result <- a$get_item("...")
  expect_equal(as.integer(result$data), 1:8)
})

test_that("BloscCodec with zstd compressor", {
  skip_if_not_installed("blosc")
  a <- zarr_create(shape = c(100), chunks = c(50), dtype = "<f8",
                   store = MemoryStore$new(),
                   compressor = BloscCodec$new(cname = "zstd", clevel = 3))
  data <- array(as.double(1:100), dim = 100)
  a$set_item("...", data)
  result <- a$get_item("...")
  expect_equal(result$data, data)
})

test_that("BloscCodec config round-trip", {
  skip_if_not_installed("blosc")
  codec <- BloscCodec$new(cname = "lz4", clevel = 5, shuffle = TRUE)
  config <- codec$get_config()
  expect_equal(as.character(config$id), "blosc")
  expect_equal(as.character(config$cname), "lz4")
  expect_equal(as.integer(config$clevel), 5L)
  restored <- get_codec(config)
  expect_s3_class(restored, "BloscCodec")
  expect_equal(as.character(restored$cname), "lz4")
})

test_that("BloscCodec requires blosc package", {
  # This test would be hard to run since we can't unload a package.
  # Just verify the constructor works when blosc is installed.
  skip_if_not_installed("blosc")
  codec <- BloscCodec$new()
  expect_s3_class(codec, "BloscCodec")
  expect_equal(codec$cname, "lz4")
  expect_equal(codec$clevel, 5)
})

test_that("BloscCodec 2D array round-trip with multiple chunks", {
  skip_if_not_installed("blosc")
  a <- array(as.double(1:100), dim = c(10, 10))
  z <- zarr_create_array(data = a, shape = dim(a), chunks = c(5, 5),
                         dtype = "<f8",
                         compressor = BloscCodec$new(cname = "lz4"))
  result <- z$get_item("...")
  expect_equal(result$data, a)
})

test_that("BloscCodec warn_if_unk_args warns for unrecognized params", {
  skip_if_not_installed("blosc")
  expect_warning(
    BloscCodec$new(cname = "lz4", unknown_param = TRUE),
    "unrecognized"
  )
})

# --- V3 Blosc round-trip ---

test_that("V3 array with blosc compressor round-trips", {
  skip_if_not_installed("blosc")
  store <- MemoryStore$new()
  z <- zarr_create(shape = c(8L), dtype = "<i4", zarr_format = 3L,
                   compressor = BloscCodec$new(cname = "lz4", clevel = 5),
                   store = store)
  data <- array(c(10L, 20L, 30L, 40L, 50L, 60L, 70L, 80L), dim = 8)
  z$set_item("...", data)
  result <- z$get_item("...")
  expect_equal(as.integer(result$data), c(10L, 20L, 30L, 40L, 50L, 60L, 70L, 80L))
})

# --- Codec base class ---

test_that("Base Codec get_config returns unbox(NA)", {
  codec <- Codec$new()
  config <- codec$get_config()
  expect_true(is.na(config))
})

# --- Codec config round-trip tests ---

test_that("ZstdCodec config round-trip", {
  codec <- ZstdCodec$new(level = 3)
  config <- codec$get_config()
  expect_equal(config$id, jsonlite::unbox("zstd"))
  expect_equal(config$level, jsonlite::unbox(3))
  restored <- get_codec(config)
  expect_s3_class(restored, "ZstdCodec")
  expect_equal(restored$level, 3)
})

test_that("ZlibCodec config round-trip", {
  codec <- ZlibCodec$new(level = 6)
  config <- codec$get_config()
  expect_equal(config$id, jsonlite::unbox("zlib"))
  restored <- get_codec(config)
  expect_s3_class(restored, "ZlibCodec")
})

test_that("GzipCodec config round-trip", {
  codec <- GzipCodec$new(level = 6)
  config <- codec$get_config()
  expect_equal(config$id, jsonlite::unbox("gzip"))
  restored <- get_codec(config)
  expect_s3_class(restored, "GzipCodec")
})

test_that("Bz2Codec config round-trip", {
  codec <- Bz2Codec$new()
  config <- codec$get_config()
  expect_equal(config$id, jsonlite::unbox("bz2"))
  restored <- get_codec(config)
  expect_s3_class(restored, "Bz2Codec")
})

test_that("LzmaCodec config round-trip", {
  codec <- LzmaCodec$new()
  config <- codec$get_config()
  expect_equal(config$id, jsonlite::unbox("lzma"))
  restored <- get_codec(config)
  expect_s3_class(restored, "LzmaCodec")
})

# --- Edge cases ---

test_that("get_codec with unknown id raises error", {
  expect_error(get_codec(list(id = "nonexistent")), "Unknown codec")
})

test_that("get_default_compressor returns ZstdCodec", {
  codec <- get_default_compressor()
  expect_s3_class(codec, "ZstdCodec")
})

test_that("Base Codec is a no-op pass-through", {
  codec <- Codec$new()
  data <- as.raw(c(1, 2, 3))
  expect_equal(codec$encode(data), data)
  expect_equal(codec$decode(data), data)
})

# --- Full array round-trip with each compressor ---

test_that("Array round-trip with ZlibCodec compressor", {
  a <- zarr_create(
    shape = c(4), chunks = c(4), dtype = "<f8",
    store = MemoryStore$new(), compressor = ZlibCodec$new()
  )
  data <- array(c(1.5, 2.5, 3.5, 4.5), dim = c(4))
  a$set_item("...", data)
  result <- a$get_item("...")
  expect_equal(result$data, data)
})

test_that("Array round-trip with GzipCodec compressor", {
  a <- zarr_create(
    shape = c(4), chunks = c(4), dtype = "<f8",
    store = MemoryStore$new(), compressor = GzipCodec$new()
  )
  data <- array(c(1.5, 2.5, 3.5, 4.5), dim = c(4))
  a$set_item("...", data)
  result <- a$get_item("...")
  expect_equal(result$data, data)
})

test_that("Array round-trip with Bz2Codec compressor", {
  a <- zarr_create(
    shape = c(4), chunks = c(4), dtype = "<f8",
    store = MemoryStore$new(), compressor = Bz2Codec$new()
  )
  data <- array(c(1.5, 2.5, 3.5, 4.5), dim = c(4))
  a$set_item("...", data)
  result <- a$get_item("...")
  expect_equal(result$data, data)
})

test_that("Array round-trip with LzmaCodec compressor", {
  a <- zarr_create(
    shape = c(4), chunks = c(4), dtype = "<f8",
    store = MemoryStore$new(), compressor = LzmaCodec$new()
  )
  data <- array(c(1.5, 2.5, 3.5, 4.5), dim = c(4))
  a$set_item("...", data)
  result <- a$get_item("...")
  expect_equal(result$data, data)
})

test_that("Array round-trip with no compressor", {
  a <- zarr_create(
    shape = c(4), chunks = c(4), dtype = "<f8",
    store = MemoryStore$new(), compressor = NA
  )
  data <- array(c(1.5, 2.5, 3.5, 4.5), dim = c(4))
  a$set_item("...", data)
  result <- a$get_item("...")
  expect_equal(result$data, data)
})

test_that("BloscCodec get_config includes blocksize and round-trips", {
  skip_if_not_installed("blosc")
  codec <- BloscCodec$new(cname = "lz4", clevel = 5, shuffle = TRUE, blocksize = 512)
  config <- codec$get_config()
  expect_equal(as.integer(config$blocksize), 512L)

  # round-trip through get_codec
  restored <- get_codec(config)
  expect_s3_class(restored, "BloscCodec")
  expect_equal(restored$blocksize, 512L)
})

test_that("BloscCodec NA blocksize normalizes to 0", {
  skip_if_not_installed("blosc")
  codec <- BloscCodec$new(blocksize = NA)
  expect_equal(codec$blocksize, 0L)
  config <- codec$get_config()
  expect_equal(as.integer(config$blocksize), 0L)
})

#' @keywords internal
char_vec_to_raw <- function(char_vec, basic_type, num_chars, byte_order) {
  # Reference: https://stat.ethz.ch/R-manual/R-devel/library/base/html/iconv.html
  if(basic_type == "S") {
    iconv_to <- "UTF-8"
    num_bytes_per_char <- 1
  } else if(basic_type == "U") {
    if(byte_order == "little") {
      iconv_to <- "UTF-32LE"
    } else if(byte_order == "big") {
      iconv_to <- "UTF-32BE"
    } else {
      stop("Got unexpected byte_order in char_vec_to_raw()")
    }
    num_bytes_per_char <- 4
  } else {
    stop("Unexpected basic_type in char_vec_to_raw()")
  }
  num_bytes <- num_chars * num_bytes_per_char

  list_of_raw <- iconv(char_vec, to = iconv_to, toRaw = TRUE)

  buf <- raw(length = length(list_of_raw) * num_bytes)

  for(i in seq_len(length(list_of_raw))) {
    raw_vec_i <- list_of_raw[[i]]

    if(length(raw_vec_i) > num_bytes * num_bytes_per_char) {
      stop("Unexpected length of raw_vec_i in char_vec_to_raw(): string probably too long for specified dtype")
    }

    offset_i_start <- (i-1) * num_bytes + 1
    offset_i_stop <- offset_i_start + length(raw_vec_i) - 1

    # Byte order is already handled by iconv via UTF-32LE/UTF-32BE selection above.
    buf[offset_i_start:offset_i_stop] <- raw_vec_i
  }

  return(buf)
}

#' @keywords internal
raw_to_char_vec <- function(raw_vec, basic_type, num_chars, byte_order) {
  # Reference: https://stat.ethz.ch/R-manual/R-devel/library/base/html/iconv.html
  if(basic_type == "S") {
    iconv_from <- "UTF-8"
    num_bytes_per_char <- 1
  } else if(basic_type == "U") {
    if(byte_order == "little") {
      iconv_from <- "UTF-32LE"
    } else if(byte_order == "big") {
      iconv_from <- "UTF-32BE"
    } else {
      stop("Got unexpected byte_order in raw_to_char_vec()")
    }
    num_bytes_per_char <- 4
  } else {
    stop("Unexpected basic_type in raw_to_char_vec()")
  }
  num_bytes <- num_chars * num_bytes_per_char

  null_byte <- as.raw(rep(0x00, times = num_bytes_per_char))

  list_of_raw <- list()
  for(i in seq_len(length(raw_vec) / num_bytes)) {

    offset_i_start <- (i-1) * num_bytes + 1
    offset_i_stop <- offset_i_start + num_bytes - 1

    # We get the raw vector for each string in the array.
    raw_vec_i <- raw_vec[offset_i_start:offset_i_stop]

    if(num_bytes_per_char == 1) {
      # Since every character only uses one byte, we can simply remove all null bytes.
      raw_vec_i <- raw_vec_i[raw_vec_i != null_byte]
    } else {
      # Since each character uses multiple bytes, we need to remove only those characters
      # for which _all_ bytes are null bytes.

      # Convert the raw vector to an int32 vector and remove zeros, then convert back to raw,
      # so that we only remove characters that contain all null bytes.
      int_vec_i <- readBin(
        con = raw_vec_i,
        what = "integer",
        size = num_bytes_per_char,
        n = num_chars,
        signed = TRUE,
        endian = byte_order
      )
      int_vec_i <- int_vec_i[int_vec_i != 0]
      raw_vec_i <- writeBin(
        object = int_vec_i,
        con = raw(),
        size = num_bytes_per_char,
        endian = byte_order
      )
    }

    # We append the raw vector for this string to the list of raw vectors.
    # This is the input format required by `iconv`.
    list_of_raw <- append(list_of_raw, list(raw_vec_i))
  }
  char_vec <- iconv(list_of_raw, from = iconv_from, to = "UTF-8", toRaw = FALSE)
  return(char_vec)
}

library(pizzarr)

test_that("NestedArray can be created from character array", {
    a <- array(data=c("a", "b", "c", "d"), dim=c(2, 2))
    na <- NestedArray$new(data = a, shape = dim(a), dtype = "|S1")

    selection <- na$get(list(slice(1, 2), slice(1, 2)))

    expect_equal(selection$data, a)
})

test_that("char_vec_to_raw works for S dtype, length 1", {
    a <- c("a", "b", "c", "d")
    raw_a <- char_vec_to_raw(a, "S", 1, "little")

    expect_equal(raw_a, as.raw(c(
        0x61,
        0x62,
        0x63,
        0x64
    )))
})

test_that("raw_to_char_vec works for S dtype, length 1", {
    char_vec <- raw_to_char_vec(as.raw(c(
        0x61,
        0x62,
        0x63,
        0x64
    )), "S", 1, "little")

    expect_equal(char_vec, c("a", "b", "c", "d"))
})

test_that("char_vec_to_raw works for S dtype, length 7", {
    a <- c("a", "b", "c", "d")
    raw_a <- char_vec_to_raw(a, "S", 7, "little")

    expect_equal(raw_a, as.raw(c(
        0x61, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x62, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x63, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x64, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
    )))
})

test_that("raw_to_char_vec works for S dtype, length 7", {
    char_vec <- raw_to_char_vec(as.raw(c(
        0x61, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x62, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x63, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x64, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
    )), "S", 7, "little")

    expect_equal(char_vec, c("a", "b", "c", "d"))
})

test_that("char_vec_to_raw works for U dtype, little endian", {
    a <- c("a", "b", "c", "d")
    raw_a <- char_vec_to_raw(a, "U", 2, "little")

    expect_equal(raw_a, as.raw(c(
        0x61, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x62, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x63, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x64, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
    )))
})

test_that("raw_to_char_vec works for U dtype, little endian", {
    char_vec <- raw_to_char_vec(as.raw(c(
        0x61, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x62, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x63, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x64, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
    )), "U", 2, "little")

    expect_equal(char_vec, c("a", "b", "c", "d"))
})

test_that("char_vec_to_raw works for U dtype, big endian", {
    a <- c("a", "b", "c", "d")
    raw_a <- char_vec_to_raw(a, "U", 2, "big")

    expect_equal(raw_a, as.raw(c(
        0x00, 0x00, 0x00, 0x61, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x62, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x63, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x64, 0x00, 0x00, 0x00, 0x00
    )))
})

test_that("NestedArray of characters can be created from raw array, S dtype", {
    a <- array(data=c("a", "b", "c", "d"), dim=c(4))
    na <- NestedArray$new(data = as.raw(c(
        0x61,
        0x62,
        0x63,
        0x64
    )), shape = dim(a), dtype = "|S1")

    selection <- na$get(list(slice(1, 4)))

    expect_equal(selection$data, a)
})

test_that("NestedArray of characters can be created from raw array, U dtype, little endian", {
    a <- array(data=c("a", "b", "c", "d"), dim=c(4))
    na <- NestedArray$new(data = as.raw(c(
        0x61, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x62, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x63, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x64, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
    )), shape = dim(a), dtype = "<U2")

    selection <- na$get(list(slice(1, 4)))

    expect_equal(selection$data, a)
})

test_that("NestedArray of characters can be created from raw array, U dtype, big endian", {
    a <- array(data=c("a", "b", "c", "d"), dim=c(4))
    na <- NestedArray$new(data = as.raw(c(
        0x00, 0x00, 0x00, 0x61, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x62, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x63, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x64, 0x00, 0x00, 0x00, 0x00
    )), shape = dim(a), dtype = ">U2")

    selection <- na$get(list(slice(1, 4)))

    expect_equal(selection$data, a)
})

test_that("NestedArray of strings can be converted to a raw array, S dtype", {
    a <- array(data=c("a", "b", "c", "d"), dim=c(4))
    na <- NestedArray$new(data = a, shape = dim(a), dtype = "|S1")

    na_as_raw <- na$flatten_to_raw()

    expect_equal(na_as_raw, as.raw(c(
        0x61,
        0x62,
        0x63,
        0x64
    )))
})

#' @keywords internal
zero_based_to_one_based <- function(selection, shape) {

  # drop this since we could do it for arbitrary indices
  # if(!all(vapply(selection, is_slice, logical(length = 1)))) 
  #   stop("selection must be a list of slices")
  
  selection_list <- list()
  
  for(i in seq_len(length(selection))) {
    
    # get selection
    sel <- selection[[i]]
    
    # for slice
    if(inherits(sel, "Slice")){
     
      # We assume the selection uses zero-based indexing,
      # and internally convert to R-based / 1-based indexing
      # before accessing data on the internal self$data.
      sel_start <- sel$start + 1 # Add one, since R indexing is zero-based.
      sel_stop <- sel$stop # Do not subtract one, since R indexing is inclusive.
      sel_step <- sel$step
      if(is.na(sel_step)) sel_step <- 1
      if(sel_start < 1) {
        sel_start <- 1
        warning("IndexError: slice start index out of bounds (too low), clamped to 1")
      }
      if(sel_start > shape[i]) {
        sel_start <- shape[i]
        warning("IndexError: slice start index out of bounds (too high), clamped to ", shape[i])
      }
      if(sel_stop < 1) {
        sel_stop <- 1
        warning("IndexError: slice stop index out of bounds (too low), clamped to 1")
      }
      if(sel_stop > shape[i]) {
        sel_stop <- shape[i]
        warning("IndexError: slice stop index out of bounds (too high), clamped to ", shape[i])
      }
      selection_list <- append(selection_list, list(seq(from = sel_start, 
                                                        to = sel_stop, 
                                                        by = sel_step))) 
    } else if(is.numeric(sel)) {
      sel <- sel + 1
      selection_list <- append(selection_list, list(sel))
    } else {
      stop("Unsupported selection type")
    }
  }
  return(selection_list)
}

# Reference: https://github.com/gzuidhof/zarr.js/blob/292804/src/nestedArray/index.ts#L9

#' The Zarr NestedArray class.
#' @title NestedArray Class
#' @docType class
#' @description
#' Represents a multi-dimensional array that can be
#' accessed and subsetted via list of Slice instances.
#' @format [R6::R6Class]
#' @rdname NestedArray
#' @importFrom R6 R6Class
#' @export
NestedArray <- R6::R6Class("NestedArray",
  private = list(
    is_zero_dim = NULL,
    dtype_basic_type = NULL,
    dtype_byte_order = NULL,
    dtype_num_bytes = NULL,
    dtype_num_items = NULL
  ),
  public = list(
    #' @field shape The shape of the array.
    shape = NULL,
    #' @field dtype The Zarr dtype of the array, as a string like ">f8".
    dtype = NULL,
    #' @field dtype_obj The Zarr dtype of the array, as a Dtype instance.
    dtype_obj = NULL,
    #' @field data The array contents as a base R array.
    data = NULL,
    #' @description
    #' Create a new NestedArray instance.
    #' @param data The data to initialize the array with.
    #' Either NULL, base R array, base R vector (numeric/logical),
    #' scalar, or raw vector.
    #' @param shape The shape of the array.
    #' @param dtype The Zarr dtype of the array, as a string like ">f8".
    #' @param order The order of the array, either "C" or "F". Only used
    #' when `data` is a raw vector. Optional.
    #' @return A `NestedArray` instance.
    initialize = function(data, shape = NA, dtype = NA, order = NA) {
      if(is.null(shape) || (!is.list(shape) && is_na(shape))) {
        if(is.raw(data)) {
          stop("Cannot infer shape from raw data, please provide shape explicitly")
        }
        shape <- dim(data)
      } else {
        shape <- normalize_shape(shape)
      }
      if(is_na(dtype) && (is.numeric(data) || is.logical(data))) {
        self$dtype_obj <- Dtype$new(get_dtype_from_array(data))
      } else if("Dtype" %in% class(dtype)) {
        self$dtype_obj <- dtype
      } else if(is.character(dtype)) {
        self$dtype_obj <- Dtype$new(dtype)
        if(self$dtype_obj$is_object) {
          stop("Object dtype was initialized from string in NestedArray, so object_codec is missing.")
        }
      } else {
        stop("dtype must be NA, string/character vector, or Dtype instance")
      }
      self$shape <- shape

      private$is_zero_dim <- (is.null(shape) || length(shape) == 0)

      if(is.null(data)) {
        # Create empty array.

        dtype_rtype <- self$dtype_obj$get_rtype()

        if (inherits(dtype_rtype, "integer64")) {
          self$data <- array(data = double(), dim = shape)
          class(self$data) <- "integer64"
        } else {
          self$data <- array(data = dtype_rtype, dim = shape)
        }
      } else if(!is.raw(data) && is.null(self$shape)) {
        # Create zero-dimensional array.

        self$data <- data # TODO?
      } else if(!is.raw(data) && (is.array(data) || is.vector(data) || inherits(data, "integer64")) && is.atomic(data)) {
        # Create array from R atomic vector or array().
        num_shape_elements <- compute_size(shape)
        # Check that data array has same shape as expected
        if(!is.null(dim(data)) && all(ensure_integer_vec(dim(data)) == ensure_integer_vec(shape))) {
          self$data <- data
        } else {
          astype_func <- self$dtype_obj$get_asrtype()
          is_i64 <- inherits(data, "integer64") || is_int64_dtype(self$dtype_obj)
          coerced <- astype_func(data)

          # Data array did not have the expected shape, so we need to reshape it.
          if(!is_na(order) && order == "C") {
            ordered_shape <- shape[rev(seq_len(length(shape)))]
            array_from_vec <- array(data = unclass(coerced), dim = ordered_shape)
            array_from_vec <- aperm(array_from_vec, rev(seq_len(length(shape))))
          } else {
            array_from_vec <- array(data = unclass(coerced), dim = shape)
          }
          if (is_i64 && has_bit64()) class(array_from_vec) <- "integer64"
          self$data <- array_from_vec
        }
      } else if(is.raw(data)) {
        # Create array from a raw vector.

        num_shape_elements <- compute_size(shape)

        # Reference: https://github.com/gzuidhof/zarr.js/blob/292804/src/nestedArray/index.ts#L134
        buf <- data
        # Create from ArrayBuffer or Buffer
        
        dtype_size <- self$dtype_obj$num_bytes
        num_data_elements <- length(buf) / dtype_size
        if (num_shape_elements != num_data_elements) {
          stop('Buffer has ${numDataElements} of dtype ${dtype}, shape is too large or small')
        }

        dtype_rtype <- self$dtype_obj$get_rtype()
        dtype_signed <- self$dtype_obj$is_signed
        if(!dtype_signed && !(dtype_size == 1 || dtype_size == 2)) {
          # readBin will warn "signed = FALSE is only valid for integers of sizes 1 and 2"
          dtype_signed <- TRUE
        }

        endian <- self$dtype_obj$byte_order
        # Normalize to only "little" or "big" since this is what writeBin accepts.
        if(endian == "nr") {
          endian <- "little"
        }

        if(self$dtype_obj$basic_type %in% c("S", "U")) {
          vec_from_raw <- raw_to_char_vec(
            buf,
            self$dtype_obj$basic_type,
            self$dtype_obj$num_items,
            endian
          )
        } else if(is_int64_dtype(self$dtype_obj) && has_bit64()) {
          vec_from_raw <- raw_to_integer64(buf, num_shape_elements, endian)
        } else if(is_int64_dtype(self$dtype_obj)) {
          vec_from_raw <- raw_to_double_from_int64(buf, num_shape_elements,
                                                    endian, dtype_signed)
        } else {
          vec_from_raw <- readBin(
            con = buf,
            what = dtype_rtype,
            size = dtype_size,
            n = num_shape_elements,
            signed = dtype_signed,
            endian = endian
          )
        }
        
        is_i64 <- inherits(vec_from_raw, "integer64")

        if(private$is_zero_dim) {
          array_from_vec <- array(data = unclass(vec_from_raw), dim = c(1))
        } else {
          if(!is_na(order) && order == "C") {
            # Either "C" or "F", defining the layout of bytes within each chunk of the array.
            # "C" means row-major order, i.e., the last dimension varies fastest;
            # "F" means column-major order, i.e., the first dimension varies fastest.
            # Reference: https://zarr.readthedocs.io/en/stable/spec/v2.html#metadata
            ordered_shape <- shape[rev(seq_len(length(shape)))]
            array_from_vec <- array(data = unclass(vec_from_raw), dim = ordered_shape)
            array_from_vec <- aperm(array_from_vec, rev(seq_len(length(shape))))
          } else {
            array_from_vec <- array(data = unclass(vec_from_raw), dim = shape)
          }
        }

        if (is_i64) class(array_from_vec) <- "integer64"
        self$data <- array_from_vec
      } else if(is_scalar(data)) {
        # Create array from a scalar value.
        astype_func <- self$dtype_obj$get_asrtype()
        dtype_rtype <- self$dtype_obj$get_rtype()
        if(private$is_zero_dim) {
          self$data <- array(data=dtype_rtype, dim=c(1))
        } else {
          self$data <- array(data=dtype_rtype, dim=shape)
        }
        self$data[] <- astype_func(data)
      } else {
        #buf_len <- compute_size(shape) * get_dtype_numbytes(dtype) 
        #buf <- raw(length = buf_len)
        # TODO?
        stop("Unexpected type for data in NestedArray$initialize()")
      }
    },
    #' @description
    #' Subset the array.
    #' @param selection A list of slices.
    #' @return A new NestedArray (potentially a subset) representing the selection.
    get = function(selection) {
      selection_list <- zero_based_to_one_based(selection, self$shape)
      
      # Using do.call here seems to work the same as `abind::asub(self$data, selection_list)`
      # so we can use do.call to avoid the extra dependency.
      subset_arr <- do.call("[", append(list(self$data), selection_list))
      subset_nested_array <- NestedArray$new(subset_arr, shape = dim(subset_arr), dtype = self$dtype_obj)
      return(subset_nested_array)
    },
    #' @description
    #' Set a subset of the array.
    #' @param selection A list of slices.
    #' @param value A NestedArray or a base R array.
    #' @return `NULL` (called for side effects, modifies `self$data` in place).
    set = function(selection, value) {
      # value should be a NestedArray.
      selection_list <- zero_based_to_one_based(selection, self$shape)

      if("NestedArray" %in% class(value)) {
        value_data <- value$data
      } else if(is_scalar(value) || is.array(value)) {
        value_data <- value
      } else {
        message(value)
        stop("Got unexpected type for value in NestedArray$set()")
      }

      # Only set values if the array is not meant to be empty.
      if(sum(length(value_data)) > 0 || sum(dim(value_data)) > 0) {
        self$data <- do.call("[<-", c(list(self$data), selection_list, list(value = value_data)))
      }
    },
    #' @description
    #' Flatten the array contents.
    #' @param order Either "C", "F", or NA.
    #' @return The data as a flat vector.
    flatten = function(order = NA) {
      # Transpose first (if needed, based on the ordering).

      # "C" means row-major order, i.e., the last dimension varies fastest;
      # "F" means column-major order, i.e., the first dimension varies fastest.
      # Reference: https://zarr.readthedocs.io/en/stable/spec/v2.html#metadata

      is_i64 <- inherits(self$data, "integer64")

      if(!is_na(order) && order == "C" && !private$is_zero_dim) {
        ordered_data <- aperm(self$data, rev(seq_len(length(self$shape))))
      } else {
        ordered_data <- self$data
      }
      result <- as.vector(ordered_data)
      if (is_i64) class(result) <- "integer64"
      return(result)
    },
    #' @description
    #' Flatten the array contents and convert to a raw vector.
    #' @param order Either "C", "F", or NA.
    #' @return The data as a flat `raw()` vector (or plain vector for object dtypes).
    flatten_to_raw = function(order = NA) {
      data_as_vec <- self$flatten(order = order)

      if(self$dtype_obj$is_object) {
        # The object_codec in filters will handle the conversion to raw.
        return(data_as_vec)
      }

      endian <- self$dtype_obj$byte_order
      # Normalize to only "little" or "big" since this is what writeBin accepts.
      if(endian == "nr") {
        endian <- "little"
      }

      # "If writeBin is called with con a raw vector, it is just an indication that a raw vector should be returned."
      # Reference: https://stat.ethz.ch/R-manual/R-devel/library/base/html/readBin.html
      if(self$dtype_obj$basic_type %in% c("S", "U")) {
        buf <- char_vec_to_raw(
          data_as_vec,
          self$dtype_obj$basic_type,
          self$dtype_obj$num_items,
          endian
        )
      } else if(is_int64_dtype(self$dtype_obj) && inherits(data_as_vec, "integer64")) {
        buf <- integer64_to_raw(data_as_vec, endian)
      } else {
        buf <- writeBin(
          data_as_vec,
          con = raw(),
          size = self$dtype_obj$num_bytes,
          endian = endian
        )
      }
      return(buf)
    },
    #' @description
    #' Convert NestedArray to a base R array.
    #' @return `array()`.
    as.array = function() {
      # Consider using drop() to simplify dimensions of 1
      return(self$data)
    }
  )
)

#' S3 method for as.array
#'
#' @param x Object.
#' @param ... Not used.
#' @keywords internal
#' @export
as.array.NestedArray = function(x, ...) {
  x$as.array()
}

library(pizzarr)

test_that("NestedArray can be accessed using list of slices", {
    a <- array(1:20, dim = c(2, 10))
    na <- NestedArray$new(a, shape = c(2, 10), dtype = "<f4")

    expect_equal(na$data, a)

    # R-based indexing
    r_sel <- na$get(list(slice(1, 2), slice(1, 3)))
    # print(r_sel$data)

    # Python-based indexing

    python_sel <- na$get(list(zb_slice(0, 2), zb_slice(0, 3)))
    # print(python_sel$data)

    expect_equal(r_sel$data, python_sel$data)
})

test_that("NestedArray can be set using list of slices", {
    a <- array(1:20, dim = c(2, 10))
    na <- NestedArray$new(a, shape = c(2, 10), dtype = "<f4")

    expect_equal(na$data, a)

    new_a <- array(1:6, dim = c(2, 3))
    new_na <- NestedArray$new(new_a, shape = c(2, 3), dtype = "<f4")

    # R-based indexing
    na$set(list(slice(1, 2), slice(1, 3)), new_na)
    # print(r_sel$data)

    r_sel <- na$get(list(slice(1, 2), slice(1, 3)))

    expect_equal(r_sel$data, new_a)
})

test_that("NestedArray can be set using list of zb_slices", {
    a <- array(1:20, dim = c(2, 10))
    na <- NestedArray$new(a, shape = c(2, 10), dtype = "<f4")

    expect_equal(na$data, a)

    new_a <- array(1:6, dim = c(2, 3))
    new_na <- NestedArray$new(new_a, shape = c(2, 3), dtype = "<f4")

    # R-based indexing
    na$set(list(zb_slice(0, 2), zb_slice(0, 3)), new_na)
    # print(r_sel$data)

    python_sel <- na$get(list(zb_slice(0, 2), zb_slice(0, 3)))

    expect_equal(python_sel$data, new_a)
})

test_that("zero_based_to_one_based", {
  
  s <- list(slice(1, 11, 1))

  expect_equal(zero_based_to_one_based(s, shape = c(20, 20)), 
               list(c(1:11)))
  
  s <- list(slice(1, 11, 2))
  
  expect_equal(zero_based_to_one_based(s, shape = c(20, 20)), 
               list(seq(1, 11, 2)))
})

test_that("set array values", {

  d <- zarr_volcano()

  a <- d$get_item("volcano")

  vals <- a[1:10, 1:20]$as.array()

  new_vals <- vals * 10

  sub <- a[1:10, 1:20]

  sub$set(list(slice(1,10), slice(1,20)), new_vals)

  expect_equal(sub$as.array(),
               new_vals)

  # TODO: remove this expect-error once implemented.
  expect_error(a[1:10, 1:20]$set("...", new_vals),
                "Unsupported selection type")

  a[1:10, 1:20]$set(list(slice(1,10), slice(1,20)), new_vals)

  # it does not update the original array -- should it?
  expect_equal(a[1:10, 1:20]$as.array(),
               vals)

  a$set_item(list(slice(1,10), slice(1,20)), new_vals)

  expect_equal(a[1:10, 1:20]$as.array(),
               new_vals)
})

library(pizzarr)

test_that("can convert a NestedArray to a raw vector, little endian unsigned 4-byte int", {
    a <- array(data=1:10, dim=c(2, 5))
    na <- NestedArray$new(data = a, shape = dim(a), dtype = "<u4")
    na_raw <- na$flatten_to_raw()

    expect_equal(length(na_raw), 40)
    expected_out <- as.raw(c(
        0x01, 0x00, 0x00, 0x00,
        0x02, 0x00, 0x00, 0x00,
        0x03, 0x00, 0x00, 0x00,
        0x04, 0x00, 0x00, 0x00,
        0x05, 0x00, 0x00, 0x00,
        0x06, 0x00, 0x00, 0x00,
        0x07, 0x00, 0x00, 0x00,
        0x08, 0x00, 0x00, 0x00,
        0x09, 0x00, 0x00, 0x00,
        0x0a, 0x00, 0x00, 0x00
    ))
    expect_equal(na_raw, expected_out)
})

test_that("can convert a NestedArray to a raw vector, big endian unsigned 4-byte int", {
    a <- array(data=1:10, dim=c(2, 5))
    na <- NestedArray$new(data = a, shape = dim(a), dtype = ">u4")
    na_raw <- na$flatten_to_raw()

    expect_equal(length(na_raw), 40)
    expected_out <- as.raw(c(
        0x00, 0x00, 0x00, 0x01,
        0x00, 0x00, 0x00, 0x02,
        0x00, 0x00, 0x00, 0x03,
        0x00, 0x00, 0x00, 0x04,
        0x00, 0x00, 0x00, 0x05,
        0x00, 0x00, 0x00, 0x06,
        0x00, 0x00, 0x00, 0x07,
        0x00, 0x00, 0x00, 0x08,
        0x00, 0x00, 0x00, 0x09,
        0x00, 0x00, 0x00, 0x0a
    ))
    expect_equal(na_raw, expected_out)
})

test_that("can convert a raw vector to a NestedArray, little endian unsigned 4-byte int", {
    a_raw <- as.raw(c(
        0x01, 0x00, 0x00, 0x00,
        0x02, 0x00, 0x00, 0x00,
        0x03, 0x00, 0x00, 0x00,
        0x04, 0x00, 0x00, 0x00,
        0x05, 0x00, 0x00, 0x00,
        0x06, 0x00, 0x00, 0x00,
        0x07, 0x00, 0x00, 0x00,
        0x08, 0x00, 0x00, 0x00,
        0x09, 0x00, 0x00, 0x00,
        0x0a, 0x00, 0x00, 0x00
    ))
    na <- NestedArray$new(a_raw, shape=c(2, 5), dtype="<u4")

    expected_out <- array(data=1:10, dim=c(2, 5))
    expect_equal(na$data, expected_out)
})

test_that("can convert a raw vector to a NestedArray, big endian unsigned 4-byte int", {
    a_raw <- as.raw(c(
        0x00, 0x00, 0x00, 0x01,
        0x00, 0x00, 0x00, 0x02,
        0x00, 0x00, 0x00, 0x03,
        0x00, 0x00, 0x00, 0x04,
        0x00, 0x00, 0x00, 0x05,
        0x00, 0x00, 0x00, 0x06,
        0x00, 0x00, 0x00, 0x07,
        0x00, 0x00, 0x00, 0x08,
        0x00, 0x00, 0x00, 0x09,
        0x00, 0x00, 0x00, 0x0a
    ))
    na <- NestedArray$new(a_raw, shape=c(2, 5), dtype=">u4")

    expected_out <- array(data=1:10, dim=c(2, 5))
    expect_equal(na$data, expected_out)
})

test_that("can convert a NestedArray to a raw vector, little endian 4-byte float", {
    a <- array(data=as.double(1:10), dim=c(2, 5))
    na <- NestedArray$new(data = a, shape = dim(a), dtype = "<f4")
    na_raw <- na$flatten_to_raw()

    expect_equal(length(na_raw), 40)
    expected_out <- as.raw(c(
        0x00, 0x00, 0x80, 0x3f,
        0x00, 0x00, 0x00, 0x40,
        0x00, 0x00, 0x40, 0x40,
        0x00, 0x00, 0x80, 0x40,
        0x00, 0x00, 0xa0, 0x40,
        0x00, 0x00, 0xc0, 0x40,
        0x00, 0x00, 0xe0, 0x40,
        0x00, 0x00, 0x00, 0x41,
        0x00, 0x00, 0x10, 0x41,
        0x00, 0x00, 0x20, 0x41
    ))
    expect_equal(na_raw, expected_out)
})

test_that("can convert a NestedArray to a raw vector, big endian 4-byte float", {
    a <- array(data=as.double(1:10), dim=c(2, 5))
    na <- NestedArray$new(data = a, shape = dim(a), dtype = ">f4")
    na_raw <- na$flatten_to_raw()

    expect_equal(length(na_raw), 40)
    expected_out <- as.raw(c(
        0x3f, 0x80, 0x00, 0x00,
        0x40, 0x00, 0x00, 0x00,
        0x40, 0x40, 0x00, 0x00,
        0x40, 0x80, 0x00, 0x00,
        0x40, 0xa0, 0x00, 0x00,
        0x40, 0xc0, 0x00, 0x00,
        0x40, 0xe0, 0x00, 0x00,
        0x41, 0x00, 0x00, 0x00,
        0x41, 0x10, 0x00, 0x00,
        0x41, 0x20, 0x00, 0x00
    ))
    expect_equal(na_raw, expected_out)
})

test_that("can convert a NestedArray to a raw vector, with F ordering", {
    a <- array(data=1:10, dim=c(2, 5))
    na <- NestedArray$new(data = a, shape = dim(a), dtype = "<u4")
    na_raw <- na$flatten_to_raw(order = "F")

    # Either “C” or “F”, defining the layout of bytes within each chunk of the array.
    # “C” means row-major order, i.e., the last dimension varies fastest;
    # “F” means column-major order, i.e., the first dimension varies fastest.
    # Reference: https://zarr.readthedocs.io/en/stable/spec/v2.html#metadata

    expect_equal(length(na_raw), 40)
    expected_out <- as.raw(c(
        # Col 1
        0x01, 0x00, 0x00, 0x00,
        0x02, 0x00, 0x00, 0x00,
        # Col 2
        0x03, 0x00, 0x00, 0x00,
        0x04, 0x00, 0x00, 0x00,
        # Col 3
        0x05, 0x00, 0x00, 0x00,
        0x06, 0x00, 0x00, 0x00,
        # Col 4
        0x07, 0x00, 0x00, 0x00,
        0x08, 0x00, 0x00, 0x00,
        # Col 5
        0x09, 0x00, 0x00, 0x00,
        0x0a, 0x00, 0x00, 0x00
    ))
    expect_equal(na_raw, expected_out)
})

test_that("can convert a NestedArray to a raw vector, with C ordering", {
    a <- array(data=1:10, dim=c(2, 5))
    na <- NestedArray$new(data = a, shape = dim(a), dtype = "<u4")
    na_raw <- na$flatten_to_raw(order = "C")

    # Either “C” or “F”, defining the layout of bytes within each chunk of the array.
    # “C” means row-major order, i.e., the last dimension varies fastest;
    # “F” means column-major order, i.e., the first dimension varies fastest.
    # Reference: https://zarr.readthedocs.io/en/stable/spec/v2.html#metadata

    expect_equal(length(na_raw), 40)
    expected_out <- as.raw(c(
        # Row 1
        0x01, 0x00, 0x00, 0x00,
        0x03, 0x00, 0x00, 0x00,
        0x05, 0x00, 0x00, 0x00,
        0x07, 0x00, 0x00, 0x00,
        0x09, 0x00, 0x00, 0x00,
        # Row 2
        0x02, 0x00, 0x00, 0x00,
        0x04, 0x00, 0x00, 0x00,
        0x06, 0x00, 0x00, 0x00,
        0x08, 0x00, 0x00, 0x00,
        0x0a, 0x00, 0x00, 0x00
    ))
    expect_equal(na_raw, expected_out)
})

#' @keywords internal
path_to_prefix <- function(path) {
    # Reference: https://github.com/zarr-developers/zarr-python/blob/5dd4a0e6cdc04c6413e14f57f61d389972ea937c/zarr/_storage/store.py#L134
    # assume path already normalized
    if(!is.na(path) && stringr::str_length(path) > 0) {
        prefix <- paste0(path, '/')
    } else {
        prefix <- ""
    }
    return(prefix)
}

#' @keywords internal
contains_array <- function(store, path=NA) {
    # Return True if the store contains an array at the given logical path.
    # Checks V2 (.zarray) first, then V3 (zarr.json with node_type="array").
    path <- normalize_storage_path(path)
    prefix <- path_to_prefix(path)
    # V2 check
    key <- paste0(prefix, ARRAY_META_KEY)
    ret <- store$contains_item(key)
    if (!is.null(ret) && ret) return(TRUE)
    # V3 check
    return(contains_array_v3(store, path))
}

#' @keywords internal
contains_group <- function(store, path=NA) {
    # Return True if the store contains a group at the given logical path.
    # Checks V2 (.zgroup) first, then V3 (zarr.json with node_type="group").
    path <- normalize_storage_path(path)
    prefix <- path_to_prefix(path)
    # V2 check
    key <- paste0(prefix, GROUP_META_KEY)
    ret <- store$contains_item(key)
    if (!is.null(ret) && ret) return(TRUE)
    # V3 check
    return(contains_group_v3(store, path))
}

# V3 format detection functions.
# Reference: https://zarr-specs.readthedocs.io/en/latest/v3/core/v3.0.html
# In V3, both arrays and groups are identified by zarr.json with a node_type field.

# Fetch and parse zarr.json node_type at a given path.
# Returns the node_type string ("array" or "group") or NULL on failure.
# @keywords internal
get_v3_node_type <- function(store, path = NA) {
  path <- normalize_storage_path(path)
  prefix <- path_to_prefix(path)
  key <- paste0(prefix, ZARR_JSON)
  if (!store$contains_item(key)) return(NULL)

  tryCatch({
    raw_meta <- store$get_item(key)
    meta <- try_fromJSON(rawToChar(raw_meta), simplifyVector = FALSE)
    if (!is.null(meta$zarr_format) && meta$zarr_format == 3 &&
        !is.null(meta$node_type)) {
      meta$node_type
    } else {
      NULL
    }
  }, error = function(e) {
    NULL
  })
}

# Check if the store contains a V3 array at the given path.
# @keywords internal
contains_array_v3 <- function(store, path = NA) {
  identical(get_v3_node_type(store, path), "array")
}

# Check if the store contains a V3 group at the given path.
# @keywords internal
contains_group_v3 <- function(store, path = NA) {
  identical(get_v3_node_type(store, path), "group")
}

# Detect zarr format version at a given path.
# Reference: zarr-python checks zarr.json first (V3), then .zarray/.zgroup (V2).
# If both exist, V3 takes precedence with a warning.
#
# @param store A Store instance.
# @param path Storage path (NA for root).
# @return Integer 2L, 3L, or NA_integer_ if neither format detected.
# @keywords internal
detect_zarr_version <- function(store, path = NA) {
  path <- normalize_storage_path(path)
  prefix <- path_to_prefix(path)

  has_v3 <- store$contains_item(paste0(prefix, ZARR_JSON))
  has_v2_array <- store$contains_item(paste0(prefix, ARRAY_META_KEY))
  has_v2_group <- store$contains_item(paste0(prefix, GROUP_META_KEY))
  has_v2 <- has_v2_array || has_v2_group

  if (has_v3 && has_v2) {
    warning("Both V2 and V3 metadata found at path '", path,
            "'. Using V3.")
    return(3L)
  }
  if (has_v3) return(3L)
  if (has_v2) return(2L)
  NA_integer_
}

# Reference: https://github.com/zarr-developers/zarr-python/blob/5dd4a0/zarr/attrs.py#L7

#' The Zarr Attributes class.
#' @title Attributes Class
#' @docType class
#' @description
#' Class providing access to user attributes on an array or group.
#'
#' @format [R6::R6Class]
#' @rdname Attributes
#' @export
Attributes <- R6::R6Class("Attributes",
  private = list(

    cached_aslist = NULL,
    zarr_format = NULL,

    get_nosync = function() {
      if (!is.null(private$zarr_format) && private$zarr_format == 3L) {
        # V3: read attributes from zarr.json
        zarr_json_key <- sub(ATTRS_KEY, ZARR_JSON, self$key, fixed = TRUE)
        raw_meta <- tryCatch({
          self$store$get_item(zarr_json_key)
        }, error = function(cond) {
          if (is_key_error(cond)) return(NULL)
          stop(cond)
        })
        if (is.null(raw_meta)) return(obj_list())
        meta <- try_fromJSON(rawToChar(raw_meta), simplifyVector = FALSE)
        if (is.null(meta$attributes)) return(obj_list())
        return(meta$attributes)
      }
      # V2: read from .zattrs
      attrs_list <- tryCatch({
        # TODO: use consolidated metadata?
        return(self$store$metadata_class$decode_metadata(self$store$get_item(self$key), auto_unbox = TRUE))
      }, error = function(cond) {
        if(is_key_error(cond)) {
          return(obj_list())
        }
        stop(cond)
      })
      return(attrs_list)
    },
    put_nosync = function(d) {
      if (!is.null(private$zarr_format) && private$zarr_format == 3L) {
        # V3: read-modify-write zarr.json (attributes are embedded)
        zarr_json_key <- sub(ATTRS_KEY, ZARR_JSON, self$key, fixed = TRUE)
        raw_meta <- self$store$get_item(zarr_json_key)
        meta <- try_fromJSON(rawToChar(raw_meta), simplifyVector = FALSE)
        meta$attributes <- d
        encoded <- charToRaw(jsonlite::toJSON(meta, auto_unbox = TRUE))
        self$store$set_item(zarr_json_key, encoded)
        if (self$cache) private$cached_aslist <- d
        return(invisible(NULL))
      }
      self$store$set_item(self$key, self$store$metadata_class$encode_metadata(d, auto_unbox = TRUE))
      if(self$cache) {
        private$cached_aslist <- d
      }
    },
    set_item_nosync = function(item, value) {
      d <- private$get_nosync()
      d[[item]] <- value
      private$put_nosync(d)
    },
    del_item_nosync = function(item) {
      d <- private$get_nosync()
      d[[item]] <- NULL
      private$put_nosync(d)
    }
  ),
  public = list(
    #' @field store Attributes store, already initialized.
    store = NULL,
    #' @field key The key under which the attributes will be stored.
    key = NULL,
    #' @field read_only If True, attributes cannot be modified.
    read_only = NULL,
    #' @field cache If True (default), attributes will be cached locally.
    cache = NULL,
    #' @field synchronizer Only necessary if attributes may be modified from multiple threads or processes.
    synchronizer = NULL,
    #' @description
    #' Create a new Attributes instance.
    #' @param store ([Store])\cr
    #'   Attributes store, already initialized.
    #' @param key (`character(1)`)\cr
    #'   Key to use for attributes (`.zattrs` is default).
    #' @param read_only (`logical(1)`)\cr
    #'   Whether the attributes are read-only.
    #' @param cache (`logical(1)`)\cr
    #'   Whether to cache attributes.
    #' @param synchronizer (`ANY` or `NA`)\cr
    #'   Synchronizer object.
    #' @param zarr_format (`integer(1)` or `NULL`)\cr
    #'   Zarr format version: `2L` for V2 (`.zattrs`), `3L` for V3 (zarr.json).
    #' @return An `Attributes` instance.
    initialize = function(store, key = NA, read_only = FALSE, cache = TRUE, synchronizer = NA, zarr_format = NULL) {
      if(is_na(key)) {
        key <- ATTRS_KEY
      }
      self$store <- store
      self$key <- key
      self$read_only <- read_only
      self$cache <- cache
      private$cached_aslist <- NA
      private$zarr_format <- zarr_format

      check_cached <- try_from_zmeta(key, store)

      if(cache & !is.null(check_cached)) {
        private$cached_aslist <- check_cached
      }

      self$synchronizer <- synchronizer
    },
    #' @description
    #' convert attributes to list
    #' @return `list()`.
    to_list = function() {
      if(self$cache && !is_na(private$cached_aslist)) {
        return(private$cached_aslist)
      }
      d <- private$get_nosync()
      if(self$cache) {
        private$cached_aslist <- d
      }
      return(d)
    },
    #' @description
    #' refresh attributes
    #' @return `NULL` (called for side effects).
    refresh = function() {
      if(self$cache) {
        new_val <- private$get_nosync()
        
        private$cached_aslist <- new_val
      }
    },
    #' @description
    #' check if object contains item
    #' @param x Object to test.
    #' @return `logical(1)`.
    contains = function(x) {
      return(x %in% names(self$to_list()))
    },
    #' @description
    #' get attribute
    #' @param item Character attribute name.
    #' @return The attribute value.
    get_item = function(item) {
      return(self$to_list()[[item]])
    },
    #' @description
    #' set attribute
    #' @param item Character attribute name.
    #' @param value Value to add or update.
    #' @return `NULL` (called for side effects).
    set_item = function(item, value) {
      # TODO: support synchronizer
      private$set_item_nosync(item, value)
    },
    #' @description
    #' delete attribute
    #' @param item Character attribute name.
    #' @return `NULL` (called for side effects).
    del_item = function(item) {
      # TODO: support synchronizer
      private$del_item_nosync(item)
    },
    #' @description
    #' Set cached attributes from V3 embedded metadata.
    #' In V3, attributes are part of zarr.json rather than a separate .zattrs file.
    #' This method pre-populates the cache so the normal .zattrs read path is skipped.
    #' @param attrs_list A named list of attributes from V3 zarr.json.
    #' @return `NULL` (modifies cache in place).
    set_cached_v3_attrs = function(attrs_list) {
      if (is.null(attrs_list) || length(attrs_list) == 0) {
        attrs_list <- obj_list()
      }
      private$cached_aslist <- attrs_list
    }
  )
)

library(pizzarr)

test_that("Can get attributes for empty store (without throwing a KeyError)", {
    store <- MemoryStore$new()
    g <- zarr_open_group(store)
    
    expect_equal(g$get_attrs()$to_list(), obj_list())
})

test_that("Can set attributes", {
    store <- MemoryStore$new()
    g <- zarr_open_group(store)
    attrs <- g$get_attrs()

    attrs$set_item("foo", "bar")
    expect_equal(attrs$to_list(), obj_list(foo = "bar"))

    attrs$set_item("baz", "qux")
    expect_equal(attrs$to_list(), obj_list(foo = "bar", baz = "qux"))
})

test_that("Can update attributes", {
    store <- MemoryStore$new()
    g <- zarr_open_group(store)
    attrs <- g$get_attrs()

    attrs$set_item("foo", "bar")
    expect_equal(attrs$to_list(), obj_list(foo = "bar"))

    attrs$set_item("baz", "qux")
    expect_equal(attrs$to_list(), obj_list(foo = "bar", baz = "qux"))

    attrs$set_item("foo", "baz")
    expect_equal(attrs$to_list(), obj_list(foo = "baz", baz = "qux"))
})

test_that("Can delete attributes", {
    store <- MemoryStore$new()
    g <- zarr_open_group(store)
    attrs <- g$get_attrs()

    attrs$set_item("foo", "bar")
    expect_equal(attrs$to_list(), obj_list(foo = "bar"))

    attrs$set_item("baz", "qux")
    expect_equal(attrs$to_list(), obj_list(foo = "bar", baz = "qux"))

    attrs$del_item("foo")
    expect_equal(attrs$to_list(), obj_list(baz = "qux"))
    
    attrs$del_item("baz")
    expect_equal(attrs$to_list(), obj_list())
})

test_that("Can set attributes with cache FALSE", {
    store <- MemoryStore$new()
    g <- zarr_open_group(store)
    attrs <- g$get_attrs()
    attrs$cache <- FALSE

    attrs$set_item("foo", "bar")
    expect_equal(attrs$to_list(), obj_list(foo = "bar"))

    attrs$set_item("baz", "qux")
    expect_equal(attrs$to_list(), obj_list(foo = "bar", baz = "qux"))
})

test_that("Can update attributes with cache FALSE", {
    store <- MemoryStore$new()
    g <- zarr_open_group(store)
    attrs <- g$get_attrs()
    attrs$cache <- FALSE

    attrs$set_item("foo", "bar")
    expect_equal(attrs$to_list(), obj_list(foo = "bar"))

    attrs$set_item("baz", "qux")
    expect_equal(attrs$to_list(), obj_list(foo = "bar", baz = "qux"))

    attrs$set_item("foo", "baz")
    expect_equal(attrs$to_list(), obj_list(foo = "baz", baz = "qux"))
})

test_that("Can delete attributes with cache FALSE", {
    store <- MemoryStore$new()
    g <- zarr_open_group(store)
    attrs <- g$get_attrs()
    attrs$cache <- FALSE

    attrs$set_item("foo", "bar")
    expect_equal(attrs$to_list(), obj_list(foo = "bar"))

    attrs$set_item("baz", "qux")
    expect_equal(attrs$to_list(), obj_list(foo = "bar", baz = "qux"))

    attrs$del_item("foo")
    expect_equal(attrs$to_list(), obj_list(baz = "qux"))
    
    attrs$del_item("baz")
    expect_equal(attrs$to_list(), obj_list())
})

if (requireNamespace("vcr", quietly = TRUE)) {
  library("vcr") # *Required* as vcr is set up on loading
  invisible(vcr::vcr_configure(
    dir = vcr::vcr_test_path("fixtures"),
    match_requests_on = c("method", "uri")
  ))
}

# Reference: https://github.com/gzuidhof/zarr.js/blob/master/src/core/types.ts

#' @title Chunk Dimension Projection
#'
#' @description
#' A mapping from chunk to output array for a single dimension.
#' Used internally to track which items within a single chunk dimension
#' correspond to which items in the target output array.
#'
#' @format [R6::R6Class] object.
#'
#' @family Indexing classes
#' @keywords internal
ChunkDimProjection <- R6::R6Class("ChunkDimProjection",
  public = list(
    #' @field dim_chunk_index (`integer(1)`)\cr
    #'   Index of chunk.
    dim_chunk_index = NULL,
    #' @field dim_chunk_sel (`integer()` | `list()`)\cr
    #'   Selection of items from chunk array.
    dim_chunk_sel = NULL,
    #' @field dim_out_sel (`integer()` | `list()`)\cr
    #'   Selection of items in target (output) array.
    dim_out_sel = NULL,
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    #'
    #' @param dim_chunk_index (`integer(1)`)\cr
    #'   Index of chunk.
    #' @param dim_chunk_sel (`integer()` | `list()`)\cr
    #'   Selection of items from chunk array.
    #' @param dim_out_sel (`integer()` | `list()`)\cr
    #'   Selection of items in target (output) array.
    initialize = function(dim_chunk_index, dim_chunk_sel, dim_out_sel) {
      self$dim_chunk_index <- dim_chunk_index
      self$dim_chunk_sel <- dim_chunk_sel
      self$dim_out_sel <- dim_out_sel
    }
  )
)

#' @title Chunk Projection
#'
#' @description
#' A mapping of items from chunk to output array. Can be used to extract
#' items from the chunk array for loading into an output array. Can also be
#' used to extract items from a value array for setting/updating in a chunk
#' array.
#'
#' @format [R6::R6Class] object.
#'
#' @family Indexing classes
#' @keywords internal
ChunkProjection <- R6::R6Class("ChunkProjection",
  public = list(
    #' @field chunk_coords (`integer()`)\cr
    #'   Indices of chunk.
    chunk_coords = NULL,
    #' @field chunk_sel (`list()`)\cr
    #'   Selection of items from chunk array.
    chunk_sel = NULL,
    #' @field out_sel (`list()`)\cr
    #'   Selection of items in target (output) array.
    out_sel = NULL,
    #' @description
    #' Creates a new instance of this [R6][R6::R6Class] class.
    #'
    #' @param chunk_coords (`integer()`)\cr
    #'   Indices of chunk.
    #' @param chunk_sel (`list()`)\cr
    #'   Selection of items from chunk array.
    #' @param out_sel (`list()`)\cr
    #'   Selection of items in target (output) array.
    initialize = function(chunk_coords, chunk_sel, out_sel) {
      self$chunk_coords <- chunk_coords
      self$chunk_sel <- chunk_sel
      self$out_sel <- out_sel
    }
  )
)

#' @title Indexer
#'
#' @description
#' Abstract base class for chunk indexers. An indexer maps a user-level
#' array selection to a sequence of [ChunkProjection] objects, each
#' describing which items to read or write within a single chunk.
#' Concrete subclasses include [BasicIndexer] and [OrthogonalIndexer].
#'
#' @format [R6::R6Class] object.
#'
#' @family Indexing classes
#' @keywords internal
Indexer <- R6::R6Class("Indexer",
  public = list(
    #' @field shape (`integer()`)\cr
    #'   Shape of the selection.
    shape = NULL,
    #' @field drop_axes (`integer()` | `NULL`)\cr
    #'   Axes to drop from the result.
    drop_axes = NULL,
    #' @description
    #' Iterate over [ChunkProjection] objects for the selection.
    #'
    #' @return `list()` of [ChunkProjection] objects.
    iter = function() {
      # To be implemented in child class
    }
  )
)

#' @title Dimension Indexer
#'
#' @description
#' Abstract base class for per-dimension chunk indexers. A dimension
#' indexer maps a selection on a single array dimension to a sequence of
#' [ChunkDimProjection] objects. Concrete subclasses include
#' [IntDimIndexer], [SliceDimIndexer], [IntArrayDimIndexer], and
#' [BoolArrayDimIndexer].
#'
#' @format [R6::R6Class] object.
#'
#' @family Indexing classes
#' @keywords internal
DimIndexer <- R6::R6Class("DimIndexer",
  public = list(
    #' @field num_items (`integer(1)`)\cr
    #'   Number of items in the selection.
    num_items = NULL,
    #' @description
    #' Iterate over [ChunkDimProjection] objects for this dimension.
    #'
    #' @return `list()` of [ChunkDimProjection] objects.
    iter = function() {
      # To be implemented in child class
    }
  )
)

#' Abstract store for Zarr
#' @title Store Class
#' @docType class
#' @description
#' Class representing an abstract store
#'
#' @format [R6::R6Class]
#' @family Store classes
#' @rdname Store
#' @export
Store <- R6::R6Class("Store",
   private = list(
      readable = NULL,
      writeable = NULL,
      erasable = NULL,
      listable = NULL,
      store_version = NULL,
      zmetadata = NULL,
      #' @keywords internal
      listdir_from_keys = function(path) {
        stop("NotImplementedError(listdir_from_keys)")
      },
      #' @keywords internal
      rename_from_keys = function(src_path, dst_path) {
        stop("NotImplementedError(rename_from_keys)")
      },
      #' @keywords internal
      rmdir_from_keys = function(path) {
        stop("NotImplementedError(rmdir_from_keys)")
      }
   ),
   public = list(
    #' @field metadata_class The metadata encoder/decoder for this store.
    #' @keywords internal
    metadata_class = NULL,
    #' @description 
    #' Create a `Store` object 
    initialize = function() {
      private$readable <- TRUE
      private$writeable <- TRUE
      private$erasable <- TRUE
      private$listable <- TRUE
      private$store_version <- 2
      self$metadata_class <- Metadata2$new()
    },
    #' @description
    #' Test if Store is readable.
    #' @return Logical.
    is_readable = function() {
      return(private$readable)
    },
    #' @description
    #' Test if Store is writeable.
    #' @return Logical.
    is_writeable = function() {
      return(private$writeable)
    },
    #' @description
    #' Test if Store is eraseable.
    #' @return Logical.
    is_erasable = function() {
      return(private$erasable)
    },
    #' @description
    #' Test if Store is listable.
    #' @return Logical.
    is_listable = function() {
      return(private$listable)
    },
    #' @description
    #' Close the store.
    #' @return `NULL`.
    close = function() {
      # Do nothing by default
    },
    #' @description
    #' List the store directory.
    #' @param path character path.
    #' @return Character vector of keys.
    listdir = function(path=NA) {
      if(is.na(path)) {
        path <- ""
      }
      path <- normalize_storage_path(path)
      return(private$listdir_from_keys(path))
    },
    #' @description
    #' Rename a Store path.
    #' @param src_path character source path.
    #' @param dst_path character destination path.
    #' @return `NULL` (called for side effects).
    rename = function(src_path, dst_path) {
      if(!self$is_erasable()) {
        stop("Store is not erasable, cannot call 'rename'")
      }
      private$rename_from_keys(src_path, dst_path)
    },
    #' @description
    #' Remove a path within a Store.
    #' @param path character path.
    #' @return `NULL` (called for side effects).
    rmdir = function(path) {
      if(!self$is_erasable()) {
        stop("Store is not erasable, cannot call 'rmdir'")
      }
      path <- normalize_storage_path(path)
      private$rmdir_from_keys(path)
    },
     #' @description
     #' Get an item from the store.
     #' @param key The item key.
     #' @return The item data in a vector of type raw.
     get_item = function(key) {

     },
     #' @description
     #' Set an item in the store.
     #' @param key The item key.
     #' @param value The item value as a vector of type raw.
     #' @return `NULL` (called for side effects).
     set_item = function(key, value) {

     },
     #' @description
     #' Determine whether the store contains an item.
     #' @param key The item key.
     #' @return A boolean value.
     contains_item = function(key) {

     },
    #' @description
    #' Delete an item from the store.
    #' @param key The item key.
    #' @return `NULL` (called for side effects).
    delete_item = function(key) {
      stop("NotImplementedError(delete_item)")
    },
    #' @description
    #' Get consolidated metadata if it exists.
    #' @return A list or `NULL`.
    get_consolidated_metadata = function() {
      return(private$zmetadata)
    },
    #' @description
    #' Return a store identifier for zarrs dispatch.
    #'
    #' Subclasses override this to return a filesystem path or URL.
    #' Returns `NULL` by default (signals: use R-native path).
    #' @return A character string or `NULL`.
    get_store_identifier = function() {
      NULL
    },
    #' @description
    #' Print a human-readable summary of the store.
    #' @param ... Ignored.
    #' @return `self` (invisibly).
    print = function(...) {
      cat(paste0("<", class(self)[1], ">\n"))
      invisible(self)
    }
   )
)

# Reference: https://github.com/zarr-developers/zarr_implementations/blob/c0bd932/generate_data/js/src/fsstore.js#L7

#' DirectoryStore for Zarr
#' @title DirectoryStore Class
#' @docType class
#' @description
#' Store class using directories and files on a standard file system.
#'
#' @format [R6::R6Class] inheriting from [Store].
#' @family Store classes
#' @rdname DirectoryStore
#' @export
DirectoryStore <- R6::R6Class("DirectoryStore",
  inherit = Store,
  public = list(
    #' @field root The path to the root of the store.
    #' @keywords internal
    root = NULL,
    #' @description
    #' Create a new file system store.
    #' @param root The path to the root of the store.
    #' @return A new `DirectoryStore` object.
    initialize = function(root) {
      super$initialize()
      self$root <- root
      if(!dir.exists(root)) {
        dir.create(root, recursive = TRUE, showWarnings = FALSE)
      }
    },
    #' @description
    #' Get an item from the store.
    #' @param key The item key.
    #' @return The item data in a vector of type raw.
    get_item = function(key) {
      fp <- file.path(self$root, key)
      if(!file.exists(fp)) {
        stop("KeyError: ", key)
      }
      fp_size <- file.info(fp)$size
      fp_pointer <- file(fp, "rb")
      on.exit(close(fp_pointer), add = TRUE)
      fp_data <- readBin(fp_pointer, what = "raw", n = fp_size)
      return(fp_data)
    },
    #' @description
    #' Set an item in the store.
    #' @param key The item key.
    #' @param value The item value as a vector of type raw.
    #' @return `NULL` (called for side effects).
    set_item = function(key, value) {
      fp <- file.path(self$root, key)
      dir.create(dirname(fp), recursive = TRUE, showWarnings = FALSE)
      fp_pointer <- file(fp, "wb")
      on.exit(close(fp_pointer), add = TRUE)
      writeBin(value, fp_pointer)
    },
    #' @description
    #' Determine whether the store contains an item.
    #' @param key The item key.
    #' @return A boolean value.
    contains_item = function(key) {
      fp <- file.path(self$root, key)
      return(file.exists(fp))
    },
    #' @description
    #' Delete an item from the store.
    #' @param key The item key.
    #' @return `NULL` (called for side effects).
    delete_item = function(key) {
      fp <- file.path(self$root, key)
      if(file.exists(fp)) {
        file.remove(fp)
      }
    },
    #' @description
    #' Remove a path within a Store.
    #' @param path Character path.
    #' @return `NULL` (called for side effects).
    rmdir = function(path=NA) {
      path <- normalize_storage_path(path)
      dir_path <- self$root
      if(!is.na(path)) {
        dir_path <- file.path(self$root, path)
      }
      if(dir.exists(dir_path)) {
        unlink(dir_path, recursive = TRUE)
      }
    },
    #' @description
    #' List the store directory.
    #' @param path Character path.
    #' @return `character()` vector of entries.
    listdir = function(path=NA) {
      if(is_na(path)) {
        dir_path <- self$root
      } else {
        dir_path <- file.path(self$root, path)
      }
      if(!dir.exists(dir_path)) {
        stop("KeyError: ", path)
      }
      dir_list <- sort(list.files(dir_path, full.names = FALSE, all.files = TRUE, include.dirs = TRUE))
      dir_list <- dir_list[!dir_list %in% c(".", "..")]
      return(dir_list)
    },
    #' @description
    #' Print a human-readable summary of the store.
    #' @description
    #' Return the absolute filesystem path for zarrs dispatch.
    #' @return A character string.
    get_store_identifier = function() {
      normalizePath(self$root, mustWork = FALSE)
    },
    #' @param ... Ignored.
    #' @return `self` (invisibly).
    print = function(...) {
      cat(paste0("<DirectoryStore> file://", self$root, "\n"))
      invisible(self)
    }
  )
)


# Reference: https://github.com/zarr-developers/zarr-python/blob/a5dfc3b4/zarr/storage.py#L512

#' MemoryStore for Zarr
#' @title MemoryStore Class
#' @docType class
#' @description
#' Store class that uses a hierarchy of list objects,
#' thus all data will be held in main memory.
#'
#' @format [R6::R6Class] inheriting from [Store].
#' @family Store classes
#' @rdname MemoryStore
#' @export
MemoryStore <- R6::R6Class("MemoryStore",
   inherit = Store,
   public = list(
     #' @field root The root list for the store.
     #' @keywords internal
     root = NULL,
     #' @description
     #' Create a new memory store.
     #' @return A new `MemoryStore` object.
     initialize = function() {
      super$initialize()
       self$root <- obj_list()
     },
     #' @description
     #' Get the parent of an item.
     #' @keywords internal
     #' @param item The item key.
     #' @return A list with the keys `parent` and `key`.
     get_parent = function(item) {
       parent <- self$root
       segments <- strsplit(item, "/")[[1]]
       if (length(segments) > 1) {
         for(k in segments[1:(length(segments)-1)]) {
           parent <- parent[[k]]
           if(!is.list(parent)) {
             stop("KeyError: ", item)
           }
         }
       }
       return(list(parent = parent, key = segments[length(segments)]))
     },
     #' @description
     #' Get an item from the store.
     #' @param item The item key.
     #' @return The item data in a vector of type raw.
     get_item = function(item=NA) {
       if(is_na(item)) {
        return(self$root)
       }
       parent_and_key <- self$get_parent(item)
       parent <- parent_and_key$parent
       key <- parent_and_key$key

       if(key %in% names(parent)) {
         value <- parent[[key]]
       } else {
         stop("KeyError: ", item)
       }
       return(value)
     },
     #' @description
     #' Set an item in the store.
     #' @param item The item key.
     #' @param value The item value as a vector of type raw.
     #' @return `NULL` (called for side effects).
     set_item = function(item, value) {
       segments <- strsplit(item, "/")[[1]]
       if(length(segments) > 1) {
         for(i in 1:(length(segments)-1)) {
           k <- segments[i]
           if(i == 1 && k %in% names(self$root)) {
             if(!is.list(self$root[[k]])) {
               stop("KeyError: ", item)
             }
           } else if(i > 1 && k %in% names(self$root[[segments[1:(i-1)]]])) {
             if(!is.list(self$root[[segments[1:i]]])) {
               stop("KeyError: ", item)
             }
           } else {
             self$root[[segments[1:i]]] <- obj_list()
           }
         }
       }
       self$root[[segments]] <- value
     },
     #' @description
     #' Determine whether the store contains an item.
     #' @param item The item key.
     #' @return A boolean value.
     contains_item = function(item) {
       result <- tryCatch({
         parent_and_key <- self$get_parent(item)
         parent <- parent_and_key$parent
         key <- parent_and_key$key
         if(key %in% names(parent)) {
           value <- parent[[key]]
           return(!is.list(value))
         } else {
           return(FALSE)
         }
       }, error = function(e) {
         return(FALSE)
       })
       return(result)
     },
     #' @description
     #' List the store directory.
     #' @param path Character path.
     #' @return `character()` vector of entries.
     listdir = function(path=NA) {
      item <- self$get_item(path)
      if(!is.list(item)) {
        stop("KeyError: ", path)
      }
      return(sort(names(item)))
     },
     #' @description
     #' Delete an item from the store.
     #' @param key The item key.
     #' @return `NULL` (called for side effects).
     delete_item = function(key) {
      self$set_item(key, NULL)
     },
     #' @description
     #' Remove a path within a Store.
     #' @param item Character item.
     #' @return `NULL` (called for side effects).
     rmdir = function(item) {
      if(is_na(item) || item == "") {
        self$root <- obj_list()
      } else {
        self$set_item(item, NULL)
      }
     },
     #' @description
     #' Print a human-readable summary of the store.
     #' @param ... Ignored.
     #' @return `self` (invisibly).
     print = function(...) {
       cat("<MemoryStore>\n")
       invisible(self)
     }
   )
)

#' @keywords internal
item_to_key <- function(item) {
  # Remove leading slash if necessary.
  if(substr(item, 1, 1) == "/") {
    key <- substr(item, 2, length(item))
  } else {
    key <- item
  }
  key
}

# Reference: https://github.com/manzt/zarrita.js/blob/main/packages/storage/src/fetch.ts

#' @title HttpStore Class
#' @docType class
#' @description
#' Store class that uses HTTP requests.
#' Read-only. Depends on the `crul` package.
#' @format [R6::R6Class] inheriting from [Store].
#' @family Store classes
#' @rdname HttpStore
#' @importFrom memoise memoise timeout
#' @export
HttpStore <- R6::R6Class("HttpStore",
  inherit = Store,
  private = list(
    url = NULL,
    base_path = NULL,
    domain = NULL,
    options = NULL,
    headers = NULL,
    client = NULL,
    zmetadata = NULL,
    make_request_memoized = NULL,
    cache_enabled = NULL,
    cache_time_seconds = NULL,
    make_request = function(item) {
      key <- item_to_key(item)
      path <- paste(private$base_path, key, sep="/")

      ret <- try_from_zmeta(key, self)

      if(!is.null(ret)) return(ret)

      tryCatch(private$client$get(path = path),
               error = function(e) {
                 warning("Can't proceed, web request failed for '", key,
                         "'. Error was: ", conditionMessage(e))
                 NULL
               })
    },
    memoize_make_request = function() {
      if(private$cache_enabled) {
        private$make_request_memoized <-  memoise(
          function(key) private$make_request(key), 
          ~timeout(private$cache_time_seconds)
        )
      } else {
        private$make_request_memoized <- private$make_request
      }
    },
    get_zmetadata = function() {
      res <- private$make_request(".zmetadata")
      
      if(!is.null(res$status_code) && res$status_code == 200) {
        out <- try_fromJSON(res$parse("UTF-8"))
      } else out <- NULL
      
      return(out)
    }
  ),
  public = list(
    #' @description
    #' Create a `HttpStore` object
    #' @param url (`character(1)`)\cr
    #'   URL of the store.
    #' @param options (`list()` or `NA`)\cr
    #'   Options passed to crul.
    #' @param headers (`list()` or `NA`)\cr
    #'   Headers passed to crul.
    #' @return A new `HttpStore` object.
    initialize = function(url, options = NA, headers = NA) {
      super$initialize()
      # Remove trailing slash if necessary.
      if(substr(url, nchar(url), nchar(url)) == "/") {
        private$url <- substr(url, 1, nchar(url)-1)
      } else {
        private$url <- url
      }
      private$options <- options
      private$headers <- headers

      segments <- stringr::str_split(private$url, "/")[[1]]
      private$domain <- paste(segments[1:3], collapse="/")

      # Support both cases in which the store is located at the root
      # of the domain, or located under some base_path.

      private$base_path <- ifelse(length(segments) == 3, "", 
                                  paste(segments[4:length(segments)], collapse="/"))
      
      if(!requireNamespace("crul", quietly = TRUE)) {
        stop("HttpStore requires the crul package")
      }

      private$client <- crul::HttpClient$new(
        url = private$domain,
        opts = private$options,
        headers = private$headers
      )

      private$cache_time_seconds <- getOption("pizzarr.http_store_cache_time_seconds")
      private$cache_enabled <- private$cache_time_seconds > 0

      private$memoize_make_request()
      
      private$zmetadata <- private$get_zmetadata()
    },
    #' @description
    #' Get an item from the store.
    #' @param item The item key.
    #' @return The item data in a vector of type raw.
    get_item = function(item) {
      res <- private$make_request_memoized(item)
      return(res$content)
    },
    #' @description
    #' Determine whether the store contains an item.
    #' @param item The item key.
    #' @return A boolean value.
    contains_item = function(item) {
      # use consolidated metadata if it exists
      if(!is.null(try_from_zmeta(item_to_key(item), self))) {
        return(TRUE)
      } else if(!is.null(self$get_consolidated_metadata())) {
        return(FALSE)
      } else {
        res <- private$make_request_memoized(item)
        
        return(!is.null(res$status_code) && res$status_code == 200)        
      }

    },
    #' @description
    #' Fetches .zmetadata from the store evaluates its names
    #' @return Character vector of unique keys that do not start with a `.`.
    listdir = function() {

      if(!is.null(private$zmetadata)) {
        tryCatch({
          out <- names(private$zmetadata$metadata) |>
            stringr::str_subset("^\\.", negate = TRUE) |>
            stringr::str_split("/") |>
            vapply(\(x) head(x, 1), "") |>
            unique()
        }, error = \(e) warning("\n\nError parsing .zmetadata:\n\n", e))
      } else {
        out <- NULL
        message(".zmetadata not found for this http store. Can't listdir")
      }
      
      return(out)
      
    },
    #' @description
    #' Get cache time of http requests.
    #' @return `numeric(1)`.
    get_cache_time_seconds = function() {
      return(private$cache_time_seconds)
    },
    #' @description
    #' Set cache time of http requests.
    #' @param seconds Number of seconds until cache is invalid -- 0 for no cache.
    #' @return `NULL` (called for side effects).
    set_cache_time_seconds = function(seconds) {
      private$cache_time_seconds <- seconds
      # We need to re-memoize.
      private$memoize_make_request()
    },
    #' @description
    #' Print a human-readable summary of the store.
    #' @description
    #' Return the store URL for zarrs dispatch.
    #' @return A character string.
    get_store_identifier = function() {
      private$url
    },
    #' @param ... Ignored.
    #' @return `self` (invisibly).
    print = function(...) {
      cat(paste0("<HttpStore> ", private$url, "\n"))
      invisible(self)
    }
  )
)

#' S3 Store for Zarr (zarrs backend)
#' @title S3Store Class
#' @docType class
#' @description
#' Thin store wrapper for S3 URLs. All I/O is delegated to the zarrs Rust
#' backend via `object_store`. Requires the `s3` compiled feature
#' (r-universe tier).
#'
#' @format [R6::R6Class]
#' @family Store classes
#' @rdname S3Store
#' @export
S3Store <- R6::R6Class("S3Store",
  inherit = Store,
  private = list(
    url = NULL
  ),
  public = list(
    #' @description
    #' Create an `S3Store`.
    #' @param url Character. S3 URL (e.g., `"s3://bucket/prefix"`).
    initialize = function(url) {
      super$initialize()
      if (!grepl("^s3://", url)) {
        stop("S3Store requires an s3:// URL, got: ", url)
      }
      private$url <- url
      private$writeable <- FALSE
      private$erasable <- FALSE
      private$listable <- FALSE
    },
    #' @description
    #' Return the S3 URL for zarrs dispatch.
    #' @return A character string.
    get_store_identifier = function() {
      private$url
    },
    #' @description
    #' Print a human-readable summary of the store.
    #' @param ... Ignored.
    #' @return `self` (invisibly).
    print = function(...) {
      cat(paste0("<S3Store> ", private$url, "\n"))
      invisible(self)
    }
  )
)

#' GCS Store for Zarr (zarrs backend)
#' @title GcsStore Class
#' @docType class
#' @description
#' Thin store wrapper for Google Cloud Storage URLs. All I/O is delegated
#' to the zarrs Rust backend via `object_store`. Requires the `gcs`
#' compiled feature (r-universe tier).
#'
#' @format [R6::R6Class]
#' @family Store classes
#' @rdname GcsStore
#' @export
GcsStore <- R6::R6Class("GcsStore",
  inherit = Store,
  private = list(
    url = NULL
  ),
  public = list(
    #' @description
    #' Create a `GcsStore`.
    #' @param url Character. GCS URL (e.g., `"gs://bucket/prefix"`).
    initialize = function(url) {
      super$initialize()
      if (!grepl("^gs://", url)) {
        stop("GcsStore requires a gs:// URL, got: ", url)
      }
      private$url <- url
      private$writeable <- FALSE
      private$erasable <- FALSE
      private$listable <- FALSE
    },
    #' @description
    #' Return the GCS URL for zarrs dispatch.
    #' @return A character string.
    get_store_identifier = function() {
      private$url
    },
    #' @description
    #' Print a human-readable summary of the store.
    #' @param ... Ignored.
    #' @return `self` (invisibly).
    print = function(...) {
      cat(paste0("<GcsStore> ", private$url, "\n"))
      invisible(self)
    }
  )
)

# --- zarrs S3 store tests ---
# These tests require network access and the s3 compiled feature.
# Skipped on CRAN and when offline.

test_that("s3 feature is compiled", {
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")
  features <- pizzarr_compiled_features()
  skip_if(!("s3" %in% features), "s3 feature not compiled")
  expect_true("s3" %in% features)
  expect_true("object_store" %in% features)
})

test_that("zarrs reads metadata from S3 store", {
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")
  skip_if(!("s3" %in% pizzarr_compiled_features()), "s3 feature not compiled")
  skip_on_cran()
  skip_if_offline()

  # OME-Zarr SciVis bonsai V2 (zstd, uint8, public, us-east-1)
  url <- "s3://ome-zarr-scivis/v0.4/64x0/bonsai.ome.zarr"

  meta <- zarrs_open_array_metadata(url, "scale0/bonsai")
  expect_equal(meta$zarr_format, 2L)
  expect_true(length(meta$shape) > 0)

  zarrs_close_store(url)
})

test_that("zarrs reads subset from S3 store", {
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")
  skip_if(!("s3" %in% pizzarr_compiled_features()), "s3 feature not compiled")
  skip_on_cran()
  skip_if_offline()

  url <- "s3://ome-zarr-scivis/v0.4/64x0/bonsai.ome.zarr"

  meta <- zarrs_open_array_metadata(url, "scale0/bonsai")
  ndim <- length(meta$shape)

  # Read a small subset: first element along each dimension
  ranges <- lapply(seq_len(ndim), function(i) c(0L, 1L))
  result <- zarrs_get_subset(url, "scale0/bonsai", ranges, NULL)
  expect_true(length(result$data) == 1)
  expect_equal(length(result$shape), ndim)

  zarrs_close_store(url)
})

test_that("S3Store class works", {
  s3 <- S3Store$new("s3://bucket/prefix")
  expect_equal(s3$get_store_identifier(), "s3://bucket/prefix")
  expect_false(s3$is_writeable())
  expect_output(print(s3), "<S3Store> s3://bucket/prefix")
  expect_error(S3Store$new("https://example.com"), "s3://")
})

test_that("can_use_zarrs_write returns FALSE for S3Store", {
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")

  s3 <- S3Store$new("s3://bucket/prefix")
  mock_indexer <- list(dim_indexers = list(
    structure(list(step = 1L), class = "SliceDimIndexer")
  ))
  class(mock_indexer) <- "BasicIndexer"

  expect_false(can_use_zarrs_write(mock_indexer, s3))
})

library(pizzarr)

test_that("Zarr DirectoryStore, can listdir", {
  store <- DirectoryStore$new(tempdir())
  store$set_item("hello/there/a", c(0xbeef))
  store$set_item("hello/there/b", c(0xbeef))
  store$set_item("hello/world/a", c(0xdead))
  store$set_item("hello/world/b", c(0xdead))

  expect_true("hello" %in% store$listdir())
  expect_true("there" %in% store$listdir("hello"))
  expect_true("world" %in% store$listdir("hello"))
  expect_true("a" %in% store$listdir("hello/there"))
  expect_true("b" %in% store$listdir("hello/there"))
})

test_that("Zarr DirectoryStore, can rmdir", {
  store <- DirectoryStore$new(tempdir())
  store$set_item("hello/there/a", c(0xbeef))
  store$set_item("hello/there/b", c(0xbeef))
  store$set_item("hello/world/a", c(0xdead))
  store$set_item("hello/world/b", c(0xdead))

  store$rmdir("hello/there")
  f <- function() store$listdir("hello/there")
  expect_error(f())
  expect_true("world" %in% store$listdir("hello"))
  expect_false("there" %in% store$listdir("hello"))

  store$rmdir("hello")
  expect_false("hello" %in% store$listdir())
})

test_that("DirectoryStore get_item on missing key raises error", {
  store <- DirectoryStore$new(tempfile())
  expect_error(store$get_item("nonexistent"))
})

test_that("DirectoryStore is readable, writeable, and listable", {
  store <- DirectoryStore$new(tempfile())
  expect_true(store$is_readable())
  expect_true(store$is_writeable())
  expect_true(store$is_listable())
})

test_that("DirectoryStore set_item overwrites existing key", {
  d <- tempfile()
  store <- DirectoryStore$new(d)
  store$set_item("key", charToRaw("value1"))
  store$set_item("key", charToRaw("value2"))
  expect_equal(rawToChar(store$get_item("key")), "value2")
})

test_that("DirectoryStore contains_item returns FALSE for missing key", {
  store <- DirectoryStore$new(tempfile())
  expect_false(store$contains_item("nonexistent"))
})

test_that("DirectoryStore contains_item returns TRUE for existing key", {
  store <- DirectoryStore$new(tempfile())
  store$set_item("mykey", charToRaw("myvalue"))
  expect_true(store$contains_item("mykey"))
})

# --- zarrs GCS store tests ---
# Tests for GCS-hosted data and the GcsStore class.
# The gs:// protocol requires GCP credentials; public GCS data is
# tested via HTTPS (storage.googleapis.com) which exercises the
# same blosc codec path.

test_that("gcs feature is compiled", {
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")
  features <- pizzarr_compiled_features()
  skip_if(!("gcs" %in% features), "gcs feature not compiled")
  expect_true("gcs" %in% features)
  expect_true("object_store" %in% features)
})

test_that("zarrs reads blosc-compressed GCS data via HTTPS", {
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")
  skip_if(!("blosc" %in% pizzarr_compiled_features()), "blosc feature not compiled")
  skip_on_cran()
  skip_if_offline()

  # Pangeo ECCO_basins via public HTTPS endpoint (blosc/lz4 compressed)
  url <- "https://storage.googleapis.com/pangeo-data/ECCO_basins.zarr"

  meta <- zarrs_open_array_metadata(url, "basin_mask")
  expect_equal(meta$zarr_format, 2L)
  expect_true(length(meta$shape) > 0)

  # Read a small subset
  ndim <- length(meta$shape)
  ranges <- lapply(seq_len(ndim), function(i) c(0L, 1L))
  result <- zarrs_get_subset(url, "basin_mask", ranges, NULL)
  expect_true(length(result$data) == 1)
  expect_equal(length(result$shape), ndim)

  zarrs_close_store(url)
})

test_that("zarrs reads from gs:// with credentials", {
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")
  skip_if(!("gcs" %in% pizzarr_compiled_features()), "gcs feature not compiled")
  skip_on_cran()
  skip_if_offline()
  # gs:// requires GCP credentials (env vars or application default)
  skip_if(!nzchar(Sys.getenv("GOOGLE_APPLICATION_CREDENTIALS")) &&
          !nzchar(Sys.getenv("GOOGLE_SERVICE_ACCOUNT")),
          "no GCP credentials configured")

  url <- "gs://pangeo-data/ECCO_basins.zarr"

  meta <- zarrs_open_array_metadata(url, "basin_mask")
  expect_equal(meta$zarr_format, 2L)

  ndim <- length(meta$shape)
  ranges <- lapply(seq_len(ndim), function(i) c(0L, 1L))
  result <- zarrs_get_subset(url, "basin_mask", ranges, NULL)
  expect_true(length(result$data) == 1)

  zarrs_close_store(url)
})

test_that("GcsStore class works", {
  gcs <- GcsStore$new("gs://bucket/prefix")
  expect_equal(gcs$get_store_identifier(), "gs://bucket/prefix")
  expect_false(gcs$is_writeable())
  expect_output(print(gcs), "<GcsStore> gs://bucket/prefix")
  expect_error(GcsStore$new("https://example.com"), "gs://")
})

test_that("can_use_zarrs_write returns FALSE for GcsStore", {
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")

  gcs <- GcsStore$new("gs://bucket/prefix")
  mock_indexer <- list(dim_indexers = list(
    structure(list(step = 1L), class = "SliceDimIndexer")
  ))
  class(mock_indexer) <- "BasicIndexer"

  expect_false(can_use_zarrs_write(mock_indexer, gcs))
})

library(pizzarr)

test_that("Zarr MemoryStore, can get and set a value at the top level", {
  store <- MemoryStore$new()

  store_contains <- store$contains_item("hello")
  expect_equal(store_contains, FALSE)

  store$set_item("hello", c(0xbeef))

  store_contains <- store$contains_item("hello")
  expect_equal(store_contains, TRUE)

  value <- store$get_item("hello")
  expect_equal(value, c(0xbeef))
})

test_that("Zarr MemoryStore, throws error if setting recursively over a non-list value", {
  should_error <- function() {
    store <- MemoryStore$new()
    store$set_item("hello", c(0xbeef))
    store$set_item("hello/there", c(0xdead))
  }
  expect_error(should_error())
})

test_that("Zarr MemoryStore, can set and get nested values", {
  store <- MemoryStore$new()
  store$set_item("hello/there", c(0xbeef))
  store$set_item("hello/world", c(0xdead))

  value_there <- store$get_item("hello/there")
  expect_equal(value_there, c(0xbeef))

  value_world <- store$get_item("hello/world")
  expect_equal(value_world, c(0xdead))
})
test_that("Zarr MemoryStore, can set and get twice nested values", {
  store <- MemoryStore$new()
  store$set_item("hello/there/a", c(0xbeef))
  store$set_item("hello/there/b", c(0xbeef))
  store$set_item("hello/world/a", c(0xdead))
  store$set_item("hello/world/b", c(0xdead))

  value_there <- store$get_item("hello/there")
  expect_equal(value_there, list(a = 0xbeef, b = 0xbeef))

  value_world <- store$get_item("hello/world")
  expect_equal(value_world, list(a = 0xdead, b = 0xdead))
})

test_that("Zarr MemoryStore, can listdir", {
  store <- MemoryStore$new()
  store$set_item("hello/there/a", c(0xbeef))
  store$set_item("hello/there/b", c(0xbeef))
  store$set_item("hello/world/a", c(0xdead))
  store$set_item("hello/world/b", c(0xdead))

  expect_equal(store$listdir(), c("hello"))
  expect_equal(store$listdir("hello"), c("there", "world"))
  expect_equal(store$listdir("hello/there"), c("a", "b"))
})

test_that("Zarr MemoryStore, can rmdir", {
  store <- MemoryStore$new()
  store$set_item("hello/there/a", c(0xbeef))
  store$set_item("hello/there/b", c(0xbeef))
  store$set_item("hello/world/a", c(0xdead))
  store$set_item("hello/world/b", c(0xdead))

  store$rmdir("hello/there")
  f <- function() store$listdir("hello/there")
  expect_error(f())
  expect_equal(store$listdir("hello"), c("world"))

  store$rmdir("hello")
  expect_equal(store$listdir(), character(0))
})

test_that("MemoryStore is readable, writeable, and listable", {
  store <- MemoryStore$new()
  expect_true(store$is_readable())
  expect_true(store$is_writeable())
  expect_true(store$is_listable())
})

test_that("MemoryStore set_item overwrites existing key", {
  store <- MemoryStore$new()
  store$set_item("key", charToRaw("value1"))
  store$set_item("key", charToRaw("value2"))
  expect_equal(rawToChar(store$get_item("key")), "value2")
})

test_that("MemoryStore contains_item returns FALSE for missing key", {
  store <- MemoryStore$new()
  expect_false(store$contains_item("nonexistent"))
})

test_that("bare Store stubs raise NotImplementedError", {
  store <- Store$new()
  expect_error(store$listdir(), "NotImplementedError\\(listdir_from_keys\\)")
  expect_error(store$rename("a", "b"), "NotImplementedError\\(rename_from_keys\\)")
  expect_error(store$rmdir("a"), "NotImplementedError\\(rmdir_from_keys\\)")
})

library(pizzarr)

# --- Dimension separator behavior tests ---

test_that("Array with '/' dimension separator writes and reads correctly (MemoryStore)", {
  store <- MemoryStore$new()
  a <- zarr_create(
    shape = c(4, 4), chunks = c(2, 2),
    store = store, dimension_separator = "/"
  )
  data <- array(1:16, dim = c(4, 4))
  a$set_item("...", data)
  result <- a$get_item("...")
  expect_equal(result$data, data)
})

test_that("Array with '.' dimension separator writes and reads correctly (MemoryStore)", {
  store <- MemoryStore$new()
  a <- zarr_create(
    shape = c(4, 4), chunks = c(2, 2),
    store = store, dimension_separator = "."
  )
  data <- array(1:16, dim = c(4, 4))
  a$set_item("...", data)
  result <- a$get_item("...")
  expect_equal(result$data, data)
})

test_that("Default dimension separator is '.'", {
  store <- DirectoryStore$new(tempfile())
  a <- zarr_create(shape = c(4, 4), chunks = c(2, 2), store = store)
  data <- array(1:16, dim = c(4, 4))
  a$set_item("...", data)
  # Chunks should use "." separator: "0.0", "0.1", "1.0", "1.1"
  expect_true(store$contains_item("0.0"))
  expect_true(store$contains_item("0.1"))
  expect_true(store$contains_item("1.0"))
  expect_true(store$contains_item("1.1"))
})

test_that("Chunk keys use '/' separator when specified (DirectoryStore)", {
  store <- DirectoryStore$new(tempfile())
  a <- zarr_create(
    shape = c(4, 4), chunks = c(2, 2),
    store = store, dimension_separator = "/"
  )
  data <- array(1:16, dim = c(4, 4))
  a$set_item("...", data)
  # Chunks should use "/" separator: "0/0", "0/1", "1/0", "1/1"
  expect_true(store$contains_item("0/0"))
  expect_true(store$contains_item("0/1"))
  expect_true(store$contains_item("1/0"))
  expect_true(store$contains_item("1/1"))
})

test_that("1D array chunk keys are plain integers", {
  store <- DirectoryStore$new(tempfile())
  a <- zarr_create(shape = c(6), chunks = c(3), store = store)
  data <- array(1:6, dim = c(6))
  a$set_item("...", data)
  expect_true(store$contains_item("0"))
  expect_true(store$contains_item("1"))
})

test_that("Dimension separator is preserved in metadata round-trip (DirectoryStore)", {
  d <- tempfile()
  store <- DirectoryStore$new(d)
  a <- zarr_create(
    shape = c(4, 4), chunks = c(2, 2),
    store = store, dimension_separator = "/"
  )
  data <- array(1:16, dim = c(4, 4))
  a$set_item("...", data)
  # Reopen from same directory
  b <- zarr_open_array(d, mode = "r")
  result <- b$get_item("...")
  expect_equal(result$data, data)
})

# Reference: https://github.com/gzuidhof/zarr.js/blob/292804/src/core/slice.ts#L78

#' Abstract Slice object
#' @title Slice Class
#' @docType class
#' @description
#' Class representing an slice of a ZARR store
#' @noRd
#' @keywords internal
Slice <- R6::R6Class("Slice",
  public = list(
    #' @field start start index
    start = NULL,
    #' @field stop stop index
    stop = NULL,
    #' @field step step size
    step = NULL,
    #' @description Create a new `Slice` object
    #' @param start integer start index of slice
    #' @param stop integer stop index of slice
    #' @param step integer step size of slice
    initialize = function(start, stop = NA, step = NA) {
      if(is.na(stop) && is.na(step)) {
        stop <- start
        start <- NA
      }
      if(!is.na(start) && start == ":") {
        start <- NA
      }
      if(!is.na(stop) && stop == ":") {
        stop <- NA
      }
      self$start <- start
      self$stop <- stop
      self$step <- step
    },
    #' @description
    #' This method takes a single integer argument `length_param` and computes information about the
    #' slice that the slice object would describe if applied to a sequence of `length_param` items.
    #' It returns a tuple of three integers; respectively these are the start and stop indices
    # and the step or stride length of the slice. Missing or out-of-bounds indices are handled
    # in a manner consistent with regular slices.
    #' @param length_param integer length parameter for calculation of slices
    indices = function(length_param) {
      start <- 0
      stop <- 0
      step <- 0

      MAX_SAFE_INTEGER <- .Machine$integer.max

      if(is.na(self$step)) {
        step <- 1
      } else {
        step <- self$step
      }

      if(is.na(self$start)) {
        if(step < 0) {
          start <- MAX_SAFE_INTEGER
        } else {
          start <- 0
        }
      } else {
        start <- self$start
        if(start < 0) {
          start <- start + length_param
        }
      }

      if(is.na(self$stop)) {
        if(step < 0) {
          stop <- -MAX_SAFE_INTEGER
        } else {
          stop <- MAX_SAFE_INTEGER
        }
      } else {
        stop <- self$stop
        if(stop < 0) {
          stop <- stop + length_param
        }
      }

      s <- adjust_indices(start, stop, step, length_param)
      start <- s[1]
      stop <- s[2]
      step <- s[3]
      length_param <- s[4]

      if(step == 0) {
        stop("step size 0 is invalid")
      }
      return(c(start, stop, step, length_param))
    }
  )
)

#' Convenience function for the internal Slice R6 class constructor.
#' @param start The start index.
#' @param stop The stop index.
#' @param step The step size. Negative values reverse the direction of the
#'   slice, matching Python/NumPy semantics (e.g., `step = -1` iterates
#'   backward from `start` toward `stop`).
#' @param zero_based The index of the dimension. By default, FALSE for R-like behavior.
#' @return A Slice instance with the specified parameters.
#' @examples
#' g <- zarr_volcano()
#' v <- g$get_item("volcano")
#'
#' # Reverse the first 5 columns of row 1 (zero-based: 5:0:-1)
#' v$get_orthogonal_selection(list(zb_slice(0, 1), zb_slice(5, 0, -1)))
#'
#' # Full reverse of row 1 (zero-based: -1::-1)
#' v$get_orthogonal_selection(list(zb_slice(0, 1), zb_slice(-1, NA, -1)))
#' @export
slice <- function(start, stop = NA, step = NA, zero_based = FALSE) {
  start_offset <- ifelse(zero_based, 0, -1)
  stop_offset <- ifelse(zero_based, 0, 0)
  if(!is_na(start) && is.numeric(start)) {
    start <- start + start_offset
  }
  if(!is_na(stop) && is.numeric(stop)) {
    stop <- stop + stop_offset
  }
  # Assumed to be zero-based
  # and stop-inclusive
  return(Slice$new(
    start = start,
    stop = stop,
    step = step
  ))
}

#' Convenience function for the internal Slice R6 class constructor
#' with zero-based indexing and exclusive stop index.
#' @inheritParams slice
#' @inherit slice return
#' @examples
#' # Equivalent to Python's arr[5:0:-1]
#' zb_slice(5, 0, -1)
#'
#' # Equivalent to Python's arr[-1::-1] (full reverse)
#' zb_slice(-1, NA, -1)
#' @export
zb_slice <- function(start, stop = NA, step = NA) {
  return(slice(start, stop, step, zero_based = TRUE))
}

#' Check if a value is a Slice instance.
#' @param s The value to check.
#' @return TRUE if the value is a Slice instance, FALSE otherwise.
#' @export
is_slice <- function(s) {
  if(class(s)[[1]] == "Slice") {
    return(TRUE)
  }
  return(FALSE)
}

#' @keywords internal
is_positive_slice <- function(s) {
  if(is_slice(s) && (is_na(s$step) || s$step >= 1)) {
    return(TRUE)
  }
  return(FALSE)
}

#' @keywords internal
is_basic_selection <- function(selection) {
  selection <- ensure_list(selection)
  # Reference: https://github.com/gzuidhof/zarr.js/blob/master/src/core/indexing.ts#L170
  for(i in seq_along(selection)) {
    s <- selection[[i]]
    if(!(is.numeric(s) || is_positive_slice(s))) {
      return(FALSE)
    }
  }
  return(TRUE)
}

#' @keywords internal
adjust_indices <- function(start, stop, step, length_param) {
  if(start < 0) {
    start <- start + length_param
    if(start < 0) {
      if(step < 0) {
        start <- -1
      } else {
        start <- 0
      }
    }
  } else if(start >= length_param) {
    if(step < 0) {
      start <- length_param - 1
    } else {
      start <- length_param
    }
  }
  if(stop < 0) {
    stop <- stop + length_param
    if(stop < 0) {
      if(step < 0) {
        stop <- -1
      } else {
        stop <- 0
      }
    }
  } else if(stop >= length_param) {
    if(step < 0) {
      stop <- length_param - 1
    } else {
      stop <- length_param
    }
  }
  if(step < 0) {
    if(stop < start) {
      length_param <- floor((start - stop - 1) / (-step) + 1)
      return(c(start, stop, step, length_param))
    }
  } else {
    if(start < stop) {
      length_param <- floor((stop - start - 1) / step + 1)
      return(c(start, stop, step, length_param))
    }
  }
  return(c(start, stop, step, 0))
}

#' @keywords internal
is_total_slice <- function(item, shape) {
  # Reference: https://github.com/gzuidhof/zarr.js/blob/15e3a3f00eb19f0133018fb65f002311ea53bb7c/src/util.ts#L129

  if (is.null(item) || is_na(item)) {
    return(TRUE)
  }
  if (is_scalar(item)) {
    item <- as.numeric(item)
  }

  for (i in seq_len(min(length(item), length(shape)))) {
    it <- item[[i]]

    if (is.null(it) || is_na(it)) {
      # continue
    } else {
      if (is_slice(it)) {
        s <- it
        is_step_one <- s$step == 1 || is.null(s$step) || is_na(s$step)

        if ((is.null(s$start) || is_na(s$start)) && (is.null(s$stop) || is_na(s$stop)) && is_step_one) {
          # continue
        } else {
          if ((as.numeric(s$stop) - as.numeric(s$start)) == shape[i] && is_step_one) {
            # continue
          } else {
            return(FALSE)
          }
        }
      } else {
        return(FALSE)
      }
    }
  }
  return(TRUE)
}

#' @keywords internal
is_contiguous_slice <- function(s) {
  # Reference: https://github.com/gzuidhof/zarr.js/blob/15e3a3f00eb19f0133018fb65f002311ea53bb7c/src/core/indexing.ts#L149
  if(is_slice(s) && (is_na(s$step) || s$step == 1)) {
    return(TRUE)
  }
  return(FALSE)
}

#' @keywords internal
is_contiguous_selection <- function(selection) {
  # Reference: https://github.com/gzuidhof/zarr.js/blob/15e3a3f00eb19f0133018fb65f002311ea53bb7c/src/core/indexing.ts#L157
  selection <- ensure_list(selection)
  for(i in seq_len(length(selection))) {
    s <- selection[[i]]
    if(!(is_integer_vec(s) || is_contiguous_slice(s) || (is.character(s) && s == "..."))) {
      return(FALSE)
    }
  }
  return(TRUE)
}

library(pizzarr)

test_that("Slice$indices() behavior matches Python slice().indices()", {
    # Python slice.indices() only returns (start, stop, step).
    # (Zarr.js returns (start, stop, step, adjusted_length).)
    expect_equal(Slice$new(0, 10, 1)$indices(10)[1:3], c(0, 10, 1))
    expect_equal(Slice$new(0, 10, 1)$indices(11)[1:3], c(0, 10, 1))
    expect_equal(Slice$new(0, 10, 1)$indices(12)[1:3], c(0, 10, 1))
    expect_equal(Slice$new(0, 10, 2)$indices(12)[1:3], c(0, 10, 2))

    expect_equal(Slice$new(1, 10, 2)$indices(5)[1:3], c(1, 5, 2))

    expect_equal(Slice$new(9, 10, 2)$indices(5)[1:3], c(5, 5, 2))
    expect_equal(Slice$new(10, 10, 2)$indices(5)[1:3], c(5, 5, 2))
    expect_equal(Slice$new(10, 10, 2)$indices(10)[1:3], c(10, 10, 2))
    expect_equal(Slice$new(10, 10, 2)$indices(11)[1:3], c(10, 10, 2))

    expect_equal(Slice$new(0, 10, -1)$indices(12)[1:3], c(0, 10, -1))

    expect_equal(Slice$new(-1, 10, -1)$indices(14)[1:3], c(13, 10, -1))

    expect_equal(Slice$new(0, 1, 1)$indices(1)[1:3], c(0, 1, 1))
})

test_that("zb_slice gets converted to Slice", {
    s1 <- zb_slice(0, 10, 1)
    s2 <- Slice$new(0, 10, 1)
    expect_equal(s1$start, s2$start)
    expect_equal(s1$stop, s2$stop)
    expect_equal(s1$step, s2$step)
})

test_that("slice gets converted to zb_slice", {
    s1 <- slice(1, 9, 1)
    s2 <- zb_slice(0, 9, 1)
    expect_equal(s1$start, s2$start)
    expect_equal(s1$stop, s2$stop)
    expect_equal(s1$step, s2$step)
})

test_that("step size greater than 1", {
  g <- zarr_volcano()
  
  s1 <- slice(1, 11, 1)
  s2 <- slice(1, 11, 2)
  s3 <- slice(1, 11, 3)
  
  d1 <- g$get_item("volcano")$get_item(list(s1, s1))$data
  d2 <- g$get_item("volcano")$get_item(list(s2, s2))$data
  d3 <- g$get_item("volcano")$get_item(list(s3, s3))$data
  
  expect_equal(dim(d1), c(11, 11))
  expect_equal(dim(d2), c(6, 6))
  expect_equal(dim(d3), c(4, 4))

  # ground truth for the volcano section with step size 2
  a2 <- t(array(data=c(
    100, 101, 101, 101, 100, 101,
    102, 103, 103, 103, 102, 103,
    104, 105, 105, 105, 104, 104,
    105, 106, 107, 107, 106, 105,
    107, 108, 109, 109, 108, 108,
    109, 110, 111, 111, 110, 112
  ), dim=c(6, 6)))
  expect_equal(d2, a2)

  # ground truth for the volcano section with step size 3
  a3 <- t(array(data=c(
    100, 101, 101, 100,
    103, 104, 104, 103,
    105, 107, 107, 105,
    108, 110, 110, 108
  ), dim=c(4, 4)))
  expect_equal(d3, a3)
})

test_that("negative step slicing matches zarr-python behavior", {
  # Use volcano data - first row has 19 values we can slice
  g <- zarr_volcano()
  volcano_arr <- g$get_item("volcano")

  # Get the full first row (1D slice) for reference
  full_row <- volcano_arr$get_orthogonal_selection(
    list(zb_slice(0, 1), zb_slice(NA))
  )$data[1,]

  # Negative step: reverse a portion of the first row
  # Python equivalent: arr[0, 5:0:-1]
  result <- volcano_arr$get_orthogonal_selection(
    list(zb_slice(0, 1), zb_slice(5, 0, -1))
  )
  expect_equal(as.vector(result$data), rev(full_row[2:6]))

  # Negative step with step size -2
  # Python equivalent: arr[0, 10:2:-2]
  result2 <- volcano_arr$get_orthogonal_selection(
    list(zb_slice(0, 1), zb_slice(10, 2, -2))
  )
  expect_equal(as.vector(result2$data), full_row[c(11, 9, 7, 5)])

  # Negative start index with negative step
  # Python equivalent: arr[0, -5:0:-1]
  ncols <- length(full_row)
  result3 <- volcano_arr$get_orthogonal_selection(
    list(zb_slice(0, 1), zb_slice(-5, 0, -1))
  )
  neg5_idx <- ncols - 5 + 1  # R 1-based index for Python's -5
  expect_equal(as.vector(result3$data), rev(full_row[2:neg5_idx]))

  # Full reverse: arr[0, -1::-1] reverses the entire row
  result4 <- volcano_arr$get_orthogonal_selection(
    list(zb_slice(0, 1), zb_slice(-1, NA, -1))
  )
  expect_equal(as.vector(result4$data), rev(full_row))

  # Negative step also works through get_item (basic selection routing)
  result5 <- volcano_arr$get_item(list(zb_slice(0, 1), zb_slice(5, 0, -1)))
  expect_equal(as.vector(result5$data), rev(full_row[2:6]))
})

library(pizzarr)

# =============================================================================
# Gap 1: adjust_indices / is_total_slice / is_contiguous_slice / is_positive_slice
# =============================================================================

# --- adjust_indices ---
# Reference: Python slice.indices() semantics

test_that("adjust_indices clamps negative start for positive step", {
  res <- adjust_indices(-20, 5, 1, 10)
  expect_equal(res[1], 0)
  expect_equal(res[3], 1)
})

test_that("adjust_indices clamps negative start for negative step", {
  res <- adjust_indices(-20, 0, -1, 10)
  expect_equal(res[1], -1)
})

test_that("adjust_indices clamps start >= length for positive step", {
  res <- adjust_indices(15, 20, 1, 10)
  expect_equal(res[1], 10)
})

test_that("adjust_indices clamps start >= length for negative step", {
  res <- adjust_indices(15, 0, -1, 10)
  expect_equal(res[1], 9)
})

test_that("adjust_indices clamps negative stop for positive step", {
  res <- adjust_indices(0, -20, 1, 10)
  expect_equal(res[2], 0)
})

test_that("adjust_indices clamps negative stop for negative step", {
  res <- adjust_indices(9, -20, -1, 10)
  expect_equal(res[2], -1)
})

test_that("adjust_indices clamps stop >= length for positive step", {
  res <- adjust_indices(0, 20, 1, 10)
  expect_equal(res[2], 10)
})

test_that("adjust_indices clamps stop >= length for negative step", {
  res <- adjust_indices(9, 20, -1, 10)
  expect_equal(res[2], 9)
})

test_that("adjust_indices computes length for positive step", {
  # start=2, stop=8, step=2 -> length = floor((8-2-1)/2 + 1) = 3
  res <- adjust_indices(2, 8, 2, 10)
  expect_equal(res[4], 3)
})

test_that("adjust_indices computes length for negative step", {
  # start=8, stop=2, step=-1 -> length = floor((8-2-1)/1 + 1) = 6
  res <- adjust_indices(8, 2, -1, 10)
  expect_equal(res[4], 6)
})

test_that("adjust_indices returns 0 length when start >= stop for positive step", {
  res <- adjust_indices(5, 3, 1, 10)
  expect_equal(res[4], 0)
})

test_that("adjust_indices returns 0 length when stop >= start for negative step", {
  res <- adjust_indices(3, 5, -1, 10)
  expect_equal(res[4], 0)
})

test_that("adjust_indices handles full range", {
  res <- adjust_indices(0, 10, 1, 10)
  expect_equal(res, c(0, 10, 1, 10))
})

test_that("adjust_indices handles negative start resolved within range", {
  # start=-3 + length=10 = 7, still in range
  res <- adjust_indices(-3, 10, 1, 10)
  expect_equal(res[1], 7)
  expect_equal(res[4], 3)
})

test_that("adjust_indices handles negative stop resolved within range", {
  # stop=-3 + length=10 = 7
  res <- adjust_indices(0, -3, 1, 10)
  expect_equal(res[2], 7)
  expect_equal(res[4], 7)
})

test_that("adjust_indices with step=3 computes correct length", {
  # start=1, stop=10, step=3 -> items at 1,4,7 -> length=3
  res <- adjust_indices(1, 10, 3, 10)
  expect_equal(res[4], 3)
})

test_that("adjust_indices with negative step=(-2) computes correct length", {
  # start=9, stop=0, step=-2 -> items at 9,7,5,3,1 -> length=5
  res <- adjust_indices(9, 0, -2, 10)
  expect_equal(res[4], 5)
})

# --- is_total_slice ---

test_that("is_total_slice returns TRUE for NULL", {
  expect_true(is_total_slice(NULL, c(10)))
})

test_that("is_total_slice returns TRUE for NA", {
  expect_true(is_total_slice(NA, c(10)))
})

test_that("is_total_slice returns FALSE for numeric scalar", {
  # A numeric scalar gets treated as non-slice
  expect_false(is_total_slice(5, c(10)))
})

# --- is_contiguous_slice ---

test_that("is_contiguous_slice TRUE for step=1", {
  s <- zb_slice(0, 5, 1)
  expect_true(is_contiguous_slice(s))
})

test_that("is_contiguous_slice TRUE for NA step", {
  s <- zb_slice(0, 5)
  expect_true(is_contiguous_slice(s))
})

test_that("is_contiguous_slice FALSE for step=2", {
  s <- zb_slice(0, 5, 2)
  expect_false(is_contiguous_slice(s))
})

test_that("is_contiguous_slice FALSE for non-slice", {
  expect_false(is_contiguous_slice(5))
})

# --- is_positive_slice ---

test_that("is_positive_slice TRUE for step=1", {
  s <- zb_slice(0, 5, 1)
  expect_true(is_positive_slice(s))
})

test_that("is_positive_slice TRUE for NA step", {
  s <- zb_slice(0, 5)
  expect_true(is_positive_slice(s))
})

test_that("is_positive_slice FALSE for negative step", {
  s <- zb_slice(5, 0, -1)
  expect_false(is_positive_slice(s))
})

test_that("is_positive_slice FALSE for non-slice", {
  expect_false(is_positive_slice(5))
})

# --- is_total_slice with list of Slices ---

test_that("is_total_slice TRUE for list of full slices", {
  s1 <- zb_slice(0, 10, 1)
  s2 <- zb_slice(0, 20, 1)
  expect_true(is_total_slice(list(s1, s2), c(10, 20)))
})

test_that("is_total_slice TRUE for list of NA-start/stop slices", {
  s1 <- zb_slice(NA, NA, 1)
  s2 <- zb_slice(NA, NA)
  expect_true(is_total_slice(list(s1, s2), c(10, 20)))
})

test_that("is_total_slice FALSE for partial slice in list", {
  s1 <- zb_slice(0, 5, 1)
  expect_false(is_total_slice(list(s1), c(10)))
})

# --- is_contiguous_selection ---

test_that("is_contiguous_selection TRUE for ellipsis string", {
  expect_true(is_contiguous_selection(list("...")))
})

test_that("is_contiguous_selection TRUE for integer vec", {
  expect_true(is_contiguous_selection(list(c(1L, 2L, 3L))))
})

test_that("is_contiguous_selection TRUE for contiguous slice", {
  s <- zb_slice(0, 5, 1)
  expect_true(is_contiguous_selection(list(s)))
})

test_that("is_contiguous_selection FALSE for non-contiguous slice", {
  s <- zb_slice(0, 5, 2)
  expect_false(is_contiguous_selection(list(s)))
})

# --- is_basic_selection ---

test_that("is_basic_selection TRUE for positive slice", {
  s <- zb_slice(0, 5, 1)
  expect_true(is_basic_selection(s))
})

test_that("is_basic_selection TRUE for numeric scalar", {
  expect_true(is_basic_selection(3))
})

test_that("is_basic_selection FALSE for negative step slice", {
  s <- zb_slice(5, 0, -1)
  expect_false(is_basic_selection(s))
})

# --- Slice$indices edge cases ---

test_that("Slice$indices with step=0 errors", {
  s <- Slice$new(0, 5, 0)
  expect_error(s$indices(10), "step size 0")
})

test_that("Slice$indices with negative step and NA start/stop", {
  # Python: slice(None, None, -1).indices(5) => (4, -1, -1)
  s <- Slice$new(NA, NA, -1)
  res <- s$indices(5)
  expect_equal(res[3], -1)
  expect_equal(res[4], 5)
})

test_that("Slice$indices with negative step and NA stop", {
  # Python: slice(3, None, -1).indices(5) => (3, -1, -1)
  s <- Slice$new(3, NA, -1)
  res <- s$indices(5)
  expect_equal(res[1], 3)
  expect_equal(res[4], 4)
})

test_that("Slice$indices with positive step and NA start", {
  # Python: slice(None, 5, 1).indices(10) => (0, 5, 1)
  s <- Slice$new(NA, 5, 1)
  res <- s$indices(10)
  expect_equal(res[1], 0)
  expect_equal(res[2], 5)
  expect_equal(res[4], 5)
})

# User-facing configuration for pizzarr concurrency settings.
#
# Three settings control zarrs parallelism and HTTP behaviour:
#   - nthreads: rayon thread pool size (set-once per process)
#   - concurrent_target: codec concurrency level (dynamic)
#   - http_batch_range_requests: multipart range toggle (at store creation)
#
# Each is backed by an R option (pizzarr.*) and an env var (PIZZARR_*).
# The env var → option mapping is handled by init_options() in options.R.
# This file provides the user-facing pizzarr_config() getter/setter and
# the internal apply_zarrs_config() called from .onLoad.

#' Get or set pizzarr configuration
#'
#' Controls parallelism and HTTP behaviour for the zarrs backend.
#' Called with no arguments, returns the current settings as a named list.
#' Called with arguments, sets the specified options and applies them to
#' the Rust backend immediately.
#'
#' @param nthreads Integer or NULL. Number of threads for the rayon thread
#'   pool. NULL uses all CPUs (the default). The pool can only be initialised
#'   once per R session; later changes require a restart. Use the
#'   \code{PIZZARR_NTHREADS} environment variable for reliable session-level
#'   control.
#' @param concurrent_target Integer or NULL. Codec concurrency level — how
#'   many codec operations zarrs runs in parallel within a single read/write
#'   call. NULL uses the zarrs default (CPU count). Can be changed at any time.
#' @param http_batch_range_requests Logical or NULL. Whether HTTP stores use
#'   multipart range requests (default TRUE). Set to FALSE for servers with
#'   incomplete multipart range support. Takes effect on the next
#'   \code{zarr_open()} or \code{zarrs_get_subset()} call that opens a new
#'   HTTP store; existing cached stores are not affected (use
#'   \code{zarrs_close_store(url)} to force re-creation).
#' @return When called with no arguments, a named list of current settings.
#'   When called with arguments, the previous values (invisibly).
#' @export
pizzarr_config <- function(nthreads = NULL, concurrent_target = NULL,
                           http_batch_range_requests = NULL) {

  getting <- is.null(nthreads) && is.null(concurrent_target) &&
    is.null(http_batch_range_requests)

  if (getting) {
    out <- list(
      nthreads = getOption("pizzarr.nthreads"),
      concurrent_target = getOption("pizzarr.concurrent_target"),
      http_batch_range_requests = getOption("pizzarr.http_batch_range_requests",
                                            TRUE)
    )
    # When zarrs is available, report the actual Rust-side thread count.
    if (.pizzarr_env$zarrs_available) {
      info <- zarrs_runtime_info()
      out$nthreads <- info$nthreads
      out$concurrent_target <- info$codec_concurrent_target
    }
    return(out)
  }

  # --- setter path ---
  prev <- list(
    nthreads = getOption("pizzarr.nthreads"),
    concurrent_target = getOption("pizzarr.concurrent_target"),
    http_batch_range_requests = getOption("pizzarr.http_batch_range_requests",
                                          TRUE)
  )

  if (!is.null(nthreads)) {
    nthreads <- as.integer(nthreads)
    if (is.na(nthreads) || nthreads < 1L) {
      stop("nthreads must be a positive integer")
    }
    options(pizzarr.nthreads = nthreads)
    if (.pizzarr_env$zarrs_available) {
      ok <- zarrs_set_nthreads(nthreads)
      if (!ok) {
        warning(
          "rayon thread pool already initialised; nthreads change takes ",
          "effect on restart. Set PIZZARR_NTHREADS env var for reliable ",
          "session-level control.",
          call. = FALSE
        )
      }
    } else {
      message("zarrs backend not available; nthreads stored but not applied.")
    }
  }

  if (!is.null(concurrent_target)) {
    concurrent_target <- as.integer(concurrent_target)
    if (is.na(concurrent_target) || concurrent_target < 1L) {
      stop("concurrent_target must be a positive integer")
    }
    options(pizzarr.concurrent_target = concurrent_target)
    if (.pizzarr_env$zarrs_available) {
      zarrs_set_codec_concurrent_target(concurrent_target)
    } else {
      message(
        "zarrs backend not available; concurrent_target stored but not applied."
      )
    }
  }

  if (!is.null(http_batch_range_requests)) {
    http_batch_range_requests <- as.logical(http_batch_range_requests)
    options(pizzarr.http_batch_range_requests = http_batch_range_requests)
    if (.pizzarr_env$zarrs_available) {
      zarrs_set_http_batch_range_requests(http_batch_range_requests)
    } else {
      message(
        "zarrs backend not available; http_batch_range_requests stored ",
        "but not applied."
      )
    }
  }

  invisible(prev)
}

# Apply R options to the Rust backend.
#
# Called from .onLoad after init_options() and zarrs availability check.
# Reads the three pizzarr.* options and calls the corresponding Rust
# setters. Silently skips NULL values (= use defaults).
#
# @keywords internal
apply_zarrs_config <- function() {
  ct <- getOption("pizzarr.concurrent_target")
  if (!is.null(ct)) {
    ct <- as.integer(ct)
    if (!is.na(ct) && ct >= 1L) {
      zarrs_set_codec_concurrent_target(ct)
    }
  }

  nt <- getOption("pizzarr.nthreads")
  if (!is.null(nt)) {
    nt <- as.integer(nt)
    if (!is.na(nt) && nt >= 1L) {
      zarrs_set_nthreads(nt)
    }
    # If FALSE (pool already init), silently accept — .onLoad should not warn.
  }

  batch <- getOption("pizzarr.http_batch_range_requests")
  if (!is.null(batch)) {
    zarrs_set_http_batch_range_requests(as.logical(batch))
  }
}

library(pizzarr)

has_zarrs <- .pizzarr_env$zarrs_available

# --- pizzarr_config() getter ---

test_that("pizzarr_config() returns named list with expected keys", {
  cfg <- pizzarr_config()
  expect_true(is.list(cfg))
  expect_true("nthreads" %in% names(cfg))
  expect_true("concurrent_target" %in% names(cfg))
  expect_true("http_batch_range_requests" %in% names(cfg))
})

test_that("pizzarr_config() nthreads is positive integer when zarrs available", {
  skip_if(!has_zarrs, "zarrs backend not available")
  cfg <- pizzarr_config()
  expect_true(is.integer(cfg$nthreads))
  expect_true(cfg$nthreads > 0L)
})

# --- pizzarr_config() setter: concurrent_target ---

test_that("pizzarr_config(concurrent_target) round-trips via zarrs", {
  skip_if(!has_zarrs, "zarrs backend not available")

  # Save and restore original
  orig <- zarrs_runtime_info()$codec_concurrent_target
  on.exit({
    zarrs_set_codec_concurrent_target(orig)
    options(pizzarr.concurrent_target = NULL)
  }, add = TRUE)

  prev <- pizzarr_config(concurrent_target = 2L)
  expect_equal(zarrs_runtime_info()$codec_concurrent_target, 2L)

  # Returns previous value invisibly
  expect_true(is.list(prev))
})

# --- pizzarr_config() setter: http_batch_range_requests ---

test_that("pizzarr_config(http_batch_range_requests) sets R option", {
  saved <- options(pizzarr.http_batch_range_requests = TRUE)
  on.exit(do.call(options, saved), add = TRUE)

  pizzarr_config(http_batch_range_requests = FALSE)
  expect_false(getOption("pizzarr.http_batch_range_requests"))

  pizzarr_config(http_batch_range_requests = TRUE)
  expect_true(getOption("pizzarr.http_batch_range_requests"))
})

# --- zarrs_runtime_info includes nthreads ---

test_that("zarrs_runtime_info includes nthreads field", {
  skip_if(!has_zarrs, "zarrs backend not available")
  info <- zarrs_runtime_info()
  expect_true("nthreads" %in% names(info))
  expect_true(info$nthreads > 0L)
})

# --- zarrs_set_nthreads validates input ---

test_that("zarrs_set_nthreads rejects non-positive input", {
  skip_if(!has_zarrs, "zarrs backend not available")
  expect_error(zarrs_set_nthreads(0L), "positive")
  expect_error(zarrs_set_nthreads(-1L), "positive")
})

# --- pure-R tier ---

test_that("pizzarr_config works without zarrs backend", {
  # Temporarily pretend zarrs is unavailable
  orig <- .pizzarr_env$zarrs_available
  on.exit(.pizzarr_env$zarrs_available <- orig, add = TRUE)
  .pizzarr_env$zarrs_available <- FALSE

  saved <- options(pizzarr.concurrent_target = NULL)
  on.exit(do.call(options, saved), add = TRUE)

  expect_message(
    pizzarr_config(concurrent_target = 3L),
    "not available"
  )
  expect_equal(getOption("pizzarr.concurrent_target"), 3L)
})

test_that("pizzarr_config(nthreads) messages when zarrs unavailable", {
  orig <- .pizzarr_env$zarrs_available
  on.exit(.pizzarr_env$zarrs_available <- orig, add = TRUE)
  .pizzarr_env$zarrs_available <- FALSE

  saved <- options(pizzarr.nthreads = NULL)
  on.exit(do.call(options, saved), add = TRUE)

  expect_message(pizzarr_config(nthreads = 4L), "not available")
  expect_equal(getOption("pizzarr.nthreads"), 4L)
})

test_that("pizzarr_config(http_batch_range_requests) messages when zarrs unavailable", {
  orig <- .pizzarr_env$zarrs_available
  on.exit(.pizzarr_env$zarrs_available <- orig, add = TRUE)
  .pizzarr_env$zarrs_available <- FALSE

  saved <- options(pizzarr.http_batch_range_requests = TRUE)
  on.exit(do.call(options, saved), add = TRUE)

  expect_message(pizzarr_config(http_batch_range_requests = FALSE), "not available")
  expect_false(getOption("pizzarr.http_batch_range_requests"))
})

utils::globalVariables(c(
  "volcano",
  # zarrs functions — called inside is_zarrs_available() guards,
  # defined in extendr-wrappers.R (r-universe) or absent (CRAN)
  "zarrs_create_array", "zarrs_runtime_info",
  "zarrs_set_nthreads", "zarrs_set_codec_concurrent_target",
  "zarrs_set_http_batch_range_requests"
))

#' pizzarr demo data
#' @details
#' For directory stores, unzips the store to a temporary directory
#' and returns the resulting path.
#' 
#' @param dataset character defining which demo dataset is desired, 
#' If NULL, all are returned
#' @param outdir character directory path to store sample zarr stores
#' 
#' @return path to ready to use zarr store
#' @export
#' @examplesIf requireNamespace("crul", quietly = TRUE)
#'
#' sample_dir <- tools::R_user_dir("pizzarr")
#'
#' clean <- !dir.exists(sample_dir)
#'
#' zarr_samples <- pizzarr_sample(outdir = sample_dir)
#'
#' #printing without system path for example
#' gsub(sample_dir, "...", zarr_samples, fixed = TRUE)
#'
#' # clean up if you don't want to keep them for next time
#' if(clean) unlink(sample_dir, recursive = TRUE)
#'
pizzarr_sample <- function(dataset = NULL, 
                           outdir = file.path(tools::R_user_dir("pizzarr"), 
                                              "pizzarr_sample")) {
  # will unzip here
  tdir <- outdir
  dir.create(tdir, showWarnings = FALSE, recursive = TRUE)
  
  # source data
  sdir <- system.file("extdata", package = "pizzarr")
  zarr_zips <- list.files(sdir, pattern = ".zarr.zip", 
                         full.names = TRUE, recursive = TRUE)
  
  avail <- gsub(paste0(sdir, "/"), "", gsub(".zip", "", zarr_zips))
  
  # if dataset is specified select it
  if(!is.null(dataset)) {
    f <- grepl(dataset, zarr_zips, fixed = TRUE)
    zarr_zips <- zarr_zips[f]

    # if dataset not found, stop and print available datasets
    if(length(zarr_zips) == 0) {
      stop("Dataset not found\n\tMust be one of:\n\t  \"",
           paste(avail, collapse = "\"\n\t  \""), "\"")
    }

    avail <- avail[f]

    # If multiple matches and query doesn't specify a version,
    # prefer the v2 variant for backward compatibility.
    if(length(zarr_zips) > 1 && !grepl("_v[0-9]", dataset)) {
      v2 <- grepl("_v2", zarr_zips, fixed = TRUE)
      if(any(v2)) {
        zarr_zips <- zarr_zips[v2]
        avail <- avail[v2]
      }
    }
  }
  
  # in case zarr_zips is all, loop over them and unzip
  for(z in seq_along(zarr_zips)) {
    
    need <- !file.exists(file.path(tdir, avail[z]))
    
    if(file.size(zarr_zips[z]) == 0 & need) {

      if(!requireNamespace("crul", quietly = TRUE)) {
        stop("Downloading sample data requires the 'crul' package.")
      }

      new_z <- file.path(tdir, basename(zarr_zips[z]))

      url <- paste0("https://github.com/zarr-developers/pizzarr/raw/refs/heads/main/docs/data/",
                     basename(zarr_zips[z]))
      client <- crul::HttpClient$new(url = url)
      res <- client$get(disk = new_z)
      res$raise_for_status()

      zarr_zips[z] <- new_z
    }
     
    if(need)
      utils::unzip(zarr_zips[z], exdir = file.path(tdir, dirname(avail[z])))
  }
  
  return(file.path(tdir, avail))
  
}

#' Create a demo Zarr group containing R's volcano dataset
#'
#' Writes the \code{\link[datasets]{volcano}} matrix into a temporary
#' DirectoryStore as a Zarr array named \code{"volcano"} and returns the
#' opened group.
#'
#' @return A \code{ZarrGroup} containing a single array called \code{"volcano"}.
#' @export
#' @examples
#' g <- zarr_volcano()
#' v <- g$get_item("volcano")
#' image(v$get_item("...")$data, main = "Maunga Whau Volcano")
zarr_volcano <- function() {
  dir <- file.path(tempdir(TRUE), "volcano.zarr")

  unlink(dir, recursive = TRUE, force = TRUE)

  z <- DirectoryStore$new(dir)

  za <- zarr_create(dim(volcano), path = "volcano", store = z, overwrite = TRUE)

  za$set_item("...", volcano)

  g <- zarr_open_group(z)

  g$get_attrs()$set_item("tile", "volcano")

  g
}

#' Create an empty named list
#'
#' A helper function to construct an empty list which converts to a JSON object rather than a JSON array.
#'
#' @param ... A variable number of list entries.
#' @return A named list.
#'
#' @keywords internal
#' @export
#' @examples
#' default_window <- obj_list(
#'   min = 0,
#'   max = 255
#' )
obj_list <- function(...) {
  retval <- stats::setNames(list(), character(0))
  param_list <- list(...)
  for(key in names(param_list)) {
    retval[[key]] = param_list[[key]]
  }
  retval
}

#' @keywords internal
zip_numeric <- function(a, b) {
  result <- list()
  for(i in seq_len(length(a))) {
    result <- append(result, list(c(a[i], b[i])))
  }
  return(result)
}

#' @keywords internal
check_selection_length <- function(selection, shape) {
  if(length(selection) > length(shape)) {
    stop('TooManyIndicesError')
  }
}


#' @keywords internal
filter_list <- function(l, pred) {
  result <- list()
  for(item in l) {
    if(pred(item)) {
      result <- append(result, item)
    }
  }
  return(result)
}

#' Convert user selections, potentially containing "...", to a list of slices
#' that can be used internally.
#' @param selection The user-provided selection list.
#' @param shape The shape of the array, to be used to fill in ellipsis values.
#' @returns A list of selections with ellipsis values converted to NA.
#' @keywords internal
replace_ellipsis <- function(selection, shape) {
  # Reference: https://github.com/gzuidhof/zarr.js/blob/master/src/core/indexing.ts#L67

  selection <- ensure_list(selection)
  
  ellipsis_index <- 0
  num_ellipsis <- 0
  for(i in seq_along(selection)) {
    if(selection[i] == "...") {
      ellipsis_index <- i
      num_ellipsis <- num_ellipsis + 1
    }
  }

  if(num_ellipsis > 1) {
    stop("RangeError(an index can only have a single ellipsis ('...'))")
  }

  if(num_ellipsis == 1) {
    # Count how many items to left and right of ellipsis
    num_items_left <- ellipsis_index - 1
    num_items_right <- length(selection) - (num_items_left + 1)
    num_items <- length(selection) - 1 # All non-ellipsis items
    if(num_items >= length(shape)) {
      # Ellipsis does nothing, just remove it
      selection <- filter_list(selection, function(item) {
        if(is.na(item) || item != "...") {
          return(TRUE)
        }
        return(FALSE)
      })
    } else {
      # Replace ellipsis with as many slices are needed for the number of dims
      num_new_items <- length(shape) - num_items
      new_item <- selection[seq_len(num_items_left)]
      for(i in seq_len(num_new_items)) {
        new_item <- append(new_item, NA)
      }
      if(num_items_right > 0) {
        new_item_right <- selection[seq(from = length(selection) - num_items_right + 1, to = length(selection))]
        for(i in seq_along(new_item_right)) {
          new_item <- append(new_item, new_item_right[i])
        }
      }
      selection <- new_item
    }
  }
  # Fill out seleciton if not  completely specified
  if(length(selection) < length(shape)) {
    num_missing <- length(shape) - length(selection)
    for(i in seq_len(num_missing)) {
      selection <- append(selection, NA)
    }
  }

  check_selection_length(selection, shape)
  return(selection)
}

#' Compute Size
#' @param shape A shape vector
#' @returns The product of shape elements.
#' @keywords internal
compute_size <- function(shape) {
  result <- 1
  for(val in shape) {
    result <- result * val
  }
  return(result)
}

#' Check if a value, potentially a vector, is NA
#'
#' @param val The value to check
#' @return Whether the value is NA
#' @keywords internal
is_na <- function(val) {
  if(is.environment(val)) {
    return(FALSE)
  }
  if(length(val) != 1) {
    # Including when val is integer(0), character(0), etc.
    return(FALSE)
  } else {
    return(is.na(val))
  }
}

#' Fill in an R array with a single scalar value.
#' @keywords internal
#' @param chunk The R array to fill.
#' @param value The scalar value (after is_scalar() check).
chunk_fill <- function(chunk, value) {
  # Chunk is an R array()
  # Value is a scalar (after is_scalar() check)

  # Need to do equivalent of chunk.fill(value) in JS
  chunk[] <- value
  chunk
}

#' Check if an error is a KeyError.
#' @param e The error to check.
#' @return TRUE if the error is a KeyError, FALSE otherwise.
#' @export
is_key_error <- function(e) {
  return(grepl("KeyError", e$message))
}

#' @keywords internal
get_list_product_aux <- function(dim_indexer_iterables, i, partial_results) {
  dim_results <- dim_indexer_iterables[[i]]
  result <- list()
  for(d in dim_results) {
    if(length(partial_results) == 0) {
      result <- append(result, list(d))
    } else {
      for(p in partial_results) {
        result <- append(result, list(append(p, list(d))))
      }
    }
  }
  return(result)
}

#' Generate a product of lists.
#' @param dim_indexer_iterables A list of lists.
#' @return A list of lists.
#' @keywords internal
get_list_product <- function(dim_indexer_iterables) {
  # Reference: https://docs.python.org/3/library/itertools.html#itertools.product
  partial_results <- list()
  for(i in seq_len(length(dim_indexer_iterables))) {
    partial_results <- get_list_product_aux(dim_indexer_iterables, i, partial_results)
  }
  return(partial_results)
}

#' Check if the bit64 package is available
#' @return Logical.
#' @keywords internal
has_bit64 <- function() {
  requireNamespace("bit64", quietly = TRUE)
}

#' Check if a dtype represents a 64-bit integer type
#' @param dtype_obj A Dtype instance.
#' @return Logical.
#' @keywords internal
is_int64_dtype <- function(dtype_obj) {
  dtype_obj$basic_type %in% c("i", "u") && dtype_obj$num_bytes == 8
}

#' Convert raw bytes to integer64 vector
#' @param buf Raw vector of 8-byte little-endian integers.
#' @param n Number of elements to read.
#' @param endian "little" or "big".
#' @return A bit64::integer64 vector.
#' @keywords internal
raw_to_integer64 <- function(buf, n, endian = "little") {
  # integer64 stores int64 bit patterns in double's 8 bytes
  vec <- readBin(con = buf, what = double(), size = 8, n = n, endian = endian)
  class(vec) <- "integer64"
  vec
}

#' Convert raw bytes to double from int64 (fallback when bit64 unavailable)
#' @param buf Raw vector of 8-byte integers.
#' @param n Number of elements to read.
#' @param endian "little" or "big".
#' @param signed Logical; if TRUE, treat as signed int64.
#' @return A double vector (precision loss possible for values > 2^53).
#' @keywords internal
raw_to_double_from_int64 <- function(buf, n, endian = "little", signed = TRUE) {
  # Read as signed 4-byte ints (unsigned flag only valid for sizes 1-2)
  vals <- readBin(con = buf, what = integer(), size = 4, n = n * 2,
                  signed = TRUE, endian = endian)
  # vals are interleaved: lo[1], hi[1], lo[2], hi[2], ...
  idx_lo <- seq(1, n * 2, by = 2)
  idx_hi <- seq(2, n * 2, by = 2)
  # Convert to unsigned double: negative int32 means high bit set
  lo_vals <- ifelse(vals[idx_lo] < 0,
                    as.double(vals[idx_lo]) + 2^32,
                    as.double(vals[idx_lo]))
  hi_vals <- ifelse(vals[idx_hi] < 0,
                    as.double(vals[idx_hi]) + 2^32,
                    as.double(vals[idx_hi]))
  result <- lo_vals + hi_vals * 2^32
  if (signed) {
    neg <- hi_vals >= 2^31
    if (any(neg)) {
      result[neg] <- result[neg] - 2^64
    }
  }
  result
}

#' Convert integer64 vector to raw bytes
#' @param x A bit64::integer64 vector.
#' @param endian "little" or "big".
#' @return A raw vector.
#' @keywords internal
integer64_to_raw <- function(x, endian = "little") {
  # Strip class so writeBin sees plain doubles (which hold the int64 bits)
  writeBin(unclass(x), con = raw(), size = 8, endian = endian)
}

test_that("is_int64_dtype identifies 64-bit integer dtypes", {
  d_i8 <- Dtype$new("<i8")
  d_u8 <- suppressWarnings(Dtype$new("<u8"))
  d_i4 <- Dtype$new("<i4")
  d_f8 <- Dtype$new("<f8")

  expect_true(is_int64_dtype(d_i8))
  expect_true(is_int64_dtype(d_u8))
  expect_false(is_int64_dtype(d_i4))
  expect_false(is_int64_dtype(d_f8))
})

test_that("uint64 dtype warns about signed overflow", {
  expect_warning(Dtype$new("<u8"), "uint64")
})

test_that("get_dtype_rtype returns integer64 for 8-byte ints when bit64 available", {
  skip_if_not_installed("bit64")
  rtype <- get_dtype_rtype("i", num_bytes = 8)
  expect_s3_class(rtype, "integer64")
})

test_that("get_dtype_rtype returns integer for 4-byte ints", {
  rtype <- get_dtype_rtype("i", num_bytes = 4)
  expect_true(is.integer(rtype))
})

test_that("get_dtype_asrtype returns as.integer64 for int64 dtype", {
  skip_if_not_installed("bit64")
  func <- get_dtype_asrtype("<i8")
  expect_identical(func, bit64::as.integer64)
})

test_that("raw_to_integer64 round-trips correctly", {
  skip_if_not_installed("bit64")

  vals <- bit64::as.integer64(c(0, 1, -1, 2147483648, -2147483649))
  raw_bytes <- integer64_to_raw(vals, endian = "little")
  result <- raw_to_integer64(raw_bytes, n = 5, endian = "little")
  expect_true(inherits(result, "integer64"))
  expect_true(all(result == vals))
})

test_that("raw_to_integer64 handles big endian", {
  skip_if_not_installed("bit64")

  vals <- bit64::as.integer64(c(42, 123456789012345))
  raw_bytes <- integer64_to_raw(vals, endian = "big")
  result <- raw_to_integer64(raw_bytes, n = 2, endian = "big")
  expect_true(all(result == vals))
})

test_that("NestedArray round-trips int64 data from raw", {
  skip_if_not_installed("bit64")

  vals <- bit64::as.integer64(c(1, 2, 3000000000, 4000000000))
  dtype <- Dtype$new("<i8")
  raw_bytes <- integer64_to_raw(vals, endian = "little")

  na <- NestedArray$new(data = raw_bytes, shape = c(4), dtype = dtype)
  expect_s3_class(na$data, "integer64")
  data_vec <- na$data
  dim(data_vec) <- NULL
  expect_true(all(data_vec == vals))

  # Round-trip back to raw
  raw_out <- na$flatten_to_raw()
  expect_equal(raw_out, raw_bytes)
})

test_that("NestedArray round-trips int64 2D array", {
  skip_if_not_installed("bit64")

  vals <- bit64::as.integer64(c(10, 20, 30, 40, 50, 60))
  dtype <- Dtype$new("<i8")
  raw_bytes <- integer64_to_raw(vals, endian = "little")

  na <- NestedArray$new(data = raw_bytes, shape = c(2, 3), dtype = dtype)
  expect_s3_class(na$data, "integer64")
  expect_equal(dim(na$data), c(2, 3))

  # Round-trip
  raw_out <- na$flatten_to_raw()
  expect_equal(raw_out, raw_bytes)
})

test_that("int64 NestedArray handles C order", {
  skip_if_not_installed("bit64")

  vals <- bit64::as.integer64(c(1, 2, 3, 4, 5, 6))
  dtype <- Dtype$new("<i8")
  raw_bytes <- integer64_to_raw(vals, endian = "little")

  na <- NestedArray$new(data = raw_bytes, shape = c(2, 3), dtype = dtype,
                        order = "C")
  expect_s3_class(na$data, "integer64")

  # Round-trip with C order
  raw_out <- na$flatten_to_raw(order = "C")
  expect_equal(raw_out, raw_bytes)
})

test_that("normalize_fill_value works for int64 dtype", {
  skip_if_not_installed("bit64")

  dtype <- Dtype$new("<i8")
  fv <- normalize_fill_value(0, dtype)
  expect_s3_class(fv, "integer64")
  expect_true(fv == bit64::as.integer64(0))

  fv2 <- normalize_fill_value(42, dtype)
  expect_s3_class(fv2, "integer64")
  expect_true(fv2 == bit64::as.integer64(42))
})

test_that("int64 zarr array round-trip via DirectoryStore", {
  skip_if_not_installed("bit64")

  dir <- file.path(tempdir(TRUE), "test_int64.zarr")
  on.exit(unlink(dir, recursive = TRUE), add = TRUE)

  store <- DirectoryStore$new(dir)
  za <- zarr_create(shape = c(4), chunks = c(4), dtype = "<i8",
                    store = store, overwrite = TRUE)

  vals <- bit64::as.integer64(c(1, 2, 3000000000, -3000000000))
  za$set_item("...", vals)

  result <- za$get_item("...")$data
  expect_s3_class(result, "integer64")
  result_vec <- result
  dim(result_vec) <- NULL
  expect_true(all(result_vec == vals))
})

test_that("int64 zarr array round-trip via MemoryStore", {
  skip_if_not_installed("bit64")

  store <- MemoryStore$new()
  za <- zarr_create(shape = c(6), chunks = c(3), dtype = "<i8",
                    store = store, overwrite = TRUE)

  vals <- bit64::as.integer64(c(100, 200, 300, 400, 500, 600))
  za$set_item("...", vals)

  result <- za$get_item("...")$data
  expect_s3_class(result, "integer64")
  result_vec <- result
  dim(result_vec) <- NULL
  expect_true(all(result_vec == vals))
})

library(pizzarr)

sample_dir <- tools::R_user_dir("pizzarr")
clean <- !dir.exists(sample_dir)

test_that("demo data", {
  skip_if_not_installed("crul")

  demo_data <- pizzarr_sample()
  
  expect_true(all(dir.exists(demo_data)))
  
  expect_true(all(grepl("zarr$", demo_data)))

  expect_error(pizzarr_sample("borked"), "Dataset not found")
  
  expect_true(grepl("bcsd_v2.zarr", pizzarr_sample("bcsd")))
  expect_true(grepl("bcsd_v3.zarr", pizzarr_sample("bcsd_v3")))
  
})

test_that("create_zarray_meta does not throw error when simple dtype is valid", {
  res <- create_zarray_meta(dtype = Dtype$new("|u1"), order = "C", fill_value = 0, dimension_separator = ".")
  expect_equal(class(res$dtype)[[1]], "scalar")
  expect_equal(as.character(res$dtype), "|u1")
})

test_that("create_zarray_meta throws error when simple dtype byteorder is invalid", {
  f <- function() create_zarray_meta(dtype = Dtype$new(":u1"), order = "C", fill_value = 0, dimension_separator = ".")
  expect_error(f())
})

test_that("create_zarray_meta throws error when simple dtype basic type is invalid", {
  f <- function() create_zarray_meta(dtype = Dtype$new("<Z1"), order = "C", fill_value = 0, dimension_separator = ".")
  expect_error(f())
})

test_that("create_zarray_meta throws error when order is invalid", {
  f <- function() create_zarray_meta(dtype = Dtype$new("|u1"), order = "A", fill_value = 0, dimension_separator = ".")
  expect_error(f())
})

test_that("create_zarray_meta throws error when dimension separator is invalid", {
  f <- function() create_zarray_meta(dtype = Dtype$new("|u1"), order = "C", fill_value = 0, dimension_separator = "___")
  expect_error(f())
})


test_that("create_zarray_meta throws error when fill_value is invalid for float dtype", {
  f <- function() create_zarray_meta(dtype = Dtype$new("<f2"), order = "C", fill_value = "foo")
  expect_error(f())
})


test_that("zip_numeric", {
  res <- zip_numeric(c(1, 2, 3), c(4, 5, 6))
  expect_equal(res, list(
    c(1, 4),
    c(2, 5),
    c(3, 6)
  ))
})

# Reference: https://github.com/gzuidhof/zarr.js/blob/master/test/core/indexing.test.ts#L25
# [expected, selection, shape]

# 1D, single item
test_that("replace_ellipsis [[0], 0, [100]]", {
  res <- replace_ellipsis(as_scalar(0), list(100))
  expect_equal(res, list(0))
})

# 1D
test_that("replace_ellipsis [[null], ellipsis, [100]]", {
  res <- replace_ellipsis("...", list(100))
  expect_equal(res, list(NA))
})

test_that("replace_ellipsis [[null], [null], [100]]", {
  res <- replace_ellipsis(list(NA), list(100))
  expect_equal(res, list(NA))
})

test_that("replace_ellipsis [[null], [null, ellipsis], [100]]", {
  res <- replace_ellipsis(list(NA, "..."), list(100))
  expect_equal(res, list(NA))
})

test_that("replace_ellipsis [[null], [ellipsis, null], [100]]", {
  res <- replace_ellipsis(list("...", NA), list(100))
  expect_equal(res, list(NA))
})

test_that("replace_ellipsis [[zb_slice(0, 5)], zb_slice(0, 5), [100]]", {
  res <- replace_ellipsis(zb_slice(0, 5), list(100))
  expect_equal(res, list(zb_slice(0, 5)))
})

test_that("replace_ellipsis [[zb_slice(null)], zb_slice(:), [100]]", {
  res <- replace_ellipsis(zb_slice(":"), list(100))
  expect_equal(res, list(zb_slice(NA)))
  expect_equal(res[[1]]$start, NA)
  expect_equal(res[[1]]$stop, NA)
  expect_equal(res[[1]]$step, NA)
})

test_that("replace_ellipsis [[zb_slice(null)], zb_slice(:, :), [100]]", {
  res <- replace_ellipsis(zb_slice(":", ":"), list(100))
  expect_equal(res, list(zb_slice(NA)))
})

# 2D, single item
test_that("replace_ellipsis [[0, 0], [0, 0], [100, 100]]", {
  res <- replace_ellipsis(list(0, 0), list(100, 100))
  expect_equal(res, list(0, 0))
})

test_that("replace_ellipsis [[0, 0], [0, 0], [100, 100]] with as_scalar", {
  res <- replace_ellipsis(list(as_scalar(0), as_scalar(0)), list(100, 100))
  expect_equal(res, list(as_scalar(0), as_scalar(0)))
})

test_that("replace_ellipsis [[-1, 1], [-1, 1], [100, 100]]", {
  res <- replace_ellipsis(list(-1, 1), list(100, 100))
  expect_equal(res, list(-1, 1))
})

# 2D, single col/row
test_that("replace_ellipsis [[0, null], [0, null], [100, 100]]", {
  res <- replace_ellipsis(list(0, NA), list(100, 100))
  expect_equal(res, list(0, NA))
})

test_that("replace_ellipsis [[0, zb_slice(null)], [0, zb_slice(null)], [100, 100]]", {
  res <- replace_ellipsis(list(0, zb_slice(NA)), list(100, 100))
  expect_equal(res, list(0, zb_slice(NA)))
})

test_that("replace_ellipsis [[null, 0], [null, 0], [100, 100]]", {
  res <- replace_ellipsis(list(NA, 0), list(100, 100))
  expect_equal(res, list(NA, 0))
})

# 2D
test_that("replace_ellipsis [[null, null], ellipsis, [100, 100]]", {
  res <- replace_ellipsis("...", list(100, 100))
  expect_equal(res, list(NA, NA))
})

test_that("replace_ellipsis [[null, null], [null], [100, 100]]", {
  res <- replace_ellipsis(list(NA), list(100, 100))
  expect_equal(res, list(NA, NA))
})

test_that("replace_ellipsis [[null, null], [null, null], [100, 100]]", {
  res <- replace_ellipsis(list(NA, NA), list(100, 100))
  expect_equal(res, list(NA, NA))
})

test_that("replace_ellipsis [[null, null], [ellipsis, null], [100, 100]]", {
  res <- replace_ellipsis(list("...", NA), list(100, 100))
  expect_equal(res, list(NA, NA))
})

test_that("replace_ellipsis [[null, null], [null, ellipsis], [100, 100]]", {
  res <- replace_ellipsis(list(NA, "..."), list(100, 100))
  expect_equal(res, list(NA, NA))
})

test_that("replace_ellipsis [[null, zb_slice(null)], [ellipsis, zb_slice(null)], [100, 100]]", {
  res <- replace_ellipsis(list("...", zb_slice(NA)), list(100, 100))
  expect_equal(res, list(NA, zb_slice(NA)))
})

test_that("replace_ellipsis [[null, null], [ellipsis, null, null], [100, 100]]", {
  res <- replace_ellipsis(list("...", NA, NA), list(100, 100))
  expect_equal(res, list(NA, NA))
})

test_that("replace_ellipsis [[null, null], [null, ellipsis, null], [100, 100]]", {
  res <- replace_ellipsis(list(NA, "...", NA), list(100, 100))
  expect_equal(res, list(NA, NA))
})

test_that("replace_ellipsis [[null, null], [null, null, ellipsis], [100, 100]]", {
  res <- replace_ellipsis(list(NA, NA, "..."), list(100, 100))
  expect_equal(res, list(NA, NA))
})

# Failure tests for replace_ellipsis
test_that("replace_ellipsis errors with invalid input", {
  f1 <- function() replace_ellipsis(list("...", "..."), list(100, 100))
  expect_error(f1())
  f2 <- function() replace_ellipsis(list(0, 1, 2), list(100, 100))
  expect_error(f2())
})


test_that("get_list_product", {
  in_list <- list(
    list('A', 'B', 'C', 'D'),
    list('x', 'y')
  )
  out_list <- get_list_product(in_list)
  expect_equal(out_list, list(
    list('A', 'x'),
    list('B', 'x'),
    list('C', 'x'),
    list('D', 'x'),
    list('A', 'y'),
    list('B', 'y'),
    list('C', 'y'),
    list('D', 'y')
  ))
})

test_that("NaN in json", {
  check <- try_fromJSON('{"name": NaN}')
  
  expect_equal(check, list(name = NULL))
  
  expect_warning(try_fromJSON("borked", "test warning"), "test warning")
})

if(clean) unlink(sample_dir, recursive = TRUE)

# Reference: https://github.com/zarr-developers/zarr-python/blob/5dd4a0/zarr/indexing.py#L61

# Check whether a selection contains only integer array-likes
# This is used to determine whether zarr.array$get_item() calls vectorized (inner) indexing
is_pure_fancy_indexing <- function(selection, ndim = length(selection)) {
    
  # one dimensions case
  if(ndim == 1){
    if(is_integer_vec(selection) | is_integer_list(selection)){
      return(TRUE)
    }
  } 
  
  # check if there are any slice objects 
  no_slicing <- (length(selection) == ndim) & !(any(sapply(selection, function(s) inherits(s, "Slice"))))
  
  # check for integer vectors
  all_integer <- all(sapply(selection, function(sel){
    (is_integer(sel) | is_integer_list(sel)) | is_integer_vec(sel)
  }))
  any_integer <- any(sapply(selection, function(sel){
    # is_integer_list(sel) | is_integer_vec(sel)
    is_integer_list(sel) | is_integer_vec(sel) | is_integer(sel)
  }))
  
  # return
  return((no_slicing & all_integer) & any_integer)
}


# Reference: https://github.com/zarr-developers/zarr-python/blob/5dd4a0/zarr/indexing.py#L655

#' The Zarr OIndex class.
#' @title OIndex Class
#' @docType class
#' @description
#'  Orthogonal index class
#' @rdname OIndex
#' @keywords internal
OIndex <- R6::R6Class("OIndex",
  public = list(
    #' @field array The array instance.
    #' @keywords internal
    array = NULL,
    #' @description
    #' Create a new OIndex instance.
    #' @param array The array instance.
    #' @return An `OIndex` instance.
    initialize = function(array) {
      self$array <- array
    },
    #' @description
    #' get method for the Oindex instance
    #' @param selection Selection to apply.
    #' @return An `OIndex` instance.
    get_item = function(selection) {
      # self$array <- array$get
      self$array$get_orthogonal_selection(selection)
    }
  )
)

# Reference: https://github.com/zarr-developers/zarr-python/blob/5dd4a0/zarr/indexing.py#L811

#' The Zarr VIndex class.
#' @title VIndex Class
#' @docType class
#' @description
#'  Vectorized index class
#' @rdname VIndex
#' @keywords internal
VIndex <- R6::R6Class("VIndex",
  public = list(
    #' @field array The array instance.
    #' @keywords internal
    array = NULL,
    #' @description
    #' Create a new VIndex instance.
    #' @param array The array instance.
    #' @return A `VIndex` instance.
    initialize = function(array) {
      self$array <- array
    }
  )
)

# Reference: https://github.com/zarr-developers/zarr-python/blob/5dd4a0/zarr/indexing.py#L138

#' The Zarr IntDimIndexer class.
#' @title IntDimIndexer Class
#' @docType class
#' @description
#' Indexer for a single integer selection along one dimension.
#' @rdname IntDimIndexer
#' @keywords internal
IntDimIndexer <- R6::R6Class("IntDimIndexer",
  inherit = DimIndexer,
  public = list(
    #' @field dim_sel Normalized integer index for this dimension.
    #' @keywords internal
    dim_sel = NULL,
    #' @field dim_len Length of the array along this dimension.
    #' @keywords internal
    dim_len = NULL,
    #' @field dim_chunk_len Length of a chunk along this dimension.
    #' @keywords internal
    dim_chunk_len = NULL,
    #' @description
    #' Create a new IntDimIndexer instance.
    #' @param dim_sel Integer dimension selection.
    #' @param dim_len Integer dimension length.
    #' @param dim_chunk_len Integer dimension chunk length.
    #' @return A `IntDimIndexer` instance.
    initialize = function(dim_sel, dim_len, dim_chunk_len) {
      # Normalize
      dim_sel <- normalize_integer_selection(dim_sel, dim_len)

      self$dim_sel <- dim_sel
      self$dim_len <- dim_len
      self$dim_chunk_len <- dim_chunk_len
      self$num_items <- 1
    },
    #' @description
    #' Compute the chunk dimension projection for the single selected index.
    #' @return a `ChunkDimProjection` instance
    iter = function() {
      # TODO: use generator/yield features from async package
      dim_chunk_index <- floor(self$dim_sel / self$dim_chunk_len)
      dim_offset <- dim_chunk_index * self$dim_chunk_len
      dim_chunk_sel <- self$dim_sel - dim_offset
      dim_out_sel <- NA
      return(list(
        ChunkDimProjection$new(
          dim_chunk_index,
          dim_chunk_sel,
          dim_out_sel
        )
      ))
    }
  )
)

# Reference: https://github.com/zarr-developers/zarr-python/blob/5dd4a0/zarr/indexing.py#L163

#' The Zarr SliceDimIndexer class.
#' @title SliceDimIndexer Class
#' @docType class
#' @description
#' Indexer for a slice selection along one dimension.
#' @rdname SliceDimIndexer
#' @keywords internal
SliceDimIndexer <- R6::R6Class("SliceDimIndexer",
  inherit = DimIndexer,
  public = list(
    #' @field dim_len Dimension length.
    #' @keywords internal
    dim_len = NULL,
    #' @field dim_chunk_len Dimension chunk length.
    #' @keywords internal
    dim_chunk_len = NULL,
    #' @field num_chunks Number of chunks.
    #' @keywords internal
    num_chunks = NULL,
    #' @field start Start index.
    #' @keywords internal
    start = NULL,
    #' @field stop Stop index.
    #' @keywords internal
    stop = NULL,
    #' @field step Step size.
    #' @keywords internal
    step = NULL,
    #' @description
    #' Create a new SliceDimIndexer instance.
    #' @param dim_sel Integer dimension selection.
    #' @param dim_len Integer dimension length.
    #' @param dim_chunk_len Integer dimension chunk length.
    #' @return A `SliceDimIndexer` instance.
    initialize = function(dim_sel, dim_len, dim_chunk_len) {
      # Reference: https://github.com/gzuidhof/zarr.js/blob/292804/src/core/indexing.ts#L311
      si <- dim_sel$indices(dim_len)
      self$start <- si[1]
      self$stop <- si[2]
      self$step <- si[3]
      if(self$step < 1) {
        stop('NegativeStepError')
      }
      self$dim_len <- dim_len
      self$dim_chunk_len <- dim_chunk_len
      self$num_items <- max(0, ceiling((self$stop - self$start) / self$step))
      self$num_chunks <- ceiling(self$dim_len / self$dim_chunk_len)
    },
    #' @description
    #' Iterate over chunks that overlap the slice, yielding a ChunkDimProjection for each.
    #' @return A list of ChunkDimProjection objects.
    iter = function() {
      # TODO: use generator/yield features from async package
      dim_chunk_index_from <- floor(self$start / self$dim_chunk_len)
      dim_chunk_index_to <- ceiling(self$stop / self$dim_chunk_len)

      # START R-SPECIFIC
      if(dim_chunk_index_from == dim_chunk_index_to) {
        dim_chunk_index_to <- dim_chunk_index_to + 1
      }
      # END R-SPECIFIC

      # Iterate over chunks in range
      result <- list()
      for(dim_chunk_index in seq(from = dim_chunk_index_from, to = (dim_chunk_index_to - 1), by = 1)) {

        # Compute offsets for chunk within overall array
        dim_offset <- dim_chunk_index * self$dim_chunk_len
        dim_limit <- min(self$dim_len, (dim_chunk_index + 1) * self$dim_chunk_len)

        # Determine chunk length, accounting for trailing chunk
        dim_chunk_len <- dim_limit - dim_offset

        dim_chunk_sel_start <- 0
        dim_chunk_sel_stop <- 0
        dim_out_offset <- 0

        if(self$start < dim_offset) {
          # Selection starts before the current chunk

          dim_chunk_sel_start <- 0
          remainder <- (dim_offset - self$start) %% self$step
          if(remainder > 0) {
            dim_chunk_sel_start <- dim_chunk_sel_start + (self$step - remainder)
          }
          # Compute number of previous items, provides offset into output array
          dim_out_offset <- ceiling((dim_offset - self$start) / self$step)
        } else {
          # Selection starts within the current chunk
          dim_chunk_sel_start <- self$start - dim_offset
          dim_out_offset <- 0
        }

        if(self$stop > dim_limit) {
          # Selection ends after current chunk
          dim_chunk_sel_stop <- self$dim_chunk_len
        } else {
          # Selection ends within current chunk
          dim_chunk_sel_stop <- self$stop - dim_offset
        }

        dim_chunk_sel <- zb_slice(dim_chunk_sel_start, dim_chunk_sel_stop, self$step)
        dim_chunk_num_items <- ceiling((dim_chunk_sel_stop - dim_chunk_sel_start) / self$step)
        dim_out_sel <- zb_slice(dim_out_offset, dim_out_offset + dim_chunk_num_items)

        result <- append(result, ChunkDimProjection$new(
          dim_chunk_index,
          dim_chunk_sel,
          dim_out_sel
        ))
      }
      return(result)
    }
  )
)

# Reference: https://github.com/zarr-developers/zarr-python/blob/5dd4a0/zarr/indexing.py#L326

#' The Zarr BasicIndexer class.
#' @title BasicIndexer Class
#' @docType class
#' @description
#'  An indexer class to normalize a selection of an array and provide an iterator 
#'  of indexes over the dimensions of an array.
#' @rdname BasicIndexer
#' @keywords internal
BasicIndexer <- R6::R6Class("BasicIndexer",
  inherit = Indexer,
  public = list(
    #' @field dim_indexers List of per-dimension indexers (IntDimIndexer or SliceDimIndexer).
    #' @keywords internal
    dim_indexers = NULL,
    #' @description
    #' Create a new BasicIndexer instance.
    #' @param selection Selection as with ZarrArray: scalar, string, or Slice.
    #' @param array [ZarrArray] object that will be indexed.
    #' @return A `BasicIndexer` instance.
    initialize = function(selection, array) {

      shape <- array$get_shape()
      chunks <- array$get_chunks()

      selection <- normalize_list_selection(selection, shape)
      
      # Setup per-dimension indexers
      dim_indexers <- vector("list", length(selection))
      for(i in seq_along(selection)) {
        dim_sel <- selection[[i]]
        dim_len <- shape[i]
        dim_chunk_len <- chunks[i]

        if(is.null(dim_sel)) {
          dim_sel <- zb_slice(NA)
        }

        if(is_integer(dim_sel)) {
          dim_indexers[[i]] <- IntDimIndexer$new(dim_sel, dim_len, dim_chunk_len)
        } else if(is_slice(dim_sel)) {
          dim_indexers[[i]] <- SliceDimIndexer$new(dim_sel, dim_len, dim_chunk_len)
        } else {
          stop('Unsupported selection item for basic indexing, expected integer or slice')
        }
      }
      self$shape <- list()
      for(d in dim_indexers) {
        if(inherits(d, "SliceDimIndexer")) {
          self$shape <- append(self$shape, d$num_items)
        }
      }
      self$drop_axes <- NA

      self$dim_indexers <- dim_indexers
    },
    #' @description 
    #'   An iterator over the dimensions of an array
    #' @return A list of ChunkProjection objects
    iter = function() {
      # TODO: use generator/yield features from async package
      result <- list()

      # dim_indexers is a list of DimIndexer objects.
      # dim_indexer_iterables is a list (one per dimension)
      # of lists of IntDimIndexer or SliceDimIndexer objects.
      dim_indexer_iterables <- lapply(self$dim_indexers, function(di) di$iter())
      dim_indexer_product <- get_list_product(dim_indexer_iterables)


      for(row_i in seq_len(length(dim_indexer_product))) {
        dim_proj <- dim_indexer_product[[row_i]]

        chunk_coords <- list()
        chunk_sel <- list()
        out_sel <- list()

        if(!is.list(dim_proj)) {
          dim_proj <- list(dim_proj)
        }

        for(p in dim_proj) {
          chunk_coords <- append(chunk_coords, p$dim_chunk_index)
          chunk_sel <- append(chunk_sel, p$dim_chunk_sel)
          if(!is_na(p$dim_out_sel)) {
            out_sel <- append(out_sel, p$dim_out_sel)
          }
        }

        result <- append(result, ChunkProjection$new(
          chunk_coords,
          chunk_sel,
          out_sel
        ))
      }

      return(result)
    }
  )
)

# Reference: https://github.com/zarr-developers/zarr-python/blob/5dd4a0e6cdc04c6413e14f57f61d389972ea937c/zarr/indexing.py#L585

#' The Zarr OrthogonalIndexer class.
#' @title OrthogonalIndexer Class
#' @docType class
#' @description
#'  An indexer class to normalize a selection of an array and provide an iterator 
#'  of indexes over the dimensions of an array.
#' @rdname OrthogonalIndexer
#' @keywords internal
OrthogonalIndexer <- R6::R6Class("OrthogonalIndexer",
                            inherit = Indexer,
                            public = list(
                              #' @field dim_indexers List of per-dimension indexers (IntDimIndexer, SliceDimIndexer, IntArrayDimIndexer, or BoolArrayDimIndexer).
                              #' @keywords internal
                              dim_indexers = NULL,
                              #' @description
                              #' Create a new OrthogonalIndexer instance.
                              #' @param selection Selection as with ZarrArray: scalar, string, or Slice.
                              #' @param array [ZarrArray] object that will be indexed.
                              #' @return An `OrthogonalIndexer` instance.
                              initialize = function(selection, array) {
                                
                                shape <- array$get_shape()
                                chunks <- array$get_chunks()
                                
                                # Normalize
                                selection <- normalize_list_selection(selection, shape)
                                
                                # Setup per-dimension indexers
                                dim_indexers <- vector("list", length(selection))
                                for(i in seq_along(selection)) {
                                  dim_sel <- selection[[i]]
                                  dim_len <- shape[i]
                                  dim_chunk_len <- chunks[i]

                                  if(is.null(dim_sel)) {
                                    dim_sel <- zb_slice(NA)
                                  }

                                  if(is_integer(dim_sel)) {
                                    dim_indexers[[i]] <- IntDimIndexer$new(dim_sel, dim_len, dim_chunk_len)
                                  } else if(is_slice(dim_sel)) {
                                    if(!is.na(dim_sel$step) && dim_sel$step < 0) {
                                      # Convert negative-step slice to integer vector
                                      si <- dim_sel$indices(dim_len)
                                      n <- si[4]
                                      if(n > 0) {
                                        dim_sel <- seq(from = si[1], by = si[3], length.out = n)
                                        dim_indexers[[i]] <- IntArrayDimIndexer$new(dim_sel, dim_len, dim_chunk_len)
                                      } else {
                                        dim_indexers[[i]] <- IntArrayDimIndexer$new(integer(0), dim_len, dim_chunk_len)
                                      }
                                    } else {
                                      dim_indexers[[i]] <- SliceDimIndexer$new(dim_sel, dim_len, dim_chunk_len)
                                    }
                                  } else if(is_bool_vec(dim_sel)) {
                                    dim_indexers[[i]] <- BoolArrayDimIndexer$new(dim_sel, dim_len, dim_chunk_len)
                                  } else if(is_integer_vec(dim_sel)) {
                                    dim_indexers[[i]] <- IntArrayDimIndexer$new(dim_sel, dim_len, dim_chunk_len)
                                  } else {
                                    stop('Unsupported selection item for basic indexing, expected integer, slice, vector of integer or boolean')
                                  }
                                }
                                self$shape <- list()
                                for(d in dim_indexers) {
                                  if(!inherits(d, "IntDimIndexer")) {
                                    self$shape <- append(self$shape, d$num_items)
                                  }
                                }
                                self$drop_axes <- NA
                                
                                self$dim_indexers <- dim_indexers
                                
                              },
                              #' @description 
                              #'   An iterator over the dimensions of an array
                              #' @return A list of ChunkProjection objects
                              iter = function() {
                                
                                # TODO: use generator/yield features from async package
                                result <- list()
                                
                                # dim_indexers is a list of DimIndexer objects.
                                # dim_indexer_iterables is a list (one per dimension)
                                # of lists of IntDimIndexer or SliceDimIndexer objects.
                                dim_indexer_iterables <- lapply(self$dim_indexers, function(di) di$iter())
                                dim_indexer_product <- get_list_product(dim_indexer_iterables)
                                
                                
                                for(row_i in seq_len(length(dim_indexer_product))) {
                                  dim_proj <- dim_indexer_product[[row_i]]
                                  
                                  chunk_coords <- list()
                                  chunk_sel <- list()
                                  out_sel <- list()
                                  
                                  if(!is.list(dim_proj)) {
                                    dim_proj <- list(dim_proj)
                                  }
                                  
                                  for(p in dim_proj) {
                                    chunk_coords <- append(chunk_coords, p$dim_chunk_index)
                                    # chunk_sel <- append(chunk_sel, p$dim_chunk_sel)
                                    chunk_sel <- append(chunk_sel, list(p$dim_chunk_sel))
                                    if(!is_na(p$dim_out_sel)) {
                                      # out_sel <- append(out_sel, p$dim_out_sel)
                                      out_sel <- append(out_sel, list(p$dim_out_sel))
                                    }
                                  }
                                  
                                  result <- append(result, ChunkProjection$new(
                                    chunk_coords,
                                    chunk_sel,
                                    out_sel
                                  ))
                                }
                                
                                return(result)
                              }
                            )
)

# Reference: https://github.com/zarr-developers/zarr-python/blob/5dd4a0e6cdc04c6413e14f57f61d389972ea937c/zarr/indexing.py#L424

#' The Order class.
#' @title Order Class
#' @docType class
#' @description
#'  Determines the sort order of an integer array selection (increasing, decreasing, or unordered).
#' @rdname Order
#' @keywords internal
Order <- R6::R6Class("Order",
                     public = list(
                       #' @field UNKNOWN UNKNOWN
                       #' @keywords internal
                       UNKNOWN = 0,
                       #' @field INCREASING INCREASING
                       #' @keywords internal
                       INCREASING = 1,
                       #' @field DECREASING DECREASING
                       #' @keywords internal
                       DECREASING = 2,
                       #' @field UNORDERED UNORDERED
                       #' @keywords internal
                       UNORDERED = 3,
                       #' @description
                       #' checking order of numbers
                       #' @param a Vector of numbers.
                       check = function(a){
                         diff_a <- diff(a)
                         diff_positive <- diff_a >= 0
                         n_diff_positive <- sum(diff_positive)
                         all_increasing <- n_diff_positive == length(diff_positive)
                         any_increasing <- n_diff_positive > 0
                         if(all_increasing){
                           return(Order$public_fields$INCREASING)
                         } else if(any_increasing) {
                           return(Order$public_fields$UNORDERED)
                         } else{
                           return(Order$public_fields$DECREASING)
                         }
                       })
                     )

# Reference: https://github.com/zarr-developers/zarr-python/blob/5dd4a0e6cdc04c6413e14f57f61d389972ea937c/zarr/indexing.py#L457

#' The Zarr IntArrayDimIndexer class.
#' @title IntArrayDimIndexer Class
#' @docType class
#' @description
#'  Indexer for an integer array selection along one dimension, supporting fancy/advanced indexing.
#' @rdname IntArrayDimIndexer
#' @keywords internal
IntArrayDimIndexer <- R6::R6Class("IntArrayDimIndexer",
                               inherit = DimIndexer,
                               public = list(
                                 #' @field dim_len Dimension length.
                                 #' @keywords internal
                                 dim_len = NULL,
                                 #' @field dim_chunk_len Dimension chunk length.
                                 #' @keywords internal
                                 dim_chunk_len = NULL,
                                 #' @field num_chunks Number of chunks.
                                 #' @keywords internal
                                 num_chunks = NULL,
                                 #' @field dim_sel Selection on dimension.
                                 #' @keywords internal
                                 dim_sel = NULL,
                                 #' @field dim_out_sel Output array indices corresponding to each selected element, reordered when selection is not increasing.
                                 #' @keywords internal
                                 dim_out_sel = NULL,
                                 #' @field order Memory order.
                                 #' @keywords internal
                                 order = NULL,
                                 #' @field chunk_nitems Number of items per chunk.
                                 #' @keywords internal
                                 chunk_nitems = NULL, 
                                 #' @field dim_chunk_ixs Chunks that should be visited.
                                 #' @keywords internal
                                 dim_chunk_ixs = NULL,
                                 #' @field chunk_nitems_cumsum Offsets into the output array.
                                 #' @keywords internal
                                 chunk_nitems_cumsum = NULL, 
                                 #' @description
                                 #' Create a new IntArrayDimIndexer instance.
                                 #' @param dim_sel Integer dimension selection.
                                 #' @param dim_len Integer dimension length.
                                 #' @param dim_chunk_len Integer dimension chunk length.
                                 #' @param sel_order order
                                 #' @return A `IntArrayDimIndexer` instance.
                                 initialize = function(dim_sel, dim_len, dim_chunk_len, sel_order = Order$public_fields$UNKNOWN) {

                                   # Normalize
                                   dim_sel <- sapply(dim_sel, normalize_integer_selection, dim_len = dim_len)
                                   self$dim_sel <- dim_sel
                                   
                                   # store attributes 
                                   self$dim_len <- dim_len
                                   self$dim_chunk_len <- dim_chunk_len
                                   self$num_items <- length(dim_sel)
                                   self$num_chunks <- ceiling(self$dim_len / self$dim_chunk_len)
                                   
                                   # dim_sel_chunk <- ceiling(dim_sel / dim_chunk_len) # pre zb_int() implementation
                                   dim_sel_chunk <- floor(dim_sel / dim_chunk_len)
                                   
                                   # determine order of indices
                                   if(sel_order == Order$public_fields$UNKNOWN)
                                     sel_order <- Order$public_methods$check(dim_sel)
                                   self$order <- sel_order
                                   
                                   if(self$order == Order$public_fields$INCREASING){
                                     self$dim_sel <-  dim_sel
                                   } else if(self$order == Order$public_fields$DECREASING) {
                                     self$dim_sel = rev(dim_sel)
                                     self$dim_out_sel = rev(seq(1,self$num_items))
                                     # self$dim_out_sel = rev(seq(0,self$num_items-1)) # Python based indexing
                                   } else {
                                     # sort indices to group by chunk
                                     self$dim_out_sel = order(dim_sel_chunk)
                                     self$dim_sel <- dim_sel[self$dim_out_sel]
                                     # self$dim_out_sel <- self$dim_out_sel - 1 # Python based indexing
                                   }
                                   
                                   # precompute number of selected items for each chunk
                                   # self$chunk_nitems <- tabulate(dim_sel_chunk, nbins = self$num_chunks) # pre zb_int() implementation
                                   self$chunk_nitems <- tabulate(dim_sel_chunk + 1, nbins = self$num_chunks)
                                   
                                   # find chunks that we need to visit
                                   self$dim_chunk_ixs = which(self$chunk_nitems != 0)
                                   
                                   # compute offsets into the output array
                                   self$chunk_nitems_cumsum = cumsum(self$chunk_nitems)
                                   
                                 },
                                 #' @description 
                                 #'   An iterator over the dimensions of an array
                                 #' @return A list of ChunkProjection objects
                                 iter = function() {
                                   
                                   # Iterate over chunks in range
                                   result <- list()
                                   for(dim_chunk_ix in self$dim_chunk_ixs) {
                                     
                                     # find region in output
                                     # if (dim_chunk_ix == 0) {
                                     if (dim_chunk_ix == 1) { 
                                       start <- 0
                                     } else {
                                       start <- self$chunk_nitems_cumsum[dim_chunk_ix - 1]
                                     }
                                     stop <- self$chunk_nitems_cumsum[dim_chunk_ix]
                                     
                                     # START R-SPECIFIC
                                     if(start == stop) {
                                       stop <- stop + 1
                                     }
                                     # END R-SPECIFIC
                                     
                                     if (self$order == Order$public_fields$INCREASING) {
                                       dim_out_sel <- seq(start, stop - 1)
                                     } else {
                                       dim_out_sel <- self$dim_out_sel[(start + 1):stop]
                                       # START R-SPECIFIC
                                       dim_out_sel <- dim_out_sel - 1
                                       # END R-SPECIFIC
                                     }

                                     # START R-SPECIFIC
                                     dim_chunk_ix <- dim_chunk_ix - 1
                                     # END R-SPECIFIC
                                     
                                     # find region in chunk
                                     dim_offset <- dim_chunk_ix * self$dim_chunk_len
                                     # dim_chunk_sel <- self$dim_sel[(start + 1):stop] - dim_offset - 1 # pre zb_int implementation()
                                     dim_chunk_sel <- self$dim_sel[(start + 1):stop] - dim_offset

                                     # # START R-SPECIFIC
                                     # dim_chunk_ix <- dim_chunk_ix - 1
                                     # # END R-SPECIFIC

                                     result <- append(result, ChunkDimProjection$new(
                                       dim_chunk_ix,
                                       dim_chunk_sel,
                                       dim_out_sel
                                     ))
                                   }
                                   return(result)
                                 }
                               )
)

# Reference: https://github.com/zarr-developers/zarr-python/blob/4a3bbf1cbb89e90ea9ca4d6d75dae23ed4b957c9/src/zarr/core/indexing.py#L581
#' The Zarr BoolArrayDimIndexer class.
#' @title BoolArrayDimIndexer Class
#' @docType class
#' @description
#'  Indexer for a boolean array selection along one dimension, selecting elements where the mask is TRUE.
#' @rdname BoolArrayDimIndexer
#' @keywords internal
BoolArrayDimIndexer <- R6::R6Class("BoolArrayDimIndexer",
                                   inherit = DimIndexer,
                                   public = list(
                                     #' @field dim_sel Selection on dimension.
                                     #' @keywords internal
                                     dim_sel = NULL,
                                     #' @field dim_len Dimension length.
                                     #' @keywords internal
                                     dim_len = NULL,
                                     #' @field dim_chunk_len Dimension chunk length.
                                     #' @keywords internal
                                     dim_chunk_len = NULL,
                                     #' @field num_chunks Number of chunks.
                                     #' @keywords internal
                                     num_chunks = NULL,
                                     #' @field chunk_nitems Number of items per chunk.
                                     #' @keywords internal
                                     chunk_nitems = NULL, 
                                     #' @field chunk_nitems_cumsum Offsets into the output array.
                                     #' @keywords internal
                                     chunk_nitems_cumsum = NULL, 
                                     #' @field dim_chunk_ixs Chunks that should be visited.
                                     #' @keywords internal
                                     dim_chunk_ixs = NULL,
                                     #' @field dim_out_sel Output array indices for the selected elements.
                                     #' @keywords internal
                                     dim_out_sel = NULL,
                                     #' @description
                                     #' Create a new BoolArrayDimIndexer instance.
                                     #' @param dim_sel Integer dimension selection.
                                     #' @param dim_len Integer dimension length.
                                     #' @param dim_chunk_len Integer dimension chunk length.
                                     #' @return A `BoolArrayDimIndexer` instance.
                                     initialize = function(dim_sel, dim_len, dim_chunk_len) {
                                       
                                       # check selection length
                                       if(length(dim_sel) != dim_len)
                                         stop(paste0("IndexError: Boolean vector has the wrong length for dimension; expected ", dim_len, ", got ", length(dim_sel)))
                                       
                                       # precompute number of selected items for each chunk
                                       num_chunks <- ceiling(dim_len / dim_chunk_len)
                                       chunk_nitems <- rep(0, num_chunks)
                                       for(dim_chunk_ix in 1:num_chunks){
                                         dim_offset <- ((dim_chunk_ix - 1) * dim_chunk_len) + 1
                                         # START R-SPECIFIC
                                         dim_offset_limits <- dim_offset+dim_chunk_len-1
                                         dim_offset_limits <- ifelse(dim_offset_limits > length(dim_sel), length(dim_sel), dim_offset_limits)
                                         # STOP R-SPECIFIC
                                         
                                         chunk_nitems[dim_chunk_ix] <- sum(dim_sel[dim_offset:dim_offset_limits] != 0)
                                       }
                                       
                                       # compute offsets into the output array
                                       chunk_nitems_cumsum <- cumsum(chunk_nitems)
                                       num_items <- rev(chunk_nitems_cumsum)[1]
                                       
                                       # find chunks that we need to visit
                                       dim_chunk_ixs <- which(chunk_nitems != 0)
                                       
                                       # store attributes
                                       self$dim_sel <- dim_sel
                                       self$dim_len <- dim_len
                                       self$dim_chunk_len <- dim_chunk_len
                                       self$num_chunks <- num_chunks
                                       self$chunk_nitems <- chunk_nitems
                                       self$chunk_nitems_cumsum <- chunk_nitems_cumsum
                                       self$num_items <- num_items
                                       self$dim_chunk_ixs <- dim_chunk_ixs
                                     },
                                     #' @description 
                                     #'   An iterator over the dimensions of an array
                                     #' @return A list of ChunkProjection objects
                                     iter = function() {
                                       
                                       # Iterate over chunks in range
                                       result <- list()
                                       for(dim_chunk_ix in self$dim_chunk_ixs) {
                                         
                                         # find region in chunk
                                         dim_offset <- ((dim_chunk_ix - 1) * self$dim_chunk_len) + 1
                                         dim_chunk_sel <- self$dim_sel[dim_offset:(dim_offset+self$dim_chunk_len-1)]
                                         
                                         # pad out if final chunk
                                         if(length(dim_chunk_sel) < length(self$dim_chunk_len)){
                                           tmp <- rep(FALSE, self$dim_chunk_len)
                                           tmp[1:length(dim_chunk_sel)] <- dim_chunk_sel
                                           dim_chunk_sel <- tmp
                                         }
                                         
                                         # find region in output
                                         if (dim_chunk_ix == 1) { 
                                           start <- 0
                                         } else {
                                           start <- self$chunk_nitems_cumsum[dim_chunk_ix - 1]
                                         }
                                         stop <- self$chunk_nitems_cumsum[dim_chunk_ix]
                                         
                                         # START R-SPECIFIC
                                         if(start == stop) {
                                           stop <- stop + 1
                                         }
                                         # END R-SPECIFIC
                                         
                                         # get out selection
                                         dim_out_sel <- seq(start, stop - 1)
                                         
                                         # make boolean as integer, specific to pizzarr
                                         dim_chunk_sel <- which(dim_chunk_sel) - 1
                                         
                                         # START R-SPECIFIC
                                         dim_chunk_ix <- dim_chunk_ix - 1
                                         # END R-SPECIFIC
                                         
                                         result <- append(result, ChunkDimProjection$new(
                                           dim_chunk_ix,
                                           dim_chunk_sel,
                                           dim_out_sel
                                         ))
                                         
                                       }
                                       
                                       return(result)
                                     }
                                   )
)

test_that("selection functionality", {
  
  a <- zarr_volcano()$get_item("volcano")
  
  sub_a <- a$get_item(list(slice(1, 10), "..."))
  
  expect_equal(sub_a$shape, c(10, a$get_shape()[2]))
  
  sub_a <- a$get_item(list(":", slice(1, 10)))
  
  expect_equal(sub_a$shape, c(a$get_shape()[1], 10))
  
  sub_a <- a$get_item(list(1, "..."))
  
  expect_equal(sub_a$shape, c(1, a$get_shape()[2]))
  
  sub_a <- a$get_item(list("...", slice(1, 10)))
  
  expect_equal(sub_a$shape, c(a$get_shape()[1], 10))
  
})

test_that("checking pure fancy indexing", {
  
  # check integer input, list vector or scalar
  expect_equal(is_pure_fancy_indexing(list(1:3, 1, 1:2)), TRUE)
  expect_equal(is_pure_fancy_indexing(list(1:5, 1, 1:2)), TRUE)
  expect_equal(is_pure_fancy_indexing(list(3, 1, 10)), TRUE)
  expect_equal(is_pure_fancy_indexing(c(1,5,2)), TRUE)
  expect_equal(is_pure_fancy_indexing(list(3, list(1,2,3), 10)), TRUE)
  
  # checking slices
  expect_equal(is_pure_fancy_indexing(list(1:5, 1, slice(1,2))), FALSE)
  expect_equal(is_pure_fancy_indexing(list(1:5, slice(1,3), slice(1,2))), FALSE)
  
  # checking non-integers
  expect_equal(is_pure_fancy_indexing(list(1:5, 1.2, 1:2)), FALSE)
  expect_equal(is_pure_fancy_indexing(c(1,5,2.4)), FALSE)
  expect_equal(is_pure_fancy_indexing(list(3, list(1,2,3.3), 10)), FALSE)
  
})

test_that("Order class", {
  
  # fields
  expect_equal(Order$public_fields$UNKNOWN, 0)
  expect_equal(Order$public_fields$INCREASING,1)
  expect_equal(Order$public_fields$DECREASING,2)
  expect_equal(Order$public_fields$UNORDERED,3)
  
  # methods
  expect_equal(Order$public_methods$check(1:5),1)
  expect_equal(Order$public_methods$check(seq(2,10,2)),1)
  expect_equal(Order$public_methods$check(10:1),2)
  expect_equal(Order$public_methods$check(c(1,5,2,6)),3)
  expect_error(Order$public_methods$check(c("a", 2)))
})

# Reference: https://github.com/zarr-developers/zarr-python/blob/5dd4a0e6cdc04c6413e14f57f61d389972ea937c/zarr/meta.py#L14

#' @title Zarr V2 Metadata Codec
#'
#' @description
#' Handles encoding and decoding of Zarr V2 metadata. Provides methods
#' to parse raw JSON bytes into R lists and to serialize metadata back
#' to JSON bytes, for both array and group nodes. Array metadata includes
#' shape, chunks, dtype, compressor, fill_value, order, and filters.
#' Group metadata contains only the zarr_format field.
#'
#' @format [R6::R6Class] object.
#'
#' @family Metadata classes
#' @keywords internal
Metadata2 <- R6::R6Class("Metadata2",
    private = list(
        ZARR_FORMAT = 2
    ),
    public = list(
        #' @description
        #' Parse metadata from raw bytes or return as-is if already a list.
        #'
        #' @param s (`raw()` | `list()` | `NULL`)\cr
        #'   Raw JSON bytes, a pre-parsed list, or `NULL`.
        #' @param auto_unbox (`logical(1)`)\cr
        #'   Passed to [jsonlite::toJSON()]. Default `FALSE`.
        #' @return `list()` or `NULL`.
        decode_metadata = function(s, auto_unbox=FALSE) {
            if(is.list(s) || is.null(s)) {
                return(s)
            } else {
                return(try_fromJSON(rawToChar(s), simplifyVector = FALSE))
            }
        },
        #' @description
        #' Encode metadata list to raw JSON bytes.
        #'
        #' @param meta (`list()`)\cr
        #'   Metadata to encode.
        #' @param auto_unbox (`logical(1)`)\cr
        #'   Passed to [jsonlite::toJSON()]. Default `FALSE`.
        #' @return `raw()`.
        encode_metadata = function(meta, auto_unbox=FALSE) {
            return(charToRaw(jsonlite::toJSON(meta, auto_unbox = auto_unbox)))
        },
        #' @description
        #' Decode and validate V2 array metadata.
        #'
        #' @param s (`raw()` | `list()` | `NULL`)\cr
        #'   Raw JSON bytes or pre-parsed list.
        #' @return `list()` or `NULL`.
        decode_array_metadata = function(s) {
            meta <- self$decode_metadata(s)
            if(!is.null(meta)) validate_v2_meta(meta)
            return(meta)
        },
        #' @description
        #' Decode and validate V2 group metadata.
        #'
        #' @param s (`raw()` | `list()` | `NULL`)\cr
        #'   Raw JSON bytes or pre-parsed list.
        #' @return `list()` or `NULL`.
        decode_group_metadata = function(s) {
            meta <- self$decode_metadata(s)
            if(!is.null(meta)) validate_v2_meta(meta)
            return(meta)
        },
        #' @description
        #' Encode array metadata to raw JSON bytes, setting zarr_format to 2.
        #'
        #' @param meta (`list()`)\cr
        #'   Array metadata list.
        #' @return `raw()`.
        encode_array_metadata = function(meta) {
            clean_meta <- meta
            clean_meta[['zarr_format']] <- jsonlite::unbox(private$ZARR_FORMAT)
            return(self$encode_metadata(clean_meta))
        },
        #' @description
        #' Encode group metadata to raw JSON bytes, setting zarr_format to 2.
        #'
        #' @param meta (`list()` | `NA`)\cr
        #'   Ignored; a fresh group metadata list is created.
        #' @return `raw()`.
        encode_group_metadata = function(meta = NA) {
            meta <- obj_list()
            meta[['zarr_format']] <- jsonlite::unbox(private$ZARR_FORMAT)
            return(self$encode_metadata(meta))
        }
    )
)

validate_v2_meta <- function(meta) {
  if(meta$zarr_format != 2) stop("unsupported zarr format ", meta$zarr_format)
}

#' @keywords internal
try_from_zmeta <- function(key, store) {
  store$get_consolidated_metadata()$metadata[[key]]
}

#' Encode a fill_value for V3 zarr.json.
#' V3 requires special float values as JSON strings: NaN, Infinity, -Infinity.
#' Normal numeric values are returned as unboxed scalars.
#' @param fill_value The fill value to encode.
#' @return A value suitable for inclusion in a jsonlite::toJSON call.
#' @keywords internal
decode_fill_value_v3 <- function(fill_value) {
  if (is.character(fill_value) && length(fill_value) == 1) {
    if (fill_value == "NaN") return(NaN)
    if (fill_value == "Infinity") return(Inf)
    if (fill_value == "-Infinity") return(-Inf)
  }
  fill_value
}

encode_fill_value_v3 <- function(fill_value) {
  if (is.numeric(fill_value) && length(fill_value) == 1) {
    # NOTE: in R, is.na(NaN) == TRUE, so check is.nan() first
    if (is.nan(fill_value)) return(jsonlite::unbox("NaN"))
    if (is.infinite(fill_value) && fill_value > 0) return(jsonlite::unbox("Infinity"))
    if (is.infinite(fill_value) && fill_value < 0) return(jsonlite::unbox("-Infinity"))
    if (is.na(fill_value)) return(jsonlite::unbox(0))
    return(jsonlite::unbox(fill_value))
  }
  if (is.logical(fill_value) && length(fill_value) == 1) {
    return(jsonlite::unbox(fill_value))
  }
  jsonlite::unbox(fill_value)
}

try_fromJSON <- function(json, warn_message = "Error parsing json was",
                         simplifyVector = FALSE) {
  tryCatch(
    jsonlite::fromJSON(json, simplifyVector),
    error = \(e) {
      # Bare (unquoted) NaN is invalid JSON but appears in some V2 metadata.
      # Only replace unquoted NaN — quoted "NaN" is valid JSON (V3 fill_value).
      if (grepl("(?<!\")NaN(?!\")", json, perl = TRUE)) {
        tryCatch(
          jsonlite::fromJSON(gsub("(?<!\")NaN(?!\")", "null", json, perl = TRUE),
                             simplifyVector),
          error = \(e2) { warning("\n\n", warn_message, "\n\n", e2); NULL }
        )
      } else {
        warning("\n\n", warn_message, "\n\n", e)
        NULL
      }
    }
  )
}

#' @title Zarr V3 Metadata Codec
#'
#' @description
#' Handles decoding of Zarr V3 metadata. In V3, each node has a single
#' `zarr.json` file containing `zarr_format`, `node_type`, and all
#' metadata (shape, data_type, codecs, chunk_grid, chunk_key_encoding,
#' fill_value, attributes, etc.). This class is read-only; no encode
#' methods are provided.
#'
#' @format [R6::R6Class] object.
#'
#' @family Metadata classes
#' @keywords internal
Metadata3 <- R6::R6Class("Metadata3",
  private = list(
    ZARR_FORMAT = 3
  ),
  public = list(
    #' @description
    #' Parse metadata from raw bytes or return as-is if already a list.
    #'
    #' @param s (`raw()` | `list()` | `NULL`)\cr
    #'   Raw JSON bytes, a pre-parsed list, or `NULL`.
    #' @return `list()` or `NULL`.
    decode_metadata = function(s) {
      if (is.list(s) || is.null(s)) {
        return(s)
      } else {
        return(try_fromJSON(rawToChar(s), simplifyVector = FALSE))
      }
    },
    #' @description
    #' Decode and validate V3 array metadata from `zarr.json`.
    #' V3 arrays have `node_type = "array"` and contain shape, data_type,
    #' chunk_grid, chunk_key_encoding, codecs, fill_value, and attributes.
    #'
    #' @param s (`raw()` | `list()` | `NULL`)\cr
    #'   Raw JSON bytes or pre-parsed list.
    #' @return `list()` or `NULL`.
    decode_array_metadata = function(s) {
      meta <- self$decode_metadata(s)
      if (!is.null(meta)) validate_v3_meta(meta, expected_node_type = "array")
      return(meta)
    },
    #' @description
    #' Decode and validate V3 group metadata from `zarr.json`.
    #' V3 groups have `node_type = "group"` and may contain attributes.
    #'
    #' @param s (`raw()` | `list()` | `NULL`)\cr
    #'   Raw JSON bytes or pre-parsed list.
    #' @return `list()` or `NULL`.
    decode_group_metadata = function(s) {
      meta <- self$decode_metadata(s)
      if (!is.null(meta)) validate_v3_meta(meta, expected_node_type = "group")
      return(meta)
    },
    #' @description
    #' Encode V3 array metadata to raw JSON bytes.
    #' Expects a list built by `create_v3_array_meta()` with all scalars
    #' already wrapped in `jsonlite::unbox()`.
    #'
    #' @param meta (`list()`)\cr
    #'   V3 array metadata list.
    #' @return `raw()`.
    encode_array_metadata = function(meta) {
      return(charToRaw(jsonlite::toJSON(meta, auto_unbox = FALSE)))
    },
    #' @description
    #' Encode V3 group metadata to raw JSON bytes.
    #' Produces `{"zarr_format":3,"node_type":"group","attributes":{}}`.
    #'
    #' @param meta (`list()` | `NA`)\cr
    #'   Ignored; a fresh group metadata list is created.
    #' @param attributes (`list()`)\cr
    #'   Attributes to embed in zarr.json. Default empty.
    #' @return `raw()`.
    encode_group_metadata = function(meta = NA, attributes = NULL) {
      group_meta <- list(
        zarr_format = jsonlite::unbox(3L),
        node_type = jsonlite::unbox("group"),
        attributes = if (is.null(attributes)) obj_list() else attributes
      )
      return(charToRaw(jsonlite::toJSON(group_meta, auto_unbox = FALSE)))
    }
  )
)

# Validate basic V3 metadata requirements.
# @param meta Parsed metadata list from zarr.json.
# @param expected_node_type Optional. "array" or "group".
validate_v3_meta <- function(meta, expected_node_type = NULL) {
  if (is.null(meta$zarr_format) || meta$zarr_format != 3) {
    stop("Expected zarr_format=3 but got ", meta$zarr_format)
  }
  if (!is.null(expected_node_type) && meta$node_type != expected_node_type) {
    stop("Expected node_type='", expected_node_type,
         "' but got '", meta$node_type, "'")
  }
}

#' Create a list of zarray metadata.
#' @inheritParams zarr_create
#' @return A list.
#' @keywords internal
create_zarray_meta <- function(shape = NA, chunks = NA, dtype = NA, compressor = NA, fill_value = NA, order = NA, filters = NA, dimension_separator = NA) {
  # Reference: https://zarr.readthedocs.io/en/stable/spec/v2.html#metadata
  if(is.na(dimension_separator)) {
    dimension_separator <- "."
  } else if(!(dimension_separator %in% c(".", "/"))) {
    stop("dimension_separator must be '.' or '/'.")
  }
  if(is_na(compressor)) {
    compressor <- jsonlite::unbox(compressor)
  } else if(!is_na(compressor) && !("id" %in% names(compressor))) {
    stop("compressor must contain an 'id' property when not null.")
  }
  if(is_na(filters)) {
    filters <- jsonlite::unbox(filters)
  }
  if(!(order %in% c("C", "F"))) {
    stop("order must be 'C' or 'F'.")
  }
  is_simple_dtype <- (!dtype$is_structured)
  dtype_str <- dtype$dtype
  if(is_simple_dtype) {
    dtype_byteorder <- dtype$byte_order
    dtype_basictype <- dtype$basic_type
    # Validation occurs in Dtype constructor.

    if(dtype_basictype == "f") {
      if(!is.numeric(fill_value) && !(fill_value %in% c("NaN", "Infinity", "-Infinity"))) {
        stop("fill_value must be NaN, Infinity, or -Infinity when dtype is float")
      }
    }
    if(dtype_basictype == "S" && !is.na(fill_value)) {
      if(!is.character(fill_value)) {
        stop("fill_value must be a character string for string dtype")
      }
    }
  } else {
    # Structured dtypes are not yet supported for validation.
  }

  if(is.null(shape)) {
    shape <-jsonlite::unbox(NA)
  }

  if(!is.null(shape) && !all(is.na(shape))) {
    if(!is.numeric(shape) || any(shape < 0)) {
      stop("shape must be a numeric vector with non-negative values")
    }
  }
  if(!is.null(chunks) && !all(is.na(chunks))) {
    if(!is.numeric(chunks) || any(chunks <= 0)) {
      stop("chunks must be a numeric vector with positive values")
    }
  }
  zarray_meta <- list(
    zarr_format = jsonlite::unbox(2),
    shape = shape,
    chunks = chunks,
    dtype = jsonlite::unbox(dtype_str),
    compressor = compressor,
    fill_value = jsonlite::unbox(fill_value),
    order = jsonlite::unbox(order),
    filters = filters,
    dimension_separator = jsonlite::unbox(dimension_separator)
  )
  return(zarray_meta)
}

#' Create a V3 array metadata list.
#' Parallel to create_zarray_meta() for V2. Builds the zarr.json metadata
#' for a V3 array node. All scalar values are wrapped with jsonlite::unbox()
#' so the result can be encoded with toJSON(auto_unbox = FALSE).
#'
#' @param shape Integer vector. Array shape.
#' @param chunks Integer vector. Chunk shape.
#' @param dtype Character. V2-style dtype string (e.g., "<f8", "|b1").
#' @param compressor Codec object or NA.
#' @param fill_value Fill value (numeric, logical, or NA).
#' @param order Character. "C" or "F" (unused in V3 codec pipeline but stored).
#' @param filters List of Codec objects or NA.
#' @param chunk_key_encoding List or NULL. If NULL, defaults to default "/" encoding.
#' @param attributes List or NULL. Embedded attributes. Default empty.
#' @param dimension_names Character vector or NULL. Omitted from zarr.json if NULL.
#' @return A list ready for jsonlite::toJSON encoding.
#' @keywords internal
create_v3_array_meta <- function(shape, chunks, dtype, compressor, fill_value,
                                  order, filters, chunk_key_encoding = NULL,
                                  attributes = NULL, dimension_names = NULL) {
  codecs <- build_v3_codec_pipeline(compressor, filters, dtype)
  dtype_info <- v2_dtype_to_v3_dtype(dtype)

  if (is.null(chunk_key_encoding)) {
    chunk_key_encoding <- list(
      name = jsonlite::unbox("default"),
      configuration = list(separator = jsonlite::unbox("/"))
    )
  }

  meta <- list(
    zarr_format = jsonlite::unbox(3L),
    node_type = jsonlite::unbox("array"),
    shape = shape,
    data_type = jsonlite::unbox(dtype_info$data_type),
    chunk_grid = list(
      name = jsonlite::unbox("regular"),
      configuration = list(chunk_shape = chunks)
    ),
    chunk_key_encoding = chunk_key_encoding,
    codecs = codecs,
    fill_value = encode_fill_value_v3(fill_value),
    attributes = if (is.null(attributes)) obj_list() else attributes
  )

  if (!is.null(dimension_names)) {
    meta$dimension_names <- dimension_names
  }

  meta
}

library(pizzarr)

# --- Metadata2 encode/decode round-trip ---

test_that("Metadata2 encode/decode round-trips a simple list", {
  m <- Metadata2$new()
  data <- list(foo = "bar", num = 42)
  encoded <- m$encode_metadata(data)
  expect_type(encoded, "raw")
  decoded <- m$decode_metadata(encoded)
  expect_equal(decoded$foo, list("bar"))
  expect_equal(decoded$num, list(42))
})

test_that("Metadata2 decode_metadata handles list input (passthrough)", {
  m <- Metadata2$new()
  data <- list(foo = "bar")
  result <- m$decode_metadata(data)
  expect_equal(result, data)
})

test_that("Metadata2 decode_metadata handles NULL input", {
  m <- Metadata2$new()
  result <- m$decode_metadata(NULL)
  expect_null(result)
})

# --- Array metadata ---

test_that("Metadata2 encode_array_metadata adds zarr_format=2", {
  m <- Metadata2$new()
  meta <- list(
    shape = list(10, 20),
    chunks = list(5, 5),
    dtype = jsonlite::unbox("<f8"),
    compressor = jsonlite::unbox(NA),
    fill_value = jsonlite::unbox(0),
    order = jsonlite::unbox("C"),
    filters = jsonlite::unbox(NA)
  )
  encoded <- m$encode_array_metadata(meta)
  decoded <- m$decode_array_metadata(encoded)
  expect_equal(decoded$zarr_format, 2)
})

# --- Group metadata ---

test_that("Metadata2 encode/decode group metadata round-trips", {
  m <- Metadata2$new()
  encoded <- m$encode_group_metadata()
  decoded <- m$decode_group_metadata(encoded)
  expect_equal(decoded$zarr_format, 2)
})

# --- Validation ---

test_that("validate_v2_meta rejects zarr_format != 2", {
  expect_error(validate_v2_meta(list(zarr_format = 3)), "unsupported zarr format")
  expect_error(validate_v2_meta(list(zarr_format = 1)), "unsupported zarr format")
})

test_that("validate_v2_meta accepts zarr_format = 2", {
  expect_silent(validate_v2_meta(list(zarr_format = 2)))
})

# --- create_zarray_meta ---

test_that("create_zarray_meta with valid inputs returns correct structure", {
  dtype <- Dtype$new("<f8")
  meta <- create_zarray_meta(
    shape = c(10, 20),
    chunks = c(5, 5),
    dtype = dtype,
    compressor = NA,
    fill_value = 0,
    order = "C",
    filters = NA,
    dimension_separator = "."
  )
  expect_equal(meta$shape, c(10, 20))
  expect_equal(meta$chunks, c(5, 5))
  expect_equal(meta$order, jsonlite::unbox("C"))
  expect_equal(meta$dimension_separator, jsonlite::unbox("."))
})

test_that("create_zarray_meta accepts '/' dimension_separator", {
  dtype <- Dtype$new("<f8")
  meta <- create_zarray_meta(
    shape = c(10), chunks = c(5), dtype = dtype,
    compressor = NA, fill_value = 0, order = "C",
    filters = NA, dimension_separator = "/"
  )
  expect_equal(meta$dimension_separator, jsonlite::unbox("/"))
})

test_that("create_zarray_meta rejects invalid dimension_separator", {
  dtype <- Dtype$new("<f8")
  expect_error(
    create_zarray_meta(
      shape = c(10), chunks = c(5), dtype = dtype,
      compressor = NA, fill_value = 0, order = "C",
      filters = NA, dimension_separator = "-"
    ),
    "dimension_separator"
  )
})

test_that("create_zarray_meta rejects invalid order", {
  dtype <- Dtype$new("<f8")
  expect_error(
    create_zarray_meta(
      shape = c(10), chunks = c(5), dtype = dtype,
      compressor = NA, fill_value = 0, order = "X",
      filters = NA, dimension_separator = "."
    ),
    "order"
  )
})

test_that("create_zarray_meta rejects negative shape values", {
  dtype <- Dtype$new("<i4")
  expect_error(
    create_zarray_meta(
      shape = c(10, -1), chunks = c(5, 5), dtype = dtype,
      compressor = NA, fill_value = 0, order = "C",
      filters = NA, dimension_separator = "."
    ),
    "shape"
  )
})

test_that("create_zarray_meta rejects non-positive chunks", {
  dtype <- Dtype$new("<i4")
  expect_error(
    create_zarray_meta(
      shape = c(10), chunks = c(0), dtype = dtype,
      compressor = NA, fill_value = 0, order = "C",
      filters = NA, dimension_separator = "."
    ),
    "chunks"
  )
})

test_that("create_zarray_meta validates float fill_value", {
  dtype <- Dtype$new("<f4")
  # Valid fill values for float dtype
  expect_silent(create_zarray_meta(
    shape = c(4), chunks = c(4), dtype = dtype,
    compressor = NA, fill_value = 0, order = "C",
    filters = NA, dimension_separator = "."
  ))
  expect_silent(create_zarray_meta(
    shape = c(4), chunks = c(4), dtype = dtype,
    compressor = NA, fill_value = "NaN", order = "C",
    filters = NA, dimension_separator = "."
  ))
  expect_silent(create_zarray_meta(
    shape = c(4), chunks = c(4), dtype = dtype,
    compressor = NA, fill_value = "Infinity", order = "C",
    filters = NA, dimension_separator = "."
  ))
  expect_silent(create_zarray_meta(
    shape = c(4), chunks = c(4), dtype = dtype,
    compressor = NA, fill_value = "-Infinity", order = "C",
    filters = NA, dimension_separator = "."
  ))
})

# v2 store keys
ARRAY_META_KEY <- '.zarray'
GROUP_META_KEY <- '.zgroup'
ATTRS_KEY <- '.zattrs'

# v3 store key
# V3 uses a single zarr.json file per node, replacing .zarray/.zgroup/.zattrs.
ZARR_JSON <- 'zarr.json'

# Internal utility functions for converting between Zarr and R data types.

#' @keywords internal
get_dtype_from_array <- function(a) {
  RTYPE_DTYPE_MAPPING <- list(
    "logical" = "|b1",
    "integer" = "<i4",
    "double" = "<f8",
    "character" = "|S8" # TODO: how many bytes to use here?
  )
  rtype_str <- typeof(a)
  return(RTYPE_DTYPE_MAPPING[[rtype_str]])
}

#' @keywords internal
is_structured_dtype <- function(dtype) {
  if(is.character(dtype) && length(dtype) == 1) {
    return(FALSE)
  }
  return(TRUE)
}

#' @keywords internal
get_dtype_parts <- function(dtype) {
  # TODO: support object dtype (without digits required in regex)
  dtype_regex <- "^(\\||>|<)(b|i|u|f|c|m|M|S|U|V|O)(\\d+)?"
  if(stringr::str_detect(dtype, dtype_regex)) {
    dtype_matches <- stringr::str_match(dtype, dtype_regex)
    basic_type <- dtype_matches[1,3]
    if(basic_type == "U") {
      byte_multiplier <- 4
    } else {
      byte_multiplier <- 1
    }
    if(!is.null(dtype_matches[1,4]) && !is.na(dtype_matches[1,4])) {
      num_items <- as.integer(dtype_matches[1,4])
      num_bytes <- num_items * byte_multiplier
    } else {
      # Support object dtype
      num_items <- NA
      num_bytes <- 4
    }
    result <- list(
      dtype_str = dtype,
      byte_order = dtype_matches[1,2],
      basic_type = dtype_matches[1,3],
      num_bytes = num_bytes,
      num_items = num_items
    )
    return(result)
  } else {
    return(NA)
  }
}

#' @keywords internal
check_dtype_support <- function(dtype_parts) {
  if(!is_na(dtype_parts) && dtype_parts$basic_type %in% c("b", "i", "u", "f", "S", "U", "O")) {
    return(TRUE)
  }
  stop(paste("Unsupported dtype:", dtype_parts))
  return(FALSE)
}

#' @keywords internal
get_dtype_rtype <- function(basic_type, num_bytes = NULL) {

  # 64-bit integers: use bit64::integer64 if available, else double
  if (basic_type %in% c("i", "u") && !is.null(num_bytes) && num_bytes == 8) {
    if (has_bit64()) {
      return(bit64::integer64())
    } else {
      warning("bit64 package not installed; reading 64-bit integers as double ",
              "(precision loss possible for values > 2^53)")
      return(double())
    }
  }

  # Reference: https://github.com/gzuidhof/zarr.js/blob/292804/src/nestedArray/types.ts#L32
  BASICTYPE_RTYPE_MAPPING <- list(
    "b" = logical(),
    "u" = integer(),
    "i" = integer(),
    "f" = double(),
    "S" = character(),
    "U" = character(),
    "O" = character() # TODO: will object always be character?
  )

  return(BASICTYPE_RTYPE_MAPPING[[basic_type]])
}

#' @keywords internal
get_dtype_endianness <- function(dtype) {
  dtype_parts <- get_dtype_parts(dtype)

  DTYPE_ENDIANNESS_MAPPING <- list(
    "|" = "nr",
    "<" = "little",
    ">" = "big"
  )
  return(DTYPE_ENDIANNESS_MAPPING[[dtype_parts$byte_order]])
}

#' @keywords internal
get_dtype_numbytes <- function(dtype) {
  dtype_parts <- get_dtype_parts(dtype)
  return(dtype_parts$num_bytes)
}

#' @keywords internal
get_dtype_signed <- function(dtype) {
  dtype_parts <- get_dtype_parts(dtype)

  DTYPE_SIGNED_MAPPING <- list(
    "b" = FALSE,
    "u" = FALSE,
    "i" = TRUE,
    "f" = TRUE,
    "S" = FALSE, # TODO: is this correct?
    "U" = FALSE,  # TODO: is this correct?
    "O" = FALSE
  )
  return(DTYPE_SIGNED_MAPPING[[dtype_parts$basic_type]])
}

#' @keywords internal
get_dtype_asrtype <- function(dtype) {
  dtype_parts <- get_dtype_parts(dtype)
  check_dtype_support(dtype_parts)

  # 64-bit integers: use bit64::as.integer64 if available, else as.double
  if (dtype_parts$basic_type %in% c("i", "u") && dtype_parts$num_bytes == 8) {
    if (has_bit64()) {
      return(bit64::as.integer64)
    } else {
      return(as.double)
    }
  }

  # Reference: https://github.com/gzuidhof/zarr.js/blob/292804/src/nestedArray/types.ts#L32
  DTYPE_RTYPE_MAPPING <- list(
    "b" = as.logical,
    "u" = as.integer,
    "i" = as.integer,
    "f" = as.double,
    "S" = as.character,
    "U" = as.character,
    "O" = as.character
  )
  return(DTYPE_RTYPE_MAPPING[[dtype_parts$basic_type]])
}

# Convert V3 data_type string to V2 numpy-style dtype string.
# Reference: https://zarr-specs.readthedocs.io/en/latest/v3/core/v3.0.html#data-types
# V3 uses simple string names like "float64"; we convert to numpy-style strings
# like "<f8" so the existing Dtype class works unchanged.
#
# @param v3_dtype Character. V3 data type name (e.g., "float64", "int16", "bool").
# @param endian Character. Byte order from the V3 "endian"/"bytes" codec.
#   Either "little" (default) or "big". Single-byte types ignore this.
# @return Character. V2-style dtype string (e.g., "<f8", "|b1").
# @keywords internal
v3_dtype_to_v2_dtype <- function(v3_dtype, endian = "little") {
  # Mapping: V3 name -> list(v2 = base string, fixed_order = byte order or NULL)
  # Single-byte types have fixed byte order "|" (not relevant).
  # Multi-byte types get byte order prefix from the endian codec.
  V3_DTYPE_MAP <- list(
    "bool"    = list(v2 = "b1",  fixed_order = "|"),
    "int8"    = list(v2 = "i1",  fixed_order = "|"),
    "int16"   = list(v2 = "i2",  fixed_order = NULL),
    "int32"   = list(v2 = "i4",  fixed_order = NULL),
    "int64"   = list(v2 = "i8",  fixed_order = NULL),
    "uint8"   = list(v2 = "u1",  fixed_order = "|"),
    "uint16"  = list(v2 = "u2",  fixed_order = NULL),
    "uint32"  = list(v2 = "u4",  fixed_order = NULL),
    "uint64"  = list(v2 = "u8",  fixed_order = NULL),
    "float32" = list(v2 = "f4",  fixed_order = NULL),
    "float64" = list(v2 = "f8",  fixed_order = NULL)
  )

  entry <- V3_DTYPE_MAP[[v3_dtype]]
  if (is.null(entry)) {
    stop("Unsupported V3 data_type: '", v3_dtype,
         "'. Supported types: ", paste(names(V3_DTYPE_MAP), collapse = ", "))
  }

  # Single-byte types use fixed "|" prefix; multi-byte types use endian codec
  prefix <- if (!is.null(entry$fixed_order)) {
    entry$fixed_order
  } else if (endian == "big") {
    ">"
  } else {
    "<"
  }

  paste0(prefix, entry$v2)
}

# Convert V2 numpy-style dtype string to V3 data_type name and endian.
# Inverse of v3_dtype_to_v2_dtype.
#
# @param dtype Character. V2 dtype string (e.g., "<f8", "|b1", ">i4").
# @return Named list with:
#   $data_type - V3 data type string (e.g., "float64", "bool")
#   $endian    - "little", "big", or NA_character_ (for single-byte types)
# @keywords internal
v2_dtype_to_v3_dtype <- function(dtype) {
  dtype_parts <- get_dtype_parts(dtype)

  V2_TO_V3_MAP <- list(
    "b1" = "bool",
    "i1" = "int8",
    "i2" = "int16",
    "i4" = "int32",
    "i8" = "int64",
    "u1" = "uint8",
    "u2" = "uint16",
    "u4" = "uint32",
    "u8" = "uint64",
    "f4" = "float32",
    "f8" = "float64"
  )

  key <- paste0(dtype_parts$basic_type, dtype_parts$num_bytes)
  v3_name <- V2_TO_V3_MAP[[key]]
  if (is.null(v3_name)) {
    stop("Cannot convert V2 dtype '", dtype, "' to V3 data_type: unsupported type '", key, "'")
  }

  # "|" prefix means byte order not applicable (single-byte types: bool, int8, uint8)
  endian_str <- get_dtype_endianness(dtype)
  endian <- if (endian_str == "nr") NA_character_ else endian_str

  list(data_type = v3_name, endian = endian)
}

#' The Zarr Dtype class.
#' @title Dtype Class
#' @docType class
#' @description
#'  A data type object (an instance of Dtype class) describes how 
#'  the bytes in the fixed-size block of memory corresponding to an array 
#'  item should be interpreted.
#' @references https://numpy.org/doc/stable/reference/arrays.dtypes.html
#' @format [R6::R6Class]
#' @rdname Dtype
#' @export
Dtype <- R6::R6Class("Dtype",
  public = list(
    #' @field dtype The original dtype string, like "<f4".
    dtype = NULL,
    #' @field byte_order The byte order of the dtype, either "little", "big", or "nr".
    byte_order = NULL,
    #' @field basic_type The basic type of the dtype, like "f".
    basic_type = NULL,
    #' @field num_bytes The number of bytes of the dtype.
    num_bytes = NULL,
    #' @field num_items The number of items of the dtype.
    num_items = NULL,
    #' @field is_signed Whether the dtype is signed. Logical/boolean.
    is_signed = NULL,
    #' @field is_structured Whether the dtype is structured. Logical/boolean.
    is_structured = NULL,
    #' @field is_object Whether the dtype is an object. Logical/boolean.
    is_object = NULL,
    #' @field object_codec The object codec instance.
    object_codec = NULL,
    #' @description
    #' Create a new Dtype instance.
    #' @param dtype The original dtype string, like "<f4".
    #' @param object_codec The object codec instance.
    #' @return A `Dtype` instance.
    initialize = function(dtype, object_codec = NA) {
      self$dtype <- dtype

      dtype_parts <- get_dtype_parts(dtype)
      check_dtype_support(dtype_parts)
      self$byte_order <-  get_dtype_endianness(dtype)
      self$basic_type <- dtype_parts$basic_type
      self$num_bytes <- dtype_parts$num_bytes
      self$num_items <- dtype_parts$num_items

      self$is_signed <- get_dtype_signed(dtype)
      self$is_structured <- is_structured_dtype(dtype)
      self$is_object <- (self$basic_type == "O")

      if (self$basic_type == "u" && self$num_bytes == 8) {
        warning("uint64 dtype: bit64 only supports signed int64; ",
                "values >= 2^63 will overflow")
      }

      self$object_codec <- object_codec
    },
    #' @description
    #' Get the R coercion function name for this dtype.
    #' @return Character string (e.g., `"as.double"`).
    get_asrtype = function() {
      return(get_dtype_asrtype(self$dtype))
    },
    #' @description
    #' Get the R base type for this dtype.
    #' @return An R prototype value (e.g., `integer()`, `double()`, or `bit64::integer64()`).
    get_rtype = function() {
      return(get_dtype_rtype(self$basic_type, self$num_bytes))
    },
    #' @description
    #' Get a constructor function for typed arrays of this dtype.
    #' @return A function that takes `dim` and returns an array.
    get_typed_array_ctr = function() {
      rtype <- self$get_rtype()
      if (inherits(rtype, "integer64")) {
        return(function(dim) {
          a <- array(data = double(), dim = dim)
          class(a) <- "integer64"
          a
        })
      }
      return(function(dim) array(data = rtype, dim = dim))
    }
  )
)

library(pizzarr)

test_that("dtype regexes work for one-digit number of bytes", {
    dtype_parts <- get_dtype_parts("|u4")
    expect_equal(dtype_parts$basic_type, "u")
    expect_equal(dtype_parts$byte_order, "|")
    expect_equal(dtype_parts$num_bytes, 4)
})

test_that("dtype regexes work for multi-digit number of bytes", {
    dtype_parts <- get_dtype_parts("<f99")
    expect_equal(dtype_parts$basic_type, "f")
    expect_equal(dtype_parts$byte_order, "<")
    expect_equal(dtype_parts$num_bytes, 99)
})

test_that("check_dtype_support throws error for non-matched regex", {
    dtype_parts <- get_dtype_parts("bad")

    expect_equal(is_na(dtype_parts), TRUE)

    f <- function() {
        check_dtype_support(dtype_parts)
    }
    expect_error(f())
})

test_that("get_dtype_rtype works", {
    dtype_rtype <- get_dtype_rtype("f")
    expect_equal(dtype_rtype, double())

    dtype_rtype <- get_dtype_rtype("u")
    expect_equal(dtype_rtype, integer())

    dtype_rtype <- get_dtype_rtype("b")
    expect_equal(dtype_rtype, logical())
})

test_that("get_dtype_endianness works", {
    dtype_endianness <- get_dtype_endianness("<f8")
    expect_equal(dtype_endianness, "little")

    dtype_endianness <- get_dtype_endianness(">f8")
    expect_equal(dtype_endianness, "big")

    dtype_endianness <- get_dtype_endianness("|b1")
    expect_equal(dtype_endianness, "nr")
})

test_that("get_dtype_numbytes works", {
    numbytes <- get_dtype_numbytes("<f8")
    expect_equal(numbytes, 8)

    numbytes <- get_dtype_numbytes("|S12")
    expect_equal(numbytes, 12)
})

test_that("get_dtype_signed works", {
    is_signed <- get_dtype_signed("<f8")
    expect_equal(is_signed, TRUE)

   is_signed <- get_dtype_signed("<u4")
    expect_equal(is_signed, FALSE)
})

test_that("get_dtype_from_array works", {
    zarr_dtype <- get_dtype_from_array(array(data=as.double(1:10), dim=c(2, 5)))
    check_dtype_support(get_dtype_parts(zarr_dtype))
    zarr_dtype <- get_dtype_from_array(array(data=as.integer(1:10), dim=c(2, 5)))
    check_dtype_support(get_dtype_parts(zarr_dtype))

    zarr_dtype <- get_dtype_from_array(array(data=as.logical(1, 0, 1, 0), dim=c(2, 2)))
    expect_false(is_na(get_dtype_parts(zarr_dtype)))
})

test_that("get_dtype_asrtype works", {
    astype_func <- get_dtype_asrtype("<f8")
    a <- astype_func(as.integer(1:10))
    expect_equal(typeof(a), "double")

    astype_func <- get_dtype_asrtype("<i4")
    a <- astype_func(as.double(1:10))
    expect_equal(typeof(a), "integer")

    astype_func <- get_dtype_asrtype("|S8")
    a <- astype_func(as.double(1:10))
    expect_equal(typeof(a), "character")
})

# --- Dtype class tests ---

test_that("Dtype class parses float64", {
  d <- Dtype$new("<f8")
  expect_equal(d$byte_order, "little")
  expect_equal(d$basic_type, "f")
  expect_equal(d$num_bytes, 8)
  expect_true(d$is_signed)
  expect_false(d$is_structured)
  expect_false(d$is_object)
})

test_that("Dtype class parses boolean", {
  d <- Dtype$new("|b1")
  expect_equal(d$byte_order, "nr")
  expect_equal(d$basic_type, "b")
  expect_equal(d$num_bytes, 1)
})

test_that("Dtype class parses unsigned int", {
  d <- Dtype$new("|u1")
  expect_equal(d$basic_type, "u")
  expect_equal(d$num_bytes, 1)
  expect_false(d$is_signed)
})

test_that("Dtype class parses object dtype", {
  d <- Dtype$new("|O")
  expect_equal(d$basic_type, "O")
  expect_true(d$is_object)
})

test_that("Dtype get_rtype returns correct R types", {
  expect_equal(Dtype$new("<f4")$get_rtype(), double())
  expect_equal(Dtype$new("<f8")$get_rtype(), double())
  expect_equal(Dtype$new("<i4")$get_rtype(), integer())
  expect_equal(Dtype$new("|b1")$get_rtype(), logical())
  expect_equal(Dtype$new("|u1")$get_rtype(), integer())
  expect_equal(Dtype$new("|S8")$get_rtype(), character())
})

test_that("Dtype get_asrtype returns correct conversion functions", {
  expect_equal(typeof(Dtype$new("<f8")$get_asrtype()(1L)), "double")
  expect_equal(typeof(Dtype$new("<i4")$get_asrtype()(1.5)), "integer")
  expect_equal(typeof(Dtype$new("|b1")$get_asrtype()(1)), "logical")
})

# --- Dtype round-trip through array creation ---

dtype_cases <- list(
  list(dtype = "|b1", data = array(c(TRUE, FALSE, TRUE, FALSE), dim = c(4))),
  list(dtype = "<i2", data = array(c(1L, 2L, 3L, 4L), dim = c(4))),
  list(dtype = "<i4", data = array(c(1L, 2L, 3L, 4L), dim = c(4))),
  list(dtype = "<f4", data = array(c(1.5, 2.5, 3.5, 4.5), dim = c(4))),
  list(dtype = "<f8", data = array(c(1.5, 2.5, 3.5, 4.5), dim = c(4))),
  list(dtype = "|u1", data = array(c(0L, 127L, 255L, 1L), dim = c(4)))
)

for (case in dtype_cases) {
  test_that(paste("dtype array round-trip:", case$dtype), {
    a <- zarr_create(
      shape = dim(case$data), dtype = case$dtype,
      store = MemoryStore$new(), compressor = NA
    )
    a$set_item("...", case$data)
    result <- a$get_item("...")
    expect_equal(result$data, case$data)
  })
}

# Generated by extendr: Do not edit by hand
# Stub for CRAN tier (no Rust shared library)

# nolint start

#' @usage NULL
NULL

#' Return compiled zarrs feature flags as a character vector.
#'
#' On the CRAN tier (pure R), this function always errors. The error
#' is caught by \code{is_zarrs_available()}, which returns \code{FALSE}.
#' @export
zarrs_compiled_features <- function() {
  stop("zarrs backend not compiled (pure R install)")
}

# nolint end

# Reference: https://github.com/gzuidhof/zarr.js/blob/292804/src/rawArray/index.ts

#' The Zarr RawArray class.
#' @title RawArray Class
#' @docType class
#' @description
#' TODO: Add export annotation / remove keywords internal annotation once implemented.
#' @rdname RawArray
#' @keywords internal
RawArray <- R6::R6Class("RawArray",
  public = list(
    #' @description
    #' Create a new RawArray instance.
    #' @return A `RawArray` instance.
    initialize = function() {
      # TODO
    }
  )
)

# TODO: int() and zb_int() now being used but do we need 'Int' Class ? 

#' Abstract Int object
#' @title Int Class
#' @docType class
#' @description
#' Class representing an indexing of a ZARR store
#' @noRd
#' @keywords internal
Int <- R6::R6Class("Int",
                     public = list(
                       #' @field start start index
                       index = NULL,
                       #' @description Create a new `Int` object
                       #' @param index integer index
                       initialize = function(index) {
                         self$index <- index
                       },
                       #' @description
                       #' This method takes a single integer argument `length_param` and computes information about the
                       #' slice that the slice object would describe if applied to a sequence of `length_param` items.
                       #' It returns a tuple of three integers; respectively these are the start and stop indices
                       # and the step or stride length of the slice. Missing or out-of-bounds indices are handled
                       # in a manner consistent with regular slices.
                       #' @param length_param integer length parameter for calculation of integer indices
                       indices = function(length_param) {
                         
                         # check length_param
                         index <- self$index[self$index <= length_param]
                         
                         # return
                         return(c(index, length_param))
                       }
                     )
)

#' Convenience function for the internal Int class constructor.
#' @param index The integer index.
#' @param zero_based The index of the dimension. By default, FALSE for R-like behavior.
#' @return A Int instance with the specified parameters.
#' @export
int <- function(index, zero_based = FALSE) {
  index_offset <- ifelse(zero_based, 0, -1)
  if(!is_na(index) && is.numeric(index)) {
    index <- index + index_offset
  }
  # Assumed to be zero-based
  # and stop-inclusive
  # return(Int$new(
  #   index = index
  # ))
  index
}

#' Convenience function for the internal Int class constructor
#' with zero-based indexing
#' @param index integer index
#' @export
zb_int <- function(index) {
  return(int(index, zero_based = TRUE))
}

library(pizzarr)

test_that("orthogonal selection", {
  a <- array(data=1:20, dim=c(2, 10))
  #      [,1] [,2] [,3] [,4] [,5] [,6] [,7] [,8] [,9] [,10]
  # [1,]    1    3    5    7    9   11   13   15   17    19
  # [2,]    2    4    6    8   10   12   14   16   18    20
  z <- zarr_create(shape=dim(a), dtype="<f4", fill_value=NA)
  z$set_item("...", a)
  
  # indices
  expect_equal(
    z$get_orthogonal_selection(list(c(0,1), c(1,2,3)))$data,
    array(data=c(3, 4, 5, 6, 7, 8), dim=c(2, 3))
  )
  expect_equal(
    z$get_orthogonal_selection(list(c(1,0),c(9,1,6)))$data,
    array(data=c(20, 19, 4, 3, 14, 13), dim=c(2, 3))
  )
  expect_error(z$get_orthogonal_selection(list(c(1,0),c(12,1,6)))$data)
  
  # int and zb_int
  expect_equal(
    z$get_orthogonal_selection(list(zb_int(c(0,1)), zb_int(c(1,2,3))))$data,
    z$get_orthogonal_selection(list(int(c(1, 2)), int(c(2,3,4))))$data)

  # TODO: do not expect error once negative indexing has been implemented.
  expect_error(z$get_orthogonal_selection(list(c(-1,1),c(12,1,6)))$data) 
  
  # mix integer vector with slices
  expect_equal(
    z$get_orthogonal_selection(list(slice(1,2),c(9,6)))$data,
    array(data=c(19, 20, 13, 14), dim=c(2, 2))
  )
  expect_equal(
    z$get_orthogonal_selection(list(c(1,0), slice(1,3)))$data,
    array(data=c(2, 1, 4, 3, 6, 5), dim=c(2, 3))
  )
  
  # boolean indexing
  ind <- rep(FALSE, 10)
  ind[c(3,5,8)] <- TRUE
  expect_equal(
    z$get_orthogonal_selection(list(c(1,0), ind))$data,
    array(data=c(6, 5, 10, 9, 16, 15), dim=c(2, 3))
  )
  expect_error(z$get_orthogonal_selection(list(c(0,1),c(TRUE,FALSE,TRUE)))$data)
})

# Reference: https://github.com/zarr-developers/zarr-python/blob/5dd4a0/zarr/core.py#L51

#' @keywords internal
format_codec <- function(codec) {
  if (is.null(codec) || is_na(codec)) return("None")
  if (is.character(codec)) return(codec)
  if (inherits(codec, "R6")) return(class(codec)[1])
  as.character(codec)
}

#' The Zarr Array class.
#' @title ZarrArray Class
#' @docType class
#' @description
#' Instantiate an array from an initialized store.
#' @param selection Selections are lists containing either scalars, strings, or Slice objects. Two character
#' literals are supported: "..." selects all remaining array dimensions and ":" selects all of a specific 
#' array dimension.
#'
#' @format [R6::R6Class]
#' @rdname ZarrArray
#' @export
ZarrArray <- R6::R6Class("ZarrArray",
  private = list(
    # store Array store, already initialized.
    #' @keywords internal
    store = NULL,
    # chunk_store Separate storage for chunks. If not provided, `store` will be used for storage of both chunks and metadata.
    #' @keywords internal
    chunk_store = NULL,
    # path Storage path. String, optional.
    #' @keywords internal
    path = NULL,
    # read_only True if array should be protected against modification.
    #' @keywords internal
    read_only = NULL,
    # synchronizer Array synchronizer. Object, optional.
    #' @keywords internal
    synchronizer = NULL,
    # cache_metadata If True (default), array configuration metadata will be cached. If False, metadata will be reloaded prior to all data access and modification.
    #' @keywords internal
    cache_metadata = NULL,
    # cache_attrs If True (default), user attributes will be cached. If False, attributes will be reloaded prior to all data access and modification.
    #' @keywords internal
    cache_attrs = NULL,
    # write_empty_chunks If True, all chunks will be stored regardless of their contents. If False (default), each chunk is compared to the array's fill value prior to storing. If a chunk is uniformly equal to the fill value, then that chunk is not be stored, and the store entry for that chunk's key is deleted.
    #' @keywords internal
    write_empty_chunks = NULL,
    # key_prefix Storage path prefix prepended to chunk keys when accessing the store. Set to path + "/" or empty string for root.
    #' @keywords internal
    key_prefix = NULL,
    # is_view TRUE if this array is a view on another array, sharing the same underlying store data.
    #' @keywords internal
    is_view = NULL,
    # attrs A MutableMapping containing user-defined attributes. Attribute values must be JSON serializable.
    #' @keywords internal
    attrs = NULL,
    # meta Parsed array metadata from .zarray (V2) or zarr.json (V3).
    #' @keywords internal
    meta = NULL,
    # shape Integer vector describing the length of each dimension of the array.
    #' @keywords internal
    shape = NULL,
    # chunks Integer vector describing the length of each dimension of a chunk.
    #' @keywords internal
    chunks = NULL,
    # dtype Dtype object representing the data type of the array.
    #' @keywords internal
    dtype = NULL,
    # fill_value A value used for uninitialized portions of the array.
    #' @keywords internal
    fill_value = NULL,
    # order A string indicating the order in which bytes are arranged within chunks ("C" for row-major, "F" for column-major).
    #' @keywords internal
    order = NULL,
    # dimension_separator Separator character used in chunk keys, typically "." (flat) or "/" (nested directory structure).
    #' @keywords internal
    dimension_separator = NULL,
    # compressor Primary compression codec used to compress chunk data before storage.
    #' @keywords internal
    compressor = NULL,
    # filters One or more codecs used to transform data prior to compression.
    #' @keywords internal
    filters = NULL,
    # zarr_format: integer, 2L or 3L. Determines which metadata format is in use.
    #' @keywords internal
    zarr_format = NULL,
    # chunk_key_encoding: list with $name and $configuration$separator (V3 only).
    # V3 default: name="default", separator="/", producing keys like "c/0/1/2".
    #' @keywords internal
    chunk_key_encoding = NULL,
    # dimension_names: character vector or NULL. Named dimensions (V3 only).
    #' @keywords internal
    dimension_names = NULL,
    # vindex Shortcut for vectorized (inner) indexing, supporting coordinate-style selection using integer arrays.
    #' @keywords internal
    vindex = NULL,
    # oindex Shortcut for orthogonal (outer) indexing, allowing independent selection per dimension using integer, slice, integer array, or boolean array.
    #' @keywords internal
    oindex = NULL,
    # (Re)load metadata from store without synchronization (file locking).
    # Detects V3 (zarr.json) vs V2 (.zarray) format automatically.
    load_metadata_nosync = function() {

      # Check for V3 zarr.json first, then fall back to V2 .zarray.
      # Reference: zarr-python checks zarr.json first for V3 format detection.
      v3_key <- paste0(private$key_prefix, ZARR_JSON)
      if (private$store$contains_item(v3_key)) {
        private$load_metadata_v3_nosync()
        return()
      }

      # V2 path (existing code, unchanged)
      private$zarr_format <- 2L

      mkey <- paste0(private$key_prefix, ARRAY_META_KEY)

      meta <- try_from_zmeta(mkey, private$store)
      
      if(is.null(meta)) {
        meta_bytes <- private$store$get_item(mkey)
        if(!is.null(meta_bytes))
          meta <- private$store$metadata_class$decode_array_metadata(meta_bytes)
      }
      
      private$meta <- meta
      
      if(is.list(meta$shape)) {
        private$shape <- as.integer(meta$shape)
      } else {
        # meta$shape might be null.
        private$shape <- meta$shape
      }
      if(is.list(meta$chunks)) {
        private$chunks <- as.integer(meta$chunks)
      } else {
        # meta$chunks might be null.
        private$chunks <- meta$chunks
      }
      private$fill_value <- meta$fill_value
      private$order <- meta$order
      if(!is.null(meta$dimension_separator) && !is.na(meta$dimension_separator)) {
        private$dimension_separator <- meta$dimension_separator
      } else {
        # V2 stores don't carry dimension separators; "." is the spec default.
        private$dimension_separator <- "."
      }
      if(is_na(meta$compressor) || is.null(meta$compressor)) {
        private$compressor <- NA
      } else {
        private$compressor <- get_codec(meta$compressor)
      }
      object_codec <- NA
      if(is_na(meta$filters) || is.null(meta$filters)) {
        private$filters <- NA
        object_codec <- NA
      } else {
        private$filters <- lapply(meta$filters, get_codec)
        if(length(private$filters) == 1) {
          object_codec <- private$filters[[1]]
        }
      }
      private$dtype <- normalize_dtype(meta$dtype, object_codec = object_codec)
    },
    # Load V3 array metadata from zarr.json.
    # Reference: https://zarr-specs.readthedocs.io/en/latest/v3/core/v3.0.html
    # V3 zarr.json contains: zarr_format, node_type, shape, data_type,
    # chunk_grid, chunk_key_encoding, codecs, fill_value, attributes,
    # dimension_names.
    load_metadata_v3_nosync = function() {
      mkey <- paste0(private$key_prefix, ZARR_JSON)
      meta_bytes <- private$store$get_item(mkey)

      # Decode and validate as V3 array
      meta3 <- Metadata3$new()
      meta <- meta3$decode_array_metadata(meta_bytes)

      private$meta <- meta
      private$zarr_format <- 3L

      # --- shape ---
      if (is.list(meta$shape)) {
        private$shape <- as.integer(meta$shape)
      } else {
        private$shape <- meta$shape
      }

      # --- chunks from chunk_grid ---
      # V3 chunk_grid must be "regular" with a chunk_shape configuration.
      if (is.null(meta$chunk_grid) || meta$chunk_grid$name != "regular") {
        stop("Only 'regular' chunk_grid is supported for V3 arrays")
      }
      chunk_shape <- meta$chunk_grid$configuration$chunk_shape
      if (is.list(chunk_shape)) {
        private$chunks <- as.integer(chunk_shape)
      } else {
        private$chunks <- chunk_shape
      }

      # --- chunk_key_encoding ---
      # V3 default: name="default", separator="/", chunk keys like "c/0/1/2"
      # V3 v2-compat: name="v2", separator=".", chunk keys like "0.1.2"
      private$chunk_key_encoding <- meta$chunk_key_encoding
      if (is.null(private$chunk_key_encoding)) {
        private$chunk_key_encoding <- list(
          name = "default",
          configuration = list(separator = "/")
        )
      }
      # Set dimension_separator for compatibility with code that reads it
      sep <- private$chunk_key_encoding$configuration$separator
      private$dimension_separator <- if (!is.null(sep)) sep else "/"

      # --- codecs -> compressor, filters, order, endian ---
      # V3 codec pipeline replaces V2 compressor + filters.
      # resolve_v3_codecs() maps V3 codecs to V2-compatible structures
      # so decode_chunk() works unchanged.
      codec_result <- resolve_v3_codecs(meta$codecs)
      private$compressor <- codec_result$compressor
      private$filters <- codec_result$filters
      private$order <- codec_result$order

      # --- dtype ---
      # V3 uses string names like "float64"; convert to V2 numpy-style
      # so the existing Dtype class works unchanged.
      v2_dtype_str <- v3_dtype_to_v2_dtype(meta$data_type,
                                            endian = codec_result$endian)

      # Check for object codec (vlen-utf8) in filters
      object_codec <- NA
      if (!is_na(private$filters)) {
        for (f in private$filters) {
          if (inherits(f, "VLenUtf8Codec")) {
            object_codec <- f
            break
          }
        }
      }
      private$dtype <- normalize_dtype(v2_dtype_str, object_codec = object_codec)

      # --- fill_value ---
      # V3 fill_value is required (not optional like V2).
      # V3 encodes NaN/Infinity/-Infinity as JSON strings; decode back to numeric.
      private$fill_value <- decode_fill_value_v3(meta$fill_value)

      # JSON arrays of strings come back as lists under simplifyVector=FALSE;
      # coerce to a character vector so the public getter returns the type
      # users expect. NULL stays NULL when the field is absent from zarr.json.
      if (is.null(meta$dimension_names)) {
        private$dimension_names <- NULL
      } else {
        private$dimension_names <- as.character(unlist(meta$dimension_names))
      }
    },
    # method_description
    # Load or reload metadata from store.
    load_metadata = function() {
      private$load_metadata_nosync()
      # TODO: support for synchronization
    },
    # method_description
    # Referesh metadata if not cached without synchronization (file locking).
    refresh_metadata_nosync = function() {
      if(!private$cache_metadata && !private$is_view) {
        private$load_metadata_nosync()
      }
    },
    # method_description
    # Refresh metadata from store if not cached.
    refresh_metadata = function() {
      if(!private$cache_metadata) {
        private$load_metadata()
      }
    },
    # method_description
    # Write metadata to store without synchronization (file locking).
    flush_metadata_nosync = function() {
      if(private$is_view) {
        stop("Operation not permitted for views")
      }

      if (!is.null(private$zarr_format) && private$zarr_format == 3L) {
        # V3: rebuild and rewrite zarr.json with updated metadata
        v3_meta <- create_v3_array_meta(
          shape = private$shape,
          chunks = private$chunks,
          dtype = private$dtype$dtype,
          compressor = private$compressor,
          fill_value = private$fill_value,
          order = private$order,
          filters = private$filters,
          chunk_key_encoding = private$chunk_key_encoding,
          attributes = if (!is.null(private$attrs)) private$attrs$to_list() else obj_list(),
          dimension_names = private$dimension_names
        )
        mkey <- paste0(private$key_prefix, ZARR_JSON)
        meta3 <- Metadata3$new()
        encoded_meta <- meta3$encode_array_metadata(v3_meta)
        private$store$set_item(mkey, encoded_meta)
        return(invisible(NULL))
      }

      # V2 path
      if(!is_na(private$compressor)) {
        compressor_config <- private$compressor$get_config()
      } else {
        compressor_config <- NA
      }
      if(!is_na(private$filters)) {
        filters_config <- list()
        for(filter in private$filters) {
          filters_config <- append(filters_config, list(filter$get_config()))
        }
      } else {
        filters_config <- NA
      }
      zarray_meta <- list(
        shape = private$shape,
        chunks = private$chunks,
        dtype = jsonlite::unbox(private$dtype$dtype),
        compressor = compressor_config,
        fill_value = jsonlite::unbox(private$fill_value),
        order = jsonlite::unbox(private$order),
        filters = filters_config,
        dimension_separator = jsonlite::unbox(private$dimension_separator)
      )
      mkey <- paste0(private$key_prefix, ARRAY_META_KEY)

      encoded_meta <- private$store$metadata_class$encode_array_metadata(zarray_meta)
      private$store$set_item(mkey, encoded_meta)
    },
    # Build the store key for a chunk at the given coordinates.
    # V2: "key_prefix/0.1.2" (dimension_separator between coords)
    # V3 default: "key_prefix/c/0/1/2" (prefix "c" + separator between coords)
    # V3 v2-compat: "key_prefix/0.1.2" (no prefix, configured separator)
    chunk_key = function(chunk_coords) {
      if (!is.null(private$zarr_format) && private$zarr_format == 3L) {
        # V3 chunk key encoding
        # Reference: https://zarr-specs.readthedocs.io/en/latest/v3/core/v3.0.html#chunk-key-encoding
        encoding_name <- private$chunk_key_encoding$name
        sep <- private$chunk_key_encoding$configuration$separator
        if (is.null(sep)) sep <- "/"

        if (encoding_name == "default") {
          # Default V3: prefix "c" + separator + coords
          # e.g., coords (0,1,2) -> "c/0/1/2"
          coord_part <- do.call(paste, c(as.list(chunk_coords), sep = sep))
          return(paste0(private$key_prefix, "c", sep, coord_part))
        } else if (encoding_name == "v2") {
          # V2-compatible: no prefix, just coords with separator
          return(paste0(private$key_prefix, do.call(paste, c(as.list(chunk_coords), sep = sep))))
        } else {
          stop("Unsupported chunk_key_encoding: ", encoding_name)
        }
      }
      # V2 chunk key (existing behavior)
      # Reference: https://github.com/zarr-developers/zarr-python/blob/5dd4a0/zarr/core.py#L2063
      return(paste0(private$key_prefix, do.call(paste, c(as.list(chunk_coords), sep = private$dimension_separator))))
    },
    # Compute the number of chunks along each dimension of the array.
    compute_cdata_shape = function() {
      # Reference: https://github.com/zarr-developers/zarr-python/blob/5dd4a0/zarr/core.py#L428
      if(is.null(private$shape)) {
        return(1)
      }
      shape <- private$shape
      chunks <- private$chunks
      cdata_shape <- list()
      for(i in seq_len(length(shape))) {
        s <- shape[i]
        c <- chunks[i]
        cdata_shape <- append(cdata_shape, ceiling(s / c))
      }
      cdata_shape <- as.numeric(cdata_shape)
      return(cdata_shape)
    },
    # method_description
    # Resize an array without synchronization (file locking)
    resize_nosync = function(...) {
      # Note: When resizing an array, the data are not rearranged in any way.
      # If one or more dimensions are shrunk, any chunks falling outside the
      # new array shape will be deleted from the underlying store.
      # Reference: https://github.com/zarr-developers/zarr-python/blob/5dd4a0/zarr/core.py#L2340
      args <- list(...)
      old_shape <- private$shape
      new_shape <- as.numeric(normalize_resize_args(old_shape, args))
      old_cdata_shape <- private$compute_cdata_shape()

      # Update metadata
      private$shape <- new_shape
      private$flush_metadata_nosync()

      # Determine the new number and arrangement of chunks
      chunks <- private$chunks
      new_cdata_shape <- list()
      for(i in seq_len(length(new_shape))) {
        s <- new_shape[i]
        c <- chunks[i]
        new_cdata_shape <- append(new_cdata_shape, ceiling(s / c))
      }
      new_cdata_shape <- as.numeric(new_cdata_shape)

      # Remove any chunks not within range
      chunk_store <- self$get_chunk_store()
      cidx_df <- do.call(expand.grid, lapply(old_cdata_shape, seq_len))
      for(row_idx in seq_len(dim(cidx_df)[1])) {
        cidx <- as.numeric(cidx_df[row_idx, ])
        if(all(as.logical(lapply(zip_numeric(cidx, new_cdata_shape), function(v) v[1] < v[2])))) {
          # pass; keep the chunk
        } else {
          key <- private$chunk_key(cidx)
          tryCatch(
            chunk_store$delete_item(key),
            error = function(e) {
              warning(paste("Failed to delete chunk", key, ":", e$message))
            }
          )
        }
      }
    },
    # Retrieve data for a zero-dimensional array.
    get_basic_selection_zd = function(selection = NA, out = NA, fields = NA) {
      # Special case basic selection for zero-dimensional array
      # Check selection is valid
      if(!is.null(selection) && selection != "...") {
        stop("err_too_many_indices(selection, ())")
      }
      selection <- ensure_list(selection)
      # Obtain encoded data for chunk
      c_key <- private$chunk_key(c(0))

      chunk_nested_array <- tryCatch({
        c_data <- self$get_chunk_store()$get_item(c_key)
        chunk_inner <- private$decode_chunk(c_data)
        # From raw.
        NestedArray$new(chunk_inner, shape = private$chunks, dtype = private$dtype, order = private$order)
      }, error = function(cond) {
        if(is_key_error(cond)) {
          # chunk not initialized
          as_dtype_func <- private$dtype$get_asrtype()
          chunk_inner <- as_dtype_func(private$fill_value)
          # From array().
          return(NestedArray$new(chunk_inner, shape = private$chunks, dtype = private$dtype, order = private$order))
        } else {
          stop(cond)
        }
      })      

      # TODO: Handle fields
      # if(!is.na(fields)) {
      #   chunk <- chunk[fields]
      # }

      # Handle selection of the scalar value via empty tuple
      if(is_na(out)) {
        out <- chunk_nested_array
      } else {
        # TODO
        #out[selection] <- as_scalar(chunk)
      }
      return(out)
    },
    # Retrieve data for an N-dimensional array using basic (integer/slice) selection.
    get_basic_selection_nd = function(selection = NA, out = NA, fields = NA) {
      # Route negative-step slices to orthogonal selection
      sel_list <- if(is.list(selection)) selection else list(selection)
      has_neg_step <- any(vapply(sel_list, function(s) {
        is_slice(s) && !is.na(s$step) && s$step < 0
      }, logical(1)))
      if(has_neg_step) {
        indexer <- OrthogonalIndexer$new(selection, self)
      } else {
        indexer <- BasicIndexer$new(selection, self)
      }
      return(private$get_selection(indexer, out = out, fields = fields))
    },
    # Retrieve data by iterating over chunks that overlap the indexer selection, extracting data into the output array.
    get_selection = function(indexer, out = NA, fields = NA) {
      # Reference: https://github.com/gzuidhof/zarr.js/blob/292804/src/core/index.ts#L304
      # We iterate over all chunks which overlap the selection and thus contain data
      # that needs to be extracted. Each chunk is processed in turn, extracting the
      # necessary data and storing into the correct location in the output array.
      out_dtype <- private$dtype
      out_shape <- indexer$shape
      out_size <- compute_size(indexer$shape)

      if(!is.na(out)) {
        # TODO: handle out provided as parameter
      } else {
        out <- NestedArray$new(NULL, shape = out_shape, dtype = out_dtype, order = private$order)
      }

      if(out_size == 0) {
        return(out)
      }

      # --- zarrs fast path ---
      if (can_use_zarrs(indexer, private$store)) {
        ranges <- selection_to_ranges(indexer)
        store_id <- private$store$get_store_identifier()
        ct <- getOption("pizzarr.concurrent_target", NULL)
        result <- tryCatch(
          zarrs_get_subset(store_id, private$path, ranges, ct),
          error = function(e) NULL
        )
        if (!is.null(result)) {
          # zarrs returns flat F-order data (transposed in Rust).
          # Reshape directly into R array with correct dims.
          zarrs_shape <- vapply(ranges, function(r) r[2L] - r[1L], integer(1))
          arr <- array(result$data, dim = zarrs_shape)
          # Squeeze scalar dims (IntDimIndexer) to match out_shape.
          # Only apply when out_shape has fewer dims (length-1 dims dropped).
          os <- unlist(out_shape)
          if (length(os) > 0 && !identical(zarrs_shape, os)) {
            dim(arr) <- os
          }
          out$data <- arr
          return(out)
        }
        # zarrs failed — fall through to R-native path
      }

      # --- R-native path ---
      parts <- indexer$iter()
      for(proj in parts) {
        private$chunk_getitem(proj$chunk_coords, proj$chunk_sel, out,
                              proj$out_sel, drop_axes = indexer$drop_axes)
      }

      # Return scalar instead of zero-dimensional array.
      if(length(out$shape) == 0) {
        return(out$data[1])
      }
      return(out)

    },
    # Set data for a zero-dimensional array.
    set_basic_selection_zd = function(selection, value, fields = NA) {
      # Reference: https://github.com/zarr-developers/zarr-python/blob/5dd4a0e6cdc04c6413e14f57f61d389972ea937c/zarr/core.py#L1625

      # check selection is valid
      selection <- ensure_list(selection)
      if(!(length(selection) == 0 || selection == "...")) {
        stop("err_too_many_indices(selection, self._shape)")
      }

      # TODO: check fields
      #check_fields(fields, self._dtype)
      #fields = check_no_multi_fields(fields)

      # obtain key for chunk
      c_key <- private$chunk_key(c(0))

      # setup chunk
      # chunk <- tryCatch({
      #   # obtain compressed data for chunk
      #   c_data <- self$get_chunk_store()$get_item(c_key)
      #   # decode chunk
      #   chunk_inner <- private$decode_chunk(c_data)
      #   return(chunk_inner)
      # }, error = function(cond) {
      #   if(is_key_error(cond)) {
      #     # chunk not initialized
      #     as_dtype_func = get_dtype_asrtype(private$dtype)
      #     chunk_inner <- as_dtype_func(private$fill_value)
      #     return(chunk_inner)
      #   } else {
      #     print(cond$message)
      #     stop("rethrow")
      #   }
      # })
      
      # TODO
      # set value
      # if fields:
      #     chunk[fields][selection] = value
      # else:
      #     chunk[selection] = value

      # TODO
      # remove chunk if write_empty_chunks is false and it only contains the fill value
      # if (not self.write_empty_chunks) and all_equal(self.fill_value, chunk):
      #     try:
      #         del self.chunk_store[ckey]
      #         return
      #     except Exception:  # pragma: no cover
      #         # deleting failed, fallback to overwriting
      #         pass
      # else:

      # encode and store

      chunk_nested_array <- NestedArray$new(as_scalar(value), shape = NULL, dtype = private$dtype, order = private$order)
      chunk_raw <- chunk_nested_array$flatten_to_raw(order = private$order)

      c_data <- private$encode_chunk(chunk_raw)
      self$get_chunk_store()$set_item(c_key, c_data)
    },
    # Set data for an N-dimensional array using basic (integer/slice) selection.
    set_basic_selection_nd = function(selection, value, fields = NA) {
      indexer <- BasicIndexer$new(selection, self)
      return(private$set_selection(indexer, value = value, fields = fields))
    },
    # Set data by iterating over chunks that overlap the indexer selection, replacing data from the value array.
    set_selection = function(indexer, value, fields = NA) {
      # Reference: https://github.com/zarr-developers/zarr-python/blob/5dd4a0/zarr/core.py#L1682
      # Reference: https://github.com/gzuidhof/zarr.js/blob/15e3a3f00eb19f0133018fb65f002311ea53bb7c/src/core/index.ts#L566

      # // We iterate over all chunks which overlap the selection and thus contain data
      # // that needs to be replaced. Each chunk is processed in turn, extracting the
      # // necessary data from the value array and storing into the chunk array.

      # // N.B., it is an important optimisation that we only visit chunks which overlap
      # // the selection. This minimises the number of iterations in the main for loop.

      selection_shape <- indexer$shape
      selection_shape_vec <- ensure_integer_vec(indexer$shape)

      if(sum(as.numeric(selection_shape)) > 0) {
        # Check value shape
        if (length(selection_shape) == 0) {
          # Setting a single value
        } else if (is_scalar(value)) {
          # Setting a scalar value
        } else if("array" %in% class(value) || (is.atomic(value) && is.vector(value))) {
          if (!is.null(dim(value)) && !all(ensure_integer_vec(dim(value)) == selection_shape_vec)) {
            stop("Shape mismatch in source array and set selection: ${dim(value)} and ${selectionShape}")
          }
          value <- NestedArray$new(value, shape = selection_shape_vec, dtype=private$dtype, order = private$order)
        } else if (inherits(value, "integer64")) {
          value <- NestedArray$new(value, shape = selection_shape_vec, dtype=private$dtype, order = private$order)
        } else if ("NestedArray" %in% class(value)) {
          if (!all(ensure_integer_vec(value$shape) == selection_shape_vec)) {
            stop("Shape mismatch in source NestedArray and set selection: ${value.shape} and ${selectionShape}")
          }
        } else {
          stop("UnsupportedOperation(object dtype requires object_codec in filters for set_item)")
        }

        # --- zarrs write fast path ---
        if (can_use_zarrs_write(indexer, private$store)) {
          ranges <- selection_to_ranges(indexer)
          store_id <- private$store$get_store_identifier()
          ct <- getOption("pizzarr.concurrent_target", NULL)
          # Extract flat data from value
          if (inherits(value, "NestedArray")) {
            write_data <- as.vector(value$data)
          } else if (is_scalar(value)) {
            write_data <- value
          } else {
            write_data <- as.vector(value)
          }
          # Rust handles F→C transpose internally; just flatten.
          # Shape is passed via ranges so Rust knows the dimensions.
          result <- tryCatch(
            zarrs_set_subset(store_id, private$path, ranges, write_data, ct),
            error = function(e) NULL
          )
          if (isTRUE(result)) return(invisible(NULL))
          # zarrs failed — fall through to R-native path
        }

        parts <- indexer$iter()
        for(proj in parts) {
          chunk_value <- private$get_chunk_value(proj, indexer, value, selection_shape)
          private$chunk_setitem(proj$chunk_coords, proj$chunk_sel, chunk_value)
        }

        return()
      }
    },
    # Decode a chunk and extract the selected region into the output array.
    process_chunk = function(out, cdata, chunk_selection, drop_axes, out_is_ndarray, fields, out_selection, partial_read_decode = FALSE) {
      # Reference: https://github.com/zarr-developers/zarr-python/blob/5dd4a0/zarr/core.py#L1755
      # TODO
    },
    # Extract the portion of a value array corresponding to a single chunk projection.
    get_chunk_value = function(proj, indexer, value, selection_shape) {
      # Reference: https://github.com/gzuidhof/zarr.js/blob/15e3a3f00eb19f0133018fb65f002311ea53bb7c/src/core/index.ts#L550
      
      # value is the full NestedArray representing the value to be set.
      # we call value.get() to get the value for the current chunk selection,
      # since the full value might span multiple chunks.
      if (length(selection_shape) == 0) {
        chunk_value <- value
      } else if (is_scalar(value)) {
        chunk_value <- value
      } else {
        chunk_value <- value$get(proj$out_sel)
        if (isTRUE(indexer$drop_axes)) {
          stop("Handling drop axes not supported yet")
        }
      }
      return(chunk_value)
    },
    # method_description
    # TODO
    chunk_buffer_to_raw_array = function(decoded_chunk) {
      # TODO
    },
    # method_description
    chunk_getitem = function(chunk_coords, chunk_selection, out, out_selection, drop_axes = NA, fields = NA) {
      # TODO
      # Reference: https://github.com/gzuidhof/zarr.js/blob/15e3a3f00eb19f0133018fb65f002311ea53bb7c/src/core/index.ts#L380

      if(length(chunk_coords) != length(private$chunks)) {
        stop("Inconsistent shapes: chunk_coords length ", length(chunk_coords),
             " != chunks length ", length(private$chunks))
      }
      c_key <- private$chunk_key(chunk_coords)

      tryCatch({
        c_data <- self$get_chunk_store()$get_item(c_key)
        decoded_chunk <- private$decode_chunk(c_data)

        if(is_contiguous_selection(out_selection) && is_total_slice(chunk_selection, private$chunks) && is.null(private$filters)) {
          out$set(out_selection, NestedArray$new(decoded_chunk, shape=private$chunks, dtype=private$dtype, order = private$order))
          return(TRUE)
        }

        # Decode chunk
        chunk <- NestedArray$new(decoded_chunk, shape=private$chunks, dtype=private$dtype, order = private$order)
        tmp <- chunk$get(chunk_selection)

        if(!is_na(drop_axes)) {
          stop("Drop axes is not supported yet")
        }
        out$set(out_selection, tmp)
      }, error = function(cond) {
        if(is_key_error(cond)) {
          # fill with scalar if cKey doesn't exist in store
          # NaN is a valid fill value; is.na(NaN)==TRUE so check is.nan first
          if ((is.numeric(private$fill_value) && is.nan(private$fill_value)) ||
              !is_na(private$fill_value)) {
            out$set(out_selection, as_scalar(private$fill_value))
          }
        } else {
          stop(cond)
        }
      })
    },
    # method_description
    # TODO
    chunk_getitems = function(lchunk_coords, lchunk_selection, out, lout_selection, drop_axes = NA, fields = NA) {
      # TODO
    },
    # method_description
    # TODO
    chunk_setitem = function(chunk_coords, chunk_selection, value, fields = NA) {
      # Reference: https://github.com/gzuidhof/zarr.js/blob/15e3a3f00eb19f0133018fb65f002311ea53bb7c/src/core/index.ts#L625
      
      # Obtain key for chunk storage
      chunk_key <- private$chunk_key(chunk_coords)

      if (is_total_slice(chunk_selection, private$chunks)) {
        # Totally replace chunk

        # Optimization: we are completely replacing the chunk, so no need
        # to access the existing chunk data

        if (is_scalar(value)) {
          chunk <- NestedArray$new(
            value,
            shape = private$chunks,
            dtype = private$dtype,
            order = private$order
          )
        } else {
          # Ensure the value data is coerced to the target dtype's R type
          as_dtype_func <- private$dtype$get_asrtype()
          coerced_data <- as_dtype_func(value$data)
          dim(coerced_data) <- dim(value$data)
          # Use actual data shape, not private$chunks: boundary chunks may be
          # smaller than the nominal chunk size so value$data won't fill a
          # full-sized chunk and would cause a recycling warning.
          chunk_shape <- if (!is.null(dim(coerced_data)) &&
                             !identical(as.integer(dim(coerced_data)),
                                        as.integer(private$chunks))) {
            dim(coerced_data)
          } else {
            private$chunks
          }
          chunk <- NestedArray$new(
            coerced_data,
            shape = chunk_shape,
            dtype = private$dtype,
            order = private$order
          )
        }
        chunk_raw <- chunk$flatten_to_raw(order = private$order)
      } else {
        # partially replace the contents of this chunk

        # Existing chunk data
        chunk_nested_array <- tryCatch({

          # Chunk is initialized if this does not error
          chunk_store_data <- self$get_chunk_store()$get_item(chunk_key)
          dbytes <- private$decode_chunk(chunk_store_data)
          # From raw.
          NestedArray$new(
            dbytes,
            shape = private$chunks,
            dtype = private$dtype,
            order = private$order
          )
        }, error = function(cond) {
          if (is_key_error(cond)) {
            # Chunk is not initialized
            dtype_constr <- private$dtype$get_typed_array_ctr()
            chunk_size <- compute_size(private$chunks)
            chunk_data <- dtype_constr(chunk_size)
            # NaN is a valid fill value; is.na(NaN)==TRUE so check is.nan first
            if ((is.numeric(private$fill_value) && is.nan(private$fill_value)) ||
                !is_na(private$fill_value)) {
              chunk_data <- chunk_fill(chunk_data, private$fill_value)
            }
            # From base R array.
            return(NestedArray$new(
              chunk_data,
              shape = private$chunks,
              dtype = private$dtype,
              order = private$order
            ))
          } else {
            stop(cond)
          }
        })
        
        # Now that we have the existing chunk data,
        # we set the new value by using the chunk_selection
        # to specify which subset to replace.
        chunk_nested_array$set(chunk_selection, value)
        chunk_raw <- chunk_nested_array$flatten_to_raw(order = private$order)
      }
      # We encode the new chunk and set it in the chunk store.
      chunk_data <- private$encode_chunk(chunk_raw)
      self$get_chunk_store()$set_item(chunk_key, chunk_data)
    },
    # method_description
    # TODO
    chunk_setitem_nosync = function(chunk_coords, chunk_selection, value, fields = NA) {
      # TODO
    },
    # method_description
    # TODO
    chunk_setitems = function(lchunk_coords, lchunk_selection, values, fields = NA) {
      # TODO
    },
    # method_description
    # TODO
    process_for_setitem = function(ckey, chunk_selection, value, fields = NA) {
      # TODO
    },
    chunk_delitem = function(ckey) {
      self$get_chunk_store()$delete_item(ckey)
    },
    # method_description
    chunk_delitems = function(ckeys) {
      for(ckey in ckeys) {
        private$chunk_delitem(ckey)
      }
    },
    # method_description
    # TODO
    decode_chunk = function(cdata, start = NA, nitems = NA, expected_shape = NA) {
      # Reference: https://github.com/zarr-developers/zarr-python/blob/5dd4a0e6cdc04c6413e14f57f61d389972ea937c/zarr/core.py#L2066
      # decompress
      if(!is_na(private$compressor)) {
        # TODO: only decode requested items
        # if (
        #     all(x is not None for x in [start, nitems])
        #     and self._compressor.codec_id == "blosc"
        # ) and hasattr(self._compressor, "decode_partial"):
        #     chunk = self._compressor.decode_partial(cdata, start, nitems)
        # else:
        chunk <- private$compressor$decode(cdata, self)
      } else {
        chunk <- cdata
      }

      # apply filters
      if(!is_na(private$filters)) {
        for (f in rev(private$filters)) {
          chunk <- f$decode(chunk)
        }
      }

      # TODO: view as numpy array with correct dtype
      # chunk <- ensure_ndarray(chunk)
      # special case object dtype, because incorrect handling can lead to
      # segfaults and other bad things happening
      # if self._dtype != object:
      #    chunk = chunk.view(self._dtype)
      # elif chunk.dtype != object:
          # If we end up here, someone must have hacked around with the filters.
          # We cannot deal with object arrays unless there is an object
          # codec in the filter chain, i.e., a filter that converts from object
          # array to something else during encoding, and converts back to object
          # array during decoding.
          # raise RuntimeError('cannot read object array without object codec')

      # ensure correct chunk shape
      return(chunk)
    },
    # method_description
    # TODO
    encode_chunk = function(chunk_as_raw) {
      # Reference: https://github.com/zarr-developers/zarr-python/blob/5dd4a0e6cdc04c6413e14f57f61d389972ea937c/zarr/core.py#L2105

      chunk <- chunk_as_raw

      # apply filters
      if(!is_na(private$filters)) {
        for(f in private$filters) {
          chunk <- f$encode(chunk)
        }
      }

      # TODO: check object encoding
      #if ensure_ndarray(chunk).dtype == object:
      #    raise RuntimeError('cannot write object array without object codec')

      # compress
      if(!is_na(private$compressor)) {
        cdata <- private$compressor$encode(chunk, self)
      } else {
        cdata <- chunk
      }

      # TODO: ensure in-memory data is immutable and easy to compare
      #if isinstance(self.chunk_store, MutableMapping):
      #    cdata = ensure_bytes(cdata)

      return(cdata)
    },
    # Append data to the array along the specified axis without synchronization.
    # Reference: https://github.com/zarr-developers/zarr-python/blob/5dd4a0/zarr/core.py#L2141
    append_nosync = function(data, axis = 1) {
      if (!is.array(data)) {
        data <- as.array(data)
      }

      old_shape <- private$shape
      data_shape <- dim(data)
      if (length(data_shape) != length(old_shape)) {
        stop("incompatible number of dimensions for append")
      }
      for (i in seq_along(old_shape)) {
        if (i != axis && data_shape[i] != old_shape[i]) {
          stop("shape mismatch on non-append axis")
        }
      }

      new_shape <- old_shape
      new_shape[axis] <- old_shape[axis] + data_shape[axis]
      do.call(self$resize, as.list(new_shape))

      selection <- list()
      for (i in seq_along(new_shape)) {
        if (i == axis) {
          selection[[i]] <- slice(old_shape[axis] + 1, new_shape[axis])
        } else {
          selection[[i]] <- slice(1, new_shape[i])
        }
      }

      self$set_item(selection, data)
    }
  ),
  public = list(
    #' @description
    #' Create a new ZarrArray instance.
    #' @param store ([Store])\cr
    #'   Array store, already initialized.
    #' @param path (`character(1)`)\cr
    #'   Storage path.
    #' @param read_only (`logical(1)`)\cr
    #'   Whether the array is read-only.
    #' @param chunk_store ([Store] or `NA`)\cr
    #'   Separate storage for chunks. If not provided, `store` will be used
    #'   for both chunks and metadata.
    #' @param synchronizer (`ANY` or `NA`)\cr
    #'   Object used to synchronize write access to the array.
    #' @param cache_metadata (`logical(1)`)\cr
    #'   Whether to cache metadata.
    #' @param cache_attrs (`logical(1)`)\cr
    #'   Whether to cache attributes.
    #' @param write_empty_chunks (`logical(1)`)\cr
    #'   Whether to write empty chunks.
    #' @return A [ZarrArray] instance.
    initialize = function(store, path = NA, read_only = FALSE, 
                          chunk_store = NA, synchronizer = NA, 
                          cache_metadata = TRUE, cache_attrs = TRUE, 
                          write_empty_chunks = TRUE) {
      private$store <- store
      private$chunk_store <- chunk_store
      if(!is.na(path) && nchar(path) > 0) {
        private$path <- normalize_storage_path(path)
        private$key_prefix <- paste0(private$path, "/")
      } else {
        private$path <- NA
        private$key_prefix <- ""
      }
      private$read_only <- read_only
      private$synchronizer <- synchronizer
      private$cache_metadata <- cache_metadata
      private$cache_attrs <- cache_attrs
      private$is_view <- FALSE
      private$write_empty_chunks <- write_empty_chunks

      private$load_metadata()

      akey <- paste0(private$key_prefix, ATTRS_KEY)
      if (!is.null(private$zarr_format) && private$zarr_format == 3L) {
        # V3: attributes are embedded in zarr.json, not in a separate .zattrs.
        # Create Attributes object and pre-populate cache from parsed metadata.
        private$attrs <- Attributes$new(store, key = akey, zarr_format = 3L)
        if (!is.null(private$meta$attributes)) {
          private$attrs$set_cached_v3_attrs(private$meta$attributes)
        }
      } else {
        # V2: attributes in separate .zattrs file (existing behavior)
        private$attrs <- Attributes$new(store, key = akey, zarr_format = 2L)
      }

      private$oindex <- OIndex$new(self)
      private$vindex <- VIndex$new(self)
    },
    #' @description
    #' get store from array.
    #' @return [Store].
    get_store = function() {
      return(private$store)
    },    
    #' @description
    #' get array path
    #' @return `character(1)` or `NA`.
    get_path = function() {
      return(private$path)
    },
    #' @description
    #' get full array name
    #' @return `character(1)` or `NA`.
    get_name = function() {
      if(!is.na(private$path)) {
        name <- private$path
        name_vec <- str_to_vec(name)
        if(name_vec[1] != "/") {
          name <- paste0("/", name)
        }
        return(name)
      }
      return(NA)
    },
    #' @description
    #' get the basename of an array
    #' @return `character(1)` or `NA`.
    get_basename = function() {
      name <- self$get_name()
      if(!is.na(name)) {
        name_segments <- stringr::str_split(name, "/")[[1]]
        return(name_segments[length(name_segments)])
      }
      return(NA)
    },
    #' @description
    #' get the read only property of an array (TRUE/FALSE)
    #' @return `logical(1)`.
    get_read_only = function() {
      return(private$read_only)
    },
    #' @description
    #' set the read only property of an array
    #' @param val (`logical(1)`)\cr
    #'   Value to set.
    #' @return `NULL` (called for side effects).
    set_read_only = function(val) {
      private$read_only <- val
    },
    #' @description
    #' get the chunk store for an array
    #' @return [Store].
    get_chunk_store = function() {
      if(is_na(private$chunk_store)) {
        return(private$store)
      } else {
        return(private$chunk_store)
      }
    },
    #' @description
    #' get the shape of an array
    #' @return `integer()`.
    get_shape = function() {
      private$refresh_metadata()
      return(private$shape)
    },
    #' @description
    #' set or reset the size of an array
    #' @param value (`integer()`)\cr
    #'   Numeric size to set.
    #' @return `NULL` (called for side effects).
    set_shape = function(value) {
      self$resize(value)
    },
    #' @description
    #' Change the shape of the array by growing or shrinking one or more dimensions.
    #' @param ... Arguments for do.call.
    #' @return `NULL` (called for side effects).
    resize = function(...) {
      args <- list(...)
      do.call(private$resize_nosync, args)
    },
    #' @description
    #' get the chunk metadata of an array
    #' @return `integer()`.
    get_chunks = function() {
      return(private$chunks)
    },
    #' @description
    #' get the Dtype of an array
    #' @return [Dtype].
    get_dtype = function() {
      return(private$dtype)
    },
    #' @description
    #' get the compressor of an array
    #' @return [Codec] or `NA`.
    get_compressor = function() {
      return(private$compressor)
    },
    #' @description
    #' get the fill value of an array
    #' @return Scalar fill value.
    get_fill_value = function() {
      return(private$fill_value)
    },
    #' @description
    #' set the fill value of an array
    #' @param val Fill value to use.
    #' @return `NULL` (called for side effects).
    set_fill_value = function(val) {
      private$fill_value <- val
      private$flush_metadata_nosync()
    },
    #' @description
    #' Set dimension names of the array (V3 only).
    #' @param names Character vector of length equal to the number of array
    #'   dimensions, or `NULL` to clear.
    #' @return `NULL` (called for side effects).
    set_dimension_names = function(names) {
      if (is.null(private$zarr_format) || private$zarr_format != 3L) {
        stop("DimensionNamesV2Error(set_dimension_names is only supported for V3 arrays)")
      }
      if (!is.null(names) && length(names) != length(private$shape)) {
        stop("DimensionNamesLengthError(length must match length(shape))")
      }
      private$dimension_names <- names
      private$flush_metadata_nosync()
    },
    #' @description
    #' get the storage order metadata of an array.
    #' @return `character(1)`, `"C"` or `"F"`.
    get_order = function() {
      return(private$order)
    },
    #' @description
    #' get the filters metadata of an array
    #' @return `list()` of [Codec] objects or `NA`.
    get_filters = function() {
      return(private$filters)
    },
    #' @description
    #' Get the synchronizer used to coordinate write access to the array.
    #' @return Synchronizer object or `NA`.
    get_synchronizer = function() {
      return(private$synchronizer)
    },
    #' @description
    #' get attributes of array
    #' @return [Attributes].
    get_attrs = function() {
      return(private$attrs)
    },
    #' @description
    #' get number of dimensions of array
    #' @return `integer(1)`.
    get_ndim = function() {
      return(length(private$shape))
    },
    #' @description
    #' Get dimension names of the array (V3 only).
    #' @return Character vector or `NULL`.
    get_dimension_names = function() {
      return(private$dimension_names)
    },
    #' @description
    #' Get the total number of elements in the array.
    #' @return `numeric(1)`.
    get_size = function() {
      # Reference: https://github.com/zarr-developers/zarr-python/blob/5dd4a0/zarr/core.py#L383
      # TODO
    },
    #' @description
    #' Get the size in bytes of each item in the array.
    #' @return `integer(1)`.
    get_itemsize = function() {
      # TODO
    },
    #' @description
    #' get number of bytes of an array
    #' @return `numeric(1)`.
    get_nbytes = function() {
      private$refresh_metadata()
      return(self$get_size() * self$get_itemsize())
    },
    #' @description
    #' Get the total number of stored bytes for the array, including metadata and compressed chunk data.
    #' @return `numeric(1)`.
    get_nbytes_stored = function() {
      # Reference: https://github.com/zarr-developers/zarr-python/blob/5dd4a0/zarr/core.py#L413
      # TODO
    },
    #' @description
    #' Get the number of chunks along each dimension of the array.
    #' @return `numeric()`.
    get_cdata_shape = function() {
      private$refresh_metadata()
      return(private$compute_cdata_shape())
    },
    #' @description
    #' Get the total number of chunks in the array.
    #' @return `integer(1)`.
    get_nchunks = function() {
      # TODO
    },
    #' @description
    #' Get the number of chunks that have been initialized with data.
    #' @return `integer(1)`.
    get_nchunks_initialized = function() {
      # TODO
    },
    #' @description
    #' get is_view metadata of array
    #' @return `logical(1)`.
    get_is_view = function() {
      return(private$is_view)
    },
    #' @description
    #' get orthogonal index of array
    #' @return [OIndex].
    get_oindex = function() {
      return(private$oindex)
    },
    #' @description
    #' get vectorized index of array
    #' @return [VIndex].
    get_vindex = function() {
      return(private$vindex)
    },
    #' @description
    #' get write empty chunks setting of array
    #' @return `logical(1)`.
    get_write_empty_chunks = function() {
      return(private$write_empty_chunks)
    },
    #' @description
    #' check if another object refers to the same array. does not check array data
    #' @param other Other object to check.
    #' @return `logical(1)`.
    equals = function(other) {
      if (!inherits(other, "ZarrArray")) return(FALSE)
      return(all(c(
        identical(private$store, other$get_store()),
        identical(private$read_only, other$get_read_only()),
        identical(private$path, other$get_path()),
        !private$is_view
      )))
    },
    #' @description
    #' Print a human-readable summary of the array.
    #' @param ... Ignored.
    #' @return `self` (invisibly).
    print = function(...) {
      nm <- self$get_name()
      if (is.na(nm)) nm <- "/"
      store_type <- class(private$store)[1]
      zf <- if (!is.null(private$zarr_format)) private$zarr_format else 2L
      cat(paste0("<ZarrArray> ", nm, "\n"))
      cat(paste0("  Shape       : (", paste(private$shape, collapse = ", "), ")\n"))
      cat(paste0("  Chunks      : (", paste(private$chunks, collapse = ", "), ")\n"))
      cat(paste0("  Data type   : ", private$dtype$dtype, "\n"))
      cat(paste0("  Fill value  : ", private$fill_value, "\n"))
      cat(paste0("  Order       : ", private$order, "\n"))
      cat(paste0("  Read-only   : ", private$read_only, "\n"))
      cat(paste0("  Compressor  : ", format_codec(private$compressor), "\n"))
      cat(paste0("  Store type  : ", store_type, "\n"))
      cat(paste0("  Zarr format : ", zf, "\n"))
      invisible(self)
    },
    #' @description
    #' Iterate over the array or a range of it, yielding successive slices along the first dimension. Uses chunk caching for efficient sequential access.
    #' @param start Start index of the iteration range (1-based, inclusive).
    #' @param end End index of the iteration range (1-based, inclusive).
    #' @return `list()` of [NestedArray] slices.
    islice = function(start = NA, end = NA) {
      # TODO
    },
    #' @description
    #' Return the length of the first dimension. Raises an error for zero-dimensional arrays.
    #' @return `integer(1)`.
    length = function() {
      if(private$shape) {
        return(private$shape[1])
      } else {
        # 0-dimensional array, same error message as numpy
        stop("length of unized object")
      }
    },
    #' @description
    #' Subset the array.
    #' @param selection Selection specifying the subset to retrieve.
    #' @return [NestedArray].
    get_item = function(selection) {
      # Reference: https://github.com/zarr-developers/zarr-python/blob/5dd4a0/zarr/core.py#L580
      # Reference: https://github.com/gzuidhof/zarr.js/blob/master/src/core/index.ts#L266
      
      if(is_pure_fancy_indexing(selection)){
        # TODO: implement vindex further for vectorized indexing
        stop("vectorized indexing is not supported yet")
        # return(self$get_vindex()$get_item(selection))
      } else {
        return(self$get_basic_selection(selection)) 
      }
    },
    #' @description
    #' Get a selection of an array based on a "basic" list of slices.
    #' @param selection Selection specifying the subset to retrieve.
    #' @param out Not yet implemented.
    #' @param fields Not yet implemented.
    #' @return [NestedArray].
    get_basic_selection = function(selection = NA, out = NA, fields = NA) {
      # Refresh metadata
      if(!private$cache_metadata) {
        private$load_metadata()
      }
      # Handle zero-dimensional arrays
      if(is.null(private$shape) || length(private$shape) == 0) {
        return(private$get_basic_selection_zd(selection, out = out, fields = fields))
      }
      return(private$get_basic_selection_nd(selection, out = out, fields = fields))
    },
    #' @description
    #' Retrieve data using orthogonal (outer) indexing.
    #' @param selection Selection specifying the subset to retrieve.
    #' @param out Not yet implemented.
    #' @param fields Not yet implemented.
    #' @return [NestedArray].
    get_orthogonal_selection = function(selection = NA, out = NA, fields = NA) {
      
      # Refresh metadata
      if(!private$cache_metadata) {
        private$load_metadata()
      }
      
      indexer <- OrthogonalIndexer$new(selection, self)
      return(private$get_selection(indexer, out = out, fields = fields))
    },
    #' @description
    #' Retrieve data using coordinate (point) indexing. Not yet implemented.
    #' @param selection Selection specifying the subset to retrieve.
    #' @param out Not yet implemented.
    #' @param fields Not yet implemented.
    #' @return [NestedArray].
    get_coordinate_selection = function(selection = NA, out = NA, fields = NA) {
      # TODO
    },
    #' @description
    #' Retrieve data using a boolean mask. Not yet implemented.
    #' @param selection Selection specifying the subset to retrieve.
    #' @param out Not yet implemented.
    #' @param fields Not yet implemented.
    #' @return [NestedArray].
    get_mask_selection = function(selection = NA, out = NA, fields = NA) {
      # TODO
    },
    #' @description
    #' Set a subset of the array.
    #' @param selection Selection specifying the subset to set.
    #' @param value The value to set, as an R array() or a Zarr NestedArray instance.
    #' @return `NULL` (called for side effects).
    set_item = function(selection, value) {
      self$set_basic_selection(selection, value)
    },
    #' @description
    #' Set a selection of the array using basic (integer/slice) indexing.
    #' @param selection Selection specifying the subset to set.
    #' @param value The value to set.
    #' @param fields Not yet implemented.
    #' @return `NULL` (called for side effects).
    set_basic_selection = function(selection, value, fields = NA) {
      # Handle zero-dimensional arrays
      if(is.null(private$shape) || length(private$shape) == 0) {
        return(private$set_basic_selection_zd(selection, value = value, fields = fields))
      }
      return(private$set_basic_selection_nd(selection, value = value, fields = fields))
    },
    #' @description
    #' Set data using orthogonal (outer) indexing. Not yet implemented.
    #' @param selection Selection specifying the subset to set.
    #' @param value Not yet implemented.
    #' @param fields Not yet implemented.
    #' @return `NULL` (called for side effects).
    set_orthogonal_selection = function(selection, value, fields = NA) {
      # TODO
    },
    #' @description
    #' Set data using coordinate (point) indexing. Not yet implemented.
    #' @param selection Selection specifying the subset to set.
    #' @param value Not yet implemented.
    #' @param fields Not yet implemented.
    #' @return `NULL` (called for side effects).
    set_coordinate_selection = function(selection, value, fields = NA) {
      # TODO
    },
    #' @description
    #' Set data using a boolean mask. Not yet implemented.
    #' @param selection Selection specifying the subset to set.
    #' @param value Not yet implemented.
    #' @param fields Not yet implemented.
    #' @return `NULL` (called for side effects).
    set_mask_selection = function(selection, value, fields = NA) {
      # TODO
    },
    #' @description
    #' Get array diagnostic info. Not yet implemented.
    #' @return `character(1)`.
    get_info = function() {
      # Reference: https://github.com/zarr-developers/zarr-python/blob/5dd4a0/zarr/core.py#L2141
      # TODO
    },
    #' @description
    #' Compute a checksum digest of the array data. Not yet implemented.
    #' @param hashname Name of hash.
    #' @return `character(1)`.
    get_digest = function(hashname = "sha1") {
      # TODO
    },
    #' @description
    #' Compute a hex-string checksum digest. Not yet implemented.
    #' @param hashname Name of hash.
    #' @return `character(1)`.
    get_hexdigest = function(hashname = "sha1") {
      # TODO
    },
    #' @description
    #' Append data to the array along the specified axis.
    #' @param data Data to append. Will be coerced to array if needed.
    #' @param axis Axis to append along (1-indexed, R convention). Default 1.
    #' @return `NULL` (called for side effects).
    append = function(data, axis = 1) {
      private$append_nosync(data, axis)
    },
    #' @description
    #' Return a view on the array with modified interpretation. Not yet implemented.
    #' @param shape Not yet implemented.
    #' @param chunks Not yet implemented.
    #' @param dtype Not yet implemented.
    #' @param fill_value Not yet implemented.
    #' @param filters Not yet implemented.
    #' @param read_only Not yet implemented.
    #' @param synchronizer Not yet implemented.
    #' @return [ZarrArray].
    view = function(shape = NA, chunks = NA, dtype = NA, fill_value = NA,
                    filters = NA, read_only = NA, synchronizer = NA) {
      # TODO
    },
    #' @description
    #' Return a view with a different dtype. Not yet implemented.
    #' @param dtype Not yet implemented.
    #' @return [ZarrArray].
    astype = function(dtype) {
      # Reference: https://github.com/zarr-developers/zarr-python/blob/5dd4a0/zarr/core.py#L2586
      # TODO
    },
    #' @description
    #' Get the dimension separator used in chunk keys.
    #' @return `character(1)`.
    get_dimension_separator = function() {
      return(private$dimension_separator)
    },
    #' @description
    #' Convert Zarr object to R array (for S3 method). Note that this loads all data into memory.
    #' @return An R `array`.
    as.array = function() {
      return(self$get_item("...")$data)
    }
  )
)

#' S3 method for custom bracket subsetting
#'
#' @param obj Object.
#' @param ... Dots.
#' @keywords internal
#' @export
`[.ZarrArray` <- function(obj, ...) {
  filters <- substitute(...())
  if(length(filters) != length(obj$get_shape())) {
    stop("This Zarr object has ", length(obj$get_shape()), " dimensions, ", length(filters), " were supplied")
  }
  filters <- manage_filters(filters)
  return(obj$get_orthogonal_selection(filters))
}

#' S3 method for custom bracket assignment
#'
#' @param obj Object.
#' @param ... Dots.
#' @param value Array or ZarrArray.
#' @keywords internal
#' @export
`[<-.ZarrArray` <- function(obj, ..., value) {
  stop("Assignment using bracket notation is not yet supported - use set_item() directly")
}

#' S3 method for as.array
#'
#' @param x Object.
#' @param ... Not used.
#' @keywords internal
#' @export
as.array.ZarrArray = function(x, ...) {
  x$as.array()
}

library(pizzarr)

test_that("no infinite loop", {
    shape <- c(27557749)
    dtype <- "<i4"
    chunks <- NA

    dtype <- normalize_dtype(dtype, object_codec = NA)
    shape <- normalize_shape(shape)
    
    dtype_itemsize <- dtype$num_bytes
    chunks <- normalize_chunks(chunks, shape, dtype_itemsize)
    expect_equal(chunks, c(215295))
})

test_that("can get array that spans multiple chunks", {
    a <- array(data=1:100, dim=c(10, 10))
    #      [,1] [,2] [,3] [,4] [,5] [,6] [,7] [,8] [,9] [,10]
    # [1,]    1    3    5    7    9   11   13   15   17    19
    # [2,]    2    4    6    8   10   12   14   16   18    20
    z <- zarr_create_array(data = a, shape=dim(a), chunks=c(5, 5), dtype="<f4", fill_value=NA)

    expect_equal(z$get_shape(), c(10, 10))
    expect_equal(z$get_chunks(), c(5, 5))

    sel <- z$get_item(list(slice(1, 10), slice(1, 10)))
    expect_equal(a, sel$data)
})

test_that("can get array that spans multiple chunks with ellipsis", {
    a <- array(data=1:100, dim=c(10, 10))
    #      [,1] [,2] [,3] [,4] [,5] [,6] [,7] [,8] [,9] [,10]
    # [1,]    1    3    5    7    9   11   13   15   17    19
    # [2,]    2    4    6    8   10   12   14   16   18    20
    z <- zarr_create_array(data = a, shape=dim(a), chunks=c(5, 5), dtype="<f4", fill_value=NA)

    expect_equal(z$get_shape(), c(10, 10))
    expect_equal(z$get_chunks(), c(5, 5))

    sel <- z$get_item("...")
    expect_equal(a, sel$data)
})

# =============================================================================
# Gap 2: ZarrArray chunk I/O — partial writes, fill values, error paths
# =============================================================================

test_that("chunk_setitem: partial chunk write reads existing data first", {
  # Write full array, then update a subset — exercises partial chunk path
  z <- zarr_create(shape = c(10L), chunks = c(5L), dtype = "<i4",
                   fill_value = 0L)
  z$set_item("...", array(1:10, dim = 10))
  # Partial write to first chunk
  z$set_item(list(slice(1, 3)), array(c(99L, 98L, 97L), dim = 3))
  result <- z$get_item("...")
  expect_equal(as.integer(result$data), c(99L, 98L, 97L, 4L, 5L, 6L, 7L, 8L, 9L, 10L))
})

test_that("chunk_getitem: missing chunk returns fill value", {
  z <- zarr_create(shape = c(10L), chunks = c(5L), dtype = "<f8",
                   fill_value = -999.0)
  # Don't write any data — chunks are missing from store
  result <- z$get_item("...")
  expect_equal(as.double(result$data), rep(-999.0, 10))
})

test_that("chunk_setitem: scalar total-replace creates full chunk from scalar", {
  z <- zarr_create(shape = c(6L), chunks = c(3L), dtype = "<i4",
                   fill_value = 0L)
  z$set_item("...", 42L)
  result <- z$get_item("...")
  expect_equal(as.integer(result$data), rep(42L, 6))
})

test_that("chunk I/O round-trip with multiple chunks (2D)", {
  a <- array(as.double(1:24), dim = c(4, 6))
  z <- zarr_create_array(data = a, shape = dim(a), chunks = c(2, 3),
                         dtype = "<f8")
  result <- z$get_item("...")
  expect_equal(result$data, a)
})

test_that("chunk I/O round-trip with non-aligned chunks (3D)", {
  a <- array(as.double(1:60), dim = c(3, 4, 5))
  z <- zarr_create_array(data = a, shape = dim(a), chunks = c(2, 2, 3),
                         dtype = "<f8")
  result <- z$get_item("...")
  expect_equal(result$data, a)
})

test_that("chunk_setitem: partial write to uninitialized chunk uses fill_value", {
  z <- zarr_create(shape = c(10L), chunks = c(5L), dtype = "<i4",
                   fill_value = 0L)
  z$set_item(list(slice(1, 2)), array(c(100L, 200L), dim = 2))
  result <- z$get_item(list(slice(1, 5)))
  expect_equal(as.integer(result$data), c(100L, 200L, 0L, 0L, 0L))
})

# =============================================================================
# End chunk I/O tests
# =============================================================================

test_that("resize smaller deletes out-of-range chunks from DirectoryStore", {
  tmp <- tempfile()
  dir.create(tmp)
  on.exit(unlink(tmp, recursive = TRUE))
  store <- DirectoryStore$new(tmp)
  z <- zarr_create(shape = c(10L), chunks = c(5L), dtype = "<i4",
                   fill_value = 0L, store = store)
  z$set_item("...", array(1:10, dim = 10))
  # Two chunks: 0, 1
  expect_true(store$contains_item("0"))
  expect_true(store$contains_item("1"))
  z$resize(5L)
  # Chunk 1 should be deleted after resize
  expect_true(store$contains_item("0"))
  expect_false(store$contains_item("1"))
})

test_that("resize smaller deletes out-of-range chunks from MemoryStore", {
  store <- MemoryStore$new()
  z <- zarr_create(shape = c(10L), chunks = c(5L), dtype = "<i4",
                   fill_value = 0L, store = store)
  z$set_item("...", array(1:10, dim = 10))
  z$resize(5L)
  result <- z$get_item("...")
  expect_equal(length(result$data), 5)
  expect_equal(as.integer(result$data), 1:5)
})

test_that("can get array where shape is not a multiple of chunk size", {
    a <- array(data=1:100, dim=c(10, 10))
    #      [,1] [,2] [,3] [,4] [,5] [,6] [,7] [,8] [,9] [,10]
    # [1,]    1    3    5    7    9   11   13   15   17    19
    # [2,]    2    4    6    8   10   12   14   16   18    20
    z <- zarr_create_array(data = a, shape=dim(a), chunks=c(3, 3), dtype="<f4", fill_value=NA)

    expect_equal(z$get_shape(), c(10, 10))
    expect_equal(z$get_chunks(), c(3, 3))

    sel <- z$get_item("...")
    expect_equal(a, sel$data)
})

library(pizzarr)

MockArray <- R6::R6Class("MockArray",
    public = list(
        shape = NULL,
        chunks = NULL,
        initialize = function(shape, chunks) {
            self$shape <- shape
            self$chunks <- chunks
        },
        get_shape = function() {
            return(self$shape)
        },
        get_chunks = function() {
            return(self$chunks)
        }
    )
)

test_that("basic indexer for array that spans multiple chunks", {
    z <- MockArray$new(shape = c(10, 10), chunks = c(5, 5))

    expect_equal(z$get_shape(), c(10, 10))
    expect_equal(z$get_chunks(), c(5, 5))

    bi <- BasicIndexer$new(list(zb_slice(0, 11), zb_slice(0, 11)), z)

    expect_equal(as.numeric(bi$shape), c(10, 10))

    expect_equal(length(bi$dim_indexers), 2)

    expect_equal("SliceDimIndexer" %in% class(bi$dim_indexers[[1]]), TRUE)
    expect_equal("SliceDimIndexer" %in% class(bi$dim_indexers[[2]]), TRUE)
    
    sdi1 <- bi$dim_indexers[[1]]
    expect_equal(sdi1$start, 0)
    expect_equal(sdi1$stop, 10)
    expect_equal(sdi1$step, 1)
    expect_equal(sdi1$dim_len, 10)
    expect_equal(sdi1$dim_chunk_len, 5)
    expect_equal(sdi1$num_items, 10)
    expect_equal(sdi1$num_chunks, 2)

    sdi2 <- bi$dim_indexers[[2]]
    expect_equal(sdi2$start, 0)
    expect_equal(sdi2$stop, 10)
    expect_equal(sdi2$step, 1)
    expect_equal(sdi2$dim_len, 10)
    expect_equal(sdi2$dim_chunk_len, 5)
    expect_equal(sdi2$num_items, 10)
    expect_equal(sdi2$num_chunks, 2)
})

test_that("basic indexer for array that spans multiple chunks where shape is not a multiple", {
    z <- MockArray$new(shape = c(10, 10), chunks = c(3, 3))

    expect_equal(z$get_shape(), c(10, 10))
    expect_equal(z$get_chunks(), c(3, 3))

    bi <- BasicIndexer$new(list(zb_slice(0, 11), zb_slice(0, 11)), z)

    expect_equal(as.numeric(bi$shape), c(10, 10))

    expect_equal(length(bi$dim_indexers), 2)

    expect_equal("SliceDimIndexer" %in% class(bi$dim_indexers[[1]]), TRUE)
    expect_equal("SliceDimIndexer" %in% class(bi$dim_indexers[[2]]), TRUE)
    
    sdi1 <- bi$dim_indexers[[1]]
    expect_equal(sdi1$start, 0)
    expect_equal(sdi1$stop, 10)
    expect_equal(sdi1$step, 1)
    expect_equal(sdi1$dim_len, 10)
    expect_equal(sdi1$dim_chunk_len, 3)
    expect_equal(sdi1$num_items, 10)
    expect_equal(sdi1$num_chunks, 4)

    sdi2 <- bi$dim_indexers[[2]]
    expect_equal(sdi2$start, 0)
    expect_equal(sdi2$stop, 10)
    expect_equal(sdi2$step, 1)
    expect_equal(sdi2$dim_len, 10)
    expect_equal(sdi2$dim_chunk_len, 3)
    expect_equal(sdi2$num_items, 10)
    expect_equal(sdi2$num_chunks, 4)
})

# =============================================================================
# Gap 9: IntDimIndexer — single-integer dimension indexing
# =============================================================================

test_that("IntDimIndexer initializes correctly", {
  idi <- IntDimIndexer$new(dim_sel = 3, dim_len = 10, dim_chunk_len = 5)
  expect_equal(idi$dim_sel, 3)
  expect_equal(idi$dim_len, 10)
  expect_equal(idi$dim_chunk_len, 5)
  expect_equal(idi$num_items, 1)
})

test_that("IntDimIndexer iter returns correct ChunkDimProjection", {
  # Index 3 in a dim of length 10 with chunk_len 5 -> chunk 0, offset 3
  idi <- IntDimIndexer$new(dim_sel = 3, dim_len = 10, dim_chunk_len = 5)
  result <- idi$iter()
  expect_equal(length(result), 1)
  proj <- result[[1]]
  expect_equal(proj$dim_chunk_index, 0)  # chunk 0
  expect_equal(proj$dim_chunk_sel, 3)     # offset within chunk
  expect_true(is.na(proj$dim_out_sel))
})

test_that("IntDimIndexer iter for index in second chunk", {
  # Index 7 in a dim of length 10 with chunk_len 5 -> chunk 1, offset 2
  idi <- IntDimIndexer$new(dim_sel = 7, dim_len = 10, dim_chunk_len = 5)
  result <- idi$iter()
  proj <- result[[1]]
  expect_equal(proj$dim_chunk_index, 1)  # chunk 1
  expect_equal(proj$dim_chunk_sel, 2)     # 7 - 5 = 2
})

test_that("IntDimIndexer iter for first element", {
  idi <- IntDimIndexer$new(dim_sel = 0, dim_len = 10, dim_chunk_len = 5)
  result <- idi$iter()
  proj <- result[[1]]
  expect_equal(proj$dim_chunk_index, 0)
  expect_equal(proj$dim_chunk_sel, 0)
})

test_that("IntDimIndexer iter for last element in chunk boundary", {
  # Index 4 (last in first chunk of size 5) -> chunk 0, offset 4
  idi <- IntDimIndexer$new(dim_sel = 4, dim_len = 10, dim_chunk_len = 5)
  result <- idi$iter()
  proj <- result[[1]]
  expect_equal(proj$dim_chunk_index, 0)
  expect_equal(proj$dim_chunk_sel, 4)
})

test_that("IntDimIndexer iter for first element of second chunk", {
  # Index 5 (first in second chunk of size 5) -> chunk 1, offset 0
  idi <- IntDimIndexer$new(dim_sel = 5, dim_len = 10, dim_chunk_len = 5)
  result <- idi$iter()
  proj <- result[[1]]
  expect_equal(proj$dim_chunk_index, 1)
  expect_equal(proj$dim_chunk_sel, 0)
})

test_that("BasicIndexer creates IntDimIndexer for integer selection", {
  z <- MockArray$new(shape = c(10, 10), chunks = c(5, 5))
  # Pass convert_integer_selection_to_slices=FALSE via normalize_list_selection
  # BasicIndexer uses normalize_list_selection which converts integers to slices by default.
  # The IntDimIndexer path is triggered when convert_integer_selection_to_slices is FALSE
  # or when selections come from non-list contexts.
  # Test that it works through BasicIndexer indirectly via single-integer selection
  bi <- BasicIndexer$new(list(zb_slice(0, 10), zb_slice(0, 10)), z)
  expect_equal(length(bi$dim_indexers), 2)
})

# =============================================================================
# End IntDimIndexer tests
# =============================================================================

test_that("basic indexer for array that spans multiple chunks where shape is not a multiple with offsets", {
    z <- MockArray$new(shape = c(10, 10), chunks = c(3, 3))

    expect_equal(z$get_shape(), c(10, 10))
    expect_equal(z$get_chunks(), c(3, 3))

    bi <- BasicIndexer$new(list(zb_slice(5, 10), zb_slice(5, 10)), z)

    expect_equal(as.numeric(bi$shape), c(5, 5))
    
    expect_equal(length(bi$dim_indexers), 2)

    expect_equal("SliceDimIndexer" %in% class(bi$dim_indexers[[1]]), TRUE)
    expect_equal("SliceDimIndexer" %in% class(bi$dim_indexers[[2]]), TRUE)
    
    sdi1 <- bi$dim_indexers[[1]]
    expect_equal(sdi1$start, 5)
    expect_equal(sdi1$stop, 10)
    expect_equal(sdi1$step, 1)
    expect_equal(sdi1$dim_len, 10)
    expect_equal(sdi1$dim_chunk_len, 3)
    expect_equal(sdi1$num_items, 5)
    expect_equal(sdi1$num_chunks, 4)

    sdi2 <- bi$dim_indexers[[2]]
    expect_equal(sdi2$start, 5)
    expect_equal(sdi2$stop, 10)
    expect_equal(sdi2$step, 1)
    expect_equal(sdi2$dim_len, 10)
    expect_equal(sdi2$dim_chunk_len, 3)
    expect_equal(sdi2$num_items, 5)
    expect_equal(sdi2$num_chunks, 4)
})

# --- zarrs HTTP store tests ---
# These tests require network access and the http_sync compiled feature.
# Skipped on CRAN and when offline.

test_that("http_sync feature is compiled", {
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")
  features <- pizzarr_compiled_features()
  expect_true("http_sync" %in% features)
})

test_that("zarrs reads from HTTPS store", {
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")
  skip_if(!("http_sync" %in% pizzarr_compiled_features()),
          "http_sync feature not compiled")
  skip_on_cran()
  skip_if_offline()

  url <- "https://raw.githubusercontent.com/DOI-USGS/rnz/main/inst/extdata/bcsd.zarr"

  # Open array metadata via zarrs
  meta <- zarrs_open_array_metadata(url, "pr")
  expect_equal(meta$zarr_format, 2L)
  expect_true(length(meta$shape) > 0)

  # Read a small subset: first element along each dimension
  ndim <- length(meta$shape)
  ranges <- lapply(seq_len(ndim), function(i) c(0L, 1L))
  result <- zarrs_get_subset(url, "pr", ranges, NULL)
  expect_true(length(result$data) == 1)
  expect_equal(length(result$shape), ndim)

  zarrs_close_store(url)
})

test_that("zarrs HTTP read matches R-native read", {
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")
  skip_if(!("http_sync" %in% pizzarr_compiled_features()),
          "http_sync feature not compiled")
  skip_on_cran()
  skip_if_offline()
  skip_if_not_installed("crul")

  url <- "https://raw.githubusercontent.com/DOI-USGS/rnz/main/inst/extdata/bcsd.zarr"

  # Read via zarrs
  z_zarrs <- zarr_open(store = HttpStore$new(url))
  arr_zarrs <- z_zarrs$get_item("pr")

  # Read via R-native
  old <- .pizzarr_env$zarrs_available
  .pizzarr_env$zarrs_available <- FALSE
  z_native <- zarr_open(store = HttpStore$new(url))
  arr_native <- z_native$get_item("pr")
  .pizzarr_env$zarrs_available <- old

  expect_equal(arr_zarrs$data, arr_native$data)

  zarrs_close_store(url)
})

test_that("can_use_zarrs_write returns FALSE for HttpStore", {
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")
  skip_if_not_installed("crul")

  hs <- HttpStore$new("https://example.com/fake.zarr")
  # Create a minimal BasicIndexer-like object for the check
  mock_indexer <- list(dim_indexers = list(
    structure(list(step = 1L), class = "SliceDimIndexer")
  ))
  class(mock_indexer) <- "BasicIndexer"

  expect_false(can_use_zarrs_write(mock_indexer, hs))
})

library(pizzarr)

sample_dir <- tools::R_user_dir("pizzarr")
clean <- !dir.exists(sample_dir)

test_that("get_basic_selection_zd", {
    # Reference: https://github.com/zarr-developers/zarr-python/blob/5dd4a0e6cdc04c6413e14f57f61d389972ea937c/zarr/tests/test_indexing.py#L70
    a <- as_scalar(42)
    z <- zarr_create(shape=dim(a), dtype="<f4", fill_value=NA)

    expect_equal(z$get_shape(), NULL)

    z$set_item("...", a)

    sel <- z$get_item("...")

    expect_equal(as.numeric(a), as.numeric(sel$data))
})

test_that("get_basic_selection_zd with anndataR IntScalar fixture", {
    root <- pizzarr_sample(file.path("fixtures", "v2", "example.zarr"))
    
    store <- DirectoryStore$new(root)
    z <- zarr_open_array(store, path = "uns/IntScalar")

    expect_equal(z$get_shape(), integer(0))

    sel <- z$get_item("...")
    scalar <- as.numeric(sel$data)

    expect_equal(scalar, 1)
})

test_that("get_basic_selection_zd with anndataR StringScalar fixture", {
    root <- pizzarr_sample(file.path("fixtures", "v2", "example.zarr"))
    
    store <- DirectoryStore$new(root)
    z <- zarr_open_array(store, path = "uns/StringScalar")

    expect_equal(z$get_shape(), integer(0))

    sel <- z$get_item("...")
    scalar <- as.character(sel$data)

    expect_equal(scalar, "A string")
})

test_that("get_basic_selection_1d", {
    # Reference: https://github.com/zarr-developers/zarr-python/blob/5dd4a0e6cdc04c6413e14f57f61d389972ea937c/zarr/tests/test_indexing.py#L70
    a <- array(data=42)
    z <- zarr_create(shape=dim(a), dtype="<f4", fill_value=NA)

    expect_equal(z$get_shape(), 1)

    z$set_item("...", a)

    sel <- z$get_item("...")

    expect_equal(a, sel$data)
})

test_that("get_basic_selection_2d - can set_item with array", {
    a <- array(data=1:20, dim=c(2, 10))
    #      [,1] [,2] [,3] [,4] [,5] [,6] [,7] [,8] [,9] [,10]
    # [1,]    1    3    5    7    9   11   13   15   17    19
    # [2,]    2    4    6    8   10   12   14   16   18    20
    z <- zarr_create(shape=dim(a), dtype="<f4", fill_value=NA)

    expect_equal(z$get_shape(), c(2, 10))
    expect_equal(z$get_chunks(), c(2, 10))

    z$set_item("...", a)

    sel <- z$get_item("...")

    expect_equal(a, sel$data)
})

test_that("get_basic_selection_2d - can set_item with NestedArray", {
    a <- array(data=1:20, dim=c(2, 10))
    #      [,1] [,2] [,3] [,4] [,5] [,6] [,7] [,8] [,9] [,10]
    # [1,]    1    3    5    7    9   11   13   15   17    19
    # [2,]    2    4    6    8   10   12   14   16   18    20
    z <- zarr_create(shape=dim(a), dtype="<f4", fill_value=NA)

    expect_equal(z$get_shape(), c(2, 10))
    expect_equal(z$get_chunks(), c(2, 10))

    z$set_item("...", NestedArray$new(data = a, shape = dim(a)))

    sel <- z$get_item("...")

    expect_equal(a, sel$data)
})

test_that("get_basic_selection_2d(zero-based) - can set_item for subset", {
    z <- zarr_create(shape=c(2, 10), dtype="<f4", fill_value=NA)

    expect_equal(z$get_shape(), c(2, 10))
    expect_equal(z$get_chunks(), c(2, 10))

    a <- array(data=1:10, dim=c(2, 5))
    #      [,0] [,1] [,2] [,3] [,4]
    # [0,]    1    3    5    7    9
    # [1,]    2    4    6    8   10

    z$set_item(list(zb_slice(0, 2), zb_slice(0, 5)), a)

    sel <- z$get_item("...")

    # fill_value=NA normalizes to 0.0 for float dtype
    expected_out <- array(data=0, dim=c(2, 10))
    expected_out[1, 1:5] <- c(1, 3, 5, 7, 9)
    expected_out[2, 1:5] <- c(2, 4, 6, 8, 10)

    expect_equal(expected_out, sel$data)
})

test_that("get_basic_selection_2d(zero-based) - can get_item for subset", {
    a <- array(data=1:20, dim=c(2, 10))
    #      [,1] [,2] [,3] [,4] [,5] [,6] [,7] [,8] [,9] [,10]
    # [1,]    1    3    5    7    9   11   13   15   17    19
    # [2,]    2    4    6    8   10   12   14   16   18    20
    z <- zarr_create(shape=dim(a), dtype="<f4", fill_value=NA)

    expect_equal(z$get_shape(), c(2, 10))
    expect_equal(z$get_chunks(), c(2, 10))

    z$set_item("...", a)

    sel <- z$get_item(list(zb_slice(0, 2), zb_slice(0, 5)))

    expected_out <- array(data=1:10, dim=c(2, 5))
    expect_equal(expected_out, sel$data)
})

test_that("get_basic_selection_2d(zero-based) - can get_item for subset with one offset on first axis", {
    a <- array(data=1:20, dim=c(2, 10))
    #      [,1] [,2] [,3] [,4] [,5] [,6] [,7] [,8] [,9] [,10]
    # [1,]    1    3    5    7    9   11   13   15   17    19
    # [2,]    2    4    6    8   10   12   14   16   18    20
    z <- zarr_create(shape=dim(a), dtype="<f4", fill_value=NA)

    expect_equal(z$get_shape(), c(2, 10))
    expect_equal(z$get_chunks(), c(2, 10))

    z$set_item("...", a)

    sel <- z$get_item(list(zb_slice(1, 3), zb_slice(0, 5)))

    expected_out <- array(data=NA, dim=c(1, 5))
    expected_out[1,] <- c(2, 4, 6, 8, 10)
    expect_equal(expected_out, sel$data)
})

test_that("get_basic_selection_2d(zero-based) - can get_item for subset with one offset on both axes", {
    a <- array(data=1:20, dim=c(2, 10))
    #      [,1] [,2] [,3] [,4] [,5] [,6] [,7] [,8] [,9] [,10]
    # [1,]    1    3    5    7    9   11   13   15   17    19
    # [2,]    2    4    6    8   10   12   14   16   18    20
    z <- zarr_create(shape=dim(a), dtype="<f4", fill_value=NA)

    expect_equal(z$get_shape(), c(2, 10))
    expect_equal(z$get_chunks(), c(2, 10))

    z$set_item("...", a)

    sel <- z$get_item(list(zb_slice(1, 2), zb_slice(1, 5)))

    expected_out <- array(data=NA, dim=c(1, 4))
    expected_out[1,] <- c(4, 6, 8, 10)
    expect_equal(expected_out, sel$data)
})

test_that("get_basic_selection_2d(one-based) - can get_item for subset with same index for start/stop", {
    a <- array(data=1:20, dim=c(2, 10))
    #      [,1] [,2] [,3] [,4] [,5] [,6] [,7] [,8] [,9] [,10]
    # [1,]    1    3    5    7    9   11   13   15   17    19
    # [2,]    2    4    6    8   10   12   14   16   18    20
    z <- zarr_create(shape=dim(a), dtype="<f4", fill_value=NA)

    expect_equal(z$get_shape(), c(2, 10))
    expect_equal(z$get_chunks(), c(2, 10))

    z$set_item("...", a)

    sel <- z$get_item(list(slice(2, 2), slice(2, 5)))

    expected_out <- array(data=NA, dim=c(1, 4))
    expected_out[1,] <- c(4, 6, 8, 10)
    expect_equal(expected_out, sel$data)
})

test_that("get_basic_selection_2d(one-based) - can get_item for subset with one offset on both axes", {
    a <- array(data=1:20, dim=c(2, 10))
    #      [,1] [,2] [,3] [,4] [,5] [,6] [,7] [,8] [,9] [,10]
    # [1,]    1    3    5    7    9   11   13   15   17    19
    # [2,]    2    4    6    8   10   12   14   16   18    20
    z <- zarr_create(shape=dim(a), dtype="<f4", fill_value=NA)

    expect_equal(z$get_shape(), c(2, 10))
    expect_equal(z$get_chunks(), c(2, 10))

    z$set_item("...", a)

    sel <- z$get_item(list(slice(1, 2), slice(1, 5)))

    expected_out <- array(data=1:10, dim=c(2, 5))
    expect_equal(expected_out, sel$data)
})

test_that("get_basic_selection_3d(one-based) - can get_item for three-dimensional array", {
    a <- array(data=1:24, dim=c(2, 3, 4))
    # , , 1
    #
    #      [,1] [,2] [,3]
    # [1,]    1    3    5
    # [2,]    2    4    6
    #
    # , , 2
    #
    #      [,1] [,2] [,3]
    # [1,]    7    9   11
    # [2,]    8   10   12
    #
    # , , 3
    #
    #      [,1] [,2] [,3]
    # [1,]   13   15   17
    # [2,]   14   16   18
    #
    # , , 4
    #
    #      [,1] [,2] [,3]
    # [1,]   19   21   23
    # [2,]   20   22   24
    z <- zarr_create(shape=dim(a), dtype="<f4", fill_value=NA)

    expect_equal(z$get_shape(), c(2, 3, 4))
    expect_equal(z$get_chunks(), c(2, 3, 4))

    z$set_item("...", a)

    sel <- z$get_item(list(slice(1, 2), slice(1, 2), slice(1, 2)))

    expected_out <- array(data=c(1, 2, 3, 4, 7, 8, 9, 10), dim=c(2, 2, 2))
    # , , 1
    #
    #      [,1] [,2]
    # [1,]    1    3
    # [2,]    2    4
    #
    # , , 2
    #
    #      [,1] [,2]
    # [1,]    7    9
    # [2,]    8   10
    expect_equal(expected_out, sel$data)
})

test_that("Can read 2D string array", {
    skip_if_not_installed("blosc")
    root <- pizzarr_sample(file.path("fixtures", "v2", "example.zarr"))
    store <- DirectoryStore$new(root)
    z <- zarr_open_array(store, path = "uns/String2D")
    nested_arr <- z$get_item("...")
    data <- nested_arr$data

    row1 <- c("row0col0", "row1col0", "row2col0", "row3col0", "row4col0", "row5col0", "row6col0", "row7col0", "row8col0", "row9col0")
    row2 <- c("row0col1", "row1col1", "row2col1", "row3col1", "row4col1", "row5col1", "row6col1", "row7col1", "row8col1", "row9col1")
    row3 <- c("row0col2", "row1col2", "row2col2", "row3col2", "row4col2", "row5col2", "row6col2", "row7col2", "row8col2", "row9col2")
    row4 <- c("row0col3", "row1col3", "row2col3", "row3col3", "row4col3", "row5col3", "row6col3", "row7col3", "row8col3", "row9col3")
    row5 <- c("row0col4", "row1col4", "row2col4", "row3col4", "row4col4", "row5col4", "row6col4", "row7col4", "row8col4", "row9col4")

    expected_arr <- t(array(data = c(row1, row2, row3, row4, row5), dim = c(10, 5)))
    expect_equal(expected_arr, data)
})

if(clean) unlink(sample_dir, recursive = TRUE)

# --- Zero-dim get/set ---

test_that("get_basic_selection on zero-dim array returns result for uninitialized chunk", {
  store <- MemoryStore$new()
  zarray_meta <- store$metadata_class$encode_array_metadata(create_zarray_meta(
    dtype = Dtype$new("<f8"),
    order = "C",
    fill_value = 9.5,
    shape = NULL,
    chunks = c(1),
    dimension_separator = "."
  ))
  store$set_item(".zarray", zarray_meta)
  a <- ZarrArray$new(store = store)
  result <- a$get_basic_selection("...")
  expect_true(!is.null(result))
})

test_that("set_basic_selection_zd writes and reads scalar value", {
  store <- MemoryStore$new()
  zarray_meta <- store$metadata_class$encode_array_metadata(create_zarray_meta(
    dtype = Dtype$new("<f8"),
    order = "C",
    fill_value = 0,
    shape = NULL,
    chunks = c(1),
    dimension_separator = "."
  ))
  store$set_item(".zarray", zarray_meta)
  a <- ZarrArray$new(store = store)
  a$set_basic_selection(list(), 3.14)
  result <- a$get_basic_selection("...")
  expect_true(!is.null(result))
})

# --- get_selection edge cases ---

test_that("get_selection returns correct length for full 1D slice", {
  a <- zarr_create(shape = c(2), chunks = c(2), store = MemoryStore$new(),
                   fill_value = 42, compressor = NA)
  result <- a$get_item("...")
  expect_equal(length(result$data), 2)
})

test_that("get_selection returns empty NestedArray when selection is empty", {
  a <- zarr_create(
    shape = c(4, 4), chunks = c(2, 2), store = MemoryStore$new(), compressor = NA
  )
  result <- a$get_item(list(slice(1, 0), slice(1, 2)))
  expect_equal(length(result$data), 0)
})

test_that("get_item raises error for vectorized (fancy) indexing", {
  a <- zarr_create(
    shape = c(4, 4), chunks = c(2, 2), store = MemoryStore$new(), compressor = NA
  )
  expect_error(a$get_item(list(c(1L, 2L), c(1L, 2L))), "vectorized indexing")
})

# --- get_basic_selection with cache_metadata=FALSE ---

test_that("get_basic_selection with cache_metadata=FALSE works", {
  d <- tempfile()
  zarr_create(shape = c(4), chunks = c(2), store = d, compressor = NA)
  b <- ZarrArray$new(store = DirectoryStore$new(d), cache_metadata = FALSE)
  result <- b$get_basic_selection("...")
  expect_equal(length(result$data), 4)
})

# --- get_orthogonal_selection ---

test_that("get_orthogonal_selection returns subset", {
  a <- zarr_create(shape = c(4, 4), chunks = c(2, 2), store = MemoryStore$new(), compressor = NA)
  data <- array(1:16, dim = c(4, 4))
  a$set_item("...", data)
  result <- a$get_orthogonal_selection(list(slice(1, 2), slice(1, 2)))
  expect_equal(dim(result$data), c(2, 2))
})

test_that("get_orthogonal_selection works with cache_metadata=FALSE", {
  d <- tempfile()
  a <- zarr_create(shape = c(4, 4), chunks = c(2, 2), store = d, compressor = NA)
  a$set_item("...", array(1:16, dim = c(4, 4)))
  b <- ZarrArray$new(store = DirectoryStore$new(d), cache_metadata = FALSE)
  result <- b$get_orthogonal_selection(list(slice(1, 2), slice(1, 2)))
  expect_equal(dim(result$data), c(2, 2))
})

# --- chunk_getitem paths ---

test_that("chunk_getitem_part2 partial-select path: sub-chunk slice read", {
  # Use a single large chunk so the whole-chunk fast path (is_total_slice) does NOT fire.
  a <- zarr_create(
    shape = c(4, 4), chunks = c(4, 4), store = MemoryStore$new(),
    compressor = NA, fill_value = 0
  )
  data <- array(1:16, dim = c(4, 4))
  a$set_item("...", data)
  result <- a$get_item(list(slice(1, 2), slice(1, 2)))
  expect_equal(dim(result$data), c(2, 2))
  expect_equal(as.vector(result$data), c(1L, 2L, 5L, 6L))
})

test_that("chunk_getitem_part2 error path: partial read of uninitialized chunk returns fill", {
  a <- zarr_create(
    shape = c(4, 4), chunks = c(4, 4), store = MemoryStore$new(),
    compressor = NA, fill_value = 7
  )
  result <- a$get_item(list(slice(1, 2), slice(1, 2)))
  expect_true(all(result$data == 7))
})

library(pizzarr)

test_that("can convert a NestedArray to a raw vector with lz4 compression", {
    skip_if_not_installed("qs")
    a <- array(data=1:10, dim=c(2, 5))
    na <- NestedArray$new(data = a, shape = dim(a), dtype = "<u4")

    # Get un-compressed raw vector.
    na_raw <- na$flatten_to_raw()

    expect_equal(length(na_raw), 40)
    expected_out <- as.raw(c(
        0x01, 0x00, 0x00, 0x00,
        0x02, 0x00, 0x00, 0x00,
        0x03, 0x00, 0x00, 0x00,
        0x04, 0x00, 0x00, 0x00,
        0x05, 0x00, 0x00, 0x00,
        0x06, 0x00, 0x00, 0x00,
        0x07, 0x00, 0x00, 0x00,
        0x08, 0x00, 0x00, 0x00,
        0x09, 0x00, 0x00, 0x00,
        0x0a, 0x00, 0x00, 0x00
    ))
    expect_equal(na_raw, expected_out)

    # Compress the raw vector.
    compressor <- suppressWarnings(Lz4Codec$new(acceleration = 1))

    compressed_raw <- compressor$encode(na_raw)

    # Un-compress the compressed vector.
    uncompressed_raw <- compressor$decode(compressed_raw)

    expect_equal(uncompressed_raw, expected_out)
})

test_that("can convert a NestedArray to a raw vector with zstd compression", {
    a <- array(data=1:10, dim=c(2, 5))
    na <- NestedArray$new(data = a, shape = dim(a), dtype = "<u4")

    # Get un-compressed raw vector.
    na_raw <- na$flatten_to_raw()

    expect_equal(length(na_raw), 40)
    expected_out <- as.raw(c(
        0x01, 0x00, 0x00, 0x00,
        0x02, 0x00, 0x00, 0x00,
        0x03, 0x00, 0x00, 0x00,
        0x04, 0x00, 0x00, 0x00,
        0x05, 0x00, 0x00, 0x00,
        0x06, 0x00, 0x00, 0x00,
        0x07, 0x00, 0x00, 0x00,
        0x08, 0x00, 0x00, 0x00,
        0x09, 0x00, 0x00, 0x00,
        0x0a, 0x00, 0x00, 0x00
    ))
    expect_equal(na_raw, expected_out)

    # Compress the raw vector.
    compressor <- ZstdCodec$new(level = 1)

    compressed_raw <- compressor$encode(na_raw)

    # Un-compress the compressed vector.
    uncompressed_raw <- compressor$decode(compressed_raw)

    expect_equal(uncompressed_raw, expected_out)
})

test_that("zarr_create with non-default compressor", {
    skip_if_not_installed("qs")

    compressor <- suppressWarnings(Lz4Codec$new(acceleration = 1))

    a <- array(data=1:20, dim=c(2, 10))
    #      [,1] [,2] [,3] [,4] [,5] [,6] [,7] [,8] [,9] [,10]
    # [1,]    1    3    5    7    9   11   13   15   17    19
    # [2,]    2    4    6    8   10   12   14   16   18    20
    z <- zarr_create(shape=dim(a), dtype="<f4", fill_value=NA, compressor = compressor)

    expect_equal(z$get_shape(), c(2, 10))
    expect_equal(z$get_chunks(), c(2, 10))

    expect_equal(as.character(z$get_compressor()$get_config()$id), "lz4")

    z$set_item("...", a)

    sel <- z$get_item(list(slice(1, 2), slice(1, 5)))

    expected_out <- array(data=1:10, dim=c(2, 5))
    expect_equal(expected_out, sel$data)
})

library(pizzarr)

test_that("1D append works", {
  z <- zarr_create(3, chunks = 3, dtype = "<f4")
  z$set_item("...", array(c(1, 2, 3)))

  z$append(array(c(4, 5)))
  expect_equal(z$get_shape(), c(5))
  expect_equal(as.numeric(z$get_item("...")$data), c(1, 2, 3, 4, 5))
})

test_that("2D append along axis 1 (rows)", {
  z <- zarr_create(c(2, 3), chunks = c(2, 3), dtype = "<f4")
  original <- array(1:6, dim = c(2, 3))
  z$set_item("...", original)

  new_row <- array(7:9, dim = c(1, 3))
  z$append(new_row, axis = 1)
  expect_equal(z$get_shape(), c(3, 3))

  result <- z$get_item("...")$data
  expect_equal(as.numeric(result[1:2, ]), as.numeric(original))
  expect_equal(as.numeric(result[3, ]), c(7, 8, 9))
})

test_that("append errors on dimension mismatch", {
  z <- zarr_create(c(2, 3), chunks = c(2, 3), dtype = "<f4")
  z$set_item("...", array(1:6, dim = c(2, 3)))

  expect_error(z$append(array(1:3)), "incompatible number of dimensions")
})

test_that("append errors on shape mismatch on non-append axis", {
  z <- zarr_create(c(2, 3), chunks = c(2, 3), dtype = "<f4")
  z$set_item("...", array(1:6, dim = c(2, 3)))

  bad_data <- array(1:4, dim = c(1, 4))
  expect_error(z$append(bad_data, axis = 1), "shape mismatch")
})

test_that("append converts non-array to array automatically", {
  a <- zarr_create(
    shape = c(4), chunks = c(2), store = MemoryStore$new(),
    compressor = NA, fill_value = 0
  )
  a$set_item("...", array(1:4, dim = c(4)))
  # Pass a plain vector, not an array — triggers as.array() path
  a$append(5:8, axis = 1)
  expect_equal(a$get_shape(), c(8))
  result <- a$get_item("...")
  expect_equal(as.vector(result$data), 1:8)
})

test_that("append raises error for wrong number of dimensions", {
  a <- zarr_create(
    shape = c(4, 2), chunks = c(2, 2), store = MemoryStore$new(), compressor = NA
  )
  expect_error(a$append(array(1:4, dim = c(4)), axis = 1), "dimensions")
})

test_that("append raises error for non-append axis shape mismatch", {
  a <- zarr_create(
    shape = c(4, 2), chunks = c(2, 2), store = MemoryStore$new(), compressor = NA
  )
  expect_error(a$append(array(1:9, dim = c(3, 3)), axis = 1), "shape mismatch")
})

if (!requireNamespace("vcr", quietly = TRUE) ||
    !requireNamespace("crul", quietly = TRUE)) {
  return(invisible())
}

vcr::use_cassette("http_base", {

  test_that("http store", {
    skip_if_not_installed("blosc")

    url <- "https://uk1s3.embassy.ebi.ac.uk/idr/zarr/v0.4/idr0062A/6001240.zarr"

    z <- pizzarr::HttpStore$new(url)
    g <- zarr_open_group(z)
    expect_equal(class(g), c("ZarrGroup", "R6"))

    attrs <- g$get_attrs()$to_list()

    expect_equal(names(attrs), c("_creator", "multiscales", "omero"))

    resolution_paths <- attrs$multiscales[[1]]$datasets[[1]]$path
    first_resolution <- resolution_paths[[1]]

    zarr_arr <- g$get_item(first_resolution)

    expect_equal(zarr_arr$get_shape(),
                 c(2L, 236L, 275L, 271L))

    expect_message(listdir_output <- z$listdir(), ".zmetadata not found for this http store. Can't listdir")

    expect_null(listdir_output)

    z_index <- 118

    nested_arr <- zarr_arr$get_item(list(slice(1, 2),
                                         slice(z_index, z_index),
                                         slice(1, 10),
                                         slice(1, 10)))

    expect_equal(nested_arr$shape,
                 c(2L, 1L, 10L, 10L))

    arr <- nested_arr$data

    expect_equal(dim(arr), nested_arr$shape)

  })

})

vcr::use_cassette("http_listdir", {

  test_that("http listdir and zmeta", {

    url<- "https://raw.githubusercontent.com/DOI-USGS/rnz/main/inst/extdata/bcsd.zarr"

    z <- pizzarr::HttpStore$new(url)
    expect_equal(class(z), c("HttpStore", "Store", "R6"))

    z$set_cache_time_seconds(1234)

    expect_equal(z$get_cache_time_seconds(), 1234)

    vars <- z$listdir()

    expect_equal(vars,
                 c("latitude", "longitude", "pr", "tas", "time"))

    g <- pizzarr::zarr_open_group(z)

    expect_equal(length(names(g$get_attrs()$to_list())), 30)

    expect_equal(g$get_item("latitude")$get_attrs()$get_item("units"), "degrees_north")

    expect_equal(names(g$get_store()$get_consolidated_metadata()$metadata),
                 names(z$get_consolidated_metadata()$metadata))

    expect_equal(dim(g$get_item("pr")$as.array()), c(12, 33, 81))
  })

})

test_that("http broken", {

  url<- "https://nogonnawork.zarr"

  expect_warning(z <- pizzarr::HttpStore$new(url), "Can't proceed, web request failed.")
  expect_equal(class(z), c("HttpStore", "Store", "R6"))

  expect_message(vars <- z$listdir(), "not found for this http store")

  expect_null(vars)

  w <- capture_warnings(g <- pizzarr::zarr_open_group(z))

  expect_true(all(grepl("Can't proceed", w)))
})

vcr::use_cassette("http_github_pattern", {

  test_that("http github pattern", {

    url <- "https://raw.githubusercontent.com/DOI-USGS/rnz/main/inst/extdata/bcsd.zarr"

    z <- zarr_open(url)

    s <- pizzarr::HttpStore$new(url)
    expect_equal(class(s), c("HttpStore", "Store", "R6"))

    g <- zarr_open_group(s, mode = "r", path = NA)

    expect_equal(z$get_attrs()$to_list(), g$get_attrs()$to_list())

    zarr_arr <- g$get_item("pr")

    expect_equal(zarr_arr$get_shape(), c(12L, 33L, 81L))

  })

})

library(pizzarr)

test_that("can set a subset of an array using a NestedArray", {
    a_orig <- array(data=1:20, dim=c(2, 10))
    #      [,1] [,2] [,3] [,4] [,5] [,6] [,7] [,8] [,9] [,10]
    # [1,]    1    3    5    7    9   11   13   15   17    19
    # [2,]    2    4    6    8   10   12   14   16   18    20
    z <- zarr_create_array(data = a_orig, shape=dim(a_orig), dtype="<f4")

    expect_equal(z$get_shape(), c(2, 10))
    expect_equal(z$get_chunks(), c(2, 10))

    s1 <- z$get_item("...")
    expect_equal(a_orig, s1$data)

    a_new <- NestedArray$new(
        data = array(data=c(96, 97, 98, 99), dim=c(2, 2)),
        shape = c(2, 2),
        dtype = "<f4"
    )

    z$set_item(list(slice(1, 2), slice(2, 3)), a_new)

    expected_out <- array(data=1:20, dim=c(2, 10))
    expected_out[1:2, 2:3] <- c(96, 97, 98, 99)
    
    s3 <- z$get_item("...")

    expect_equal(expected_out, s3$data)
})

test_that("set_selection handles scalar value assignment", {
  a <- zarr_create(shape = c(4, 4), chunks = c(2, 2), store = MemoryStore$new(),
                   fill_value = 0, compressor = NA)
  a$set_item("...", 5L)
  result <- a$get_item("...")
  expect_true(all(result$data == 5))
})

test_that("set_selection handles NestedArray value", {
  a <- zarr_create(shape = c(4, 4), chunks = c(2, 2), store = MemoryStore$new(),
                   fill_value = 0, compressor = NA)
  data <- array(1:16, dim = c(4, 4))
  na <- NestedArray$new(data, shape = c(4L, 4L), dtype = a$get_dtype(), order = "C")
  a$set_item("...", na)
  result <- a$get_item("...")
  expect_equal(result$data, data)
})

test_that("set_selection shape mismatch raises error", {
  a <- zarr_create(shape = c(4, 4), chunks = c(2, 2), store = MemoryStore$new(),
                   compressor = NA)
  wrong <- array(1:9, dim = c(3, 3))
  expect_error(a$set_item("...", wrong), "Shape mismatch")
})

test_that("can set a subset of an array using a base R array", {
    a_orig <- array(data=1:20, dim=c(2, 10))
    #      [,1] [,2] [,3] [,4] [,5] [,6] [,7] [,8] [,9] [,10]
    # [1,]    1    3    5    7    9   11   13   15   17    19
    # [2,]    2    4    6    8   10   12   14   16   18    20
    z <- zarr_create_array(data = a_orig, shape=dim(a_orig), dtype="<f4")

    expect_equal(z$get_shape(), c(2, 10))
    expect_equal(z$get_chunks(), c(2, 10))

    s1 <- z$get_item("...")
    expect_equal(a_orig, s1$data)

    a_new <- array(data=c(96, 97, 98, 99), dim=c(2, 2))

    z$set_item(list(slice(1, 2), slice(2, 3)), a_new)

    expected_out <- array(data=1:20, dim=c(2, 10))
    expected_out[1:2, 2:3] <- c(96, 97, 98, 99)

    s3 <- z$get_item("...")

    expect_equal(expected_out, s3$data)
})

# --- chunk_setitem paths ---

test_that("chunk_setitem scalar total-replace path", {
  # set_item with scalar covers all chunks; each chunk_setitem call sees is_total_slice=TRUE and is_scalar=TRUE
  a <- zarr_create(
    shape = c(4, 4), chunks = c(2, 2), store = MemoryStore$new(),
    compressor = NA, fill_value = 0
  )
  a$set_item("...", 5L)
  result <- a$get_item("...")
  expect_true(all(result$data == 5))
})

test_that("chunk_setitem partial replace works (partial chunk write)", {
  a <- zarr_create(
    shape = c(4, 4), chunks = c(4, 4), store = MemoryStore$new(),
    compressor = NA, fill_value = 0
  )
  sel <- list(slice(1, 2), slice(1, 2))
  a$set_item(sel, array(c(1L, 2L, 3L, 4L), dim = c(2, 2)))
  result <- a$get_item(sel)
  expect_equal(as.vector(result$data), c(1L, 2L, 3L, 4L))
})

test_that("chunk_setitem partial replace reads existing chunk before overwrite", {
  a <- zarr_create(
    shape = c(4, 4), chunks = c(4, 4), store = MemoryStore$new(),
    compressor = NA, fill_value = 0
  )
  a$set_item("...", array(1:16, dim = c(4, 4)))
  a$set_item(list(slice(1, 2), slice(1, 2)), array(c(99L, 99L, 99L, 99L), dim = c(2, 2)))
  result <- a$get_item("...")
  expect_equal(result$data[1, 1], 99)
  expect_equal(result$data[1, 3], 9L)  # untouched
})

test_that("chunk_setitem partial write into uninitialized chunk succeeds", {
  a <- zarr_create(
    shape = c(4, 4), chunks = c(4, 4), store = MemoryStore$new(),
    compressor = NA, fill_value = 42
  )
  a$set_item(list(slice(1, 2), slice(1, 2)), array(c(1L, 2L, 3L, 4L), dim = c(2, 2)))
  result <- a$get_item(list(slice(1, 2), slice(1, 2)))
  expect_equal(as.vector(result$data), c(1L, 2L, 3L, 4L))
})

test_that("get_chunk_value returns scalar when value is scalar", {
  a <- zarr_create(shape = c(4), chunks = c(2), store = MemoryStore$new(),
                   fill_value = 7, compressor = NA)
  a$set_item("...", 7L)
  result <- a$get_item("...")
  expect_true(all(result$data == 7))
})

test_that("set_item rejects unsupported value types with helpful error", {
  z <- zarr_create(shape = c(4L), chunks = c(4L), dtype = "<f4",
                   fill_value = 0, store = MemoryStore$new())
  expect_error(z$set_item("...", list(1, 2, 3, 4)), "object_codec")
})

test_that("encode_chunk runs without error for compressor=NA", {
  a <- zarr_create(shape = c(2), chunks = c(2), store = MemoryStore$new(), compressor = NA)
  a$set_item("...", array(c(1L, 2L), dim = c(2)))
  result <- a$get_item("...")
  expect_equal(as.vector(result$data), c(1L, 2L))
})

library(pizzarr)

test_that("can use as.array on NestedArray", {
    a <- array(data=1:20, dim=c(2, 10))
    #      [,1] [,2] [,3] [,4] [,5] [,6] [,7] [,8] [,9] [,10]
    # [1,]    1    3    5    7    9   11   13   15   17    19
    # [2,]    2    4    6    8   10   12   14   16   18    20
    z <- zarr_create(shape=dim(a), dtype="<f4", fill_value=NA)

    expect_equal(z$get_shape(), c(2, 10))
    expect_equal(z$get_chunks(), c(2, 10))

    z$set_item("...", a)

    sel <- z$get_item("...")

    expect_equal(a, as.array(sel))
})

test_that("can use as.array on ZarrArray", {
    a <- array(data=1:20, dim=c(2, 10))
    #      [,1] [,2] [,3] [,4] [,5] [,6] [,7] [,8] [,9] [,10]
    # [1,]    1    3    5    7    9   11   13   15   17    19
    # [2,]    2    4    6    8   10   12   14   16   18    20
    z <- zarr_create(shape=dim(a), dtype="<f4", fill_value=NA)

    expect_equal(z$get_shape(), c(2, 10))
    expect_equal(z$get_chunks(), c(2, 10))

    z$set_item("...", a)

    expect_equal(a, as.array(z))
})

test_that("can use bracket notation to get values from ZarrArray", {
    a <- array(data=1:20, dim=c(2, 10))
    #      [,1] [,2] [,3] [,4] [,5] [,6] [,7] [,8] [,9] [,10]
    # [1,]    1    3    5    7    9   11   13   15   17    19
    # [2,]    2    4    6    8   10   12   14   16   18    20
    z <- zarr_create(shape=dim(a), dtype="<f4", fill_value=NA)

    expect_equal(z$get_shape(), c(2, 10))
    expect_equal(z$get_chunks(), c(2, 10))

    z$set_item("...", a)

    sel <- z[1:2, 1:3]

    b <- array(data=c(1, 2, 3, 4, 5, 6), dim=c(2, 3))

    expect_equal(b, as.array(sel))
})

test_that("S3 methods of Zarr object", {
    a <- array(data=1:20, dim=c(2, 10))
    #      [,1] [,2] [,3] [,4] [,5] [,6] [,7] [,8] [,9] [,10]
    # [1,]    1    3    5    7    9   11   13   15   17    19
    # [2,]    2    4    6    8   10   12   14   16   18    20
    z <- zarr_create(shape=dim(a), dtype="<f4", fill_value=NA)

    expect_equal(z$get_shape(), c(2, 10))
    expect_equal(z$get_chunks(), c(2, 10))

    z$set_item("...", a)

    #subetting with brackets `[`
    expect_equal(z[4,5], z$get_item(list(slice(4, 4), slice(5, 5))))
    # TODO: decide if this is ultimately how we want 1:3 (stop beyond end of dimension) to behave
    expect_equal(z[1:3,5], z$get_item(list(slice(1, 3), slice(5, 5))))
    
    # with seq
    # for now seq(from, to, by) does not use slice(start,stop,step)
    expect_equal(
        z[1:2,seq(1,5,2)]$data,
        array(data=c(1, 2, 5, 6, 9, 10), dim=c(2, 3))
    )
    

    # arbitrary indices
    expect_equal(
        z[c(2,1),c(10,1,6)]$data,
        array(data=c(20, 19, 2, 1, 12, 11), dim=c(2, 3))
    )
      
    # compare regular slicing and step size
    expect_error(z[1], "This Zarr object has 2 dimensions, 1 were supplied")
    
    # test `[.ZarrArray` against get_item
    expect_equal(z[1:2,5], z$get_item(list(slice(1, 3), slice(5, 5))))
    expect_equal(z[1,seq(1,5)], z$get_item(list(slice(1, 1), slice(1, 5))))
    expect_equal(z[1,seq(1,5,2)], z$get_item(list(slice(1, 1), slice(1, 5, 2))))
    expect_equal(z[1,], z$get_item(list(slice(1, 1))))
    
    #Converting Zarr to array
    expect_equal(as.array(z), a)
})

library(pizzarr)

sample_dir <- tools::R_user_dir("pizzarr")
clean <- !dir.exists(sample_dir)

test_that("Can open Zarr group using convenience function", {
    skip_if_not_installed("qs")

    root <- pizzarr_sample(file.path("fixtures", "v2", "data.zarr"))

    g <- zarr_open_group(root)
    a <- suppressWarnings(g$get_item("1d.contiguous.lz4.i2"))

    expect_equal(a$get_shape(), c(4))
})

test_that("Can open Zarr group or array using convenience function", {
    skip_if_not_installed("qs")

    root <- pizzarr_sample(file.path("fixtures", "v2", "data.zarr"))
    g <- zarr_open(root)
    a <- suppressWarnings(zarr_open(root, path="1d.contiguous.lz4.i2"))

    expect_equal(class(g)[1], "ZarrGroup")
    expect_equal(class(a)[1], "ZarrArray")
})

test_that("Can open Zarr group and read a 1D 2-byte integer array with LZ4 compression", {
    skip_if_not_installed("qs")

    root <- pizzarr_sample(file.path("fixtures", "v2", "data.zarr"))

    store <- DirectoryStore$new(root)
    g <- ZarrGroup$new(store)
    a <- suppressWarnings(g$get_item("1d.contiguous.lz4.i2"))

    expect_equal(a$get_shape(), c(4))
    expect_equal(a$get_chunks(), c(4))

    expect_equal(as.character(a$get_compressor()$get_config()$id), "lz4")

    selection <- a$get_item("...")

    expect_equal(dim(selection$data), c(4))
    expect_equal(selection$data, array(data=c(1, 2, 3, 4), dim=c(4)))
})

test_that("Can open Zarr group and read a 1D 2-byte integer array with Zstd compression", {

    root <- pizzarr_sample(file.path("fixtures", "v2", "data.zarr"))
    
    store <- DirectoryStore$new(root)
    g <- ZarrGroup$new(store)
    a <- g$get_item("1d.contiguous.zstd.i2")

    expect_equal(a$get_shape(), c(4))
    expect_equal(a$get_chunks(), c(4))

    expect_equal(as.character(a$get_compressor()$get_config()$id), "zstd")

    selection <- a$get_item("...")

    expect_equal(dim(selection$data), c(4))
    expect_equal(selection$data, array(data=c(1, 2, 3, 4), dim=c(4)))
})

test_that("Can open Zarr group and read a 1D 2-byte integer array with Blosc compression", {

    root <- pizzarr_sample(file.path("fixtures", "v2", "data.zarr"))
    
    store <- DirectoryStore$new(root)
    g <- ZarrGroup$new(store)

    f <- function() {
        a <- g$get_item("1d.contiguous.blosc.i2")

        expect_equal(a$get_shape(), c(4))
        expect_equal(a$get_chunks(), c(4))

        expect_equal(as.character(a$get_compressor()$get_config()$id), "blosc")

        selection <- a$get_item("...")

        expect_equal(dim(selection$data), c(4))
        expect_equal(selection$data, array(data=c(1, 2, 3, 4), dim=c(4)))
    }

    if(requireNamespace("blosc", quietly=TRUE)) {
        f()
    } else {
        expect_error(f())
    }
})

test_that("Can open Zarr group and read a 1D 2-byte integer array with Zlib compression", {

    root <- pizzarr_sample(file.path("fixtures", "v2", "data.zarr"))
    
    store <- DirectoryStore$new(root)
    g <- ZarrGroup$new(store)
    a <- g$get_item("1d.contiguous.zlib.i2")

    expect_equal(a$get_shape(), c(4))
    expect_equal(a$get_chunks(), c(4))

    expect_equal(as.character(a$get_compressor()$get_config()$id), "zlib")

    selection <- a$get_item("...")

    expect_equal(dim(selection$data), c(4))
    expect_equal(selection$data, array(data=c(1, 2, 3, 4), dim=c(4)))
})

test_that("Can open Zarr group and read a 1D 2-byte integer array with no compression", {

    root <- pizzarr_sample(file.path("fixtures", "v2", "data.zarr"))
    
    store <- DirectoryStore$new(root)
    g <- ZarrGroup$new(store)
    a <- g$get_item("1d.contiguous.raw.i2")

    expect_equal(a$get_shape(), c(4))
    expect_equal(a$get_chunks(), c(4))

    expect_equal(is.na(a$get_compressor()), TRUE)

    selection <- a$get_item("...")

    expect_equal(dim(selection$data), c(4))
    expect_equal(selection$data, array(data=c(1, 2, 3, 4), dim=c(4)))
})

test_that("Can open Zarr group and read a 1D 4-byte integer array", {
    skip_if_not_installed("blosc")

    root <- pizzarr_sample(file.path("fixtures", "v2", "data.zarr"))
    
    store <- DirectoryStore$new(root)
    g <- ZarrGroup$new(store)
    a <- g$get_item("1d.contiguous.i4")

    expect_equal(a$get_shape(), c(4))
    expect_equal(a$get_chunks(), c(4))

    selection <- a$get_item("...")

    expect_equal(dim(selection$data), c(4))
    expect_equal(selection$data, array(data=c(1, 2, 3, 4), dim=c(4)))
})

test_that("Can open Zarr group and read a 1D 1-byte unsigned integer array", {
    skip_if_not_installed("blosc")

    root <- pizzarr_sample(file.path("fixtures", "v2", "data.zarr"))
    
    store <- DirectoryStore$new(root)
    g <- ZarrGroup$new(store)
    a <- g$get_item("1d.contiguous.u1")

    expect_equal(a$get_shape(), c(4))
    expect_equal(a$get_chunks(), c(4))

    selection <- a$get_item("...")

    expect_equal(dim(selection$data), c(4))
    expect_equal(selection$data, array(data=c(255, 0, 255, 0), dim=c(4)))
})

test_that("Can open Zarr group and read a 1D 4-byte float array, little endian", {
    skip_if_not_installed("blosc")

    root <- pizzarr_sample(file.path("fixtures", "v2", "data.zarr"))
    
    store <- DirectoryStore$new(root)
    g <- ZarrGroup$new(store)
    a <- g$get_item("1d.contiguous.f4.le")

    expect_equal(a$get_shape(), c(4))
    expect_equal(a$get_chunks(), c(4))

    selection <- a$get_item("...")

    expect_equal(dim(selection$data), c(4))
    expect_equal(selection$data, array(data=c(-1000.5, 0, 1000.5, 0), dim=c(4)))
})

test_that("Can open Zarr group and read a 1D 4-byte float array, big endian", {
    skip_if_not_installed("blosc")

    root <- pizzarr_sample(file.path("fixtures", "v2", "data.zarr"))
    
    store <- DirectoryStore$new(root)
    g <- ZarrGroup$new(store)
    a <- g$get_item("1d.contiguous.f4.be")

    expect_equal(a$get_shape(), c(4))
    expect_equal(a$get_chunks(), c(4))

    selection <- a$get_item("...")

    expect_equal(dim(selection$data), c(4))
    expect_equal(selection$data, array(data=c(-1000.5, 0, 1000.5, 0), dim=c(4)))
})

test_that("Can open Zarr group and read a 1D 8-byte float array", {
    skip_if_not_installed("blosc")

    root <- pizzarr_sample(file.path("fixtures", "v2", "data.zarr"))
    
    store <- DirectoryStore$new(root)
    g <- ZarrGroup$new(store)
    a <- g$get_item("1d.contiguous.f8")

    expect_equal(a$get_shape(), c(4))
    expect_equal(a$get_chunks(), c(4))

    selection <- a$get_item("...")

    expect_equal(dim(selection$data), c(4))
    expect_equal(selection$data, array(data=c(1.5, 2.5, 3.5, 4.5), dim=c(4)))
})

test_that("Can open Zarr group and read a 1D 2-byte float array, 2 chunks", {
    skip_if_not_installed("blosc")

    root <- pizzarr_sample(file.path("fixtures", "v2", "data.zarr"))
    
    store <- DirectoryStore$new(root)
    g <- ZarrGroup$new(store)
    a <- g$get_item("1d.chunked.i2")

    expect_equal(a$get_shape(), c(4))
    expect_equal(a$get_chunks(), c(2))

    selection <- a$get_item("...")

    expect_equal(dim(selection$data), c(4))
    expect_equal(selection$data, array(data=c(1, 2, 3, 4), dim=c(4)))
})

test_that("Can open Zarr group and read a 1D 2-byte float array, 2 chunks, ragged", {
    skip_if_not_installed("blosc")

    root <- pizzarr_sample(file.path("fixtures", "v2", "data.zarr"))
    
    store <- DirectoryStore$new(root)
    g <- ZarrGroup$new(store)
    a <- g$get_item("1d.chunked.ragged.i2")

    expect_equal(a$get_shape(), c(5))
    expect_equal(a$get_chunks(), c(2))

    selection <- a$get_item("...")

    expect_equal(dim(selection$data), c(5))
    expect_equal(selection$data, array(data=c(1, 2, 3, 4, 5), dim=c(5)))
})

test_that("Can open Zarr group and read a 2D 2-byte integer array", {
    skip_if_not_installed("blosc")

    root <- pizzarr_sample(file.path("fixtures", "v2", "data.zarr"))
    
    store <- DirectoryStore$new(root)
    g <- ZarrGroup$new(store)
    a <- g$get_item("2d.contiguous.i2")

    expect_equal(a$get_shape(), c(2, 2))
    expect_equal(a$get_chunks(), c(2, 2))

    selection <- a$get_item("...")

    expect_equal(dim(selection$data), c(2, 2))
    expect_equal(selection$data[1,], c(1, 2))
    expect_equal(selection$data[2,], c(3, 4))
})

test_that("Can open Zarr group and read a 2D 2-byte integer array, 2 chunks", {
    skip_if_not_installed("blosc")

    root <- pizzarr_sample(file.path("fixtures", "v2", "data.zarr"))
    
    store <- DirectoryStore$new(root)
    g <- ZarrGroup$new(store)
    a <- g$get_item("2d.chunked.i2")

    expect_equal(a$get_shape(), c(2, 2))
    expect_equal(a$get_chunks(), c(1, 1))

    selection <- a$get_item("...")

    expect_equal(dim(selection$data), c(2, 2))
    expect_equal(selection$data[1,], c(1, 2))
    expect_equal(selection$data[2,], c(3, 4))
})

test_that("Can open Zarr group and read a 2D 2-byte integer array, 2 chunks, ragged", {
    skip_if_not_installed("blosc")

    root <- pizzarr_sample(file.path("fixtures", "v2", "data.zarr"))
    
    store <- DirectoryStore$new(root)
    g <- ZarrGroup$new(store)
    a <- g$get_item("2d.chunked.ragged.i2")

    expect_equal(a$get_shape(), c(3, 3))
    expect_equal(a$get_chunks(), c(2, 2))

    selection <- a$get_item("...")

    expect_equal(dim(selection$data), c(3, 3))
    expect_equal(selection$data[1,], c(1, 2, 3))
    expect_equal(selection$data[2,], c(4, 5, 6))
    expect_equal(selection$data[3,], c(7, 8, 9))
})


test_that("Can open Zarr group and read a 1D 1-byte boolean array", {
    skip_if_not_installed("blosc")

    root <- pizzarr_sample(file.path("fixtures", "v2", "data.zarr"))
    
    store <- DirectoryStore$new(root)
    g <- ZarrGroup$new(store)
    a <- g$get_item("1d.contiguous.b1")

    expect_equal(a$get_shape(), c(4))
    expect_equal(a$get_chunks(), c(4))

    selection <- a$get_item("...")

    expect_equal(dim(selection$data), c(4))
    expect_equal(selection$data, array(data=c(TRUE, FALSE, TRUE, FALSE), dim=c(4)))
})

test_that("Can open Zarr group and read a 1D S7 string array", {
    skip_if_not_installed("blosc")

    root <- pizzarr_sample(file.path("fixtures", "v2", "data.zarr"))
    
    store <- DirectoryStore$new(root)
    g <- ZarrGroup$new(store)
    a <- g$get_item("1d.contiguous.S7")

    expect_equal(a$get_shape(), c(4))
    expect_equal(a$get_chunks(), c(4))

    selection <- a$get_item("...")

    expect_equal(dim(selection$data), c(4))
    expect_equal(selection$data, array(data=c("a", "b", "cc", "d"), dim=c(4)))
})

test_that("Can open Zarr group and read a 1D U7 string array", {
    skip_if_not_installed("blosc")

    root <- pizzarr_sample(file.path("fixtures", "v2", "data.zarr"))
    
    store <- DirectoryStore$new(root)
    g <- ZarrGroup$new(store)
    a <- g$get_item("1d.contiguous.U7")

    expect_equal(a$get_shape(), c(4))
    expect_equal(a$get_chunks(), c(4))

    selection <- a$get_item("...")

    expect_equal(dim(selection$data), c(4))
    expect_equal(selection$data, array(data=c("a", "b", "cc", "d"), dim=c(4)))
})

test_that("Can open Zarr group and read a 1D U13 little endian string array", {
    skip_if_not_installed("blosc")

    root <- pizzarr_sample(file.path("fixtures", "v2", "data.zarr"))
    
    store <- DirectoryStore$new(root)
    g <- ZarrGroup$new(store)
    a <- g$get_item("1d.contiguous.U13.le")

    expect_equal(a$get_shape(), c(4))
    expect_equal(a$get_chunks(), c(4))

    selection <- a$get_item("...")

    expect_equal(dim(selection$data), c(4))
    expect_equal(selection$data, array(data=c("a", "b", "cc", "d"), dim=c(4)))
})

test_that("Can open Zarr group and read a 1D U13 big endian string array", {
    skip_if_not_installed("blosc")

    root <- pizzarr_sample(file.path("fixtures", "v2", "data.zarr"))
    
    store <- DirectoryStore$new(root)
    g <- ZarrGroup$new(store)
    a <- g$get_item("1d.contiguous.U13.be")

    expect_equal(a$get_shape(), c(4))
    expect_equal(a$get_chunks(), c(4))

    selection <- a$get_item("...")

    expect_equal(dim(selection$data), c(4))
    expect_equal(selection$data, array(data=c("a", "b", "cc", "d"), dim=c(4)))
})

test_that("Can open Zarr group and read a 2D U7 string array", {
    skip_if_not_installed("blosc")

    root <- pizzarr_sample(file.path("fixtures", "v2", "data.zarr"))
    
    store <- DirectoryStore$new(root)
    g <- ZarrGroup$new(store)
    a <- g$get_item("2d.chunked.U7")

    expect_equal(a$get_shape(), c(2, 2))
    expect_equal(a$get_chunks(), c(1, 1))

    selection <- a$get_item("...")

    expect_equal(dim(selection$data), c(2, 2))
    expect_equal(selection$data, array(data=c("a", "cc", "b", "d"), dim=c(2, 2)))
})

test_that("Can open Zarr group and read a 1D VLen-UTF8 string array with no compression", {

    root <- pizzarr_sample(file.path("fixtures", "v2", "data.zarr"))
    
    store <- DirectoryStore$new(root)
    g <- ZarrGroup$new(store)
    a <- g$get_item("1d.contiguous.raw.vlen-utf8")

    expect_equal(a$get_shape(), c(4))
    expect_equal(a$get_chunks(), c(4))

    selection <- a$get_item("...")

    expect_equal(dim(selection$data), c(4))
    expect_equal(selection$data, array(data=c("a", "b", "cc", "d"), dim=c(4)))
})

test_that("Can open Zarr group and read a 1D VLen-UTF8 string array with Blosc compression", {
    skip_if_not_installed("blosc")

    root <- pizzarr_sample(file.path("fixtures", "v2", "data.zarr"))
    
    store <- DirectoryStore$new(root)
    g <- ZarrGroup$new(store)
    a <- g$get_item("1d.contiguous.blosc.vlen-utf8")

    expect_equal(a$get_shape(), c(4))
    expect_equal(a$get_chunks(), c(4))

    selection <- a$get_item("...")

    expect_equal(dim(selection$data), c(4))
    expect_equal(selection$data, array(data=c("a", "b", "cc", "d"), dim=c(4)))
})

test_that("Can open Zarr group and read a 2D VLen-UTF8 string array with no compression", {
    root <- pizzarr_sample(file.path("fixtures", "v2", "data.zarr"))
    
    store <- DirectoryStore$new(root)
    g <- ZarrGroup$new(store)
    a <- g$get_item("2d.chunked.raw.vlen-utf8")

    expect_equal(a$get_shape(), c(2, 2))
    expect_equal(a$get_chunks(), c(1, 1))

    selection <- a$get_item("...")

    expect_equal(dim(selection$data), c(2, 2))
    expect_equal(selection$data, array(data=c("a", "cc", "b", "d"), dim=c(2, 2)))
})

test_that("Can open Zarr group and read a 2D VLen-UTF8 string array with Blosc compression", {
    skip_if_not_installed("blosc")
    root <- pizzarr_sample(file.path("fixtures", "v2", "data.zarr"))
    
    store <- DirectoryStore$new(root)
    g <- ZarrGroup$new(store)
    a <- g$get_item("2d.chunked.blosc.vlen-utf8")

    expect_equal(a$get_shape(), c(2, 2))
    expect_equal(a$get_chunks(), c(1, 1))

    selection <- a$get_item("...")

    expect_equal(dim(selection$data), c(2, 2))
    expect_equal(selection$data, array(data=c("a", "cc", "b", "d"), dim=c(2, 2)))
})

test_that("DirectoryStore can listdir", {
    root <- pizzarr_sample(file.path("fixtures", "v2", "data.zarr"))
    
    store <- DirectoryStore$new(root)
    
    expect_equal(store$listdir("1d.chunked.i2"), c(".zarray", "0", "1"))
})

test_that("Can create 0-element 1D array", {

    store <- MemoryStore$new()
    value <- character(0)
    dims <- c(0)
    name <- "my-0-element-array-1d"

    object_codec <- VLenUtf8Codec$new()
    data <- array(data = value, dim = dims)
    a <- zarr_create_array(data, store = store, path = name, dtype = "|O", object_codec = object_codec, shape = dims)

    expect_equal(a$get_shape(), c(0))

    sel <- a$get_item("...")
    
    expect_equal(is.character(sel$data), TRUE)
    expect_equal(length(sel$data), 0)
    expect_equal(dim(sel$data), c(0))
})

test_that("Can create 0-element 2D array", {

    store <- MemoryStore$new()
    value <- character(0)
    dims <- c(0, 0)
    name <- "my-0-element-array-2d"

    object_codec <- VLenUtf8Codec$new()
    data <- array(data = value, dim = dims)
    a <- zarr_create_array(data, store = store, path = name, dtype = "|O", object_codec = object_codec, shape = dims)

    expect_equal(a$get_shape(), c(0, 0))

    sel <- a$get_item("...")
    
    expect_equal(is.character(sel$data), TRUE)
    expect_equal(length(sel$data), 0)
    expect_equal(dim(sel$data), c(0, 0))
})

if(clean) unlink(sample_dir, recursive = TRUE)

library(pizzarr)

make_array <- function(shape = c(4, 4), chunks = c(2, 2), dtype = "<i4",
                       compressor = NA, fill_value = 0, order = "C") {
  zarr_create(
    shape = shape, chunks = chunks, dtype = dtype,
    store = MemoryStore$new(), compressor = compressor,
    fill_value = fill_value, order = order
  )
}

test_that("Zarr Array can load .zarray metadata", {
  store <- MemoryStore$new()

  zarray_meta <- store$metadata_class$encode_array_metadata(create_zarray_meta(
    dtype = Dtype$new("|u1"),
    order = "C",
    fill_value = 0,
    shape = c(1, 2),
    chunks = c(3, 4),
    dimension_separator = "."
  ))

  store$set_item(".zarray", zarray_meta)

  array <- ZarrArray$new(store = store)
  shape <- array$get_shape()

  expect_equal(shape, c(1, 2))
})

test_that("load_metadata_nosync handles list-style shape/chunks from JSON", {
  d <- tempfile()
  a <- zarr_create(shape = c(3, 4), chunks = c(2, 2), store = d, compressor = NA)
  a$set_item("...", array(1:12, dim = c(3, 4)))
  b <- zarr_open_array(d, mode = "r")
  expect_equal(b$get_shape(), c(3, 4))
  expect_equal(b$get_chunks(), c(2, 2))
})

test_that("cache_metadata=FALSE reloads metadata on each access", {
  d <- tempfile()
  zarr_create(shape = c(4), chunks = c(2), store = d, compressor = NA)
  b <- ZarrArray$new(store = DirectoryStore$new(d), cache_metadata = FALSE)
  expect_equal(b$get_shape(), c(4))
  # Calling again exercises the refresh_metadata path a second time
  expect_equal(b$get_shape(), c(4))
})

test_that("Zarr Array can be resized", {
  store <- MemoryStore$new()

  zarray_meta <- store$metadata_class$encode_array_metadata(create_zarray_meta(
    dtype = Dtype$new("|u1"),
    order = "C",
    fill_value = 0,
    shape = c(4, 5),
    chunks = c(3, 4),
    dimension_separator = "."
  ))

  store$set_item(".zarray", zarray_meta)

  array <- ZarrArray$new(store = store)
  old_shape <- array$get_shape()
  expect_equal(old_shape, c(4, 5))
  array$resize(1, 2)

  new_shape <- array$get_shape()
  expect_equal(new_shape, c(1, 2))
})

test_that("flush_metadata_nosync (V2) serializes filters on resize", {
  store <- MemoryStore$new()
  zarray_meta <- store$metadata_class$encode_array_metadata(create_zarray_meta(
    dtype = Dtype$new("|u1"),
    order = "C",
    fill_value = 0,
    shape = c(4),
    chunks = c(4),
    compressor = NA,
    filters = list(list(id = "zlib", level = 1L)),
    dimension_separator = "."
  ))
  store$set_item(".zarray", zarray_meta)
  a <- ZarrArray$new(store = store)
  a$resize(8)
  expect_equal(a$get_shape(), c(8))
})

# --- Write/read integration tests ---

test_that("Array write/read round-trip, MemoryStore, no compressor", {
  a <- zarr_create(
    shape = c(4, 4), chunks = c(2, 2),
    store = MemoryStore$new(), compressor = NA
  )
  data <- array(1:16, dim = c(4, 4))
  a$set_item("...", data)
  result <- a$get_item("...")
  expect_equal(result$data, data)
})

test_that("Array write/read round-trip, MemoryStore, Zstd compressor", {
  a <- zarr_create(
    shape = c(4, 4), chunks = c(2, 2),
    store = MemoryStore$new(), compressor = ZstdCodec$new()
  )
  data <- array(1:16, dim = c(4, 4))
  a$set_item("...", data)
  result <- a$get_item("...")
  expect_equal(result$data, data)
})

test_that("Array write/read round-trip, DirectoryStore, default compressor", {
  a <- zarr_create(
    shape = c(4, 4), chunks = c(2, 2),
    store = DirectoryStore$new(tempfile())
  )
  data <- array(1:16, dim = c(4, 4))
  a$set_item("...", data)
  result <- a$get_item("...")
  expect_equal(result$data, data)
})

test_that("Array persists to DirectoryStore and reopens correctly", {
  d <- tempfile()
  data <- array(1:20, dim = c(4, 5))
  a <- zarr_create(shape = c(4, 5), chunks = c(2, 5), store = d)
  a$set_item("...", data)
  b <- zarr_open_array(d, mode = "r")
  result <- b$get_item("...")
  expect_equal(result$data, data)
})

test_that("1D array write/read round-trip", {
  a <- zarr_create(
    shape = c(10), chunks = c(5),
    store = MemoryStore$new(), compressor = NA
  )
  data <- array(1:10, dim = c(10))
  a$set_item("...", data)
  result <- a$get_item("...")
  expect_equal(result$data, data)
})

test_that("3D array write/read round-trip", {
  a <- zarr_create(
    shape = c(3, 3, 3), chunks = c(2, 2, 2),
    store = MemoryStore$new(), compressor = NA
  )
  data <- array(1:27, dim = c(3, 3, 3))
  a$set_item("...", data)
  result <- a$get_item("...")
  expect_equal(result$data, data)
})

test_that("Array with fill_value=0 returns fill for unwritten chunks", {
  a <- zarr_create(
    shape = c(4), chunks = c(2),
    store = MemoryStore$new(), fill_value = 0, compressor = NA
  )
  result <- a$get_item("...")
  expect_equal(as.vector(result$data), c(0, 0, 0, 0))
})

test_that("Array metadata properties are correct", {
  a <- zarr_create(
    shape = c(10, 20), chunks = c(5, 10), dtype = "<f4",
    store = MemoryStore$new(), compressor = ZstdCodec$new()
  )
  expect_equal(a$get_shape(), c(10, 20))
  expect_equal(a$get_chunks(), c(5, 10))
  expect_s3_class(a$get_compressor(), "ZstdCodec")
})

# --- chunk_key format ---

test_that("chunk_key produces dot-separated keys for V2", {
  a <- make_array(shape = c(4, 4), chunks = c(2, 2))
  data <- array(1:16, dim = c(4, 4))
  a$set_item("...", data)
  result <- a$get_item("...")
  expect_equal(result$data, data)
  keys <- a$get_store()$listdir(NA)
  chunk_keys <- keys[!grepl("^\\.z", keys)]
  expect_true(any(grepl("\\.", chunk_keys)))
})

# --- Accessor methods ---

test_that("get_cdata_shape works correctly", {
  a <- make_array(shape = c(5, 7), chunks = c(2, 3))
  cs <- a$get_cdata_shape()
  expect_equal(cs, c(3, 3))  # ceiling(5/2)=3, ceiling(7/3)=3
})

test_that("get_store returns the store", {
  store <- MemoryStore$new()
  a <- zarr_create(shape = c(2), chunks = c(2), store = store)
  expect_identical(a$get_store(), store)
})

test_that("get_path returns NA for root-level array", {
  a <- make_array()
  expect_true(is.na(a$get_path()))
})

test_that("get_path returns path for nested array", {
  g <- zarr_create_group(store = MemoryStore$new())
  a <- g$create_dataset("mydata", shape = c(4), chunks = c(2), dtype = "<i4")
  expect_equal(a$get_path(), "mydata")
})

test_that("get_name returns NA for root array", {
  a <- make_array()
  expect_true(is.na(a$get_name()))
})

test_that("get_name returns /path for nested array", {
  g <- zarr_create_group(store = MemoryStore$new())
  a <- g$create_dataset("mydata", shape = c(4), chunks = c(2), dtype = "<i4")
  expect_equal(a$get_name(), "/mydata")
})

test_that("get_basename returns NA for root array", {
  a <- make_array()
  expect_true(is.na(a$get_basename()))
})

test_that("get_basename returns last path segment", {
  g <- zarr_create_group(store = MemoryStore$new())
  a <- g$create_dataset("mydata", shape = c(4), chunks = c(2), dtype = "<i4")
  expect_equal(a$get_basename(), "mydata")
})

test_that("set_read_only changes read_only property", {
  a <- make_array()
  expect_false(a$get_read_only())
  a$set_read_only(TRUE)
  expect_true(a$get_read_only())
})

test_that("set_shape resizes array", {
  a <- zarr_create(shape = c(4), chunks = c(2), store = MemoryStore$new(), compressor = NA)
  a$set_shape(c(8))
  expect_equal(a$get_shape(), c(8))
})

test_that("get_fill_value returns the configured fill value", {
  a <- zarr_create(shape = c(4), chunks = c(2), store = MemoryStore$new(),
                   fill_value = 42, compressor = NA)
  expect_equal(a$get_fill_value(), 42)
})

test_that("set_fill_value updates fill value in memory", {
  a <- zarr_create(shape = c(4), chunks = c(2), store = MemoryStore$new(),
                   fill_value = 0, compressor = NA)
  a$set_fill_value(7)
  expect_equal(a$get_fill_value(), 7)
})

test_that("get_order returns storage order", {
  a <- make_array(order = "C")
  expect_equal(a$get_order(), "C")
})

test_that("get_filters returns NA when no filters set", {
  a <- make_array()
  expect_true(is_na(a$get_filters()))
})

test_that("get_synchronizer returns NA by default", {
  a <- make_array()
  expect_true(is.na(a$get_synchronizer()))
})

test_that("get_ndim returns number of dimensions", {
  a <- make_array(shape = c(3, 4, 5), chunks = c(2, 2, 2))
  expect_equal(a$get_ndim(), 3)
})

test_that("get_is_view returns FALSE for normal array", {
  a <- make_array()
  expect_false(a$get_is_view())
})

test_that("get_oindex returns OIndex object", {
  a <- make_array()
  expect_s3_class(a$get_oindex(), "OIndex")
})

test_that("get_vindex returns VIndex object", {
  a <- make_array()
  expect_s3_class(a$get_vindex(), "VIndex")
})

test_that("get_write_empty_chunks returns TRUE by default", {
  a <- make_array()
  expect_true(a$get_write_empty_chunks())
})

test_that("get_dimension_separator returns the separator", {
  a <- make_array()
  sep <- a$get_dimension_separator()
  expect_true(sep %in% c(".", "/"))
})

test_that("equals FALSE for read_only mismatch between two arrays", {
  store1 <- MemoryStore$new()
  a <- zarr_create(shape = c(4), chunks = c(2), store = store1, compressor = NA)
  store2 <- MemoryStore$new()
  b <- zarr_create(shape = c(4), chunks = c(2), store = store2, compressor = NA)
  b$set_read_only(TRUE)
  expect_false(a$equals(b))
})

test_that("equals returns TRUE for self-equality", {
  store <- MemoryStore$new()
  a <- zarr_create(shape = c(4), chunks = c(2), store = store, compressor = NA)
  expect_true(a$equals(a))
})

test_that("equals returns FALSE for different stores", {
  a <- zarr_create(shape = c(4), chunks = c(2), store = MemoryStore$new(), compressor = NA)
  b <- zarr_create(shape = c(4), chunks = c(2), store = MemoryStore$new(), compressor = NA)
  expect_false(a$equals(b))
})

test_that("equals returns FALSE for non-ZarrArray object", {
  store <- MemoryStore$new()
  a <- zarr_create(shape = c(4), chunks = c(2), store = store, compressor = NA)
  expect_false(a$equals("not an array"))
})

test_that("length returns first dimension size", {
  # Note: length() uses `if(private$shape)` which errors on multi-element vectors
  # so only works reliably on 1D arrays
  a <- zarr_create(shape = c(7), chunks = c(4), store = MemoryStore$new(), compressor = NA)
  expect_equal(a$length(), 7)
})

test_that("[<-.ZarrArray raises error", {
  a <- make_array(shape = c(4, 4), chunks = c(2, 2))
  expect_error(`[<-.ZarrArray`(a, value = 1), "not yet supported")
})

library(pizzarr)

# Unzip V3 test fixture to temp directory.
# Fixture generated with zarrita (see inst/extdata/fixtures/v3/generate-v3.py).
v3_zip <- system.file("extdata/fixtures/v3/data.zarr.zip", package = "pizzarr")
tdir <- tempfile("v3test")
dir.create(tdir)
utils::unzip(v3_zip, exdir = tdir)
v3_root <- file.path(tdir, "data.zarr")

# --- Format detection ---

test_that("detect_zarr_version returns 3 for V3 store", {
  store <- DirectoryStore$new(v3_root)
  expect_equal(detect_zarr_version(store), 3L)
})

test_that("contains_group detects V3 root group", {
  store <- DirectoryStore$new(v3_root)
  expect_true(contains_group(store))
})

test_that("contains_array detects V3 array", {
  store <- DirectoryStore$new(v3_root)
  expect_true(contains_array(store, "1d.contiguous.raw.i2"))
})

# --- Group opening ---

test_that("V3 group can be opened with zarr_open_group", {
  g <- zarr_open_group(v3_root, mode = "r")
  expect_s3_class(g, "ZarrGroup")
})

test_that("V3 group can be opened with zarr_open", {
  g <- zarr_open(v3_root, mode = "r")
  expect_s3_class(g, "ZarrGroup")
})

# --- 1D arrays ---

test_that("V3 1d contiguous raw (no compression) int16", {
  store <- DirectoryStore$new(v3_root)
  a <- ZarrArray$new(store, path = "1d.contiguous.raw.i2", read_only = TRUE)
  expect_equal(a$get_shape(), 4)
  result <- a$get_item("...")
  expect_equal(as.integer(result$data), c(1L, 2L, 3L, 4L))
})

test_that("V3 1d contiguous gzip int16", {
  store <- DirectoryStore$new(v3_root)
  a <- ZarrArray$new(store, path = "1d.contiguous.gzip.i2", read_only = TRUE)
  expect_equal(a$get_shape(), 4)
  result <- a$get_item("...")
  expect_equal(as.integer(result$data), c(1L, 2L, 3L, 4L))
})

test_that("V3 1d contiguous blosc int16", {
  skip_if_not_installed("blosc")
  store <- DirectoryStore$new(v3_root)
  a <- ZarrArray$new(store, path = "1d.contiguous.blosc.i2", read_only = TRUE)
  expect_equal(a$get_shape(), 4)
  result <- a$get_item("...")
  expect_equal(as.integer(result$data), c(1L, 2L, 3L, 4L))
})

test_that("V3 1d contiguous int32", {
  skip_if_not_installed("blosc")
  store <- DirectoryStore$new(v3_root)
  a <- ZarrArray$new(store, path = "1d.contiguous.i4", read_only = TRUE)
  expect_equal(a$get_shape(), 4)
  result <- a$get_item("...")
  expect_equal(as.integer(result$data), c(1L, 2L, 3L, 4L))
})

test_that("V3 1d contiguous uint8", {
  skip_if_not_installed("blosc")
  store <- DirectoryStore$new(v3_root)
  a <- ZarrArray$new(store, path = "1d.contiguous.u1", read_only = TRUE)
  expect_equal(a$get_shape(), 4)
  result <- a$get_item("...")
  expect_equal(as.integer(result$data), c(255L, 0L, 255L, 0L))
})

test_that("V3 1d contiguous float32 little-endian", {
  skip_if_not_installed("blosc")
  store <- DirectoryStore$new(v3_root)
  a <- ZarrArray$new(store, path = "1d.contiguous.f4.le", read_only = TRUE)
  expect_equal(a$get_shape(), 4)
  result <- a$get_item("...")
  expect_equal(as.double(result$data), c(-1000.5, 0, 1000.5, 0))
})

test_that("V3 1d contiguous float32 big-endian", {
  skip_if_not_installed("blosc")
  store <- DirectoryStore$new(v3_root)
  a <- ZarrArray$new(store, path = "1d.contiguous.f4.be", read_only = TRUE)
  expect_equal(a$get_shape(), 4)
  result <- a$get_item("...")
  expect_equal(as.double(result$data), c(-1000.5, 0, 1000.5, 0))
})

test_that("V3 1d contiguous float64", {
  skip_if_not_installed("blosc")
  store <- DirectoryStore$new(v3_root)
  a <- ZarrArray$new(store, path = "1d.contiguous.f8", read_only = TRUE)
  expect_equal(a$get_shape(), 4)
  result <- a$get_item("...")
  expect_equal(as.double(result$data), c(1.5, 2.5, 3.5, 4.5))
})

test_that("V3 1d contiguous bool", {
  skip_if_not_installed("blosc")
  store <- DirectoryStore$new(v3_root)
  a <- ZarrArray$new(store, path = "1d.contiguous.b1", read_only = TRUE)
  expect_equal(a$get_shape(), 4)
  result <- a$get_item("...")
  expect_equal(as.logical(result$data), c(TRUE, FALSE, TRUE, FALSE))
})

test_that("V3 1d chunked int16", {
  skip_if_not_installed("blosc")
  store <- DirectoryStore$new(v3_root)
  a <- ZarrArray$new(store, path = "1d.chunked.i2", read_only = TRUE)
  expect_equal(a$get_shape(), 4)
  result <- a$get_item("...")
  expect_equal(as.integer(result$data), c(1L, 2L, 3L, 4L))
})

test_that("V3 1d chunked ragged int16", {
  skip_if_not_installed("blosc")
  store <- DirectoryStore$new(v3_root)
  a <- ZarrArray$new(store, path = "1d.chunked.ragged.i2", read_only = TRUE)
  expect_equal(a$get_shape(), 5)
  result <- a$get_item("...")
  expect_equal(as.integer(result$data), c(1L, 2L, 3L, 4L, 5L))
})

# --- 2D arrays ---

test_that("V3 2d contiguous int16", {
  skip_if_not_installed("blosc")
  store <- DirectoryStore$new(v3_root)
  a <- ZarrArray$new(store, path = "2d.contiguous.i2", read_only = TRUE)
  expect_equal(a$get_shape(), c(2L, 2L))
  result <- a$get_item("...")
  # Data written as [[1,2],[3,4]] in C order (row-major).
  # R uses column-major, so check element access directly.
  expect_equal(result$data[1, 1], 1)
  expect_equal(result$data[1, 2], 2)
  expect_equal(result$data[2, 1], 3)
  expect_equal(result$data[2, 2], 4)
})

test_that("V3 2d chunked int16", {
  skip_if_not_installed("blosc")
  store <- DirectoryStore$new(v3_root)
  a <- ZarrArray$new(store, path = "2d.chunked.i2", read_only = TRUE)
  expect_equal(a$get_shape(), c(2L, 2L))
  result <- a$get_item("...")
  expect_equal(result$data[1, 1], 1)
  expect_equal(result$data[1, 2], 2)
  expect_equal(result$data[2, 1], 3)
  expect_equal(result$data[2, 2], 4)
})

test_that("V3 2d chunked ragged int16", {
  skip_if_not_installed("blosc")
  store <- DirectoryStore$new(v3_root)
  a <- ZarrArray$new(store, path = "2d.chunked.ragged.i2", read_only = TRUE)
  expect_equal(a$get_shape(), c(3L, 3L))
  result <- a$get_item("...")
  # Data written as [[1,2,3],[4,5,6],[7,8,9]] in C order.
  expect_equal(result$data[1, 1], 1)
  expect_equal(result$data[1, 3], 3)
  expect_equal(result$data[2, 1], 4)
  expect_equal(result$data[3, 3], 9)
})

# --- 3D arrays ---

test_that("V3 3d contiguous int16", {
  skip_if_not_installed("blosc")
  store <- DirectoryStore$new(v3_root)
  a <- ZarrArray$new(store, path = "3d.contiguous.i2", read_only = TRUE)
  expect_equal(a$get_shape(), c(3L, 3L, 3L))
  result <- a$get_item("...")
  # Data written as np.arange(27).reshape(3,3,3) in C order.
  # In C order: element[i,j,k] = i*9 + j*3 + k (0-based).
  # Check element access (R 1-based): data[i,j,k] = (i-1)*9 + (j-1)*3 + (k-1)
  expect_equal(result$data[1, 1, 1], 0)
  expect_equal(result$data[1, 1, 2], 1)
  expect_equal(result$data[1, 2, 1], 3)
  expect_equal(result$data[2, 1, 1], 9)
  expect_equal(result$data[3, 3, 3], 26)
  expect_equal(length(result$data), 27)
})

test_that("V3 3d chunked int16", {
  skip_if_not_installed("blosc")
  store <- DirectoryStore$new(v3_root)
  a <- ZarrArray$new(store, path = "3d.chunked.i2", read_only = TRUE)
  expect_equal(a$get_shape(), c(3L, 3L, 3L))
  result <- a$get_item("...")
  expect_equal(result$data[1, 1, 1], 0)
  expect_equal(result$data[2, 1, 1], 9)
  expect_equal(result$data[3, 3, 3], 26)
  expect_equal(length(result$data), 27)
})

test_that("V3 3d chunked mixed C-order int16", {
  skip_if_not_installed("blosc")
  store <- DirectoryStore$new(v3_root)
  a <- ZarrArray$new(store, path = "3d.chunked.mixed.i2.C", read_only = TRUE)
  expect_equal(a$get_shape(), c(3L, 3L, 3L))
  result <- a$get_item("...")
  expect_equal(result$data[1, 1, 1], 0)
  expect_equal(result$data[2, 1, 1], 9)
  expect_equal(result$data[3, 3, 3], 26)
  expect_equal(length(result$data), 27)
})

test_that("V3 3d chunked mixed F-order int16", {
  skip_if_not_installed("blosc")
  store <- DirectoryStore$new(v3_root)
  a <- ZarrArray$new(store, path = "3d.chunked.mixed.i2.F", read_only = TRUE)
  expect_equal(a$get_shape(), c(3L, 3L, 3L))
  result <- a$get_item("...")
  expect_equal(result$data[1, 1, 1], 0)
  expect_equal(result$data[2, 1, 1], 9)
  expect_equal(result$data[3, 3, 3], 26)
  expect_equal(length(result$data), 27)
})

# --- Auto-detection via zarr_open ---

test_that("zarr_open auto-detects V3 group and can read child array", {
  g <- zarr_open(v3_root, mode = "r")
  expect_s3_class(g, "ZarrGroup")
  arr <- g$get_item("1d.contiguous.raw.i2")
  expect_s3_class(arr, "ZarrArray")
  result <- arr$get_item("...")
  expect_equal(as.integer(result$data), c(1L, 2L, 3L, 4L))
})

# --- V3 open in write mode ---

test_that("V3 store opens in append mode without write guard message", {
  tdir2 <- tempfile("v3write_open")
  dir.create(tdir2)
  utils::unzip(v3_zip, exdir = tdir2)
  v3_root2 <- file.path(tdir2, "data.zarr")
  g <- zarr_open_group(v3_root2, mode = "a")
  expect_s3_class(g, "ZarrGroup")
  expect_false(g$get_read_only())
  unlink(tdir2, recursive = TRUE)
})

# Clean up
unlink(tdir, recursive = TRUE)

# =============================================================================
# V3 Write tests
# =============================================================================

# --- Basic V3 array creation and round-trip ---

test_that("zarr_create with zarr_format=3 creates zarr.json", {
  store <- MemoryStore$new()
  z <- zarr_create(shape = c(4L), dtype = "<i2", zarr_format = 3L, store = store)
  expect_true(store$contains_item("zarr.json"))
  expect_false(store$contains_item(".zarray"))
  raw_meta <- store$get_item("zarr.json")
  meta <- jsonlite::fromJSON(rawToChar(raw_meta))
  expect_equal(meta$zarr_format, 3)
  expect_equal(meta$node_type, "array")
  expect_equal(meta$data_type, "int16")
})

test_that("V3 array write and read round-trip (int32, no compression)", {
  store <- MemoryStore$new()
  z <- zarr_create(shape = c(6L), dtype = "<i4", zarr_format = 3L,
                   compressor = NA, store = store)
  data_in <- array(c(10L, 20L, 30L, 40L, 50L, 60L), dim = c(6))
  z$set_item("...", data_in)
  result <- z$get_item("...")
  expect_equal(as.integer(result$data), c(10L, 20L, 30L, 40L, 50L, 60L))
})

test_that("V3 array write and read round-trip (float64, zstd)", {
  store <- MemoryStore$new()
  z <- zarr_create(shape = c(5L), dtype = "<f8", zarr_format = 3L,
                   compressor = ZstdCodec$new(level = 1), store = store)
  data_in <- array(c(1.1, 2.2, 3.3, 4.4, 5.5), dim = c(5))
  z$set_item("...", data_in)
  result <- z$get_item("...")
  expect_equal(as.double(result$data), c(1.1, 2.2, 3.3, 4.4, 5.5), tolerance = 1e-10)
})

test_that("V3 array write and read round-trip (bool)", {
  store <- MemoryStore$new()
  z <- zarr_create(shape = c(4L), dtype = "|b1", zarr_format = 3L,
                   compressor = NA, store = store)
  data_in <- array(c(TRUE, FALSE, TRUE, FALSE), dim = c(4))
  z$set_item("...", data_in)
  result <- z$get_item("...")
  expect_equal(as.logical(result$data), c(TRUE, FALSE, TRUE, FALSE))
})

test_that("V3 array write and read round-trip (uint8)", {
  store <- MemoryStore$new()
  z <- zarr_create(shape = c(4L), dtype = "|u1", zarr_format = 3L,
                   compressor = NA, store = store)
  data_in <- array(c(0L, 127L, 128L, 255L), dim = c(4))
  z$set_item("...", data_in)
  result <- z$get_item("...")
  expect_equal(as.integer(result$data), c(0L, 127L, 128L, 255L))
})

test_that("V3 array write and read round-trip (float32)", {
  store <- MemoryStore$new()
  z <- zarr_create(shape = c(4L), dtype = "<f4", zarr_format = 3L,
                   compressor = NA, store = store)
  data_in <- array(c(-1.5, 0, 1.5, 3.14), dim = c(4))
  z$set_item("...", data_in)
  result <- z$get_item("...")
  expect_equal(as.double(result$data), c(-1.5, 0, 1.5, 3.14), tolerance = 1e-5)
})

test_that("V3 2D array write and read round-trip (gzip)", {
  store <- MemoryStore$new()
  z <- zarr_create(shape = c(3L, 4L), chunks = c(2L, 2L), dtype = "<i2",
                   zarr_format = 3L, compressor = GzipCodec$new(), store = store)
  data_in <- array(seq_len(12L), dim = c(3, 4))
  z$set_item("...", data_in)
  result <- z$get_item("...")
  expect_equal(result$data, data_in)
})

test_that("V3 3D array write and read round-trip", {
  store <- MemoryStore$new()
  z <- zarr_create(shape = c(2L, 3L, 4L), dtype = "<f8", zarr_format = 3L,
                   compressor = NA, store = store)
  data_in <- array(as.double(0:23), dim = c(2, 3, 4))
  z$set_item("...", data_in)
  result <- z$get_item("...")
  expect_equal(result$data, data_in)
})

# --- Metadata verification ---

test_that("V3 zarr.json has correct structure (bytes codec, no endian for single-byte)", {
  store <- MemoryStore$new()
  zarr_create(shape = c(4L), dtype = "|b1", zarr_format = 3L,
              compressor = NA, store = store)
  raw_meta <- store$get_item("zarr.json")
  meta <- jsonlite::fromJSON(rawToChar(raw_meta), simplifyVector = FALSE)
  expect_equal(meta$zarr_format, 3)
  expect_equal(meta$node_type, "array")
  expect_equal(meta$data_type, "bool")
  expect_equal(meta$chunk_grid$name, "regular")
  expect_equal(meta$chunk_key_encoding$name, "default")
  expect_equal(meta$chunk_key_encoding$configuration$separator, "/")
  # For bool (single-byte), bytes codec should have no endian config
  codecs <- meta$codecs
  bytes_codec <- codecs[[1]]
  expect_equal(bytes_codec$name, "bytes")
  expect_null(bytes_codec$configuration)
  expect_true(is.list(meta$attributes) && length(meta$attributes) == 0)
})

test_that("V3 zarr.json bytes codec has endian for multi-byte types", {
  store <- MemoryStore$new()
  zarr_create(shape = c(4L), dtype = "<f8", zarr_format = 3L,
              compressor = NA, store = store)
  raw_meta <- store$get_item("zarr.json")
  meta <- jsonlite::fromJSON(rawToChar(raw_meta), simplifyVector = FALSE)
  codecs <- meta$codecs
  bytes_codec <- codecs[[1]]
  expect_equal(bytes_codec$name, "bytes")
  expect_equal(bytes_codec$configuration$endian, "little")
})

test_that("V3 zarr.json codec pipeline includes compressor after bytes codec", {
  store <- MemoryStore$new()
  zarr_create(shape = c(10L), dtype = "<f8", zarr_format = 3L,
              compressor = ZstdCodec$new(level = 3), store = store)
  raw_meta <- store$get_item("zarr.json")
  meta <- jsonlite::fromJSON(rawToChar(raw_meta), simplifyVector = FALSE)
  expect_equal(length(meta$codecs), 2)
  expect_equal(meta$codecs[[1]]$name, "bytes")
  expect_equal(meta$codecs[[2]]$name, "zstd")
  expect_equal(meta$codecs[[2]]$configuration$level, 3)
})

test_that("V3 fill_value NaN is encoded as string", {
  store <- MemoryStore$new()
  zarr_create(shape = c(4L), dtype = "<f8", fill_value = NaN,
              zarr_format = 3L, compressor = NA, store = store)
  raw_meta <- store$get_item("zarr.json")
  meta <- jsonlite::fromJSON(rawToChar(raw_meta))
  expect_equal(meta$fill_value, "NaN")
})

test_that("V3 fill_value Inf is encoded as string", {
  store <- MemoryStore$new()
  zarr_create(shape = c(4L), dtype = "<f8", fill_value = Inf,
              zarr_format = 3L, compressor = NA, store = store)
  raw_meta <- store$get_item("zarr.json")
  meta <- jsonlite::fromJSON(rawToChar(raw_meta))
  expect_equal(meta$fill_value, "Infinity")
})

# --- Group write tests ---

test_that("zarr_create_group with zarr_format=3 creates zarr.json with node_type=group", {
  store <- MemoryStore$new()
  g <- zarr_create_group(store = store, zarr_format = 3L)
  expect_s3_class(g, "ZarrGroup")
  expect_true(store$contains_item("zarr.json"))
  expect_false(store$contains_item(".zgroup"))
  raw_meta <- store$get_item("zarr.json")
  meta <- jsonlite::fromJSON(rawToChar(raw_meta))
  expect_equal(meta$zarr_format, 3)
  expect_equal(meta$node_type, "group")
})

test_that("V3 group create_dataset creates V3 array", {
  store <- MemoryStore$new()
  g <- zarr_create_group(store = store, zarr_format = 3L)
  a <- g$create_dataset("data", shape = c(5L), dtype = "<i4", compressor = NA)
  expect_s3_class(a, "ZarrArray")
  expect_true(store$contains_item("data/zarr.json"))
  expect_false(store$contains_item("data/.zarray"))
})

test_that("V3 group create_group creates nested V3 group", {
  store <- MemoryStore$new()
  g <- zarr_create_group(store = store, zarr_format = 3L)
  sub <- g$create_group("sub")
  expect_s3_class(sub, "ZarrGroup")
  expect_true(store$contains_item("sub/zarr.json"))
  expect_false(store$contains_item("sub/.zgroup"))
})

# --- Attribute write tests ---

test_that("V3 array attributes are stored in zarr.json, not .zattrs", {
  store <- MemoryStore$new()
  z <- zarr_create(shape = c(4L), dtype = "<i4", zarr_format = 3L,
                   compressor = NA, store = store)
  z$get_attrs()$set_item("units", "meters")
  expect_false(store$contains_item(".zattrs"))
  raw_meta <- store$get_item("zarr.json")
  meta <- jsonlite::fromJSON(rawToChar(raw_meta), simplifyVector = FALSE)
  expect_equal(meta$attributes$units, "meters")
})

test_that("V3 array attributes round-trip: write, re-open, read back", {
  store <- MemoryStore$new()
  z <- zarr_create(shape = c(4L), dtype = "<i4", zarr_format = 3L,
                   compressor = NA, store = store)
  z$get_attrs()$set_item("description", "test array")
  z$get_attrs()$set_item("version", 42L)

  # Re-open
  z2 <- ZarrArray$new(store)
  attrs <- z2$get_attrs()$to_list()
  expect_equal(attrs$description, "test array")
  expect_equal(attrs$version, 42L)
})

test_that("V3 group attributes are in zarr.json", {
  store <- MemoryStore$new()
  g <- zarr_create_group(store = store, zarr_format = 3L)
  g$get_attrs()$set_item("generator", "pizzarr")
  raw_meta <- store$get_item("zarr.json")
  meta <- jsonlite::fromJSON(rawToChar(raw_meta), simplifyVector = FALSE)
  expect_equal(meta$attributes$generator, "pizzarr")
})

# --- Resize and append tests ---

test_that("V3 array resize updates zarr.json shape", {
  store <- MemoryStore$new()
  z <- zarr_create(shape = c(4L), dtype = "<i4", zarr_format = 3L,
                   compressor = NA, store = store)
  z$resize(c(8L))
  raw_meta <- store$get_item("zarr.json")
  meta <- jsonlite::fromJSON(rawToChar(raw_meta))
  expect_equal(meta$shape, 8)
})

test_that("V3 array append round-trip", {
  store <- MemoryStore$new()
  z <- zarr_create(shape = c(3L), dtype = "<f8", zarr_format = 3L,
                   compressor = NA, store = store)
  z$set_item("...", array(c(1.0, 2.0, 3.0), dim = 3))
  z$append(array(c(4.0, 5.0), dim = 2))
  expect_equal(z$get_shape(), 5)
  result <- z$get_item("...")
  expect_equal(as.double(result$data), c(1.0, 2.0, 3.0, 4.0, 5.0))
})

# --- DirectoryStore V3 write ---

test_that("V3 write to DirectoryStore creates correct file layout", {
  tdir3 <- tempfile("v3dirwrite")
  dir.create(tdir3)
  store <- DirectoryStore$new(tdir3)
  g <- zarr_create_group(store = store, zarr_format = 3L)
  a <- g$create_dataset("arr", shape = c(4L), dtype = "<i4", compressor = NA)
  data_in <- array(c(1L, 2L, 3L, 4L), dim = 4)
  a$set_item("...", data_in)

  # Check file layout: zarr.json at root and arr/
  expect_true(file.exists(file.path(tdir3, "zarr.json")))
  expect_true(file.exists(file.path(tdir3, "arr", "zarr.json")))
  expect_false(file.exists(file.path(tdir3, ".zgroup")))
  expect_false(file.exists(file.path(tdir3, "arr", ".zarray")))

  # Chunk key uses c/0 format (default V3 encoding with "/" separator)
  expect_true(file.exists(file.path(tdir3, "arr", "c", "0")))

  # Read back
  store2 <- DirectoryStore$new(tdir3)
  g2 <- zarr_open_group(store2, mode = "r")
  a2 <- g2$get_item("arr")
  result <- a2$get_item("...")
  expect_equal(as.integer(result$data), c(1L, 2L, 3L, 4L))

  unlink(tdir3, recursive = TRUE)
})

# =============================================================================
# zarr-python 3.x interop tests
# Fixture generated with zarr-python 3.1.5
# (see inst/extdata/fixtures/v3/generate-v3-zarr-python.py)
# =============================================================================

zp_zip <- system.file("extdata/fixtures/v3/zarr_python_v3.zarr.zip",
                       package = "pizzarr")
zp_dir <- tempfile("zp_v3test")
dir.create(zp_dir)
utils::unzip(zp_zip, exdir = zp_dir)
zp_root <- file.path(zp_dir, "zarr_python_v3.zarr")

test_that("zarr-python V3 store opens as group", {
  g <- zarr_open(zp_root, mode = "r")
  expect_s3_class(g, "ZarrGroup")
  attrs <- g$get_attrs()$to_list()
  expect_equal(attrs$generator, "zarr-python")
  expect_equal(attrs$zarr_python_version, "3.1.5")
})

test_that("zarr-python V3 int32 1d (gzip)", {
  g <- zarr_open(zp_root, mode = "r")
  a <- g$get_item("int32_1d")
  expect_equal(a$get_shape(), 6)
  expect_equal(as.integer(a$as.array()), c(10L, 20L, 30L, 40L, 50L, 60L))
})

test_that("zarr-python V3 float64 1d (zstd)", {
  g <- zarr_open(zp_root, mode = "r")
  a <- g$get_item("float64_1d")
  expect_equal(a$get_shape(), 5)
  expect_equal(as.double(a$as.array()), c(1.1, 2.2, 3.3, 4.4, 5.5))
})

test_that("zarr-python V3 bool 1d (uncompressed)", {
  g <- zarr_open(zp_root, mode = "r")
  a <- g$get_item("bool_1d")
  expect_equal(a$get_shape(), 4)
  expect_equal(as.logical(a$as.array()), c(TRUE, FALSE, TRUE, FALSE))
})

test_that("zarr-python V3 uint8 1d (uncompressed, chunked)", {
  g <- zarr_open(zp_root, mode = "r")
  a <- g$get_item("uint8_1d")
  expect_equal(a$get_shape(), 4)
  expect_equal(as.integer(a$as.array()), c(0L, 127L, 128L, 255L))
})

test_that("zarr-python V3 float32 1d (blosc)", {
  skip_if_not_installed("blosc")
  g <- zarr_open(zp_root, mode = "r")
  a <- g$get_item("float32_1d")
  expect_equal(a$get_shape(), 4)
  result <- as.double(a$as.array())
  expect_equal(result[1], -1.5)
  expect_equal(result[2], 0.0)
  expect_equal(result[3], 1.5)
  expect_equal(result[4], 3.14, tolerance = 1e-5)
})

test_that("zarr-python V3 int16 2d (gzip)", {
  g <- zarr_open(zp_root, mode = "r")
  a <- g$get_item("int16_2d")
  expect_equal(a$get_shape(), c(4, 3))
  m <- a$as.array()
  expect_equal(m[1, 1], 0)
  expect_equal(m[1, 3], 2)
  expect_equal(m[4, 3], 11)
})

test_that("zarr-python V3 float64 3d (zstd)", {
  g <- zarr_open(zp_root, mode = "r")
  a <- g$get_item("float64_3d")
  expect_equal(a$get_shape(), c(2, 3, 4))
  arr <- a$as.array()
  expect_equal(arr[1, 1, 1], 0)
  expect_equal(arr[2, 3, 4], 23)
})

test_that("zarr-python V3 fill value works", {
  g <- zarr_open(zp_root, mode = "r")
  a <- g$get_item("with_fill")
  expect_equal(a$get_shape(), 6)
  d <- as.double(a$as.array())
  expect_equal(d[1:3], c(1.0, 2.0, 3.0))
  expect_equal(d[4:6], c(-9999.0, -9999.0, -9999.0))
})

test_that("zarr-python V3 ragged 2d (gzip)", {
  g <- zarr_open(zp_root, mode = "r")
  a <- g$get_item("ragged_2d")
  expect_equal(a$get_shape(), c(5, 7))
  m <- a$as.array()
  expect_equal(m[1, 1], 0)
  expect_equal(m[5, 7], 34)
  expect_equal(length(m), 35)
})

test_that("zarr-python V3 nested group with attributes", {
  g <- zarr_open(zp_root, mode = "r")
  sub <- g$get_item("var_group")
  expect_s3_class(sub, "ZarrGroup")
  attrs <- sub$get_attrs()$to_list()
  expect_equal(attrs$description, "Group with multiple variables")
  expect_equal(attrs$source, "zarr-python 3.x test generation")
})

test_that("zarr-python V3 sub-group array with attributes (temperature)", {
  g <- zarr_open(zp_root, mode = "r")
  temp <- g$get_item("var_group")$get_item("temperature")
  expect_equal(temp$get_shape(), c(3, 4))
  ta <- temp$get_attrs()$to_list()
  expect_equal(ta$units, "K")
  expect_equal(ta$long_name, "Temperature")
  m <- temp$as.array()
  expect_equal(m[1, 1], 280.1, tolerance = 1e-1)
  expect_equal(m[3, 4], 292.2, tolerance = 1e-1)
})

test_that("zarr-python V3 sub-group array (pressure, gzip)", {
  g <- zarr_open(zp_root, mode = "r")
  pres <- g$get_item("var_group")$get_item("pressure")
  expect_equal(pres$get_shape(), c(3, 4))
  m <- pres$as.array()
  expect_equal(m[1, 1], 101325.0)
  expect_equal(m[3, 4], 101270.0)
})

# Clean up
unlink(zp_dir, recursive = TRUE)

# --- V3 write: flush_metadata_nosync ---

test_that("flush_metadata_nosync (V3) writes updated zarr.json", {
  d <- tempfile()
  a <- zarr_create(shape = c(4, 4), chunks = c(2, 2), store = d,
                   compressor = NA, zarr_format = 3L)
  a$resize(6, 4)
  b <- zarr_open_array(d, mode = "r")
  expect_equal(b$get_shape(), c(6, 4))
})

# --- V3 metadata loading ---

test_that("V3 load_metadata_v3_nosync populates shape/chunks/dtype/fill_value", {
  d <- tempfile()
  a <- zarr_create(shape = c(4, 4), chunks = c(2, 2), dtype = "<f8",
                   store = d, compressor = NA, fill_value = 0,
                   zarr_format = 3L)
  a$set_item("...", array(1:16 * 1.0, dim = c(4, 4)))
  b <- zarr_open_array(d, mode = "r")
  expect_equal(b$get_shape(), c(4, 4))
  expect_equal(b$get_chunks(), c(2, 2))
  result <- b$get_item("...")
  expect_equal(result$data, array(1:16 * 1.0, dim = c(4, 4)))
})

test_that("V3 chunk_key generates c/0/0 style keys for default encoding", {
  d <- tempfile()
  a <- zarr_create(shape = c(4, 4), chunks = c(2, 2), store = d,
                   compressor = NA, zarr_format = 3L)
  data <- array(1:16, dim = c(4, 4))
  a$set_item("...", data)
  result <- a$get_item("...")
  expect_equal(result$data, data)
})

test_that("V3 array with default chunk_key_encoding uses slash separator", {
  d <- tempfile()
  a <- zarr_create(
    shape = c(4), chunks = c(2), store = d,
    compressor = NA, zarr_format = 3L
  )
  data <- array(1:4, dim = c(4))
  a$set_item("...", data)
  b <- zarr_open_array(d, mode = "r")
  result <- b$get_item("...")
  expect_equal(as.integer(result$data), 1:4)
  expect_equal(b$get_dimension_separator(), "/")
})

# =============================================================================
# Gap 3: Additional V3 metadata loading tests
# =============================================================================

test_that("V3 load_metadata_v3_nosync handles different dtypes", {
  for (dt in c("<i2", "<i4", "<f4", "<f8", "|u1", "|b1")) {
    store <- MemoryStore$new()
    z <- zarr_create(shape = c(4L), dtype = dt, zarr_format = 3L,
                     compressor = NA, store = store)
    reopened <- ZarrArray$new(store, read_only = TRUE)
    expect_equal(reopened$get_shape(), 4)
  }
})

test_that("V3 load_metadata_v3_nosync with gzip codec", {
  store <- MemoryStore$new()
  z <- zarr_create(shape = c(8L), dtype = "<f8", zarr_format = 3L,
                   compressor = GzipCodec$new(), store = store)
  z$set_item("...", array(as.double(1:8), dim = 8))
  reopened <- ZarrArray$new(store, read_only = TRUE)
  result <- reopened$get_item("...")
  expect_equal(as.double(result$data), as.double(1:8))
})

test_that("V3 load_metadata_v3_nosync with zstd codec", {
  store <- MemoryStore$new()
  z <- zarr_create(shape = c(8L), dtype = "<i4", zarr_format = 3L,
                   compressor = ZstdCodec$new(level = 1), store = store)
  z$set_item("...", array(1:8, dim = 8))
  reopened <- ZarrArray$new(store, read_only = TRUE)
  result <- reopened$get_item("...")
  expect_equal(as.integer(result$data), 1:8)
})

test_that("V3 load_metadata_v3_nosync with 2D array", {
  store <- MemoryStore$new()
  z <- zarr_create(shape = c(3L, 4L), chunks = c(2L, 2L), dtype = "<f8",
                   zarr_format = 3L, compressor = NA, store = store)
  data <- array(as.double(1:12), dim = c(3, 4))
  z$set_item("...", data)
  reopened <- ZarrArray$new(store, read_only = TRUE)
  expect_equal(reopened$get_shape(), c(3, 4))
  expect_equal(reopened$get_chunks(), c(2, 2))
  result <- reopened$get_item("...")
  expect_equal(result$data, data)
})

test_that("V3 metadata contains correct fill_value for NaN", {
  store <- MemoryStore$new()
  z <- zarr_create(shape = c(4L), dtype = "<f8", fill_value = NaN,
                   zarr_format = 3L, compressor = NA, store = store)
  reopened <- ZarrArray$new(store, read_only = TRUE)
  # NaN fill_value round-trips through V3 encode/decode
  expect_true(is.nan(reopened$get_fill_value()))
  # Uninitialized chunks should contain NaN
  result <- reopened$get_item("...")
  expect_true(all(is.nan(result$data)))
})

# =============================================================================
# Gap 5: ZarrGroup$initialize V3 path
# =============================================================================

test_that("V3 group init populates metadata correctly", {
  store <- MemoryStore$new()
  g <- zarr_create_group(store = store, zarr_format = 3L)
  expect_s3_class(g, "ZarrGroup")
  expect_equal(g$get_name(), "/")
  expect_false(g$get_read_only())
})

test_that("V3 group re-open preserves attributes", {
  store <- MemoryStore$new()
  g <- zarr_create_group(store = store, zarr_format = 3L)
  g$get_attrs()$set_item("description", "test group")
  # Re-open
  g2 <- zarr_open_group(store, mode = "r")
  attrs <- g2$get_attrs()$to_list()
  expect_equal(attrs$description, "test group")
})

test_that("V3 group get_path returns stored path", {
  store <- MemoryStore$new()
  g <- zarr_create_group(store = store, zarr_format = 3L)
  expect_equal(g$get_path(), "")
})

test_that("V3 group get_meta returns metadata", {
  store <- MemoryStore$new()
  g <- zarr_create_group(store = store, zarr_format = 3L)
  meta <- g$get_meta()
  expect_false(is.null(meta))
})

test_that("V3 group with nested arrays and groups", {
  store <- MemoryStore$new()
  g <- zarr_create_group(store = store, zarr_format = 3L)
  sub <- g$create_group("sub")
  a <- sub$create_dataset("data", shape = c(4L), dtype = "<i4", compressor = NA)
  a$set_item("...", array(1:4, dim = 4))
  # Re-open and traverse
  g2 <- zarr_open_group(store, mode = "r")
  sub2 <- g2$get_item("sub")
  expect_s3_class(sub2, "ZarrGroup")
  a2 <- sub2$get_item("data")
  expect_equal(as.integer(a2$get_item("...")$data), 1:4)
})

test_that("V3 group GroupNotFoundError on empty store", {
  store <- MemoryStore$new()
  expect_error(zarr_open_group(store, mode = "r"), "GroupNotFoundError")
})

test_that("V3 vlen-utf8 string array loads correctly via load_metadata_v3_nosync", {
  v3_zip <- system.file("extdata/fixtures/v3/data.zarr.zip", package = "pizzarr")
  tdir <- tempfile("v3vlen")
  dir.create(tdir)
  utils::unzip(v3_zip, exdir = tdir)
  v3_root <- file.path(tdir, "data.zarr")
  store <- DirectoryStore$new(v3_root)
  if (contains_array(store, "1d.contiguous.raw.i2")) {
    a <- ZarrArray$new(store, path = "1d.contiguous.raw.i2", read_only = TRUE)
    expect_equal(a$get_shape(), 4)
  } else {
    skip("fixture not found")
  }
})

# --- dimension_names threading (PR 0B) ---

test_that("V3 zarr_create with dimension_names round-trips on reopen", {
  store <- MemoryStore$new()
  z <- zarr_create(shape = c(2L, 3L), dtype = "<f8", zarr_format = 3L,
                   compressor = NA, store = store,
                   dimension_names = c("x", "y"))
  expect_equal(z$get_dimension_names(), c("x", "y"))

  # Re-open from store and confirm dimension_names persisted in zarr.json.
  z2 <- ZarrArray$new(store)
  expect_equal(z2$get_dimension_names(), c("x", "y"))
})

test_that("V3 set_dimension_names attaches names post-creation and round-trips", {
  store <- MemoryStore$new()
  z <- zarr_create(shape = c(2L, 3L), dtype = "<f8", zarr_format = 3L,
                   compressor = NA, store = store)
  expect_null(z$get_dimension_names())

  z$set_dimension_names(c("a", "b"))
  expect_equal(z$get_dimension_names(), c("a", "b"))

  z2 <- ZarrArray$new(store)
  expect_equal(z2$get_dimension_names(), c("a", "b"))
})

test_that("set_dimension_names errors on V2 arrays", {
  store <- MemoryStore$new()
  z <- zarr_create(shape = c(2L, 3L), dtype = "<f8", zarr_format = 2L,
                   compressor = NA, store = store)
  expect_error(z$set_dimension_names(c("x", "y")), "DimensionNamesV2Error")
})

test_that("zarr_create errors when dimension_names length != length(shape)", {
  store <- MemoryStore$new()
  expect_error(
    zarr_create(shape = c(4L, 5L), dtype = "<f8", zarr_format = 3L,
                compressor = NA, store = store,
                dimension_names = c("x")),
    "DimensionNamesLengthError"
  )
})

test_that("zarr_create errors when dimension_names is passed with zarr_format = 2L", {
  store <- MemoryStore$new()
  expect_error(
    zarr_create(shape = c(4L, 5L), dtype = "<f8", zarr_format = 2L,
                compressor = NA, store = store,
                dimension_names = c("x", "y")),
    "DimensionNamesV2Error"
  )
})

test_that("V3 group create_dataset threads dimension_names through ...", {
  store <- MemoryStore$new()
  g <- zarr_create_group(store = store, zarr_format = 3L)
  a <- g$create_dataset("data", shape = c(2L, 3L, 4L), dtype = "<f8",
                        compressor = NA,
                        dimension_names = c("x", "y", "t"))
  expect_equal(a$get_dimension_names(), c("x", "y", "t"))
})

#' Convert a string into a character vector.
#'
#' @param s The string.
#' @return A vector where each element is an individual character.
#' @keywords internal
str_to_vec <- function(s) {
  return(stringr::str_split(s, pattern = "")[[1]])
}

#' @keywords internal
normalize_list_selection <- function(selection, shape, convert_integer_selection_to_slices = TRUE) {
  # Reference: https://github.com/gzuidhof/zarr.js/blob/292804/src/core/indexing.ts#L45
  selection <- replace_ellipsis(selection, shape)

  for(i in seq_along(selection)) {
    dim_sel <- selection[[i]]
    if(is_integer(dim_sel)){
      if(convert_integer_selection_to_slices) {
        selection[[i]] <- zb_slice(dim_sel, dim_sel + 1, 1)
      } else {
        selection[[i]] <- normalize_integer_selection(dim_sel, shape[i])
      }
    } else if(is_integer_vec(dim_sel)) {
      selection[[i]] <- sapply(dim_sel, normalize_integer_selection, dim_len = shape[i])
    } else if(is_slice(dim_sel)) {
      selection[[i]] <- dim_sel
    } else if(is_bool_vec(dim_sel)) {
      selection[[i]] <- selection[[i]]
    } else if(!is.null(dim_sel) && !is.environment(dim_sel) &&
              (is.na(dim_sel) || dim_sel == ":")) {
      selection[[i]] <- zb_slice(NA, NA, 1)
    }
  }
  return(selection)
}

#' @keywords internal
normalize_integer_selection <- function(dim_sel, dim_len) {
  # Reference: https://github.com/gzuidhof/zarr.js/blob/master/src/core/indexing.ts#L110

  # Normalize type to int
  dim_sel <- as_scalar(dim_sel)

  # TODO: do we really need this for R type array indexing ?
  # handle wraparound
  if(dim_sel < 0) {
    dim_sel <- dim_len + dim_sel
  }

  # TODO: do we need to normalize R indexing or Python indexing here ?
  # handle out of bounds
  if(dim_sel >= dim_len || dim_sel < 0) {
  # if(dim_sel > dim_len || dim_sel < 1) { # pre zb_int implementation
    stop('BoundsCheckError(dim_len)')
  }

  return(dim_sel)
}

#' @keywords internal
normalize_resize_args <- function(old_shape, args) {
  if(length(args) == 1) {
    new_shape <- args[1]
  } else {
    new_shape <- args
  }
  new_shape <- as.list(new_shape)
  if(length(new_shape) != length(old_shape)) {
    stop('new shape must have same number of dimensions')
  }
  return(new_shape)
}

#' @keywords internal
normalize_storage_path <- function(path) {
  # Reference: https://github.com/gzuidhof/zarr.js/blob/29280463ff2f275c31c1fa0f002daa947b8f09b2/src/util.ts#L32

  if(!is.na(path)) {
    # convert backslash to forward slash
    path <- gsub("\\\\", "/", path)
    # strip leading and trailing slashes
    path <- gsub("^/+|/+$", "", path)
    # collapse any repeated slashes
    path <- gsub("/+", "/", path)

    # don't allow path segments with just '.' or '..'
    if (nchar(path) > 0) {
      segments <- strsplit(path, "/", fixed = TRUE)[[1]]
      if (any(segments %in% c(".", ".."))) {
        stop("path containing '.' or '..' segment not allowed")
      }
    }
  } else {
    path <- ""
  }
  return(path)
}

#' @keywords internal
normalize_shape <- function(shape) {
  # Reference: https://github.com/gzuidhof/zarr.js/blob/292804/src/util.ts#L69
  if(!is.null(shape)) {
    shape <- ensure_integer_vec(shape)
    return(floor(shape))
  }
  return(shape)
}

#' @keywords internal
normalize_dtype <- function(dtype, object_codec = NA) {
  # Reference: https://github.com/zarr-developers/zarr-python/blob/5dd4a0e6cdc04c6413e14f57f61d389972ea937c/zarr/util.py#L152

  if(is_na(dtype)) {
    # np.dtype(None) returns 'float64'
    if(!is_na(object_codec)) {
      stop("expected object_codec to be NA due to NA dtype")
    }
    return(Dtype$new("<f8"))
  }

  # Construct Dtype instance.
  # convenience API for object arrays
  if("Dtype" %in% class(dtype)) {
    return(dtype)
  }
  
  if(is.character(dtype)) {
    # Filter list was NA but there could be non-NA object_codec parameter.
    return(Dtype$new(dtype, object_codec = object_codec))
  }

  stop("dtype must be NA, string/character vector, or Dtype instance")
}

#' @keywords internal
guess_chunks <- function(shape, typesize) {
  # Reference: https://github.com/zarr-developers/zarr-python/blob/5dd4a0e6cdc04c6413e14f57f61d389972ea937c/zarr/util.py#L69

  # Guess an appropriate chunk layout for an array, given its shape and
  # the size of each element in bytes.  Will allocate chunks only as large
  # as MAX_SIZE.  Chunks are generally close to some power-of-2 fraction of
  # each axis, slightly favoring bigger values for the last index.
  # Undocumented and subject to change without warning.

  CHUNK_BASE <- 256*1024  # Multiplier by which chunks are adjusted
  CHUNK_MIN <- 128*1024  # Soft lower limit (128k)
  CHUNK_MAX <- 64*1024*1024  # Hard upper limit

  ndims <- length(shape)
  # require chunks to have non-zero length for all dimensions
  chunks <- rep(1, ndims)
  for(i in seq_along(shape)) {
    if(shape[i] >= 1) {
      chunks[i] <- shape[i]
    }
  }

  # Determine the optimal chunk size in bytes using a PyTables expression.
  # This is kept as a float.
  dset_size <- prod(chunks) * typesize
  target_size <- CHUNK_BASE * (2 ** log10(dset_size / (1024. * 1024)))

  if (target_size > CHUNK_MAX) {
      target_size <- CHUNK_MAX
  } else if (target_size < CHUNK_MIN) {
      target_size <- CHUNK_MIN
  }

  idx <- 0
  while (TRUE) {
      # Repeatedly loop over the axes, dividing them by 2.  Stop when:
      # 1a. We're smaller than the target chunk size, OR
      # 1b. We're within 50% of the target chunk size, AND
      # 2. The chunk is smaller than the maximum chunk size

      chunk_bytes <- prod(chunks)*typesize

      if ((chunk_bytes < target_size || abs(chunk_bytes-target_size)/target_size < 0.5) && chunk_bytes < CHUNK_MAX) {
        break
      }

      if (prod(chunks) == 1) {
        break  # Element size larger than CHUNK_MAX
      }

      chunks[idx %% ndims + 1] <- ceiling(chunks[idx %% ndims + 1] / 2.0)
      idx <- idx + 1
  }

  return(as.integer(chunks))
}

#' @keywords internal
normalize_chunks <- function(chunks, shape, typesize) {
  # Reference: https://github.com/zarr-developers/zarr-python/blob/5dd4a0e6cdc04c6413e14f57f61d389972ea937c/zarr/util.py#L115

  # N.B., expect shape already normalized

  # handle auto-chunking
  if(is_na(chunks) || isTRUE(chunks)) {
    return(guess_chunks(shape, typesize))
  }

  # handle no chunking
  if (isFALSE(chunks)) {
    return(shape)
  }

  # handle 1D convenience form
  if(is_scalar(chunks)) {
    chunks <- rep(as.integer(chunks), length(shape))
  }

  # handle bad dimensionality
  if(length(chunks) > length(shape)) {
    stop("too many dimensions in chunks")
  }

  # handle underspecified chunks
  if(length(chunks) < length(shape)) {
    # assume chunks across remaining dimensions
    chunks <- c(chunks, shape[(length(chunks)+1):length(shape)])
  }

  # handle None or -1 in chunks
  if (-1 %in% chunks || NA %in% chunks) {
    for(i in seq_along(chunks)) {
      if(chunks[i] == -1 || is.na(chunks[i])) {
        chunks[i] <- shape[i]
      }
    }
  }

  if(!is.numeric(chunks)) {
    stop("chunks must be numeric")
  }

  return(chunks)
}

#' @keywords internal
normalize_store_arg <- function(store, storage_options=NA, mode=NA) {
  if(is.na(mode)) {
    mode <- "r"
  }

  if(is_na(store)) {
    return(MemoryStore$new())
  }

  if(is.character(store)) {
    if(grepl("^https?://", store)) {
      return(HttpStore$new(store))
    }
    if(grepl("://", store, fixed=TRUE) || grepl("::", store, fixed=TRUE)) {
      # TODO: return FSStore
    }
    if(endsWith(store, ".zip")) {
      # TODO: return ZipStore
      # return(ZipStore$new(store, mode=mode))
    }
    return(DirectoryStore$new(store))
  } else {
    if(!("Store" %in% class(store))) {
      stop("store must be NA, a string, or a Store instance")
    }
  }
  return(store)
}

#' @keywords internal
normalize_order <- function(order) {
  order <- toupper(order)
  if (!(order %in% c('C', 'F'))) {
    stop("order must be either 'C' or 'F'")
  }
  return(order)
}

#' @keywords internal
normalize_object_codec <- function(dtype, object_codec) {
  if (dtype$is_object) {
    if (is_na(object_codec)) {
      stop("missing object_codec for object array")
    }
  } else if (!is_na(object_codec)) {
    warning("an object_codec is only needed for object arrays")
  }
  return(object_codec)
}

#' @keywords internal
normalize_fill_value <- function(fill_value, dtype) {
  if (is_na(fill_value)) { # or dtype.hasobject:
        # no fill value
        # pass
  } else {
    rtype <- dtype$get_rtype()
    if (fill_value == 0) {
      if(is.logical(rtype)) {
        fill_value <- FALSE
      } else if(inherits(rtype, "integer64")) {
        fill_value <- bit64::as.integer64(0)
      } else if(is.integer(rtype)) {
        fill_value <- 0L
      } else if(is.double(rtype)) {
        fill_value <- 0.0
      }
    } else if(is.character(rtype)) {
      if(!is.character(fill_value)) {
        stop("fill_value {!r} is not valid for dtype {}; must be a 'unicode string'")
      }
    } else {
      if(is.logical(rtype)) {
        fill_value <- TRUE
      } else if(inherits(rtype, "integer64")) {
        fill_value <- bit64::as.integer64(fill_value)
      } else if(is.integer(rtype)) {
        fill_value <- as.integer(fill_value)
      } else if(is.double(rtype)) {
        fill_value <- as.double(fill_value)
      } else {
        stop("fill_value {!r} is not valid for dtype {}")
      }
    }
  }
  return(fill_value)
}

library(pizzarr)

test_that("normalize_storage_path changes backslashes to forward slashes", {
  res <- normalize_storage_path("test\\nested\\dirs")
  expect_equal(res, "test/nested/dirs")
})

test_that("normalize_storage_path without leading nor trailing slashes", {
  res <- normalize_storage_path("test")
  expect_equal(res, "test")
})

test_that("normalize_storage_path with leading slash", {
  res <- normalize_storage_path("/test")
  expect_equal(res, "test")
})

test_that("normalize_storage_path with trailing slash", {
  res <- normalize_storage_path("test/")
  expect_equal(res, "test")
})

test_that("normalize_storage_path with both leading and trailing slashes", {
  res <- normalize_storage_path("/test//")
  expect_equal(res, "test")
})

test_that("normalize_storage_path collapses repeated slashes", {
  res <- normalize_storage_path("/test//this")
  expect_equal(res, "test/this")
})

test_that("normalize_resize_args throws error when different number of dimensions", {
  f <- function() normalize_resize_args(list(1, 2, 3), list(4, 5))
  expect_error(f())
})

test_that("normalize_resize_args", {
  res <- normalize_resize_args(c(1, 2), c(2, 1))
  expect_equal(res, list(2, 1))
})
test_that("normalize_resize_args with int argument", {
  res <- normalize_resize_args(c(1), 2)
  expect_equal(res, list(2))
})
test_that("normalize_resize_args with int argument", {
  res <- normalize_resize_args(c(1), list(2))
  expect_equal(res, list(2))
})

test_that("normalize_integer_selection with valid input", {
  # Reference: https://github.com/gzuidhof/zarr.js/blob/292804/test/core/indexing.test.ts#L12
  res <- normalize_integer_selection(1, 100)
  expect_equal(res, as_scalar(1))
  res <- normalize_integer_selection(-1, 100)
  expect_equal(res, as_scalar(99))
})

test_that("normalize_integer_selection with invalid input", {
  # Reference: https://github.com/gzuidhof/zarr.js/blob/292804/test/core/indexing.test.ts#L12
  
  # TODO: should normalization be performed based  R indexing integers or python ?
  # Artur: I think it has to be in Python indexing since zero_based_to_one_based() is called later when R arrays are filled from chunk selection
  normalize_integer_selection(99, 100)
  f1 <- function() normalize_integer_selection(100, 100)
  expect_error(f1())
  f2 <- function() normalize_integer_selection(1000, 100)
  expect_error(f2())
  f3 <- function() normalize_integer_selection(-1000, 100)
  expect_error(f3())
})

test_that("normalize_list_selection handles slice objects without crashing", {
  # Regression: is.na() on Slice R6 objects causes

  # "invalid 'x' type in 'x || y'" in non-interactive contexts (e.g. knitr)
  s1 <- slice(1, 3)
  s2 <- slice(2, 5)
  shape <- c(10, 10)
  result <- normalize_list_selection(list(s1, s2), shape)
  expect_true(is_slice(result[[1]]))
  expect_true(is_slice(result[[2]]))
})

test_that("normalize_list_selection handles mix of slice and integer", {
  s1 <- slice(1, 3)
  shape <- c(10, 10)
  result <- normalize_list_selection(list(s1, 2L), shape)
  expect_true(is_slice(result[[1]]))
  expect_true(is_slice(result[[2]]))  # integer gets converted to slice
})

test_that("normalize_chunks", {
  res <- normalize_chunks(c(10), c(100), 1)
  expect_equal(res, c(10))

  res <- normalize_chunks(as_scalar(10), c(100), 1)
  expect_equal(res, c(10))

  res <- normalize_chunks(c(10, 10), c(100, 10), 1)
  expect_equal(res, c(10, 10))

  res <- normalize_chunks(as_scalar(10), c(100, 10), 1)
  expect_equal(res, c(10, 10))

  res <- normalize_chunks(c(10, NA), c(100, 10), 1)
  expect_equal(res, c(10, 10))

  res <- normalize_chunks(as_scalar(30), c(100, 20, 10), 1)
  expect_equal(res, c(30, 30, 30))

  res <- normalize_chunks(c(30), c(100, 20, 10), 1)
  expect_equal(res, c(30, 20, 10))

  res <- normalize_chunks(c(30, NA), c(100, 20, 10), 1)
  expect_equal(res, c(30, 20, 10))

  res <- normalize_chunks(c(30, NA, NA), c(100, 20, 10), 1)
  expect_equal(res, c(30, 20, 10))

  res <- normalize_chunks(c(30, 20, NA), c(100, 20, 10), 1)
  expect_equal(res, c(30, 20, 10))

  res <- normalize_chunks(c(30, 20, 10), c(100, 20, 10), 1)
  expect_equal(res, c(30, 20, 10))

  f1 <- function() normalize_chunks("foo", c(100), 1)
  expect_error(f1())

  f2 <- function() normalize_chunks(c(100, 10), c(100), 1)
  expect_error(f2())

  # test auto-chunking
  res <- normalize_chunks(NA, c(100), 1)
  expect_equal(res, c(100))

  res <- normalize_chunks(as_scalar(-1), c(100), 1)
  expect_equal(res, c(100))

  res <- normalize_chunks(c(30, -1, NA), c(100, 20, 10), 1)
  expect_equal(res, c(30, 20, 10))
})

# Reference: https://github.com/zarr-developers/zarr-python/blob/5dd4a0/zarr/hierarchy.py#L39

#' The Zarr Group class.
#' @title ZarrGroup Class
#' @docType class
#' @description
#' Instantiate a group from an initialized store.
#'
#' @format [R6::R6Class]
#' @rdname ZarrGroup
#' @export
ZarrGroup <- R6::R6Class("ZarrGroup",
  private = list(
    # store A MutableMapping providing the underlying storage for the group.
    #' @keywords internal
    store = NULL,
    # path Storage path for this group within the store.
    #' @keywords internal
    path = NULL,
    # read_only TRUE if modification operations are not permitted.
    #' @keywords internal
    read_only = NULL,
    # chunk_store Separate storage for array chunks. If not provided, the main store handles both chunks and metadata.
    #' @keywords internal
    chunk_store = NULL,
    # cache_attrs If TRUE (default), user attributes will be cached for faster access.
    #' @keywords internal
    cache_attrs = NULL,
    # synchronizer Object used to synchronize write access to the group and its arrays.
    #' @keywords internal
    synchronizer = NULL,
    # key_prefix Storage path prefix prepended to keys when accessing the store. Set to path + "/" or empty string for root.
    #' @keywords internal
    key_prefix = NULL,
    # meta Parsed group metadata from .zgroup (V2) or zarr.json (V3).
    #' @keywords internal
    meta = NULL,
    # attrs A MutableMapping containing user-defined attributes. Attribute values must be JSON serializable.
    #' @keywords internal
    attrs = NULL,
    # zarr_format Integer 2L or 3L. Detected from the store on initialization.
    #' @keywords internal
    zarr_format = NULL,
    item_path = function(item) {
      # Reference: https://github.com/zarr-developers/zarr-python/blob/5dd4a0e6cdc04c6413e14f57f61d389972ea937c/zarr/hierarchy.py#L302
      # Check if the first character is a forward slash
      is_absolute <- !is_na(item) && is.character(item) && substr(item, 1, 1) == "/"
      path <- normalize_storage_path(item)
      if(!is_absolute && !is_na(private$path)) {
        path <- paste0(private$key_prefix, path)
      }
      return(path)
    },
    create_group_nosync = function(name, overwrite = FALSE) {
      path <- private$item_path(name)

      # create terminal group (inherit parent's zarr_format)
      init_group(
        private$store,
        path = path,
        overwrite = overwrite,
        chunk_store = self$get_chunk_store(),
        zarr_format = if (!is.null(private$zarr_format)) private$zarr_format else 2L
      )

      # create group instance
      return(ZarrGroup$new(
        private$store,
        path = path,
        read_only = private$read_only,
        chunk_store = self$get_chunk_store(),
        synchronizer = private$synchronizer,
        cache_attrs = private$attrs$cache
      ))
    },
    print_tree = function(prefix, level, depth) {
      if (!is.na(level) && depth >= level) return(invisible(NULL))

      path <- if (is_na(private$path) || private$path == "") "" else private$path
      listdir_key <- if (path == "") NA else path
      entries <- tryCatch(private$store$listdir(listdir_key), error = function(e) character(0))
      # Skip metadata keys
      entries <- entries[!grepl("^\\.", entries) & entries != "zarr.json"]

      # Filter to zarr nodes (arrays and groups)
      children <- character(0)
      for (e in entries) {
        child_path <- if (path == "") e else paste0(path, "/", e)
        if (contains_array(private$store, child_path) ||
            contains_group(private$store, child_path)) {
          children <- c(children, e)
        }
      }

      n <- length(children)
      for (i in seq_along(children)) {
        is_last <- (i == n)
        connector <- if (is_last) "\u2514\u2500\u2500 " else "\u251C\u2500\u2500 "
        child_path <- if (path == "") children[i] else paste0(path, "/", children[i])

        if (contains_array(private$store, child_path)) {
          arr <- ZarrArray$new(private$store, path = child_path, read_only = TRUE)
          shape_str <- paste0("(", paste(arr$get_shape(), collapse = ", "), ")")
          dtype_str <- arr$get_dtype()$dtype
          cat(paste0(prefix, connector, children[i], " ", shape_str, " ", dtype_str, "\n"))
        } else {
          cat(paste0(prefix, connector, children[i], "\n"))
          child_group <- ZarrGroup$new(
            private$store, path = child_path, read_only = TRUE
          )
          child_prefix <- paste0(prefix, if (is_last) "    " else "\u2502   ")
          child_group$.__enclos_env__$private$print_tree(child_prefix, level, depth + 1L)
        }
      }
    },
    create_dataset_nosync = function(name, data = NA, ...) {
      path <- private$item_path(name)

      kwargs <- list(...)

      if(is.null(kwargs[['synchronizer']])) {
        synchronizer <- private$synchronizer
      } else {
        synchronizer <- kwargs[['synchronizer']]
      }
      if(is.null(kwargs[['cache_attrs']])) {
        cache_attrs <- private$attrs$cache
      } else {
        cache_attrs <- kwargs[['cache_attrs']]
      }

      # Inherit parent group's zarr_format unless explicitly overridden
      grp_zarr_format <- if (!is.null(private$zarr_format)) private$zarr_format else 2L

      if(is_na(data)) {
        a <- zarr_create(
          store = private$store,
          path = path,
          chunk_store = self$get_chunk_store(),
          synchronizer = synchronizer,
          cache_attrs = cache_attrs,
          zarr_format = grp_zarr_format,
          ...
        )
      } else {
        a <- zarr_create_array(
          store = private$store,
          path = path,
          data = data,
          chunk_store = self$get_chunk_store(),
          synchronizer = synchronizer,
          cache_attrs = cache_attrs,
          zarr_format = grp_zarr_format,
          ...
        )
      }
      return(a)
    }
  ),
  public = list(
    #' @description
    #' Create a new ZarrGroup instance.
    #' @param store ([Store])\cr
    #'   Group store, already initialized.
    #' @param path (`character(1)`)\cr
    #'   Storage path.
    #' @param read_only (`logical(1)`)\cr
    #'   Whether the group is read-only.
    #' @param chunk_store ([Store] or `NA`)\cr
    #'   Separate storage for chunks.
    #' @param synchronizer (`ANY` or `NA`)\cr
    #'   Object used to synchronize write access.
    #' @param cache_attrs (`logical(1)`)\cr
    #'   Whether to cache attributes.
    #' @return A `ZarrGroup` instance.
    initialize = function(store, path = NA, read_only = FALSE, 
                          chunk_store = NA, cache_attrs = TRUE, 
                          synchronizer = NA) {
      private$store <- store
      private$path <- normalize_storage_path(path)
      if(!is_na(private$path) && private$path != "") {
        private$key_prefix <- paste0(private$path, "/")
      } else {
        private$key_prefix <- ""
      }
      private$read_only <- read_only
      private$chunk_store <- chunk_store

      # guard conditions
      # contains_array() checks both V2 and V3 formats
      if(contains_array(store, path)) {
        stop("ContainsArrayError(path)")
      }

      # Detect V3 vs V2 and load metadata accordingly
      v3_key <- paste0(private$key_prefix, ZARR_JSON)
      if (store$contains_item(v3_key)) {
        # V3 group: read zarr.json, validate as group
        private$zarr_format <- 3L
        meta_bytes <- store$get_item(v3_key)
        meta3 <- Metadata3$new()
        private$meta <- meta3$decode_group_metadata(meta_bytes)

        # V3 attributes are embedded in zarr.json
        a_key <- paste0(private$key_prefix, ATTRS_KEY)
        private$attrs <- Attributes$new(store, key = a_key, read_only = read_only,
                                        cache = cache_attrs, synchronizer = synchronizer,
                                        zarr_format = 3L)
        if (!is.null(private$meta$attributes)) {
          private$attrs$set_cached_v3_attrs(private$meta$attributes)
        }
      } else {
        # V2 group: existing logic unchanged
        private$zarr_format <- 2L
        m_key <- paste0(private$key_prefix, GROUP_META_KEY)

        # use consolidated metadata if exists
        meta <- try_from_zmeta(m_key, store)

        if(!is.null(meta)) {
          private$meta <- meta
        } else {
          # initialize metadata
          meta_bytes <- tryCatch({
            meta_bytes <- store$get_item(m_key)
          }, error = function(cond) {
            if(is_key_error(cond)) {
              stop("GroupNotFoundError(path) in Group$new")
            } else {
              stop(cond$message)
            }
          })

          if(!is.null(meta_bytes))
            private$meta <- private$store$metadata_class$decode_group_metadata(meta_bytes)
        }

        # V2 attributes in separate .zattrs file
        a_key <- paste0(private$key_prefix, ATTRS_KEY)
        private$attrs <- Attributes$new(store, key = a_key, read_only = read_only,
                                        cache = cache_attrs, synchronizer = synchronizer,
                                        zarr_format = 2L)
      }
    },
    #' @description
    #' Get group store
    #' @return [Store].
    get_store = function() {
      return(private$store)
    },
    #' @description
    #' Get group path
    #' @return `character(1)`.
    get_path = function() {
      return(private$path)
    },
    #' @description
    #' Get group metadata
    #' @return `list()` or `NULL`.
    get_meta = function() {
      return(private$meta)
    },
    #' @description
    #' Get group name
    #' @return `character(1)`.
    get_name = function() {
      if(!is_na(private$path)) {
        name <- private$path
        if(substr(name, 1, 1) != "/") {
          name <- paste0("/", name)
        }
        return(name)
      }
      return("/")
    },
    #' @description
    #' Is store read only?
    #' @return `logical(1)`.
    get_read_only = function() {
      return(private$read_only)
    },
    #' @description
    #' Get group chunk store
    #' @return [Store].
    get_chunk_store = function() {
      if(is_na(private$chunk_store)) {
        return(private$store)
      }
      return(private$chunk_store)
    },
    #' @description
    #' Get group synchronizer
    #' @return Synchronizer object or `NA`.
    get_synchronizer = function() {
      return(private$synchronizer)
    },
    #' @description
    #' Get group attributes
    #' @return [Attributes].
    get_attrs = function() {
      return(private$attrs)
    },
    #' @description
    #' Test for group membership.
    #' @param item Character item to test for.
    #' @return `logical(1)`.
    contains_item = function(item) {
      path <- private$item_path(item)
      return(contains_array(private$store, path) || contains_group(private$store, path))
    },
    #' @description
    #' Obtain a group member.
    #' @param item Character item to retrieve.
    #' @return [ZarrArray] or [ZarrGroup].
    get_item = function(item) {
      if(is.null(item)) {
        #for case with no internet
        stop("item can not be null")
      }
      path <- private$item_path(item)
      if(contains_array(private$store, path)) {
        return(ZarrArray$new(
          private$store,
          path = path,
          read_only = private$read_only,
          chunk_store = self$get_chunk_store(),
          synchronizer = private$synchronizer,
          cache_attrs = private$attrs$cache
        ))
      } else if(contains_group(private$store, path)) {
        return(ZarrGroup$new(
          private$store,
          path = path,
          read_only = private$read_only,
          chunk_store = self$get_chunk_store(),
          synchronizer = private$synchronizer,
          cache_attrs = private$attrs$cache
        ))
      } else {
        stop("KeyError: item")
      }
    },
    #' @description
    #' create a group
    #' @param name Character group name.
    #' @param overwrite Logical overwrite.
    #' @return [ZarrGroup].
    create_group = function(name, overwrite = FALSE) {
      return(private$create_group_nosync(name, overwrite = overwrite))
    },
    #' @description
    #' Create a dataset (array) within this group.
    #' @param name Character dataset name.
    #' @param data Data to add to group.
    #' @param ... Extra arguments to pass to `zarr_create()` or `array()`.
    #' @return [ZarrArray].
    create_dataset = function(name, data = NA, ...) {
      return(private$create_dataset_nosync(name, data = data, ...))
    },
    #' @description
    #' Print a human-readable summary of the group.
    #' @param ... Ignored.
    #' @return `self` (invisibly).
    print = function(...) {
      nm <- self$get_name()
      store_type <- class(private$store)[1]
      zf <- if (!is.null(private$zarr_format)) private$zarr_format else 2L
      # Count zarr node members
      path <- if (is_na(private$path) || private$path == "") "" else private$path
      listdir_key <- if (path == "") NA else path
      entries <- tryCatch(private$store$listdir(listdir_key), error = function(e) character(0))
      entries <- entries[!grepl("^\\.", entries) & entries != "zarr.json"]
      n_members <- 0L
      for (e in entries) {
        child_path <- if (path == "") e else paste0(path, "/", e)
        if (contains_array(private$store, child_path) ||
            contains_group(private$store, child_path)) {
          n_members <- n_members + 1L
        }
      }
      cat(paste0("<ZarrGroup> ", nm, "\n"))
      cat(paste0("  Store type  : ", store_type, "\n"))
      cat(paste0("  Zarr format : ", zf, "\n"))
      cat(paste0("  Read-only   : ", private$read_only, "\n"))
      cat(paste0("  No. members : ", n_members, "\n"))
      invisible(self)
    },
    #' @description
    #' Print an ASCII tree showing the group hierarchy.
    #' @param level Maximum depth to display. `NA` (default) shows all levels.
    #' @return `self` (invisibly).
    tree = function(level = NA) {
      nm <- self$get_name()
      cat(nm, "\n")
      private$print_tree("", level, 0L)
      invisible(self)
    }
    # TODO: convenience functions like zeros, ones, empty, full, ...
  )
)

library(pizzarr)


test_that("create group", {
    z <- zarr_create_group()

    name <- z$get_name()

    expect_equal(name, "/")
})

test_that("create nested groups", {
    g1 <- zarr_create_group()
    g2 <- g1$create_group("foo")
    g3 <- g2$create_group("bar")

    name <- g3$get_name()
    expect_equal(name, "/foo/bar")
})

test_that("create array within a nested group", {
    g1 <- zarr_create_group()
    g2 <- g1$create_group("foo")
    g3 <- g2$create_group("bar")

    data <- array(data=1:10, dim=c(2, 5))
    a <- g3$create_dataset("baz", data=data, shape=dim(data))

    expect_equal(a$get_shape(), c(2, 5))
    expect_equal(a$get_name(), "/foo/bar/baz")

    zb_sel <- a$get_item(list(zb_slice(0, 2), zb_slice(0, 2)))
    ob_sel <- a$get_item(list(slice(1, 2), slice(1, 2)))

    expect_equal(zb_sel$data, ob_sel$data)
})

test_that("create group with overwrite = TRUE", {
    store <- MemoryStore$new()
    z <- zarr_create_group(store = store, path = "foo")

    z2 <- zarr_create_group(store = store, path = "foo", overwrite = TRUE)

    name <- z$get_name()
    expect_equal(name, "/foo")

    name2 <- z2$get_name()
    expect_equal(name2, "/foo")
})


test_that("open group twice with MemoryStore", {
    store <- MemoryStore$new()
    z <- zarr_open_group(store = store, path = "foo")

    z2 <- zarr_open_group(store = store, path = "foo")

    name <- z$get_name()
    expect_equal(name, "/foo")

    name2 <- z2$get_name()
    expect_equal(name2, "/foo")
})

test_that("open group twice with DirectoryStore", {
    store <- DirectoryStore$new(tempdir())
    z <- zarr_open_group(store = store, path = "foo")

    z2 <- zarr_open_group(store = store, path = "foo")

    name <- z$get_name()
    expect_equal(name, "/foo")

    name2 <- z2$get_name()
    expect_equal(name2, "/foo")
})

# --- contains_item tests ---

test_that("ZarrGroup contains_item finds arrays and subgroups", {
  store <- MemoryStore$new()
  g <- zarr_open_group(store)
  g$create_dataset("arr1", data = array(1:4, dim = c(4)), shape = c(4))
  g$create_group("subgroup")

  expect_true(g$contains_item("arr1"))
  expect_true(g$contains_item("subgroup"))
  expect_false(g$contains_item("nonexistent"))
})

# --- Mixed hierarchy navigation ---

test_that("Group get_item returns correct type for arrays vs groups", {
  store <- MemoryStore$new()
  g <- zarr_open_group(store)
  g$create_dataset("myarray", data = array(1:4, dim = c(4)), shape = c(4))
  g$create_group("mygroup")

  item_a <- g$get_item("myarray")
  expect_s3_class(item_a, "ZarrArray")

  item_g <- g$get_item("mygroup")
  expect_s3_class(item_g, "ZarrGroup")
})

test_that("Group get_item on missing key raises error", {
  store <- MemoryStore$new()
  g <- zarr_open_group(store)
  expect_error(g$get_item("nonexistent"))
})

# --- Group attributes ---

test_that("Group attributes set and get", {
  store <- MemoryStore$new()
  g <- zarr_open_group(store)
  g$get_attrs()$set_item("mykey", "myvalue")
  expect_equal(g$get_attrs()$get_item("mykey"), "myvalue")
})

test_that("Group attributes round-trip through reopen", {
  store <- MemoryStore$new()
  g <- zarr_open_group(store)
  g$get_attrs()$set_item("mykey", "myvalue")

  g2 <- zarr_open_group(store, mode = "r")
  expect_equal(g2$get_attrs()$get_item("mykey"), "myvalue")
})

# --- Group properties ---

test_that("Group get_store returns the store", {
  store <- MemoryStore$new()
  g <- zarr_open_group(store)
  expect_identical(g$get_store(), store)
})

test_that("Group get_read_only reflects mode", {
  store <- MemoryStore$new()
  g <- zarr_open_group(store)
  expect_false(g$get_read_only())
})

# zarrs dispatch helpers
#
# R-side functions for deciding whether to use the zarrs fast path
# and converting pizzarr selection objects to zarrs ranges.

#' Check whether the zarrs path can handle this selection
#'
#' Returns TRUE for BasicIndexer with contiguous SliceDimIndexer
#' (step==1) and/or IntDimIndexer on every dimension. Step>1 slices,
#' IntArrayDimIndexer (fancy indexing), and OrthogonalIndexer fall
#' through to R-native.
#'
#' @param indexer A BasicIndexer or OrthogonalIndexer object.
#' @param store A Store object.
#' @return Logical scalar.
#' @keywords internal
can_use_zarrs <- function(indexer, store) {
  if (!.pizzarr_env$zarrs_available) return(FALSE)
  store_id <- store$get_store_identifier()
  if (is.null(store_id)) return(FALSE)  # MemoryStore

  # Check cloud store features are compiled
  features <- pizzarr_compiled_features()
  if (inherits(store, "S3Store") && !("s3" %in% features)) return(FALSE)
  if (inherits(store, "GcsStore") && !("gcs" %in% features)) return(FALSE)

  # BasicIndexer with contiguous SliceDimIndexer or IntDimIndexer on every dim
  if (!inherits(indexer, "BasicIndexer")) return(FALSE)
  all(vapply(indexer$dim_indexers, function(di) {
    if (inherits(di, "SliceDimIndexer")) {
      # Reject step>1 — zarrs retrieves contiguous blocks only
      di$step == 1L
    } else {
      inherits(di, "IntDimIndexer")
    }
  }, logical(1)))
}

#' Check whether the zarrs write path can handle this selection
#'
#' Same eligibility as read. Separate function for future extensibility
#' (e.g., read-only remote stores).
#'
#' @param indexer A BasicIndexer or OrthogonalIndexer object.
#' @param store A Store object.
#' @return Logical scalar.
#' @keywords internal
can_use_zarrs_write <- function(indexer, store) {
  # Remote stores are read-only for now
  if (inherits(store, "HttpStore")) return(FALSE)
  if (inherits(store, "S3Store")) return(FALSE)
  if (inherits(store, "GcsStore")) return(FALSE)
  can_use_zarrs(indexer, store)
}

#' Check whether the zarrs create path can handle this array
#'
#' Returns TRUE when zarrs is available, the store is writable and
#' has a store identifier, and the dtype is numeric/logical (not
#' object or string).
#'
#' @param store A Store object.
#' @param dtype A Dtype object.
#' @return Logical scalar.
#' @keywords internal
can_use_zarrs_create <- function(store, dtype) {
  if (!.pizzarr_env$zarrs_available) return(FALSE)
  store_id <- store$get_store_identifier()
  if (is.null(store_id)) return(FALSE)  # MemoryStore
  if (inherits(store, "HttpStore")) return(FALSE)  # read-only
  if (inherits(store, "S3Store")) return(FALSE)  # read-only for now
  if (inherits(store, "GcsStore")) return(FALSE)  # read-only for now
  if (dtype$is_object) return(FALSE)
  if (dtype$basic_type %in% c("S", "U")) return(FALSE)
  TRUE
}

#' Map an R compressor object to a zarrs codec preset string
#'
#' Returns one of \code{"none"}, \code{"gzip"}, \code{"blosc"},
#' \code{"zstd"}, or \code{NA_character_} if the compressor is not
#' recognised (signalling fallback to R-native).
#'
#' @param compressor A codec object or NA.
#' @param compressor_config A config list or NA.
#' @return Character scalar.
#' @keywords internal
compressor_to_preset <- function(compressor, compressor_config) {
  if (is_na(compressor) || is.null(compressor)) return("none")
  if (is.list(compressor_config)) {
    id <- compressor_config$id
    if (id %in% c("zlib", "gzip")) return("gzip")
    if (identical(id, "blosc")) return("blosc")
    if (identical(id, "zstd")) return("zstd")
  }
  NA_character_
}

#' Convert pizzarr indexer to zarrs ranges
#'
#' @param indexer A BasicIndexer with SliceDimIndexer or IntDimIndexer dims.
#' @return List of integer vectors.
#' @keywords internal
selection_to_ranges <- function(indexer) {
  lapply(indexer$dim_indexers, function(di) {
    if (inherits(di, "IntDimIndexer")) {
      # Scalar selection: single-element range [i, i+1)
      as.integer(c(di$dim_sel, di$dim_sel + 1L))
    } else {
      as.integer(c(di$start, di$stop))
    }
  })
}

# Tests for zarrs dispatch helpers (R/zarrs-dispatch.R)

# --- can_use_zarrs_create() ---

test_that("can_use_zarrs_create returns FALSE when zarrs unavailable", {
  orig <- .pizzarr_env$zarrs_available
  on.exit(.pizzarr_env$zarrs_available <- orig)
  .pizzarr_env$zarrs_available <- FALSE

  d <- tempfile()
  dir.create(d)
  on.exit(unlink(d, recursive = TRUE), add = TRUE)
  ds <- DirectoryStore$new(d)
  dtype <- Dtype$new("<f8")
  expect_false(can_use_zarrs_create(ds, dtype))
})

test_that("can_use_zarrs_create returns FALSE for MemoryStore", {
  ms <- MemoryStore$new()
  dtype <- Dtype$new("<f8")
  expect_false(can_use_zarrs_create(ms, dtype))
})

test_that("can_use_zarrs_create returns FALSE for HttpStore", {
  hs <- HttpStore$new("http://example.com")
  dtype <- Dtype$new("<f8")
  expect_false(can_use_zarrs_create(hs, dtype))
})

test_that("can_use_zarrs_create returns FALSE for S3Store", {
  s3 <- S3Store$new("s3://bucket/prefix")
  dtype <- Dtype$new("<f8")
  expect_false(can_use_zarrs_create(s3, dtype))
})

test_that("can_use_zarrs_create returns FALSE for GcsStore", {
  gcs <- GcsStore$new("gs://bucket/prefix")
  dtype <- Dtype$new("<f8")
  expect_false(can_use_zarrs_create(gcs, dtype))
})

test_that("can_use_zarrs_create returns FALSE for object dtype", {
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")
  d <- tempfile()
  dir.create(d)
  on.exit(unlink(d, recursive = TRUE))
  ds <- DirectoryStore$new(d)
  dtype <- Dtype$new("|O")
  expect_false(can_use_zarrs_create(ds, dtype))
})

test_that("can_use_zarrs_create returns FALSE for string dtype", {
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")
  d <- tempfile()
  dir.create(d)
  on.exit(unlink(d, recursive = TRUE))
  ds <- DirectoryStore$new(d)
  dtype <- Dtype$new("<U10")
  expect_false(can_use_zarrs_create(ds, dtype))
})

test_that("can_use_zarrs_create returns TRUE for numeric DirectoryStore", {
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")
  d <- tempfile()
  dir.create(d)
  on.exit(unlink(d, recursive = TRUE))
  ds <- DirectoryStore$new(d)
  dtype <- Dtype$new("<f8")
  expect_true(can_use_zarrs_create(ds, dtype))
})

# --- compressor_to_preset() ---

test_that("compressor_to_preset returns 'none' for NA", {
  expect_equal(compressor_to_preset(NA, NA), "none")
})

test_that("compressor_to_preset returns 'none' for NULL", {
  expect_equal(compressor_to_preset(NULL, NULL), "none")
})

test_that("compressor_to_preset maps zlib to gzip", {
  expect_equal(compressor_to_preset("x", list(id = "zlib")), "gzip")
})

test_that("compressor_to_preset maps gzip to gzip", {
  expect_equal(compressor_to_preset("x", list(id = "gzip")), "gzip")
})

test_that("compressor_to_preset maps blosc to blosc", {
  expect_equal(compressor_to_preset("x", list(id = "blosc")), "blosc")
})

test_that("compressor_to_preset maps zstd to zstd", {
  expect_equal(compressor_to_preset("x", list(id = "zstd")), "zstd")
})

test_that("compressor_to_preset returns NA for unknown compressor", {
  result <- compressor_to_preset("x", list(id = "lz4"))
  expect_true(is.na(result))
})

# --- can_use_zarrs with step > 1 ---

test_that("can_use_zarrs returns FALSE for step > 1 slice", {
  skip_if(!.pizzarr_env$zarrs_available, "zarrs backend not available")
  d <- tempfile()
  on.exit(unlink(d, recursive = TRUE))
  z <- zarr_create(store = d, shape = 10L, chunks = 5L, dtype = "<f8")
  ds <- DirectoryStore$new(d)
  indexer <- BasicIndexer$new(list(slice(1L, 10L, 2L)), z)
  expect_false(can_use_zarrs(indexer, ds))
})

# transforming filters to be passed to ZarrArray$get_orthogonal_selection()
# 
# a:b => slice(a,b)
# seq(from, to, by) => slice(start, stop, step) ? for now indices of seq(from, to, by) are passed to get_orthogonal_selection (check below, TODO)
# c(a,b,c) => c(a,b,c), combine elements are passed as indices or boolean
# empty dimension => return everything
# 
manage_filters <- function(filters) {
  lapply(filters, function(x) {
    # Proceed based on type of filter
    if(typeof(x) == "symbol") {
      # When empty dimension, return everything
      if(x == "") {
        return(NULL)
      } else {
        # TODO: is eval() always the solution here ?, e.g. ind <- c(TRUE, FALSE, TRUE, TRUE) then eval(ind) if typeof(ind) == "symbol"
        return(eval(x))
      }
    } else if(typeof(x) == "double") {
      # Return single value for dimension
      return(slice(x, x))
    } else if(typeof(x) == "language") {
      x <- as.list(x)
      # Return a range (supplied via : or seq())
      if(x[[1]] == ":") {
        return(slice(x[[2]], x[[3]]))
      } else if(x[[1]] == "seq") {
        # TODO: do we need slicing for this case ? otherwise implement slice(start, stop, step)
        arg_names <- names(x)
        from <- ifelse("from" %in% arg_names, x[[which("from" == arg_names)]], x[[2]])
        to <- ifelse("to" %in% arg_names, x[[which("to" == arg_names)]], x[[3]])
        if(length(x) > 3) {
          by <- ifelse("by" %in% arg_names, x[[which("by" == arg_names)]], x[[4]])
          return(int(seq(from, to, by)))
        } else {
          by <- NA
          return(int(seq(from, to)))
        }
        return(int(seq(from, to, by)))
      } else if(x[[1]] == "c") {
        # return elements of the combine function as indices
        check_func <- sapply(x, function(y) {
          !is.function(eval(y))
        })
        x <- x[check_func]
        
        # correct for integer or boolean vectors
        x <- sapply(x, function(y) {
          if(is.numeric(y))
            return(int(floor(y))) 
          if(is_bool(y))
            return(y)
        })
        return(x)
      } else {
        stop("Unsupported filter '", as.character(x), "' supplied")
      }
    } else {
      stop("Unsupported filter '", as.character(x), "' supplied")
    }
  })
}

#!/usr/bin/env bash
# Build a CRAN-ready tarball without the Rust src/ directory.
#
# The CRAN tier ships pure R. This script copies the package tree to a
# temp dir, invokes tools/cran-strip.sh to remove Rust-related files,
# and runs R CMD build on the clean copy.
#
# Usage:
#   bash tools/cran-build.sh
#
# Output:
#   pizzarr_<version>.tar.gz in the current directory

set -euo pipefail

PKG_DIR="$(cd "$(dirname "$0")/.." && pwd)"
PKG_NAME="$(grep '^Package:' "$PKG_DIR/DESCRIPTION" | sed 's/Package: *//')"
PKG_VERSION="$(grep '^Version:' "$PKG_DIR/DESCRIPTION" | sed 's/Version: *//')"

TMPDIR="$(mktemp -d)"
trap 'rm -rf "$TMPDIR"' EXIT

echo "=== Building CRAN tarball for ${PKG_NAME} ${PKG_VERSION} ==="
echo "    Source: $PKG_DIR"
echo "    Temp:   $TMPDIR"

# Copy package tree, skipping .git and the Rust target dir (can be >1 GB).
# We delete src/ entirely below, so skip the Rust target dir now.
mkdir -p "$TMPDIR/$PKG_NAME"
for item in "$PKG_DIR"/*; do
  base="$(basename "$item")"
  case "$base" in
    .git) continue ;;
    src)
      mkdir -p "$TMPDIR/$PKG_NAME/src"
      for sub in "$item"/*; do
        subbase="$(basename "$sub")"
        if [ "$subbase" = "rust" ]; then
          mkdir -p "$TMPDIR/$PKG_NAME/src/rust"
          for rsub in "$sub"/*; do
            rsubbase="$(basename "$rsub")"
            [ "$rsubbase" = "target" ] && continue
            cp -r "$rsub" "$TMPDIR/$PKG_NAME/src/rust/"
          done
        else
          cp -r "$sub" "$TMPDIR/$PKG_NAME/src/"
        fi
      done
      ;;
    *) cp -r "$item" "$TMPDIR/$PKG_NAME/" ;;
  esac
done
# Copy dotfiles
for item in "$PKG_DIR"/.[!.]*; do
  base="$(basename "$item")"
  [ "$base" = ".git" ] && continue
  cp -r "$item" "$TMPDIR/$PKG_NAME/" 2>/dev/null || true
done

# Strip Rust-related files from the tmp copy.
bash "$PKG_DIR/tools/cran-strip.sh" "$TMPDIR/$PKG_NAME"

echo "=== Running R CMD build ==="
cd "$TMPDIR"

# Find R on PATH or use R_HOME
if command -v R &> /dev/null; then
  R_CMD="R"
elif [ -n "${R_HOME:-}" ]; then
  R_CMD="${R_HOME}/bin/R"
else
  # Common Windows location
  R_CMD="$(ls -d /c/Users/*/AppData/Local/Programs/R/R-*/bin/R.exe 2>/dev/null | tail -1)"
  if [ -z "$R_CMD" ]; then
    echo "ERROR: R not found. Set R_HOME or add R to PATH."
    exit 1
  fi
fi

"$R_CMD" CMD build "$PKG_NAME"

# Copy tarball to original directory
cp "${PKG_NAME}_${PKG_VERSION}.tar.gz" "$PKG_DIR/"

echo "=== Done: ${PKG_NAME}_${PKG_VERSION}.tar.gz ==="

#!/usr/bin/env bash
# Strip Rust-related files from a pizzarr source tree in place.
#
# The CRAN tier ships pure R. This script removes src/, the configure
# scripts, extendr machinery, and related metadata so what's left is
# equivalent to what would be shipped to CRAN.
#
# Used by:
#   tools/cran-build.sh        — on a tmp copy, before R CMD build
#   .github/workflows/rhub.yaml — on the GHA checkout, before run-check
#
# Usage:
#   bash tools/cran-strip.sh [pkg-dir]
#
# Default pkg-dir is $PWD. The script expects DESCRIPTION and NAMESPACE
# in pkg-dir. Uses only POSIX grep/awk — no sed -i portability traps.

set -euo pipefail

PKG_DIR="${1:-$PWD}"
cd "$PKG_DIR"

if [ ! -f DESCRIPTION ] || [ ! -f NAMESPACE ]; then
  echo "ERROR: $PKG_DIR is not an R package root (missing DESCRIPTION or NAMESPACE)" >&2
  exit 1
fi

# Remove Rust/extendr source and build hooks.
rm -rf src/
rm -f configure configure.win
rm -f cleanup cleanup.win
rm -f tools/config.R tools/msrv.R

# DESCRIPTION: drop the multi-line SystemRequirements field and the
# Config/rextendr/version line.
awk '
  /^SystemRequirements:/ { skip = 1; next }
  skip && /^[[:space:]]/ { next }
  { skip = 0; print }
' DESCRIPTION > DESCRIPTION.tmp
mv DESCRIPTION.tmp DESCRIPTION

grep -v '^Config/rextendr/version:' DESCRIPTION > DESCRIPTION.tmp
mv DESCRIPTION.tmp DESCRIPTION

# Replace extendr-wrappers.R with a minimal stub. No DLL exists on the
# CRAN tier, so the only surviving export is zarrs_compiled_features(),
# which errors; is_zarrs_available() catches that and returns FALSE.
if [ -f R/extendr-wrappers.R ]; then
  cat > R/extendr-wrappers.R << 'WRAPPER_EOF'
# Generated by extendr: Do not edit by hand
# Stub for CRAN tier (no Rust shared library)

# nolint start

#' @usage NULL
NULL

#' Return compiled zarrs feature flags as a character vector.
#'
#' On the CRAN tier (pure R), this function always errors. The error
#' is caught by \code{is_zarrs_available()}, which returns \code{FALSE}.
#' @export
zarrs_compiled_features <- function() {
  stop("zarrs backend not compiled (pure R install)")
}

# nolint end
WRAPPER_EOF
fi

# NAMESPACE: drop useDynLib, drop exports of Rust-only zarrs_* functions.
# zarrs_compiled_features is the only zarrs_ export that survives.
grep -v '^useDynLib' NAMESPACE > NAMESPACE.tmp
mv NAMESPACE.tmp NAMESPACE

awk '
  /^export\(zarrs_/ && !/zarrs_compiled_features/ { next }
  { print }
' NAMESPACE > NAMESPACE.tmp
mv NAMESPACE.tmp NAMESPACE

# Remove Rd pages for Rust-only functions.
rm -f man/zarrs_close_store.Rd man/zarrs_create_array.Rd \
      man/zarrs_get_subset.Rd man/zarrs_node_exists.Rd \
      man/zarrs_open_array_metadata.Rd man/zarrs_runtime_info.Rd \
      man/zarrs_set_codec_concurrent_target.Rd \
      man/zarrs_set_http_batch_range_requests.Rd \
      man/zarrs_set_nthreads.Rd man/zarrs_set_subset.Rd

# Include vignettes in the shipped tarball. They are .Rbuildignored on
# the r-universe tier, so strip that line for the CRAN tier.
if [ -f .Rbuildignore ]; then
  grep -v '^\^vignettes\$' .Rbuildignore > .Rbuildignore.tmp
  mv .Rbuildignore.tmp .Rbuildignore
fi

# Drop tools/ if now empty.
rmdir tools 2>/dev/null || true

---
title: "Basics"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Basics}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
  knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  out.width = "100%"
)
sample_dir <- tools::R_user_dir("pizzarr")
clean <- !dir.exists(sample_dir)
```

## Load the package

```{r}
library(pizzarr)
```

## Create an empty ZarrArray

```{r}
(a <- array(data=1:20, dim=c(2, 10)))

z <- zarr_create_empty(shape=dim(a), dtype="<f4")
```

## Create a ZarrArray based on a base R array

```{r}
z <- zarr_create_array(data = a, shape=dim(a), dtype="<f4", fill_value=NA)

# R-like one-based slicing
s1 <- z$get_item(list(slice(1, 2), slice(1, 5)))

print(s1$data)

# Python-like zero-based slicing
s2 <- z$get_item(list(zb_slice(0, 2), zb_slice(0, 5)))

print(s2$data)

```

## Create nested ZarrGroups and ZarrArrays

```{r}
g1 <- zarr_create_group()
g2 <- g1$create_group("foo")
g3 <- g2$create_group("bar")

data <- array(data=1:10, dim=c(2, 5))
a <- g3$create_dataset("baz", data=data, shape=dim(data))

print(a$get_name())

print(a$get_shape())
```

## List arrays in a zarr root group

```{r}
root <- pizzarr_sample("fixtures/v2/data.zarr")

z <- zarr_open(root)

class(z)

store <- z$get_store()

class(store)

print(store$listdir())
```

## Open a ZarrArray from a DirectoryStore (convenience)

```{r}
root <- pizzarr_sample("fixtures/v2/data.zarr")

g <- zarr_open_group(root)
a <- g$get_item("1d.contiguous.lz4.i2")

print(a$get_shape())
```

## Open a ZarrArray from a DirectoryStore

```{r}
root <- pizzarr_sample("fixtures/v2/data.zarr")
    
store <- DirectoryStore$new(root)
g <- ZarrGroup$new(store)
a <- g$get_item("1d.contiguous.lz4.i2")

print(a$get_shape())
```

## Get attributes from a root group and ZarrArray

```{r}
root <- pizzarr_sample("dog.ome.zarr")

z <- zarr_open(root)

class(z)

attrs <- z$get_attrs()$to_list()

names(attrs)

lengths(attrs$omero)

z$get_store()$listdir()

a <- z$get_item("4")

class(a)

a$get_attrs()$to_list()
```
```{r, include=FALSE}
if(clean) unlink(sample_dir, recursive = TRUE)
```

---
title: "Reading Zarr V3"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Reading Zarr V3}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  out.width = "100%"
)
sample_dir <- tools::R_user_dir("pizzarr")
clean <- !dir.exists(sample_dir)
```

pizzarr can read Zarr stores in both V2 and V3 format. This vignette
demonstrates opening the same dataset (BCSD climate observations) stored in
each format and shows the data are identical.

## Open V2 and V3 stores

```{r}
library(pizzarr)

v2_root <- pizzarr_sample("bcsd")
v3_root <- pizzarr_sample("bcsd_v3")

v2 <- zarr_open(v2_root)
v3 <- zarr_open(v3_root)
```

## Compare group attributes

```{r}
v2_attrs <- v2$get_attrs()$to_list()
v3_attrs <- v3$get_attrs()$to_list()

identical(v2_attrs, v3_attrs)
```

## Compare arrays

```{r}
# Precipitation array
v2_pr <- v2$get_item("pr")$as.array()
v3_pr <- v3$get_item("pr")$as.array()

identical(v2_pr, v3_pr)

# Temperature array
v2_tas <- v2$get_item("tas")$as.array()
v3_tas <- v3$get_item("tas")$as.array()

identical(v2_tas, v3_tas)

# Coordinate arrays
identical(v2$get_item("latitude")$as.array(), v3$get_item("latitude")$as.array())
identical(v2$get_item("longitude")$as.array(), v3$get_item("longitude")$as.array())
identical(v2$get_item("time")$as.array(), v3$get_item("time")$as.array())
```

## Inspect an array

```{r}
v3$get_item("pr")$get_shape()

v3$get_item("pr")$get_attrs()$to_list()
```

```{r, include=FALSE}
if(clean) unlink(sample_dir, recursive = TRUE)
```

---
title: "Local OME-NGFF Demo"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Local OME-NGFF Demo}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

### Demo

```r
library(pizzarr)

# The path to the root of the OME-NGFF Zarr store.
root <- pizzarr_sample("dog.ome.zarr")

# Open the OME-NGFF as a DirectoryStore.
store <- DirectoryStore$new(root)
g <- zarr_open_group(store)

# Using the OME metadata, get the path to the first resolution of the image pyramid.
attrs <- g$get_attrs()$to_list()
resolution_paths <- attrs$multiscales[[1]]$datasets[[1]]$path
first_resolution <- resolution_paths[[1]]

# Load the 3-dimensional array of RGB pixels (as a ZarrArray instance).
zarr_arr <- g$get_item(first_resolution)

print(zarr_arr$get_shape())
# [1]   3 500 750

# Load the 3-dimensional array of RGB pixels (as a NestedArray instance).
nested_arr <- zarr_arr$get_item("...")

# Extract the NestedArray contents as a base R array.
arr <- nested_arr$data

# Plot the RGB image using base R rasterImage().
dm <- dim(arr)
plot.new()
plot.window(c(0, dm[3]), c(0, dm[2]), asp = 1)
rasterImage(aperm(arr, c(2, 3, 1)) / 255, 0, 0, dm[3], dm[2])
```

---
title: "R Session Info"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{R Session Info}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

This page runs the `sessionInfo()` function and prints the results.
The output can be used to check the dependency versions that were used to run tests.

```{r include=TRUE, echo=TRUE}
library(pizzarr)
sessionInfo()
```

---
title: "Zarr Operations Cookbook"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Zarr Operations Cookbook}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  out.width = "100%"
)
```

This vignette covers common zarr array operations: persistent storage,
compression, resizing, filters, and advanced indexing.

```{r}
library(pizzarr)
```

## Persistent arrays

Create an array on disk, close the session, and reopen it later.

```{r}
path <- file.path(tempdir(), "example.zarr")

# Create a persistent array backed by a DirectoryStore
z <- zarr_open_array(
  store = path, mode = "w",
  shape = c(5, 10), chunks = c(5, 5), dtype = "<f4"
)

# Write data
z$set_item("...", array(1:50, dim = c(5, 10)))

z$get_shape()
```

Reopen the same path in read mode:

```{r}
z2 <- zarr_open_array(store = path, mode = "r")

z2$get_shape()

z2$get_item("...")$data
```

For quick save/load of an existing array:

```{r}
save_path <- file.path(tempdir(), "saved.zarr")

# Save an R array directly
zarr_save_array(save_path, zarr_create_array(
  data = volcano, shape = dim(volcano), dtype = "<f8"
))

# Reopen
z3 <- zarr_open_array(save_path, mode = "r")

all.equal(z3$as.array(), volcano)
```

## Compression

By default, pizzarr uses Zstandard compression. You can choose a different
compressor when creating an array.

### Zstandard (default)

```{r}
z_zstd <- zarr_create(
  shape = c(100, 100), dtype = "<f4",
  compressor = ZstdCodec$new(level = 3)
)

z_zstd$get_compressor()$get_config()
```

### Gzip

Gzip compression is interoperable with zarr-python and other implementations,
but is slower than Zstandard because R lacks an in-memory gzip API.
For best write performance, prefer `ZstdCodec`.

```{r}
z_gzip <- zarr_create(
  shape = c(100, 100), dtype = "<f4",
  compressor = GzipCodec$new(level = 5)
)

z_gzip$get_compressor()$get_config()
```

### Blosc (with algorithm selection)

```{r}
z_blosc <- zarr_create(
  shape = c(100, 100), dtype = "<f4",
  compressor = BloscCodec$new(cname = "lz4", clevel = 5, shuffle = TRUE)
)

z_blosc$get_compressor()$get_config()
```

### No compression

```{r}
z_none <- zarr_create(
  shape = c(100, 100), dtype = "<f4",
  compressor = NA
)

is.na(z_none$get_compressor())
```

## Resizing arrays

Arrays can be resized after creation. Data in the overlapping region is
preserved; new regions are filled with the fill value.

```{r}
z <- zarr_create(
  shape = c(5, 10), chunks = c(5, 5),
  dtype = "<i4", fill_value = 0L,
  compressor = "default"
)

z$set_item("...", array(1:50, dim = c(5, 10)))

z$get_shape()

# Grow the array
z$resize(10, 20)

z$get_shape()

# Original data is preserved in the top-left corner
z[1:5, 1:10]$data

# New region is filled with fill_value
z[6:10, 1:5]$data
```

Shrinking removes chunks that fall outside the new shape:

```{r}
z$resize(3, 4)

z$get_shape()

z$get_item("...")$data
```

## Appending data

Use `append()` to grow an array along an axis, adding new data at the end.
This is equivalent to zarr-python's `z.append(data, axis=0)`, but uses
R's 1-based axis indexing (axis 1 = first dimension).

```{r}
z <- zarr_create(
  shape = c(3, 4), chunks = c(3, 4),
  dtype = "<i4", fill_value = 0L
)

z$set_item("...", array(1:12, dim = c(3, 4)))

z$as.array()
```

Append new rows (axis 1, the default):

```{r}
new_rows <- array(13:20, dim = c(2, 4))

z$append(new_rows)

z$get_shape()

z$as.array()
```

Append new columns (axis 2):

```{r}
new_cols <- array(21:30, dim = c(5, 2))

z$append(new_cols, axis = 2)

z$get_shape()

z$as.array()
```

## Filters

Filters transform chunk data before compression. They are codec instances
passed as a list to the `filters` parameter. A common use case is
variable-length UTF-8 string arrays, which require `VLenUtf8Codec` as a filter.

```{r}
words <- c("alpha", "bravo", "charlie", "delta")

z_str <- zarr_create_array(
  data = array(words, dim = length(words)),
  shape = length(words), dtype = "|O",
  object_codec = VLenUtf8Codec$new()
)

z_str$get_item("...")$data

z_str$get_filters()
```

## Advanced indexing

Beyond basic slicing with `slice()` or `[`, pizzarr supports orthogonal
indexing for independent selection along each dimension.

### Setup

```{r}
z <- zarr_create_array(
  data = matrix(1:30, nrow = 5, ncol = 6),
  shape = c(5, 6), dtype = "<i4"
)

z$as.array()
```

### Basic slicing with `[`

The bracket operator uses orthogonal indexing internally:

```{r}
# Select rows 1-3, columns 2-4
z[1:3, 2:4]$data
```

### Orthogonal selection with integer arrays

Select specific rows and columns independently. Note that
`get_orthogonal_selection` uses zero-based indices (like zarr-python),
while the `[` operator uses R's one-based indexing:

```{r}
z$get_orthogonal_selection(list(c(0L, 2L, 4L), zb_slice(0, 6)))$data
```

### Boolean (mask) dimension indexing

Select dimensions using logical vectors:

```{r}
row_mask <- c(TRUE, FALSE, TRUE, FALSE, TRUE)

z$get_orthogonal_selection(list(row_mask, zb_slice(0, 6)))$data
```

### Using the OIndex object

The `$get_oindex()` accessor provides the same orthogonal indexing:

```{r}
oi <- z$get_oindex()

oi$get_item(list(c(0L, 4L), c(1L, 3L, 5L)))$data
```

### Slicing with step

Select every other row, every third column using `seq()` in bracket notation:

```{r}
z[seq(1, 5, 2), seq(1, 6, 3)]$data
```

### Ellipsis and colon shorthand

`"..."` selects all remaining dimensions; `":"` selects all along one dimension.
These work with `get_item()`:

```{r}
# All rows, column 1
z$get_item(list(":", 1))$data

# Row 1, all columns
z$get_item(list(1, "..."))$data
```

```{r, include = FALSE}
unlink(file.path(tempdir(), "example.zarr"), recursive = TRUE)
unlink(file.path(tempdir(), "saved.zarr"), recursive = TRUE)
```

---
title: "Remote OME-NGFF Demo"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Remote OME-NGFF Demo}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

### Setup

We install the `blosc` package to be able to use Blosc as the decompression codec, since the image we want to load is Blosc-compressed.

```r
install.packages("blosc")
```

### Demo

```r
library(pizzarr)

# The path to the root of the OME-NGFF Zarr store.
root <- "https://uk1s3.embassy.ebi.ac.uk/idr/zarr/v0.4/idr0062A/6001240.zarr"

# Open the OME-NGFF as an HttpStore.
store <- HttpStore$new(root)
g <- zarr_open_group(store)

# Using the OME metadata, get the path to the first resolution of the image pyramid.
attrs <- g$get_attrs()$to_list()
resolution_paths <- attrs$multiscales[[1]]$datasets[[1]]$path
first_resolution <- resolution_paths[[1]]

# Load the 4-dimensional array of pixels (as a ZarrArray instance).
zarr_arr <- g$get_item(first_resolution)

print(zarr_arr$get_shape())
# [1]   2 236 275 271

z_index <- 118

# Load a single Z-slice of pixels (as a NestedArray instance).
nested_arr <- zarr_arr$get_item(list(slice(1, 2), slice(z_index, z_index), slice(NA, NA), slice(NA, NA)))

print(nested_arr$shape)
# [1]   2   1 275 271

# Extract the NestedArray contents as a base R array.
arr <- nested_arr$data

# Convert to a pseudo-RGB array to make raster package happy.
rg_arr <- aperm(arr, c(2, 4, 3, 1))[1,,,]
rgb_arr <- array(dim=c(271, 275, 3))
rgb_arr[,,1] <- rg_arr[,,1]
rgb_arr[,,2] <- rg_arr[,,2]
rgb_arr[,,3] <- 0

# Plot the pseudo-RGB image using base R rasterImage().
plot.new()
plot.window(c(0, 271), c(0, 275), asp = 1)
rasterImage(rgb_arr / max(rgb_arr), 0, 0, 271, 275)
```

---
title: "zarrs Backend"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{zarrs Backend}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  out.width = "100%"
)
sample_dir <- tools::R_user_dir("pizzarr")
clean <- !dir.exists(sample_dir)
```

pizzarr ships in two tiers. The CRAN build is pure R --- no Rust compilation,
no system dependencies. It handles local and HTTP Zarr stores with sequential
chunk I/O via `lapply`. The r-universe build compiles in the
[zarrs](https://github.com/zarrs/zarrs) Rust crate via
[extendr](https://cran.r-project.org/package=rextendr), adding parallel decompression,
cloud-native store backends (S3, GCS), and codecs beyond what R packages provide.

The split exists because CRAN's macOS build machines ship a Rust toolchain
(rustc 1.84) that is too old for zarrs, which requires rustc >= 1.91.
r-universe builds against the latest stable toolchain, so it can compile zarrs
and distribute pre-built binaries. End users on either tier install with
`install.packages()` --- no Rust toolchain needed.

## Checking availability

```{r}
library(pizzarr)

has_zarrs <- pizzarr:::.pizzarr_env$zarrs_available
```

`pizzarr_compiled_features()` lists the feature flags compiled into the zarrs
backend. On the CRAN tier it returns `character(0)` with a message; on the
r-universe tier it returns the compiled capabilities:

```{r}
pizzarr_compiled_features()
```

The internal flag `.pizzarr_env$zarrs_available` is a logical scalar set once
at package load. Dispatch logic throughout pizzarr checks this flag to decide
whether to call into Rust or fall through to the R-native path:

```{r}
pizzarr:::.pizzarr_env$zarrs_available
```

## Upgrading to the zarrs tier

`pizzarr_upgrade()` prints the r-universe install command when zarrs is not
compiled in, or confirms that the backend is already present:

```{r}
pizzarr_upgrade()
```

The startup message that CRAN users see on `library(pizzarr)` can be silenced
with `options(pizzarr.suggest_runiverse = FALSE)`.

## Probing store metadata

The examples below require the zarrs backend. When this vignette is built
without it, the code chunks are not evaluated.

`zarrs_node_exists()` opens a filesystem store via the Rust backend, probes
for V2 and V3 metadata keys at a given path, and returns a list with three
fields: `exists` (logical), `node_type` (character), and `zarr_format`
(integer or NULL). The store handle is cached on the Rust side --- subsequent
calls to the same store path reuse it without re-opening.

### V2 store

```{r, eval=has_zarrs}
v2_root <- pizzarr_sample("fixtures/v2/data.zarr")

# Root group
zarrs_node_exists(v2_root, "")
```

```{r, eval=has_zarrs}
# An array within the store
zarrs_node_exists(v2_root, "1d.contiguous.lz4.i2")
```

```{r, eval=has_zarrs}
# A path that does not exist
zarrs_node_exists(v2_root, "does_not_exist")
```

### V3 store

V2 and V3 detection is automatic. zarrs probes for `zarr.json` first (V3),
then falls back to `.zarray` / `.zgroup` (V2):

```{r, eval=has_zarrs}
v3_root <- pizzarr_sample("fixtures/v3/data.zarr")

zarrs_node_exists(v3_root, "")
```

## Store cache management

The Rust backend holds open store handles in a process-global cache keyed by
normalized path. `zarrs_close_store()` removes a handle from the cache and
returns `TRUE`. A second call to the same path returns `FALSE` --- it was
already removed:

```{r, eval=has_zarrs}
zarrs_close_store(v2_root)
zarrs_close_store(v2_root)
```

```{r, eval=has_zarrs}
zarrs_close_store(v3_root)
```

## Array metadata

`zarrs_open_array_metadata()` opens a zarrs array and returns its metadata as a
named list. The store handle is cached, so repeated calls to the same store are
fast. The returned list contains `shape`, `chunks`, `dtype`, `r_type`,
`fill_value_json`, `zarr_format`, and `order`.

### V2 array

```{r, eval=has_zarrs}
v2_root <- pizzarr_sample("fixtures/v2/data.zarr")
zarrs_open_array_metadata(v2_root, "1d.contiguous.raw.i2")
```

### V3 array

V3 arrays work the same way. The `zarr_format` field distinguishes V2 from V3:

```{r, eval=has_zarrs}
v3_root <- pizzarr_sample("fixtures/v3/data.zarr")
zarrs_open_array_metadata(v3_root, "1d.contiguous.gzip.i2")
```

### Data type classification

The `r_type` field maps zarrs data types to R-compatible type families. zarrs
numeric types are classified as `"double"`, `"integer"`, or `"logical"` based on
what R can represent natively:

- **double**: float64 (zero-cost), float32 (widened), uint32/int64/uint64 (widened, precision risk > 2^53)
- **integer**: int32 (zero-cost), int8/int16/uint8/uint16 (widened)
- **logical**: bool

Unsupported types (strings, complex) report `"unsupported"` and fall back to the
R-native code path.

```{r, eval=has_zarrs}
zarrs_close_store(v2_root)
zarrs_close_store(v3_root)
```

## Runtime info and tuning

`zarrs_runtime_info()` reports the current zarrs configuration --- the codec
concurrency target, thread pool size, how many store handles are cached, and
which features were compiled in:

```{r, eval=has_zarrs}
zarrs_runtime_info()
```

### pizzarr_config()

`pizzarr_config()` is the main interface for viewing and changing concurrency
settings. Called with no arguments it returns the current state; with arguments
it sets the specified values:

```{r, eval=has_zarrs}
# View current settings
pizzarr_config()

# Set codec concurrency to 2 parallel operations per read/write
pizzarr_config(concurrent_target = 2L)
zarrs_runtime_info()$codec_concurrent_target
```

Three settings are available:

- **nthreads** --- rayon thread pool size. Set-once per R session (the thread
  pool can only be initialised once). For reliable session-level control, set
  the `PIZZARR_NTHREADS` environment variable before starting R.
- **concurrent_target** --- how many codec operations zarrs runs in parallel
  within a single read or write call. Can be changed at any time.
- **http_batch_range_requests** --- whether HTTP stores use multipart range
  requests (default TRUE). Set to FALSE for servers with incomplete multipart
  support. Takes effect on the next store open.

All three settings can also be configured via environment variables
(`PIZZARR_NTHREADS`, `PIZZARR_CONCURRENT_TARGET`,
`PIZZARR_HTTP_BATCH_RANGE_REQUESTS`) or R options (`pizzarr.nthreads`, etc.),
which are read at package load time. Environment variables persist across
sessions without needing `.Rprofile` edits.

The lower-level `zarrs_set_codec_concurrent_target()` function is still
available for direct use:

```{r, eval=has_zarrs}
zarrs_set_codec_concurrent_target(2L)
zarrs_runtime_info()$codec_concurrent_target
```

## Reading data via zarrs

When the zarrs backend is available and the selection is a contiguous slice
(step == 1), `ZarrArray$get_item()` dispatches reads to zarrs automatically.
zarrs handles chunk identification, parallel decompression, and codec execution
internally, bypassing pizzarr's R-native chunk loop. Scalar integer selections
(e.g., selecting a single row of a matrix) are also eligible --- they become
length-1 ranges on the Rust side. Unsupported selections (step > 1 slices,
fancy indexing, MemoryStore) fall through to the R-native path transparently.

### Basic read

```{r, eval=has_zarrs}
d <- tempfile("zarrs_vignette_")
z <- zarr_create(store = d, shape = c(100L, 50L), chunks = c(10L, 10L),
                 dtype = "<f8")
z$set_item("...", array(as.double(seq_len(5000)), dim = c(100, 50)))

# Re-open and read a subset --- zarrs handles the chunk I/O
z2 <- zarr_open(store = d)
result <- z2$get_item(list(slice(1L, 10L), slice(1L, 5L)))
dim(result$data)
```

### Direct zarrs_get_subset call

For lower-level access, `zarrs_get_subset()` reads a contiguous subset
directly via the Rust backend. Ranges are 0-based with exclusive stop, matching
zarrs conventions:

```{r, eval=has_zarrs}
result <- zarrs_get_subset(d, "", list(c(0L, 10L), c(0L, 5L)), NULL)
str(result)
```

### Concurrency control

The optional `concurrent_target` parameter (or the `pizzarr.concurrent_target`
R option) controls how many parallel codec operations zarrs uses within a single
read call. Setting it to `1L` disables parallel decompression:

```{r, eval=has_zarrs}
result <- zarrs_get_subset(d, "", list(c(0L, 10L), c(0L, 5L)), 1L)
length(result$data)
```

```{r, eval=has_zarrs}
zarrs_close_store(d)
unlink(d, recursive = TRUE)
```

## Creating arrays via zarrs

When the zarrs backend is available and the store is a writable filesystem path,
`zarr_create()` dispatches array creation to zarrs instead of building metadata
JSON in R. zarrs validates the metadata structure, writes it to the store, and
the array is ready for data. The dispatch is transparent --- the same
`zarr_create()` call works on both tiers, and unsupported configurations
(MemoryStore, object dtypes, custom filters) fall through to the R-native path.

### Transparent dispatch

The `zarr_create()` examples earlier in this vignette already use this path when
zarrs is available. The zarrs backend handles V2 and V3 formats, all 11 numeric
data types, and four codec presets:

```{r, eval=has_zarrs}
# V3 array with gzip compression
d <- tempfile("zarrs_create_vignette_")
z <- zarr_create(store = d, shape = c(20L, 10L), chunks = c(10L, 10L),
                 dtype = "<f8", zarr_format = 3L)
z
```

```{r, eval=has_zarrs}
# Confirm V3 metadata was written
file.exists(file.path(d, "zarr.json"))
```

```{r, eval=has_zarrs}
zarrs_close_store(d)
unlink(d, recursive = TRUE)
```

### Direct zarrs_create_array call

`zarrs_create_array()` provides lower-level access to the Rust creation path.
It accepts V3-style data type names (`"float64"`, `"int32"`, `"bool"`, etc.)
and a codec preset string (`"none"`, `"gzip"`, `"blosc"`, or `"zstd"`).
The return value is the same metadata list as `zarrs_open_array_metadata()`:

```{r, eval=has_zarrs}
d <- tempfile("zarrs_create_direct_")
dir.create(d)

meta <- zarrs_create_array(
  store_url = d,
  array_path = "",
  shape = c(100L, 50L),
  chunks = c(10L, 10L),
  dtype = "float64",
  codec_preset = "gzip",
  fill_value = 0.0,
  attributes_json = "{}",
  zarr_format = 3L
)
str(meta)
```

The array is immediately usable for reads and writes:

```{r, eval=has_zarrs}
zarrs_set_subset(d, "", list(c(0L, 10L), c(0L, 5L)),
                 as.double(1:50), NULL)
result <- zarrs_get_subset(d, "", list(c(0L, 10L), c(0L, 5L)), NULL)
head(result$data)
```

```{r, eval=has_zarrs}
zarrs_close_store(d)
unlink(d, recursive = TRUE)
```

### Codec presets

The zarrs creation path supports four named codec presets. Custom codec
configurations fall through to the R-native path.

| Preset | V2 compressor | V3 codec chain | Notes |
|--------|--------------|----------------|-------|
| `"none"` | null | bytes only | No compression |
| `"gzip"` | gzip, level 1 | bytes + gzip(1) | Fast, reasonable ratio |
| `"blosc"` | blosc, lz4, clevel 5 | bytes + blosc(lz4, 5) | Requires `blosc` feature |
| `"zstd"` | --- | bytes + zstd(3) | V3 only; requires `zstd` feature |

One difference from the R-native path: zarrs uses the `"gzip"` compressor id
for V2 arrays, while zarr-python uses `"zlib"`. Both produce gzip-compatible
output, and zarrs reads either id when opening existing arrays.

## Writing data via zarrs

The write path mirrors the read path. When the zarrs backend is available and
the selection qualifies (contiguous slices, filesystem-backed store),
`ZarrArray$set_item()` dispatches writes to zarrs instead of iterating over
chunks in R. zarrs encodes the data, splits it across the affected chunks, and
writes them to disk --- using its internal thread pool for parallel compression
when multiple chunks are involved.

Data type narrowing happens on the Rust side. R doubles narrow to the array's
stored type (float32, int64, uint32, etc.) and R integers narrow to smaller
integer types (int16, int8, uint8, uint16) with range checking. An out-of-range
value produces an error rather than silent truncation.

### Basic write

```{r, eval=has_zarrs}
d <- tempfile("zarrs_write_vignette_")
z <- zarr_create(store = d, shape = c(20L, 10L), chunks = c(10L, 10L),
                 dtype = "<f8")

# set_item dispatches to zarrs when eligible
z$set_item("...", array(as.double(1:200), dim = c(20, 10)))

# Read back to confirm
z2 <- zarr_open(store = d)
result <- z2$get_item(list(slice(1L, 5L), slice(1L, 3L)))
result$data
```

### Partial overwrite

Writing to a subset of an existing array works the same way. zarrs reads the
affected chunks, merges the new data, and writes them back:

```{r, eval=has_zarrs}
# Overwrite rows 3-7, columns 1-2
z$set_item(list(slice(3L, 7L), slice(1L, 2L)),
           array(rep(-1.0, 10), dim = c(5, 2)))

result <- z2$get_item(list(slice(1L, 10L), slice(1L, 3L)))
result$data
```

### Direct zarrs_set_subset call

`zarrs_set_subset()` provides lower-level access to the Rust write path.
Data is a flat vector in R's native F-order (column-major) --- the Rust
backend handles the F-to-C order conversion internally. The function returns
`TRUE` on success:

```{r, eval=has_zarrs}
# Write 10 values to the first row (0-based range [0, 1) x [0, 10))
zarrs_set_subset(d, "", list(c(0L, 1L), c(0L, 10L)),
                   as.double(101:110), NULL)

result <- zarrs_get_subset(d, "", list(c(0L, 1L), c(0L, 10L)), NULL)
result$data
```

```{r, eval=has_zarrs}
zarrs_close_store(d)
unlink(d, recursive = TRUE)
```

## HTTP reads via zarrs

When the `http_sync` feature is compiled in, the zarrs backend can read
directly from HTTP/HTTPS Zarr stores using the
[zarrs_http](https://github.com/zarrs/zarrs_http) crate. This bypasses
pizzarr's R-native `crul`-based chunk loop, giving parallel chunk decode
on remote data.

HTTP stores are read-only in zarrs --- write dispatch (`set_item`) falls
through to the R-native path automatically.

### Transparent dispatch

The zarrs fast path activates automatically when an `HttpStore`-backed
array is read with a contiguous selection. No code changes are needed
compared to the R-native path:

```{r, eval=has_zarrs && ("http_sync" %in% pizzarr_compiled_features())}
url <- "https://raw.githubusercontent.com/DOI-USGS/rnz/main/inst/extdata/bcsd.zarr"

z <- zarr_open(store = HttpStore$new(url))

# zarrs handles the HTTP reads + parallel decompression
pr <- z$get_item("pr")
pr
```

```{r, eval=has_zarrs && ("http_sync" %in% pizzarr_compiled_features())}
# Read a subset --- zarrs fetches only the chunks that overlap
result <- pr$get_item(list(slice(1L, 3L), slice(1L, 5L), slice(1L, 5L)))
dim(result$data)
```

### Direct zarrs_get_subset from HTTP

`zarrs_get_subset()` also works with HTTP URLs. The store handle is cached
on the Rust side, so repeated reads to the same URL reuse the connection:

```{r, eval=has_zarrs && ("http_sync" %in% pizzarr_compiled_features())}
meta <- zarrs_open_array_metadata(url, "pr")
str(meta[c("shape", "dtype", "zarr_format")])
```

```{r, eval=has_zarrs && ("http_sync" %in% pizzarr_compiled_features())}
# Read a single element (first along each dimension)
ranges <- lapply(seq_along(meta$shape), function(i) c(0L, 1L))
result <- zarrs_get_subset(url, "pr", ranges, NULL)
result$data
```

```{r, eval=has_zarrs && ("http_sync" %in% pizzarr_compiled_features())}
zarrs_close_store(url)
```

### Feature detection

Check whether HTTP support is compiled in with `pizzarr_compiled_features()`.
When `"http_sync"` is present, zarrs can open `http://` and `https://` URLs.
When it is absent, HTTP reads fall through to the R-native `crul`-based path:

```{r, eval=has_zarrs}
"http_sync" %in% pizzarr_compiled_features()
```

## S3 reads via zarrs

When the `s3` feature is compiled in, the zarrs backend can read from Amazon S3
buckets using the [object_store](https://docs.rs/object_store/) crate with an
async-to-sync adapter. Public buckets work without credentials (unsigned
requests). Authenticated access uses standard AWS environment variables
(`AWS_ACCESS_KEY_ID`, `AWS_SECRET_ACCESS_KEY`, `AWS_REGION`).

S3 stores are currently read-only via zarrs --- write operations fall through to
the R-native path.

```{r, eval=has_zarrs && ("s3" %in% pizzarr_compiled_features())}
# OME-Zarr bonsai dataset on AWS Open Data (V2, zstd, uint8)
s3_url <- "s3://ome-zarr-scivis/v0.4/64x0/bonsai.ome.zarr"

# Read array metadata
meta <- zarrs_open_array_metadata(s3_url, "scale0/bonsai")
str(meta[c("shape", "dtype", "zarr_format")])
```

```{r, eval=has_zarrs && ("s3" %in% pizzarr_compiled_features())}
# Read a small subset (first 4x4x4 corner)
result <- zarrs_get_subset(s3_url, "scale0/bonsai",
                           list(c(0L, 4L), c(0L, 4L), c(0L, 4L)), NULL)
str(result)
```

```{r, eval=has_zarrs && ("s3" %in% pizzarr_compiled_features())}
zarrs_close_store(s3_url)
```

### GCS and other cloud stores

GCS data hosted on Google Cloud Storage is publicly accessible via HTTPS
endpoints. The zarrs HTTP backend reads these directly:

```{r, eval=has_zarrs && ("blosc" %in% pizzarr_compiled_features())}
# Pangeo ECCO ocean basins (V2, blosc/lz4, float32)
gcs_url <- "https://storage.googleapis.com/pangeo-data/ECCO_basins.zarr"

meta <- zarrs_open_array_metadata(gcs_url, "basin_mask")
cat("Shape:", paste(meta$shape, collapse = " x "), "\n")
cat("Dtype:", meta$dtype, "\n")
```

```{r, eval=has_zarrs && ("blosc" %in% pizzarr_compiled_features())}
# Read a single basin mask slice
result <- zarrs_get_subset(gcs_url, "basin_mask",
                           list(c(0L, 1L), c(0L, 90L), c(0L, 90L)), NULL)
cat("Slice dimensions:", paste(result$shape, collapse = " x "), "\n")
```

```{r, eval=has_zarrs && ("blosc" %in% pizzarr_compiled_features())}
zarrs_close_store(gcs_url)
```

Authenticated GCS access via `gs://` URLs requires the `gcs` compiled feature
and GCP credentials (environment variables or application default credentials).
The `S3Store` and `GcsStore` R6 classes provide URL wrappers for high-level
use with `zarr_open()`:

```r
# S3 (requires s3 feature)
z <- zarr_open(store = S3Store$new("s3://bucket/path/to/store.zarr"))

# GCS (requires gcs feature + credentials)
z <- zarr_open(store = GcsStore$new("gs://bucket/path/to/store.zarr"))
```

## C/F order handling

zarrs stores data in C-order (row-major), while R uses F-order (column-major).
The Rust backend handles this conversion transparently:

- **Reads:** `zarrs_get_subset()` returns data in F-order, ready for
  `array(data, dim = shape)` with no `aperm()` needed.
- **Writes:** `zarrs_set_subset()` accepts F-order data and converts to C-order
  internally before writing to the store.

The transpose uses cache-blocked tiling for 2D arrays and output-order
iteration with incremental index tracking for higher dimensions, matching or
exceeding the performance of R's C-level `aperm()`.

```{r, include=FALSE}
if (clean) unlink(sample_dir, recursive = TRUE)
```

---
title: "Remote AnnData Demo"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Remote AnnData Demo}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

### Setup

We install the `blosc` package to be able to use Blosc as the decompression codec, since the data we want to load is Blosc-compressed.

```r
install.packages("blosc")
```

### Demo

```r
library(pizzarr)

# The path to the root of the AnnData-Zarr store.
root <- "https://data-1.vitessce.io/0.0.33/main/human-lymph-node-10x-visium/human_lymph_node_10x_visium.h5ad.zarr"

# Open the AnnData-Zarr store as an HttpStore.
store <- HttpStore$new(root)
g <- zarr_open_group(store)

obs_attrs <- g$get_item("obs")$get_attrs()$to_list()
index_colname <- obs_attrs[['_index']]
index_arr <- g$get_item(paste0("obs/", index_colname))$get_item("...")$data

umap_arr <- g$get_item("obsm/X_umap")$get_item("...")$data
cluster_arr <- g$get_item("obs/clusters")$get_item("...")$data

print(dim(umap_arr))
# [1] 3861    2

print(dim(index_arr))
# [1] 3861

print(dim(index_arr))
# [1] 3861

plot(x=umap_arr[,1], y=umap_arr[,2], col=cluster_arr)
```

---
title: "Zarr V3 Cross-Implementation Interop"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Zarr V3 Cross-Implementation Interop}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  out.width = "100%",
  message = FALSE
)
sample_dir <- tools::R_user_dir("pizzarr")
clean <- !dir.exists(sample_dir)
```

This vignette demonstrates
pizzarr reading a Zarr V3 store that was written by
[zarr-python](https://github.com/zarr-developers/zarr-python) 3.x ---
showing that the two implementations are interoperable.

The test store includes arrays with various data types, compression
codecs, chunk layouts, fill values, and nested groups. It was generated
with the script bundled in this project at:
`inst/extdata/fixtures/v3/generate-v3-zarr-python.py`.

## Summary

- **Data types**: bool, uint8, int16, int32, float32, float64
- **Compression**: uncompressed, gzip, zstd, blosc
- **Layouts**: 1D, 2D, 3D; contiguous and ragged chunks
- **Fill values**: default (0) and custom (-9999)
- **Groups**: nested groups with attributes
- **Array metadata**: per-array attributes (units, long_name)

## Open the store

```{r}
library(pizzarr)

store_path <- pizzarr_sample("fixtures/v3/zarr_python_v3.zarr")
root <- zarr_open(store_path)
```

## Root-level attributes

zarr-python embedded these attributes when it created the store:

```{r}
root$get_attrs()$to_list()
```

## Reading different data types

### Integer (int32, gzip compressed)

```{r}
a <- root$get_item("int32_1d")
a$get_shape()
a$as.array()
```

### Float64 (zstd compressed)

```{r}
a <- root$get_item("float64_1d")
a$as.array()
```

### Boolean (uncompressed)

```{r}
a <- root$get_item("bool_1d")
a$as.array()
```

### Unsigned integer (uint8, chunked)

```{r}
a <- root$get_item("uint8_1d")
a$as.array()
```

### Float32 (blosc compressed)

```{r}
a <- root$get_item("float32_1d")
a$as.array()
```

## Multi-dimensional arrays

### 2D int16 (gzip)

```{r}
a <- root$get_item("int16_2d")
a$get_shape()
a$as.array()
```

### 3D float64 (zstd)

```{r}
a <- root$get_item("float64_3d")
a$get_shape()
a$as.array()
```

## Fill values

The `with_fill` array was created with `fill_value = -9999.0` and only
the first chunk was written. The second chunk returns fill values:

```{r}
a <- root$get_item("with_fill")
a$as.array()
```

## Ragged (non-aligned) chunks

When the array shape is not evenly divisible by the chunk shape, edge
chunks are smaller. This 5x7 array uses 3x4 chunks:
```{r}
a <- root$get_item("ragged_2d")
a$get_shape()
a$as.array()
```

## Nested groups and array attributes

The store contains a sub-group with attributes and two arrays that carry
their own metadata:

```{r}
g <- root$get_item("var_group")
g$get_attrs()$to_list()
```

```{r}
temp <- g$get_item("temperature")
temp$get_attrs()$to_list()
temp$as.array()
```

```{r}
pres <- g$get_item("pressure")
pres$get_attrs()$to_list()
pres$as.array()
```

```{r, include=FALSE}
if(clean) unlink(sample_dir, recursive = TRUE)
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{compute_size}
\alias{compute_size}
\title{Compute Size}
\usage{
compute_size(shape)
}
\arguments{
\item{shape}{A shape vector}
}
\value{
The product of shape elements.
}
\description{
Compute Size
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/atomic.R
\name{is_integer_scalar}
\alias{is_integer_scalar}
\title{Check if a value is both a scalar (tagged by as_scalar) and an integer.}
\usage{
is_integer_scalar(s)
}
\description{
Check if a value is both a scalar (tagged by as_scalar) and an integer.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/indexing.R
\docType{class}
\name{SliceDimIndexer}
\alias{SliceDimIndexer}
\title{SliceDimIndexer Class}
\description{
Indexer for a slice selection along one dimension.
}
\details{
The Zarr SliceDimIndexer class.
}
\keyword{internal}
\section{Super class}{
\code{\link[pizzarr:DimIndexer]{pizzarr::DimIndexer}} -> \code{SliceDimIndexer}
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{dim_len}}{Dimension length.}

\item{\code{dim_chunk_len}}{Dimension chunk length.}

\item{\code{num_chunks}}{Number of chunks.}

\item{\code{start}}{Start index.}

\item{\code{stop}}{Stop index.}

\item{\code{step}}{Step size.}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-SliceDimIndexer-new}{\code{SliceDimIndexer$new()}}
\item \href{#method-SliceDimIndexer-iter}{\code{SliceDimIndexer$iter()}}
\item \href{#method-SliceDimIndexer-clone}{\code{SliceDimIndexer$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-SliceDimIndexer-new"></a>}}
\if{latex}{\out{\hypertarget{method-SliceDimIndexer-new}{}}}
\subsection{Method \code{new()}}{
Create a new SliceDimIndexer instance.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{SliceDimIndexer$new(dim_sel, dim_len, dim_chunk_len)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{dim_sel}}{Integer dimension selection.}

\item{\code{dim_len}}{Integer dimension length.}

\item{\code{dim_chunk_len}}{Integer dimension chunk length.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A \code{SliceDimIndexer} instance.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-SliceDimIndexer-iter"></a>}}
\if{latex}{\out{\hypertarget{method-SliceDimIndexer-iter}{}}}
\subsection{Method \code{iter()}}{
Iterate over chunks that overlap the slice, yielding a ChunkDimProjection for each.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{SliceDimIndexer$iter()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
A list of ChunkDimProjection objects.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-SliceDimIndexer-clone"></a>}}
\if{latex}{\out{\hypertarget{method-SliceDimIndexer-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{SliceDimIndexer$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/int.R
\name{zb_int}
\alias{zb_int}
\title{Convenience function for the internal Int class constructor
with zero-based indexing}
\usage{
zb_int(index)
}
\arguments{
\item{index}{integer index}
}
\description{
Convenience function for the internal Int class constructor
with zero-based indexing
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zarr-array.R
\docType{class}
\name{ZarrArray}
\alias{ZarrArray}
\title{ZarrArray Class}
\format{
\link[R6:R6Class]{R6::R6Class}
}
\description{
Instantiate an array from an initialized store.
}
\details{
The Zarr Array class.
}
\keyword{internal}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-ZarrArray-new}{\code{ZarrArray$new()}}
\item \href{#method-ZarrArray-get_store}{\code{ZarrArray$get_store()}}
\item \href{#method-ZarrArray-get_path}{\code{ZarrArray$get_path()}}
\item \href{#method-ZarrArray-get_name}{\code{ZarrArray$get_name()}}
\item \href{#method-ZarrArray-get_basename}{\code{ZarrArray$get_basename()}}
\item \href{#method-ZarrArray-get_read_only}{\code{ZarrArray$get_read_only()}}
\item \href{#method-ZarrArray-set_read_only}{\code{ZarrArray$set_read_only()}}
\item \href{#method-ZarrArray-get_chunk_store}{\code{ZarrArray$get_chunk_store()}}
\item \href{#method-ZarrArray-get_shape}{\code{ZarrArray$get_shape()}}
\item \href{#method-ZarrArray-set_shape}{\code{ZarrArray$set_shape()}}
\item \href{#method-ZarrArray-resize}{\code{ZarrArray$resize()}}
\item \href{#method-ZarrArray-get_chunks}{\code{ZarrArray$get_chunks()}}
\item \href{#method-ZarrArray-get_dtype}{\code{ZarrArray$get_dtype()}}
\item \href{#method-ZarrArray-get_compressor}{\code{ZarrArray$get_compressor()}}
\item \href{#method-ZarrArray-get_fill_value}{\code{ZarrArray$get_fill_value()}}
\item \href{#method-ZarrArray-set_fill_value}{\code{ZarrArray$set_fill_value()}}
\item \href{#method-ZarrArray-set_dimension_names}{\code{ZarrArray$set_dimension_names()}}
\item \href{#method-ZarrArray-get_order}{\code{ZarrArray$get_order()}}
\item \href{#method-ZarrArray-get_filters}{\code{ZarrArray$get_filters()}}
\item \href{#method-ZarrArray-get_synchronizer}{\code{ZarrArray$get_synchronizer()}}
\item \href{#method-ZarrArray-get_attrs}{\code{ZarrArray$get_attrs()}}
\item \href{#method-ZarrArray-get_ndim}{\code{ZarrArray$get_ndim()}}
\item \href{#method-ZarrArray-get_dimension_names}{\code{ZarrArray$get_dimension_names()}}
\item \href{#method-ZarrArray-get_size}{\code{ZarrArray$get_size()}}
\item \href{#method-ZarrArray-get_itemsize}{\code{ZarrArray$get_itemsize()}}
\item \href{#method-ZarrArray-get_nbytes}{\code{ZarrArray$get_nbytes()}}
\item \href{#method-ZarrArray-get_nbytes_stored}{\code{ZarrArray$get_nbytes_stored()}}
\item \href{#method-ZarrArray-get_cdata_shape}{\code{ZarrArray$get_cdata_shape()}}
\item \href{#method-ZarrArray-get_nchunks}{\code{ZarrArray$get_nchunks()}}
\item \href{#method-ZarrArray-get_nchunks_initialized}{\code{ZarrArray$get_nchunks_initialized()}}
\item \href{#method-ZarrArray-get_is_view}{\code{ZarrArray$get_is_view()}}
\item \href{#method-ZarrArray-get_oindex}{\code{ZarrArray$get_oindex()}}
\item \href{#method-ZarrArray-get_vindex}{\code{ZarrArray$get_vindex()}}
\item \href{#method-ZarrArray-get_write_empty_chunks}{\code{ZarrArray$get_write_empty_chunks()}}
\item \href{#method-ZarrArray-equals}{\code{ZarrArray$equals()}}
\item \href{#method-ZarrArray-print}{\code{ZarrArray$print()}}
\item \href{#method-ZarrArray-islice}{\code{ZarrArray$islice()}}
\item \href{#method-ZarrArray-length}{\code{ZarrArray$length()}}
\item \href{#method-ZarrArray-get_item}{\code{ZarrArray$get_item()}}
\item \href{#method-ZarrArray-get_basic_selection}{\code{ZarrArray$get_basic_selection()}}
\item \href{#method-ZarrArray-get_orthogonal_selection}{\code{ZarrArray$get_orthogonal_selection()}}
\item \href{#method-ZarrArray-get_coordinate_selection}{\code{ZarrArray$get_coordinate_selection()}}
\item \href{#method-ZarrArray-get_mask_selection}{\code{ZarrArray$get_mask_selection()}}
\item \href{#method-ZarrArray-set_item}{\code{ZarrArray$set_item()}}
\item \href{#method-ZarrArray-set_basic_selection}{\code{ZarrArray$set_basic_selection()}}
\item \href{#method-ZarrArray-set_orthogonal_selection}{\code{ZarrArray$set_orthogonal_selection()}}
\item \href{#method-ZarrArray-set_coordinate_selection}{\code{ZarrArray$set_coordinate_selection()}}
\item \href{#method-ZarrArray-set_mask_selection}{\code{ZarrArray$set_mask_selection()}}
\item \href{#method-ZarrArray-get_info}{\code{ZarrArray$get_info()}}
\item \href{#method-ZarrArray-get_digest}{\code{ZarrArray$get_digest()}}
\item \href{#method-ZarrArray-get_hexdigest}{\code{ZarrArray$get_hexdigest()}}
\item \href{#method-ZarrArray-append}{\code{ZarrArray$append()}}
\item \href{#method-ZarrArray-view}{\code{ZarrArray$view()}}
\item \href{#method-ZarrArray-astype}{\code{ZarrArray$astype()}}
\item \href{#method-ZarrArray-get_dimension_separator}{\code{ZarrArray$get_dimension_separator()}}
\item \href{#method-ZarrArray-as.array}{\code{ZarrArray$as.array()}}
\item \href{#method-ZarrArray-clone}{\code{ZarrArray$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-new"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-new}{}}}
\subsection{Method \code{new()}}{
Create a new ZarrArray instance.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$new(
  store,
  path = NA,
  read_only = FALSE,
  chunk_store = NA,
  synchronizer = NA,
  cache_metadata = TRUE,
  cache_attrs = TRUE,
  write_empty_chunks = TRUE
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{store}}{(\link{Store})\cr
Array store, already initialized.}

\item{\code{path}}{(\code{character(1)})\cr
Storage path.}

\item{\code{read_only}}{(\code{logical(1)})\cr
Whether the array is read-only.}

\item{\code{chunk_store}}{(\link{Store} or \code{NA})\cr
Separate storage for chunks. If not provided, \code{store} will be used
for both chunks and metadata.}

\item{\code{synchronizer}}{(\code{ANY} or \code{NA})\cr
Object used to synchronize write access to the array.}

\item{\code{cache_metadata}}{(\code{logical(1)})\cr
Whether to cache metadata.}

\item{\code{cache_attrs}}{(\code{logical(1)})\cr
Whether to cache attributes.}

\item{\code{write_empty_chunks}}{(\code{logical(1)})\cr
Whether to write empty chunks.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A \link{ZarrArray} instance.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-get_store"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-get_store}{}}}
\subsection{Method \code{get_store()}}{
get store from array.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$get_store()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
\link{Store}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-get_path"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-get_path}{}}}
\subsection{Method \code{get_path()}}{
get array path
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$get_path()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
\code{character(1)} or \code{NA}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-get_name"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-get_name}{}}}
\subsection{Method \code{get_name()}}{
get full array name
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$get_name()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
\code{character(1)} or \code{NA}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-get_basename"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-get_basename}{}}}
\subsection{Method \code{get_basename()}}{
get the basename of an array
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$get_basename()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
\code{character(1)} or \code{NA}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-get_read_only"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-get_read_only}{}}}
\subsection{Method \code{get_read_only()}}{
get the read only property of an array (TRUE/FALSE)
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$get_read_only()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
\code{logical(1)}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-set_read_only"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-set_read_only}{}}}
\subsection{Method \code{set_read_only()}}{
set the read only property of an array
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$set_read_only(val)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{val}}{(\code{logical(1)})\cr
Value to set.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{NULL} (called for side effects).
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-get_chunk_store"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-get_chunk_store}{}}}
\subsection{Method \code{get_chunk_store()}}{
get the chunk store for an array
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$get_chunk_store()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
\link{Store}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-get_shape"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-get_shape}{}}}
\subsection{Method \code{get_shape()}}{
get the shape of an array
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$get_shape()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
\code{integer()}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-set_shape"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-set_shape}{}}}
\subsection{Method \code{set_shape()}}{
set or reset the size of an array
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$set_shape(value)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{value}}{(\code{integer()})\cr
Numeric size to set.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{NULL} (called for side effects).
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-resize"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-resize}{}}}
\subsection{Method \code{resize()}}{
Change the shape of the array by growing or shrinking one or more dimensions.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$resize(...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{...}}{Arguments for do.call.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{NULL} (called for side effects).
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-get_chunks"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-get_chunks}{}}}
\subsection{Method \code{get_chunks()}}{
get the chunk metadata of an array
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$get_chunks()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
\code{integer()}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-get_dtype"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-get_dtype}{}}}
\subsection{Method \code{get_dtype()}}{
get the Dtype of an array
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$get_dtype()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
\link{Dtype}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-get_compressor"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-get_compressor}{}}}
\subsection{Method \code{get_compressor()}}{
get the compressor of an array
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$get_compressor()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
\link{Codec} or \code{NA}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-get_fill_value"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-get_fill_value}{}}}
\subsection{Method \code{get_fill_value()}}{
get the fill value of an array
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$get_fill_value()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Scalar fill value.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-set_fill_value"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-set_fill_value}{}}}
\subsection{Method \code{set_fill_value()}}{
set the fill value of an array
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$set_fill_value(val)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{val}}{Fill value to use.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{NULL} (called for side effects).
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-set_dimension_names"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-set_dimension_names}{}}}
\subsection{Method \code{set_dimension_names()}}{
Set dimension names of the array (V3 only).
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$set_dimension_names(names)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{names}}{Character vector of length equal to the number of array
dimensions, or \code{NULL} to clear.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{NULL} (called for side effects).
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-get_order"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-get_order}{}}}
\subsection{Method \code{get_order()}}{
get the storage order metadata of an array.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$get_order()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
\code{character(1)}, \code{"C"} or \code{"F"}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-get_filters"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-get_filters}{}}}
\subsection{Method \code{get_filters()}}{
get the filters metadata of an array
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$get_filters()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
\code{list()} of \link{Codec} objects or \code{NA}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-get_synchronizer"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-get_synchronizer}{}}}
\subsection{Method \code{get_synchronizer()}}{
Get the synchronizer used to coordinate write access to the array.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$get_synchronizer()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Synchronizer object or \code{NA}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-get_attrs"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-get_attrs}{}}}
\subsection{Method \code{get_attrs()}}{
get attributes of array
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$get_attrs()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
\link{Attributes}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-get_ndim"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-get_ndim}{}}}
\subsection{Method \code{get_ndim()}}{
get number of dimensions of array
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$get_ndim()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
\code{integer(1)}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-get_dimension_names"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-get_dimension_names}{}}}
\subsection{Method \code{get_dimension_names()}}{
Get dimension names of the array (V3 only).
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$get_dimension_names()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Character vector or \code{NULL}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-get_size"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-get_size}{}}}
\subsection{Method \code{get_size()}}{
Get the total number of elements in the array.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$get_size()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
\code{numeric(1)}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-get_itemsize"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-get_itemsize}{}}}
\subsection{Method \code{get_itemsize()}}{
Get the size in bytes of each item in the array.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$get_itemsize()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
\code{integer(1)}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-get_nbytes"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-get_nbytes}{}}}
\subsection{Method \code{get_nbytes()}}{
get number of bytes of an array
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$get_nbytes()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
\code{numeric(1)}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-get_nbytes_stored"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-get_nbytes_stored}{}}}
\subsection{Method \code{get_nbytes_stored()}}{
Get the total number of stored bytes for the array, including metadata and compressed chunk data.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$get_nbytes_stored()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
\code{numeric(1)}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-get_cdata_shape"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-get_cdata_shape}{}}}
\subsection{Method \code{get_cdata_shape()}}{
Get the number of chunks along each dimension of the array.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$get_cdata_shape()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
\code{numeric()}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-get_nchunks"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-get_nchunks}{}}}
\subsection{Method \code{get_nchunks()}}{
Get the total number of chunks in the array.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$get_nchunks()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
\code{integer(1)}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-get_nchunks_initialized"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-get_nchunks_initialized}{}}}
\subsection{Method \code{get_nchunks_initialized()}}{
Get the number of chunks that have been initialized with data.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$get_nchunks_initialized()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
\code{integer(1)}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-get_is_view"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-get_is_view}{}}}
\subsection{Method \code{get_is_view()}}{
get is_view metadata of array
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$get_is_view()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
\code{logical(1)}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-get_oindex"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-get_oindex}{}}}
\subsection{Method \code{get_oindex()}}{
get orthogonal index of array
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$get_oindex()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
\link{OIndex}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-get_vindex"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-get_vindex}{}}}
\subsection{Method \code{get_vindex()}}{
get vectorized index of array
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$get_vindex()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
\link{VIndex}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-get_write_empty_chunks"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-get_write_empty_chunks}{}}}
\subsection{Method \code{get_write_empty_chunks()}}{
get write empty chunks setting of array
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$get_write_empty_chunks()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
\code{logical(1)}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-equals"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-equals}{}}}
\subsection{Method \code{equals()}}{
check if another object refers to the same array. does not check array data
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$equals(other)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{other}}{Other object to check.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{logical(1)}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-print"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-print}{}}}
\subsection{Method \code{print()}}{
Print a human-readable summary of the array.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$print(...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{...}}{Ignored.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{self} (invisibly).
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-islice"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-islice}{}}}
\subsection{Method \code{islice()}}{
Iterate over the array or a range of it, yielding successive slices along the first dimension. Uses chunk caching for efficient sequential access.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$islice(start = NA, end = NA)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{start}}{Start index of the iteration range (1-based, inclusive).}

\item{\code{end}}{End index of the iteration range (1-based, inclusive).}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{list()} of \link{NestedArray} slices.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-length"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-length}{}}}
\subsection{Method \code{length()}}{
Return the length of the first dimension. Raises an error for zero-dimensional arrays.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$length()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
\code{integer(1)}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-get_item"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-get_item}{}}}
\subsection{Method \code{get_item()}}{
Subset the array.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$get_item(selection)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{selection}}{Selection specifying the subset to retrieve.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\link{NestedArray}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-get_basic_selection"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-get_basic_selection}{}}}
\subsection{Method \code{get_basic_selection()}}{
Get a selection of an array based on a "basic" list of slices.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$get_basic_selection(selection = NA, out = NA, fields = NA)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{selection}}{Selection specifying the subset to retrieve.}

\item{\code{out}}{Not yet implemented.}

\item{\code{fields}}{Not yet implemented.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\link{NestedArray}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-get_orthogonal_selection"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-get_orthogonal_selection}{}}}
\subsection{Method \code{get_orthogonal_selection()}}{
Retrieve data using orthogonal (outer) indexing.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$get_orthogonal_selection(selection = NA, out = NA, fields = NA)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{selection}}{Selection specifying the subset to retrieve.}

\item{\code{out}}{Not yet implemented.}

\item{\code{fields}}{Not yet implemented.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\link{NestedArray}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-get_coordinate_selection"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-get_coordinate_selection}{}}}
\subsection{Method \code{get_coordinate_selection()}}{
Retrieve data using coordinate (point) indexing. Not yet implemented.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$get_coordinate_selection(selection = NA, out = NA, fields = NA)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{selection}}{Selection specifying the subset to retrieve.}

\item{\code{out}}{Not yet implemented.}

\item{\code{fields}}{Not yet implemented.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\link{NestedArray}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-get_mask_selection"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-get_mask_selection}{}}}
\subsection{Method \code{get_mask_selection()}}{
Retrieve data using a boolean mask. Not yet implemented.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$get_mask_selection(selection = NA, out = NA, fields = NA)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{selection}}{Selection specifying the subset to retrieve.}

\item{\code{out}}{Not yet implemented.}

\item{\code{fields}}{Not yet implemented.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\link{NestedArray}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-set_item"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-set_item}{}}}
\subsection{Method \code{set_item()}}{
Set a subset of the array.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$set_item(selection, value)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{selection}}{Selection specifying the subset to set.}

\item{\code{value}}{The value to set, as an R array() or a Zarr NestedArray instance.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{NULL} (called for side effects).
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-set_basic_selection"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-set_basic_selection}{}}}
\subsection{Method \code{set_basic_selection()}}{
Set a selection of the array using basic (integer/slice) indexing.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$set_basic_selection(selection, value, fields = NA)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{selection}}{Selection specifying the subset to set.}

\item{\code{value}}{The value to set.}

\item{\code{fields}}{Not yet implemented.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{NULL} (called for side effects).
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-set_orthogonal_selection"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-set_orthogonal_selection}{}}}
\subsection{Method \code{set_orthogonal_selection()}}{
Set data using orthogonal (outer) indexing. Not yet implemented.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$set_orthogonal_selection(selection, value, fields = NA)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{selection}}{Selection specifying the subset to set.}

\item{\code{value}}{Not yet implemented.}

\item{\code{fields}}{Not yet implemented.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{NULL} (called for side effects).
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-set_coordinate_selection"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-set_coordinate_selection}{}}}
\subsection{Method \code{set_coordinate_selection()}}{
Set data using coordinate (point) indexing. Not yet implemented.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$set_coordinate_selection(selection, value, fields = NA)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{selection}}{Selection specifying the subset to set.}

\item{\code{value}}{Not yet implemented.}

\item{\code{fields}}{Not yet implemented.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{NULL} (called for side effects).
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-set_mask_selection"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-set_mask_selection}{}}}
\subsection{Method \code{set_mask_selection()}}{
Set data using a boolean mask. Not yet implemented.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$set_mask_selection(selection, value, fields = NA)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{selection}}{Selection specifying the subset to set.}

\item{\code{value}}{Not yet implemented.}

\item{\code{fields}}{Not yet implemented.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{NULL} (called for side effects).
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-get_info"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-get_info}{}}}
\subsection{Method \code{get_info()}}{
Get array diagnostic info. Not yet implemented.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$get_info()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
\code{character(1)}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-get_digest"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-get_digest}{}}}
\subsection{Method \code{get_digest()}}{
Compute a checksum digest of the array data. Not yet implemented.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$get_digest(hashname = "sha1")}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{hashname}}{Name of hash.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{character(1)}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-get_hexdigest"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-get_hexdigest}{}}}
\subsection{Method \code{get_hexdigest()}}{
Compute a hex-string checksum digest. Not yet implemented.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$get_hexdigest(hashname = "sha1")}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{hashname}}{Name of hash.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{character(1)}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-append"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-append}{}}}
\subsection{Method \code{append()}}{
Append data to the array along the specified axis.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$append(data, axis = 1)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{data}}{Data to append. Will be coerced to array if needed.}

\item{\code{axis}}{Axis to append along (1-indexed, R convention). Default 1.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{NULL} (called for side effects).
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-view"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-view}{}}}
\subsection{Method \code{view()}}{
Return a view on the array with modified interpretation. Not yet implemented.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$view(
  shape = NA,
  chunks = NA,
  dtype = NA,
  fill_value = NA,
  filters = NA,
  read_only = NA,
  synchronizer = NA
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{shape}}{Not yet implemented.}

\item{\code{chunks}}{Not yet implemented.}

\item{\code{dtype}}{Not yet implemented.}

\item{\code{fill_value}}{Not yet implemented.}

\item{\code{filters}}{Not yet implemented.}

\item{\code{read_only}}{Not yet implemented.}

\item{\code{synchronizer}}{Not yet implemented.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\link{ZarrArray}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-astype"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-astype}{}}}
\subsection{Method \code{astype()}}{
Return a view with a different dtype. Not yet implemented.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$astype(dtype)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{dtype}}{Not yet implemented.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\link{ZarrArray}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-get_dimension_separator"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-get_dimension_separator}{}}}
\subsection{Method \code{get_dimension_separator()}}{
Get the dimension separator used in chunk keys.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$get_dimension_separator()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
\code{character(1)}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-as.array"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-as.array}{}}}
\subsection{Method \code{as.array()}}{
Convert Zarr object to R array (for S3 method). Note that this loads all data into memory.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$as.array()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
An R \code{array}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrArray-clone"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrArray-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrArray$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/indexing.R
\docType{class}
\name{OIndex}
\alias{OIndex}
\title{OIndex Class}
\description{
Orthogonal index class
}
\details{
The Zarr OIndex class.
}
\keyword{internal}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{array}}{The array instance.}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-OIndex-new}{\code{OIndex$new()}}
\item \href{#method-OIndex-get_item}{\code{OIndex$get_item()}}
\item \href{#method-OIndex-clone}{\code{OIndex$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-OIndex-new"></a>}}
\if{latex}{\out{\hypertarget{method-OIndex-new}{}}}
\subsection{Method \code{new()}}{
Create a new OIndex instance.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{OIndex$new(array)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{array}}{The array instance.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
An \code{OIndex} instance.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-OIndex-get_item"></a>}}
\if{latex}{\out{\hypertarget{method-OIndex-get_item}{}}}
\subsection{Method \code{get_item()}}{
get method for the Oindex instance
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{OIndex$get_item(selection)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{selection}}{Selection to apply.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
An \code{OIndex} instance.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-OIndex-clone"></a>}}
\if{latex}{\out{\hypertarget{method-OIndex-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{OIndex$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zarr-array.R
\name{[.ZarrArray}
\alias{[.ZarrArray}
\title{S3 method for custom bracket subsetting}
\usage{
\method{[}{ZarrArray}(obj, ...)
}
\arguments{
\item{obj}{Object.}

\item{...}{Dots.}
}
\description{
S3 method for custom bracket subsetting
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/array-nested.R
\docType{class}
\name{NestedArray}
\alias{NestedArray}
\title{NestedArray Class}
\format{
\link[R6:R6Class]{R6::R6Class}
}
\description{
Represents a multi-dimensional array that can be
accessed and subsetted via list of Slice instances.
}
\details{
The Zarr NestedArray class.
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{shape}}{The shape of the array.}

\item{\code{dtype}}{The Zarr dtype of the array, as a string like ">f8".}

\item{\code{dtype_obj}}{The Zarr dtype of the array, as a Dtype instance.}

\item{\code{data}}{The array contents as a base R array.}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-NestedArray-new}{\code{NestedArray$new()}}
\item \href{#method-NestedArray-get}{\code{NestedArray$get()}}
\item \href{#method-NestedArray-set}{\code{NestedArray$set()}}
\item \href{#method-NestedArray-flatten}{\code{NestedArray$flatten()}}
\item \href{#method-NestedArray-flatten_to_raw}{\code{NestedArray$flatten_to_raw()}}
\item \href{#method-NestedArray-as.array}{\code{NestedArray$as.array()}}
\item \href{#method-NestedArray-clone}{\code{NestedArray$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-NestedArray-new"></a>}}
\if{latex}{\out{\hypertarget{method-NestedArray-new}{}}}
\subsection{Method \code{new()}}{
Create a new NestedArray instance.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{NestedArray$new(data, shape = NA, dtype = NA, order = NA)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{data}}{The data to initialize the array with.
Either NULL, base R array, base R vector (numeric/logical),
scalar, or raw vector.}

\item{\code{shape}}{The shape of the array.}

\item{\code{dtype}}{The Zarr dtype of the array, as a string like ">f8".}

\item{\code{order}}{The order of the array, either "C" or "F". Only used
when \code{data} is a raw vector. Optional.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A \code{NestedArray} instance.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-NestedArray-get"></a>}}
\if{latex}{\out{\hypertarget{method-NestedArray-get}{}}}
\subsection{Method \code{get()}}{
Subset the array.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{NestedArray$get(selection)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{selection}}{A list of slices.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A new NestedArray (potentially a subset) representing the selection.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-NestedArray-set"></a>}}
\if{latex}{\out{\hypertarget{method-NestedArray-set}{}}}
\subsection{Method \code{set()}}{
Set a subset of the array.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{NestedArray$set(selection, value)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{selection}}{A list of slices.}

\item{\code{value}}{A NestedArray or a base R array.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{NULL} (called for side effects, modifies \code{self$data} in place).
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-NestedArray-flatten"></a>}}
\if{latex}{\out{\hypertarget{method-NestedArray-flatten}{}}}
\subsection{Method \code{flatten()}}{
Flatten the array contents.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{NestedArray$flatten(order = NA)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{order}}{Either "C", "F", or NA.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
The data as a flat vector.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-NestedArray-flatten_to_raw"></a>}}
\if{latex}{\out{\hypertarget{method-NestedArray-flatten_to_raw}{}}}
\subsection{Method \code{flatten_to_raw()}}{
Flatten the array contents and convert to a raw vector.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{NestedArray$flatten_to_raw(order = NA)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{order}}{Either "C", "F", or NA.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
The data as a flat \code{raw()} vector (or plain vector for object dtypes).
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-NestedArray-as.array"></a>}}
\if{latex}{\out{\hypertarget{method-NestedArray-as.array}{}}}
\subsection{Method \code{as.array()}}{
Convert NestedArray to a base R array.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{NestedArray$as.array()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
\code{array()}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-NestedArray-clone"></a>}}
\if{latex}{\out{\hypertarget{method-NestedArray-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{NestedArray$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{replace_ellipsis}
\alias{replace_ellipsis}
\title{Convert user selections, potentially containing "...", to a list of slices
that can be used internally.}
\usage{
replace_ellipsis(selection, shape)
}
\arguments{
\item{selection}{The user-provided selection list.}

\item{shape}{The shape of the array, to be used to fill in ellipsis values.}
}
\value{
A list of selections with ellipsis values converted to NA.
}
\description{
Convert user selections, potentially containing "...", to a list of slices
that can be used internally.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extendr-wrappers.R
\name{zarrs_compiled_features}
\alias{zarrs_compiled_features}
\title{Return compiled zarrs feature flags as a character vector.}
\usage{
zarrs_compiled_features()
}
\description{
Called once at \code{.onLoad} to populate \code{.pizzarr_env$zarrs_available}.
The feature list is determined at compile time via \verb{cfg!} checks.
Also installs a no-op panic hook on first call.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/indexing.R
\docType{class}
\name{IntDimIndexer}
\alias{IntDimIndexer}
\title{IntDimIndexer Class}
\description{
Indexer for a single integer selection along one dimension.
}
\details{
The Zarr IntDimIndexer class.
}
\keyword{internal}
\section{Super class}{
\code{\link[pizzarr:DimIndexer]{pizzarr::DimIndexer}} -> \code{IntDimIndexer}
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{dim_sel}}{Normalized integer index for this dimension.}

\item{\code{dim_len}}{Length of the array along this dimension.}

\item{\code{dim_chunk_len}}{Length of a chunk along this dimension.}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-IntDimIndexer-new}{\code{IntDimIndexer$new()}}
\item \href{#method-IntDimIndexer-iter}{\code{IntDimIndexer$iter()}}
\item \href{#method-IntDimIndexer-clone}{\code{IntDimIndexer$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-IntDimIndexer-new"></a>}}
\if{latex}{\out{\hypertarget{method-IntDimIndexer-new}{}}}
\subsection{Method \code{new()}}{
Create a new IntDimIndexer instance.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{IntDimIndexer$new(dim_sel, dim_len, dim_chunk_len)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{dim_sel}}{Integer dimension selection.}

\item{\code{dim_len}}{Integer dimension length.}

\item{\code{dim_chunk_len}}{Integer dimension chunk length.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A \code{IntDimIndexer} instance.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-IntDimIndexer-iter"></a>}}
\if{latex}{\out{\hypertarget{method-IntDimIndexer-iter}{}}}
\subsection{Method \code{iter()}}{
Compute the chunk dimension projection for the single selected index.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{IntDimIndexer$iter()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
a \code{ChunkDimProjection} instance
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-IntDimIndexer-clone"></a>}}
\if{latex}{\out{\hypertarget{method-IntDimIndexer-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{IntDimIndexer$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stores.R
\docType{class}
\name{Store}
\alias{Store}
\title{Store Class}
\format{
\link[R6:R6Class]{R6::R6Class}
}
\description{
Class representing an abstract store
}
\details{
Abstract store for Zarr
}
\seealso{
Other Store classes: 
\code{\link{DirectoryStore}},
\code{\link{GcsStore}},
\code{\link{HttpStore}},
\code{\link{MemoryStore}},
\code{\link{S3Store}}
}
\concept{Store classes}
\keyword{internal}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{metadata_class}}{The metadata encoder/decoder for this store.}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-Store-new}{\code{Store$new()}}
\item \href{#method-Store-is_readable}{\code{Store$is_readable()}}
\item \href{#method-Store-is_writeable}{\code{Store$is_writeable()}}
\item \href{#method-Store-is_erasable}{\code{Store$is_erasable()}}
\item \href{#method-Store-is_listable}{\code{Store$is_listable()}}
\item \href{#method-Store-close}{\code{Store$close()}}
\item \href{#method-Store-listdir}{\code{Store$listdir()}}
\item \href{#method-Store-rename}{\code{Store$rename()}}
\item \href{#method-Store-rmdir}{\code{Store$rmdir()}}
\item \href{#method-Store-get_item}{\code{Store$get_item()}}
\item \href{#method-Store-set_item}{\code{Store$set_item()}}
\item \href{#method-Store-contains_item}{\code{Store$contains_item()}}
\item \href{#method-Store-delete_item}{\code{Store$delete_item()}}
\item \href{#method-Store-get_consolidated_metadata}{\code{Store$get_consolidated_metadata()}}
\item \href{#method-Store-get_store_identifier}{\code{Store$get_store_identifier()}}
\item \href{#method-Store-print}{\code{Store$print()}}
\item \href{#method-Store-clone}{\code{Store$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Store-new"></a>}}
\if{latex}{\out{\hypertarget{method-Store-new}{}}}
\subsection{Method \code{new()}}{
Create a \code{Store} object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Store$new()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Store-is_readable"></a>}}
\if{latex}{\out{\hypertarget{method-Store-is_readable}{}}}
\subsection{Method \code{is_readable()}}{
Test if Store is readable.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Store$is_readable()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Logical.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Store-is_writeable"></a>}}
\if{latex}{\out{\hypertarget{method-Store-is_writeable}{}}}
\subsection{Method \code{is_writeable()}}{
Test if Store is writeable.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Store$is_writeable()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Logical.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Store-is_erasable"></a>}}
\if{latex}{\out{\hypertarget{method-Store-is_erasable}{}}}
\subsection{Method \code{is_erasable()}}{
Test if Store is eraseable.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Store$is_erasable()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Logical.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Store-is_listable"></a>}}
\if{latex}{\out{\hypertarget{method-Store-is_listable}{}}}
\subsection{Method \code{is_listable()}}{
Test if Store is listable.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Store$is_listable()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Logical.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Store-close"></a>}}
\if{latex}{\out{\hypertarget{method-Store-close}{}}}
\subsection{Method \code{close()}}{
Close the store.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Store$close()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
\code{NULL}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Store-listdir"></a>}}
\if{latex}{\out{\hypertarget{method-Store-listdir}{}}}
\subsection{Method \code{listdir()}}{
List the store directory.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Store$listdir(path = NA)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{path}}{character path.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Character vector of keys.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Store-rename"></a>}}
\if{latex}{\out{\hypertarget{method-Store-rename}{}}}
\subsection{Method \code{rename()}}{
Rename a Store path.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Store$rename(src_path, dst_path)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{src_path}}{character source path.}

\item{\code{dst_path}}{character destination path.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{NULL} (called for side effects).
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Store-rmdir"></a>}}
\if{latex}{\out{\hypertarget{method-Store-rmdir}{}}}
\subsection{Method \code{rmdir()}}{
Remove a path within a Store.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Store$rmdir(path)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{path}}{character path.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{NULL} (called for side effects).
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Store-get_item"></a>}}
\if{latex}{\out{\hypertarget{method-Store-get_item}{}}}
\subsection{Method \code{get_item()}}{
Get an item from the store.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Store$get_item(key)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{key}}{The item key.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
The item data in a vector of type raw.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Store-set_item"></a>}}
\if{latex}{\out{\hypertarget{method-Store-set_item}{}}}
\subsection{Method \code{set_item()}}{
Set an item in the store.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Store$set_item(key, value)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{key}}{The item key.}

\item{\code{value}}{The item value as a vector of type raw.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{NULL} (called for side effects).
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Store-contains_item"></a>}}
\if{latex}{\out{\hypertarget{method-Store-contains_item}{}}}
\subsection{Method \code{contains_item()}}{
Determine whether the store contains an item.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Store$contains_item(key)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{key}}{The item key.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A boolean value.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Store-delete_item"></a>}}
\if{latex}{\out{\hypertarget{method-Store-delete_item}{}}}
\subsection{Method \code{delete_item()}}{
Delete an item from the store.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Store$delete_item(key)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{key}}{The item key.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{NULL} (called for side effects).
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Store-get_consolidated_metadata"></a>}}
\if{latex}{\out{\hypertarget{method-Store-get_consolidated_metadata}{}}}
\subsection{Method \code{get_consolidated_metadata()}}{
Get consolidated metadata if it exists.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Store$get_consolidated_metadata()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
A list or \code{NULL}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Store-get_store_identifier"></a>}}
\if{latex}{\out{\hypertarget{method-Store-get_store_identifier}{}}}
\subsection{Method \code{get_store_identifier()}}{
Return a store identifier for zarrs dispatch.

Subclasses override this to return a filesystem path or URL.
Returns \code{NULL} by default (signals: use R-native path).
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Store$get_store_identifier()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
A character string or \code{NULL}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Store-print"></a>}}
\if{latex}{\out{\hypertarget{method-Store-print}{}}}
\subsection{Method \code{print()}}{
Print a human-readable summary of the store.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Store$print(...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{...}}{Ignored.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{self} (invisibly).
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Store-clone"></a>}}
\if{latex}{\out{\hypertarget{method-Store-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Store$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zarrs-dispatch.R
\name{selection_to_ranges}
\alias{selection_to_ranges}
\title{Convert pizzarr indexer to zarrs ranges}
\usage{
selection_to_ranges(indexer)
}
\arguments{
\item{indexer}{A BasicIndexer with SliceDimIndexer or IntDimIndexer dims.}
}
\value{
List of integer vectors.
}
\description{
Convert pizzarr indexer to zarrs ranges
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/options.R
\docType{data}
\name{pizzarr_option_defaults}
\alias{pizzarr_option_defaults}
\title{pizzarr_option_defaults}
\format{
An object of class \code{list} of length 4.
}
\usage{
pizzarr_option_defaults
}
\description{
\itemize{
\item pizzarr.http_store_cache_time_seconds how long to cache web requests
\item pizzarr.nthreads number of threads for parallel codec operations
(NULL = all CPUs). Set-once: takes effect only before the first zarrs
operation. Use env var PIZZARR_NTHREADS for reliable session-level control.
\item pizzarr.concurrent_target codec concurrency level — how many codec
operations zarrs runs in parallel within a single read/write call
(NULL = zarrs default, typically CPU count). Can be changed at any time.
\item pizzarr.http_batch_range_requests whether HTTP stores use multipart
range requests (TRUE by default). Set to FALSE for servers with
incomplete multipart range support. Takes effect on next store open.
}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/slicing.R
\name{slice}
\alias{slice}
\title{Convenience function for the internal Slice R6 class constructor.}
\usage{
slice(start, stop = NA, step = NA, zero_based = FALSE)
}
\arguments{
\item{start}{The start index.}

\item{stop}{The stop index.}

\item{step}{The step size. Negative values reverse the direction of the
slice, matching Python/NumPy semantics (e.g., \code{step = -1} iterates
backward from \code{start} toward \code{stop}).}

\item{zero_based}{The index of the dimension. By default, FALSE for R-like behavior.}
}
\value{
A Slice instance with the specified parameters.
}
\description{
Convenience function for the internal Slice R6 class constructor.
}
\examples{
g <- zarr_volcano()
v <- g$get_item("volcano")

# Reverse the first 5 columns of row 1 (zero-based: 5:0:-1)
v$get_orthogonal_selection(list(zb_slice(0, 1), zb_slice(5, 0, -1)))

# Full reverse of row 1 (zero-based: -1::-1)
v$get_orthogonal_selection(list(zb_slice(0, 1), zb_slice(-1, NA, -1)))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{is_na}
\alias{is_na}
\title{Check if a value, potentially a vector, is NA}
\usage{
is_na(val)
}
\arguments{
\item{val}{The value to check}
}
\value{
Whether the value is NA
}
\description{
Check if a value, potentially a vector, is NA
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/atomic.R
\name{as_scalar}
\alias{as_scalar}
\title{Convert a value to a scalar to opt-out of R default vector casting behavior.
This uses the \code{jsonlite::unbox} function to "tag" the value as a scalar.}
\usage{
as_scalar(obj)
}
\arguments{
\item{obj}{The value to convert.}
}
\value{
The value wrapped as a scalar.
}
\description{
Convert a value to a scalar to opt-out of R default vector casting behavior.
This uses the \code{jsonlite::unbox} function to "tag" the value as a scalar.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stores.R
\docType{class}
\name{HttpStore}
\alias{HttpStore}
\title{HttpStore Class}
\format{
\link[R6:R6Class]{R6::R6Class} inheriting from \link{Store}.
}
\description{
Store class that uses HTTP requests.
Read-only. Depends on the \code{crul} package.
}
\seealso{
Other Store classes: 
\code{\link{DirectoryStore}},
\code{\link{GcsStore}},
\code{\link{MemoryStore}},
\code{\link{S3Store}},
\code{\link{Store}}
}
\concept{Store classes}
\section{Super class}{
\code{\link[pizzarr:Store]{pizzarr::Store}} -> \code{HttpStore}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-HttpStore-new}{\code{HttpStore$new()}}
\item \href{#method-HttpStore-get_item}{\code{HttpStore$get_item()}}
\item \href{#method-HttpStore-contains_item}{\code{HttpStore$contains_item()}}
\item \href{#method-HttpStore-listdir}{\code{HttpStore$listdir()}}
\item \href{#method-HttpStore-get_cache_time_seconds}{\code{HttpStore$get_cache_time_seconds()}}
\item \href{#method-HttpStore-set_cache_time_seconds}{\code{HttpStore$set_cache_time_seconds()}}
\item \href{#method-HttpStore-get_store_identifier}{\code{HttpStore$get_store_identifier()}}
\item \href{#method-HttpStore-print}{\code{HttpStore$print()}}
\item \href{#method-HttpStore-clone}{\code{HttpStore$clone()}}
}
}
\if{html}{\out{
<details><summary>Inherited methods</summary>
<ul>
<li><span class="pkg-link" data-pkg="pizzarr" data-topic="Store" data-id="close"><a href='../../pizzarr/html/Store.html#method-Store-close'><code>pizzarr::Store$close()</code></a></span></li>
<li><span class="pkg-link" data-pkg="pizzarr" data-topic="Store" data-id="delete_item"><a href='../../pizzarr/html/Store.html#method-Store-delete_item'><code>pizzarr::Store$delete_item()</code></a></span></li>
<li><span class="pkg-link" data-pkg="pizzarr" data-topic="Store" data-id="get_consolidated_metadata"><a href='../../pizzarr/html/Store.html#method-Store-get_consolidated_metadata'><code>pizzarr::Store$get_consolidated_metadata()</code></a></span></li>
<li><span class="pkg-link" data-pkg="pizzarr" data-topic="Store" data-id="is_erasable"><a href='../../pizzarr/html/Store.html#method-Store-is_erasable'><code>pizzarr::Store$is_erasable()</code></a></span></li>
<li><span class="pkg-link" data-pkg="pizzarr" data-topic="Store" data-id="is_listable"><a href='../../pizzarr/html/Store.html#method-Store-is_listable'><code>pizzarr::Store$is_listable()</code></a></span></li>
<li><span class="pkg-link" data-pkg="pizzarr" data-topic="Store" data-id="is_readable"><a href='../../pizzarr/html/Store.html#method-Store-is_readable'><code>pizzarr::Store$is_readable()</code></a></span></li>
<li><span class="pkg-link" data-pkg="pizzarr" data-topic="Store" data-id="is_writeable"><a href='../../pizzarr/html/Store.html#method-Store-is_writeable'><code>pizzarr::Store$is_writeable()</code></a></span></li>
<li><span class="pkg-link" data-pkg="pizzarr" data-topic="Store" data-id="rename"><a href='../../pizzarr/html/Store.html#method-Store-rename'><code>pizzarr::Store$rename()</code></a></span></li>
<li><span class="pkg-link" data-pkg="pizzarr" data-topic="Store" data-id="rmdir"><a href='../../pizzarr/html/Store.html#method-Store-rmdir'><code>pizzarr::Store$rmdir()</code></a></span></li>
<li><span class="pkg-link" data-pkg="pizzarr" data-topic="Store" data-id="set_item"><a href='../../pizzarr/html/Store.html#method-Store-set_item'><code>pizzarr::Store$set_item()</code></a></span></li>
</ul>
</details>
}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-HttpStore-new"></a>}}
\if{latex}{\out{\hypertarget{method-HttpStore-new}{}}}
\subsection{Method \code{new()}}{
Create a \code{HttpStore} object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{HttpStore$new(url, options = NA, headers = NA)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{url}}{(\code{character(1)})\cr
URL of the store.}

\item{\code{options}}{(\code{list()} or \code{NA})\cr
Options passed to crul.}

\item{\code{headers}}{(\code{list()} or \code{NA})\cr
Headers passed to crul.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A new \code{HttpStore} object.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-HttpStore-get_item"></a>}}
\if{latex}{\out{\hypertarget{method-HttpStore-get_item}{}}}
\subsection{Method \code{get_item()}}{
Get an item from the store.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{HttpStore$get_item(item)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{item}}{The item key.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
The item data in a vector of type raw.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-HttpStore-contains_item"></a>}}
\if{latex}{\out{\hypertarget{method-HttpStore-contains_item}{}}}
\subsection{Method \code{contains_item()}}{
Determine whether the store contains an item.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{HttpStore$contains_item(item)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{item}}{The item key.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A boolean value.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-HttpStore-listdir"></a>}}
\if{latex}{\out{\hypertarget{method-HttpStore-listdir}{}}}
\subsection{Method \code{listdir()}}{
Fetches .zmetadata from the store evaluates its names
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{HttpStore$listdir()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Character vector of unique keys that do not start with a \code{.}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-HttpStore-get_cache_time_seconds"></a>}}
\if{latex}{\out{\hypertarget{method-HttpStore-get_cache_time_seconds}{}}}
\subsection{Method \code{get_cache_time_seconds()}}{
Get cache time of http requests.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{HttpStore$get_cache_time_seconds()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
\code{numeric(1)}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-HttpStore-set_cache_time_seconds"></a>}}
\if{latex}{\out{\hypertarget{method-HttpStore-set_cache_time_seconds}{}}}
\subsection{Method \code{set_cache_time_seconds()}}{
Set cache time of http requests.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{HttpStore$set_cache_time_seconds(seconds)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{seconds}}{Number of seconds until cache is invalid -- 0 for no cache.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{NULL} (called for side effects).
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-HttpStore-get_store_identifier"></a>}}
\if{latex}{\out{\hypertarget{method-HttpStore-get_store_identifier}{}}}
\subsection{Method \code{get_store_identifier()}}{
Print a human-readable summary of the store.


Return the store URL for zarrs dispatch.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{HttpStore$get_store_identifier()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
A character string.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-HttpStore-print"></a>}}
\if{latex}{\out{\hypertarget{method-HttpStore-print}{}}}
\subsection{Method \code{print()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{HttpStore$print(...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{...}}{Ignored.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{self} (invisibly).
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-HttpStore-clone"></a>}}
\if{latex}{\out{\hypertarget{method-HttpStore-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{HttpStore$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/numcodecs.R
\name{get_codec}
\alias{get_codec}
\title{Get a codec instance from the registry.}
\usage{
get_codec(config)
}
\arguments{
\item{config}{A codec config as a named list.}
}
\value{
The instance of the codec.
}
\description{
Get a codec instance from the registry.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/numcodecs.R
\docType{class}
\name{Codec}
\alias{Codec}
\title{Codec Class}
\format{
\link[R6:R6Class]{R6::R6Class}
}
\description{
Abstract class representing a compressor.
}
\details{
Abstract compressor for Zarr
}
\seealso{
Other Codec classes: 
\code{\link{BloscCodec}},
\code{\link{Bz2Codec}},
\code{\link{GzipCodec}},
\code{\link{Lz4Codec}},
\code{\link{LzmaCodec}},
\code{\link{VLenUtf8Codec}},
\code{\link{ZlibCodec}},
\code{\link{ZstdCodec}}
}
\concept{Codec classes}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-Codec-encode}{\code{Codec$encode()}}
\item \href{#method-Codec-decode}{\code{Codec$decode()}}
\item \href{#method-Codec-get_config}{\code{Codec$get_config()}}
\item \href{#method-Codec-clone}{\code{Codec$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Codec-encode"></a>}}
\if{latex}{\out{\hypertarget{method-Codec-encode}{}}}
\subsection{Method \code{encode()}}{
Compress data.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Codec$encode(buf, zarr_arr)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{buf}}{(\code{raw()})\cr
The un-compressed data.}

\item{\code{zarr_arr}}{(\link{ZarrArray})\cr
The ZarrArray instance.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Compressed data.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Codec-decode"></a>}}
\if{latex}{\out{\hypertarget{method-Codec-decode}{}}}
\subsection{Method \code{decode()}}{
Decompress data.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Codec$decode(buf, zarr_arr)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{buf}}{(\code{raw()})\cr
The compressed data.}

\item{\code{zarr_arr}}{(\link{ZarrArray})\cr
The ZarrArray instance.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Un-compressed data.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Codec-get_config"></a>}}
\if{latex}{\out{\hypertarget{method-Codec-get_config}{}}}
\subsection{Method \code{get_config()}}{
Get codec configuration as a list.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Codec$get_config()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
A named list.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Codec-clone"></a>}}
\if{latex}{\out{\hypertarget{method-Codec-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Codec$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/atomic.R
\name{is_bool_list}
\alias{is_bool_list}
\title{Check that a value is a list of one or more boolean}
\usage{
is_bool_list(s)
}
\description{
Check that a value is a list of one or more boolean
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zarrs-bridge.R
\name{pizzarr_upgrade}
\alias{pizzarr_upgrade}
\title{Upgrade to the zarrs backend}
\usage{
pizzarr_upgrade()
}
\description{
Prints the command to install pizzarr from r-universe with the
compiled zarrs backend, or reports that zarrs is already available.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{raw_to_double_from_int64}
\alias{raw_to_double_from_int64}
\title{Convert raw bytes to double from int64 (fallback when bit64 unavailable)}
\usage{
raw_to_double_from_int64(buf, n, endian = "little", signed = TRUE)
}
\arguments{
\item{buf}{Raw vector of 8-byte integers.}

\item{n}{Number of elements to read.}

\item{endian}{"little" or "big".}

\item{signed}{Logical; if TRUE, treat as signed int64.}
}
\value{
A double vector (precision loss possible for values > 2^53).
}
\description{
Convert raw bytes to double from int64 (fallback when bit64 unavailable)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/indexing.R
\docType{class}
\name{Order}
\alias{Order}
\title{Order Class}
\description{
Determines the sort order of an integer array selection (increasing, decreasing, or unordered).
}
\details{
The Order class.
}
\keyword{internal}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{UNKNOWN}}{UNKNOWN}

\item{\code{INCREASING}}{INCREASING}

\item{\code{DECREASING}}{DECREASING}

\item{\code{UNORDERED}}{UNORDERED}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-Order-check}{\code{Order$check()}}
\item \href{#method-Order-clone}{\code{Order$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Order-check"></a>}}
\if{latex}{\out{\hypertarget{method-Order-check}{}}}
\subsection{Method \code{check()}}{
checking order of numbers
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Order$check(a)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{a}}{Vector of numbers.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Order-clone"></a>}}
\if{latex}{\out{\hypertarget{method-Order-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Order$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{raw_to_integer64}
\alias{raw_to_integer64}
\title{Convert raw bytes to integer64 vector}
\usage{
raw_to_integer64(buf, n, endian = "little")
}
\arguments{
\item{buf}{Raw vector of 8-byte little-endian integers.}

\item{n}{Number of elements to read.}

\item{endian}{"little" or "big".}
}
\value{
A bit64::integer64 vector.
}
\description{
Convert raw bytes to integer64 vector
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zarrs-dispatch.R
\name{can_use_zarrs}
\alias{can_use_zarrs}
\title{Check whether the zarrs path can handle this selection}
\usage{
can_use_zarrs(indexer, store)
}
\arguments{
\item{indexer}{A BasicIndexer or OrthogonalIndexer object.}

\item{store}{A Store object.}
}
\value{
Logical scalar.
}
\description{
Returns TRUE for BasicIndexer with contiguous SliceDimIndexer
(step==1) and/or IntDimIndexer on every dimension. Step>1 slices,
IntArrayDimIndexer (fancy indexing), and OrthogonalIndexer fall
through to R-native.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{zarr_volcano}
\alias{zarr_volcano}
\title{Create a demo Zarr group containing R's volcano dataset}
\usage{
zarr_volcano()
}
\value{
A \code{ZarrGroup} containing a single array called \code{"volcano"}.
}
\description{
Writes the \code{\link[datasets]{volcano}} matrix into a temporary
DirectoryStore as a Zarr array named \code{"volcano"} and returns the
opened group.
}
\examples{
g <- zarr_volcano()
v <- g$get_item("volcano")
image(v$get_item("...")$data, main = "Maunga Whau Volcano")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dtypes.R
\docType{class}
\name{Dtype}
\alias{Dtype}
\title{Dtype Class}
\format{
\link[R6:R6Class]{R6::R6Class}
}
\description{
A data type object (an instance of Dtype class) describes how
the bytes in the fixed-size block of memory corresponding to an array
item should be interpreted.
}
\details{
The Zarr Dtype class.
}
\references{
https://numpy.org/doc/stable/reference/arrays.dtypes.html
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{dtype}}{The original dtype string, like "<f4".}

\item{\code{byte_order}}{The byte order of the dtype, either "little", "big", or "nr".}

\item{\code{basic_type}}{The basic type of the dtype, like "f".}

\item{\code{num_bytes}}{The number of bytes of the dtype.}

\item{\code{num_items}}{The number of items of the dtype.}

\item{\code{is_signed}}{Whether the dtype is signed. Logical/boolean.}

\item{\code{is_structured}}{Whether the dtype is structured. Logical/boolean.}

\item{\code{is_object}}{Whether the dtype is an object. Logical/boolean.}

\item{\code{object_codec}}{The object codec instance.}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-Dtype-new}{\code{Dtype$new()}}
\item \href{#method-Dtype-get_asrtype}{\code{Dtype$get_asrtype()}}
\item \href{#method-Dtype-get_rtype}{\code{Dtype$get_rtype()}}
\item \href{#method-Dtype-get_typed_array_ctr}{\code{Dtype$get_typed_array_ctr()}}
\item \href{#method-Dtype-clone}{\code{Dtype$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Dtype-new"></a>}}
\if{latex}{\out{\hypertarget{method-Dtype-new}{}}}
\subsection{Method \code{new()}}{
Create a new Dtype instance.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Dtype$new(dtype, object_codec = NA)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{dtype}}{The original dtype string, like "<f4".}

\item{\code{object_codec}}{The object codec instance.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A \code{Dtype} instance.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Dtype-get_asrtype"></a>}}
\if{latex}{\out{\hypertarget{method-Dtype-get_asrtype}{}}}
\subsection{Method \code{get_asrtype()}}{
Get the R coercion function name for this dtype.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Dtype$get_asrtype()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Character string (e.g., \code{"as.double"}).
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Dtype-get_rtype"></a>}}
\if{latex}{\out{\hypertarget{method-Dtype-get_rtype}{}}}
\subsection{Method \code{get_rtype()}}{
Get the R base type for this dtype.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Dtype$get_rtype()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
An R prototype value (e.g., \code{integer()}, \code{double()}, or \code{bit64::integer64()}).
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Dtype-get_typed_array_ctr"></a>}}
\if{latex}{\out{\hypertarget{method-Dtype-get_typed_array_ctr}{}}}
\subsection{Method \code{get_typed_array_ctr()}}{
Get a constructor function for typed arrays of this dtype.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Dtype$get_typed_array_ctr()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
A function that takes \code{dim} and returns an array.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Dtype-clone"></a>}}
\if{latex}{\out{\hypertarget{method-Dtype-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Dtype$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/meta.R
\name{decode_fill_value_v3}
\alias{decode_fill_value_v3}
\title{Encode a fill_value for V3 zarr.json.
V3 requires special float values as JSON strings: NaN, Infinity, -Infinity.
Normal numeric values are returned as unboxed scalars.}
\usage{
decode_fill_value_v3(fill_value)
}
\arguments{
\item{fill_value}{The fill value to encode.}
}
\value{
A value suitable for inclusion in a jsonlite::toJSON call.
}
\description{
Encode a fill_value for V3 zarr.json.
V3 requires special float values as JSON strings: NaN, Infinity, -Infinity.
Normal numeric values are returned as unboxed scalars.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{has_bit64}
\alias{has_bit64}
\title{Check if the bit64 package is available}
\usage{
has_bit64()
}
\value{
Logical.
}
\description{
Check if the bit64 package is available
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/numcodecs.R
\name{get_default_compressor}
\alias{get_default_compressor}
\title{Get the default compressor.}
\usage{
get_default_compressor()
}
\value{
A Zstd compressor instance.
}
\description{
Get the default compressor.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/indexing-types.R
\name{Indexer}
\alias{Indexer}
\title{Indexer}
\format{
\link[R6:R6Class]{R6::R6Class} object.
}
\description{
Abstract base class for chunk indexers. An indexer maps a user-level
array selection to a sequence of \link{ChunkProjection} objects, each
describing which items to read or write within a single chunk.
Concrete subclasses include \link{BasicIndexer} and \link{OrthogonalIndexer}.
}
\seealso{
Other Indexing classes: 
\code{\link{ChunkDimProjection}},
\code{\link{ChunkProjection}},
\code{\link{DimIndexer}}
}
\concept{Indexing classes}
\keyword{internal}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{shape}}{(\code{integer()})\cr
Shape of the selection.}

\item{\code{drop_axes}}{(\code{integer()} | \code{NULL})\cr
Axes to drop from the result.}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-Indexer-iter}{\code{Indexer$iter()}}
\item \href{#method-Indexer-clone}{\code{Indexer$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Indexer-iter"></a>}}
\if{latex}{\out{\hypertarget{method-Indexer-iter}{}}}
\subsection{Method \code{iter()}}{
Iterate over \link{ChunkProjection} objects for the selection.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Indexer$iter()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
\code{list()} of \link{ChunkProjection} objects.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Indexer-clone"></a>}}
\if{latex}{\out{\hypertarget{method-Indexer-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Indexer$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/indexing.R
\docType{class}
\name{VIndex}
\alias{VIndex}
\title{VIndex Class}
\description{
Vectorized index class
}
\details{
The Zarr VIndex class.
}
\keyword{internal}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{array}}{The array instance.}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-VIndex-new}{\code{VIndex$new()}}
\item \href{#method-VIndex-clone}{\code{VIndex$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-VIndex-new"></a>}}
\if{latex}{\out{\hypertarget{method-VIndex-new}{}}}
\subsection{Method \code{new()}}{
Create a new VIndex instance.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{VIndex$new(array)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{array}}{The array instance.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A \code{VIndex} instance.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-VIndex-clone"></a>}}
\if{latex}{\out{\hypertarget{method-VIndex-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{VIndex$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{pizzarr_sample}
\alias{pizzarr_sample}
\title{pizzarr demo data}
\usage{
pizzarr_sample(
  dataset = NULL,
  outdir = file.path(tools::R_user_dir("pizzarr"), "pizzarr_sample")
)
}
\arguments{
\item{dataset}{character defining which demo dataset is desired,
If NULL, all are returned}

\item{outdir}{character directory path to store sample zarr stores}
}
\value{
path to ready to use zarr store
}
\description{
pizzarr demo data
}
\details{
For directory stores, unzips the store to a temporary directory
and returns the resulting path.
}
\examples{
\dontshow{if (requireNamespace("crul", quietly = TRUE)) withAutoprint(\{ # examplesIf}

sample_dir <- tools::R_user_dir("pizzarr")

clean <- !dir.exists(sample_dir)

zarr_samples <- pizzarr_sample(outdir = sample_dir)

#printing without system path for example
gsub(sample_dir, "...", zarr_samples, fixed = TRUE)

# clean up if you don't want to keep them for next time
if(clean) unlink(sample_dir, recursive = TRUE)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stores.R
\docType{class}
\name{GcsStore}
\alias{GcsStore}
\title{GcsStore Class}
\format{
\link[R6:R6Class]{R6::R6Class}
}
\description{
Thin store wrapper for Google Cloud Storage URLs. All I/O is delegated
to the zarrs Rust backend via \code{object_store}. Requires the \code{gcs}
compiled feature (r-universe tier).
}
\details{
GCS Store for Zarr (zarrs backend)
}
\seealso{
Other Store classes: 
\code{\link{DirectoryStore}},
\code{\link{HttpStore}},
\code{\link{MemoryStore}},
\code{\link{S3Store}},
\code{\link{Store}}
}
\concept{Store classes}
\section{Super class}{
\code{\link[pizzarr:Store]{pizzarr::Store}} -> \code{GcsStore}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-GcsStore-new}{\code{GcsStore$new()}}
\item \href{#method-GcsStore-get_store_identifier}{\code{GcsStore$get_store_identifier()}}
\item \href{#method-GcsStore-print}{\code{GcsStore$print()}}
\item \href{#method-GcsStore-clone}{\code{GcsStore$clone()}}
}
}
\if{html}{\out{
<details><summary>Inherited methods</summary>
<ul>
<li><span class="pkg-link" data-pkg="pizzarr" data-topic="Store" data-id="close"><a href='../../pizzarr/html/Store.html#method-Store-close'><code>pizzarr::Store$close()</code></a></span></li>
<li><span class="pkg-link" data-pkg="pizzarr" data-topic="Store" data-id="contains_item"><a href='../../pizzarr/html/Store.html#method-Store-contains_item'><code>pizzarr::Store$contains_item()</code></a></span></li>
<li><span class="pkg-link" data-pkg="pizzarr" data-topic="Store" data-id="delete_item"><a href='../../pizzarr/html/Store.html#method-Store-delete_item'><code>pizzarr::Store$delete_item()</code></a></span></li>
<li><span class="pkg-link" data-pkg="pizzarr" data-topic="Store" data-id="get_consolidated_metadata"><a href='../../pizzarr/html/Store.html#method-Store-get_consolidated_metadata'><code>pizzarr::Store$get_consolidated_metadata()</code></a></span></li>
<li><span class="pkg-link" data-pkg="pizzarr" data-topic="Store" data-id="get_item"><a href='../../pizzarr/html/Store.html#method-Store-get_item'><code>pizzarr::Store$get_item()</code></a></span></li>
<li><span class="pkg-link" data-pkg="pizzarr" data-topic="Store" data-id="is_erasable"><a href='../../pizzarr/html/Store.html#method-Store-is_erasable'><code>pizzarr::Store$is_erasable()</code></a></span></li>
<li><span class="pkg-link" data-pkg="pizzarr" data-topic="Store" data-id="is_listable"><a href='../../pizzarr/html/Store.html#method-Store-is_listable'><code>pizzarr::Store$is_listable()</code></a></span></li>
<li><span class="pkg-link" data-pkg="pizzarr" data-topic="Store" data-id="is_readable"><a href='../../pizzarr/html/Store.html#method-Store-is_readable'><code>pizzarr::Store$is_readable()</code></a></span></li>
<li><span class="pkg-link" data-pkg="pizzarr" data-topic="Store" data-id="is_writeable"><a href='../../pizzarr/html/Store.html#method-Store-is_writeable'><code>pizzarr::Store$is_writeable()</code></a></span></li>
<li><span class="pkg-link" data-pkg="pizzarr" data-topic="Store" data-id="listdir"><a href='../../pizzarr/html/Store.html#method-Store-listdir'><code>pizzarr::Store$listdir()</code></a></span></li>
<li><span class="pkg-link" data-pkg="pizzarr" data-topic="Store" data-id="rename"><a href='../../pizzarr/html/Store.html#method-Store-rename'><code>pizzarr::Store$rename()</code></a></span></li>
<li><span class="pkg-link" data-pkg="pizzarr" data-topic="Store" data-id="rmdir"><a href='../../pizzarr/html/Store.html#method-Store-rmdir'><code>pizzarr::Store$rmdir()</code></a></span></li>
<li><span class="pkg-link" data-pkg="pizzarr" data-topic="Store" data-id="set_item"><a href='../../pizzarr/html/Store.html#method-Store-set_item'><code>pizzarr::Store$set_item()</code></a></span></li>
</ul>
</details>
}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-GcsStore-new"></a>}}
\if{latex}{\out{\hypertarget{method-GcsStore-new}{}}}
\subsection{Method \code{new()}}{
Create a \code{GcsStore}.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{GcsStore$new(url)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{url}}{Character. GCS URL (e.g., \code{"gs://bucket/prefix"}).}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-GcsStore-get_store_identifier"></a>}}
\if{latex}{\out{\hypertarget{method-GcsStore-get_store_identifier}{}}}
\subsection{Method \code{get_store_identifier()}}{
Return the GCS URL for zarrs dispatch.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{GcsStore$get_store_identifier()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
A character string.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-GcsStore-print"></a>}}
\if{latex}{\out{\hypertarget{method-GcsStore-print}{}}}
\subsection{Method \code{print()}}{
Print a human-readable summary of the store.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{GcsStore$print(...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{...}}{Ignored.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{self} (invisibly).
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-GcsStore-clone"></a>}}
\if{latex}{\out{\hypertarget{method-GcsStore-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{GcsStore$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/atomic.R
\name{is_integer_list}
\alias{is_integer_list}
\title{Check that a value is a list of one or more integers.}
\usage{
is_integer_list(s)
}
\description{
Check that a value is a list of one or more integers.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/indexing-types.R
\name{ChunkDimProjection}
\alias{ChunkDimProjection}
\title{Chunk Dimension Projection}
\format{
\link[R6:R6Class]{R6::R6Class} object.
}
\description{
A mapping from chunk to output array for a single dimension.
Used internally to track which items within a single chunk dimension
correspond to which items in the target output array.
}
\seealso{
Other Indexing classes: 
\code{\link{ChunkProjection}},
\code{\link{DimIndexer}},
\code{\link{Indexer}}
}
\concept{Indexing classes}
\keyword{internal}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{dim_chunk_index}}{(\code{integer(1)})\cr
Index of chunk.}

\item{\code{dim_chunk_sel}}{(\code{integer()} | \code{list()})\cr
Selection of items from chunk array.}

\item{\code{dim_out_sel}}{(\code{integer()} | \code{list()})\cr
Selection of items in target (output) array.}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-ChunkDimProjection-new}{\code{ChunkDimProjection$new()}}
\item \href{#method-ChunkDimProjection-clone}{\code{ChunkDimProjection$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ChunkDimProjection-new"></a>}}
\if{latex}{\out{\hypertarget{method-ChunkDimProjection-new}{}}}
\subsection{Method \code{new()}}{
Creates a new instance of this \link[R6:R6Class]{R6} class.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ChunkDimProjection$new(dim_chunk_index, dim_chunk_sel, dim_out_sel)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{dim_chunk_index}}{(\code{integer(1)})\cr
Index of chunk.}

\item{\code{dim_chunk_sel}}{(\code{integer()} | \code{list()})\cr
Selection of items from chunk array.}

\item{\code{dim_out_sel}}{(\code{integer()} | \code{list()})\cr
Selection of items in target (output) array.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ChunkDimProjection-clone"></a>}}
\if{latex}{\out{\hypertarget{method-ChunkDimProjection-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ChunkDimProjection$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zarr-group.R
\docType{class}
\name{ZarrGroup}
\alias{ZarrGroup}
\title{ZarrGroup Class}
\format{
\link[R6:R6Class]{R6::R6Class}
}
\description{
Instantiate a group from an initialized store.
}
\details{
The Zarr Group class.
}
\keyword{internal}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-ZarrGroup-new}{\code{ZarrGroup$new()}}
\item \href{#method-ZarrGroup-get_store}{\code{ZarrGroup$get_store()}}
\item \href{#method-ZarrGroup-get_path}{\code{ZarrGroup$get_path()}}
\item \href{#method-ZarrGroup-get_meta}{\code{ZarrGroup$get_meta()}}
\item \href{#method-ZarrGroup-get_name}{\code{ZarrGroup$get_name()}}
\item \href{#method-ZarrGroup-get_read_only}{\code{ZarrGroup$get_read_only()}}
\item \href{#method-ZarrGroup-get_chunk_store}{\code{ZarrGroup$get_chunk_store()}}
\item \href{#method-ZarrGroup-get_synchronizer}{\code{ZarrGroup$get_synchronizer()}}
\item \href{#method-ZarrGroup-get_attrs}{\code{ZarrGroup$get_attrs()}}
\item \href{#method-ZarrGroup-contains_item}{\code{ZarrGroup$contains_item()}}
\item \href{#method-ZarrGroup-get_item}{\code{ZarrGroup$get_item()}}
\item \href{#method-ZarrGroup-create_group}{\code{ZarrGroup$create_group()}}
\item \href{#method-ZarrGroup-create_dataset}{\code{ZarrGroup$create_dataset()}}
\item \href{#method-ZarrGroup-print}{\code{ZarrGroup$print()}}
\item \href{#method-ZarrGroup-tree}{\code{ZarrGroup$tree()}}
\item \href{#method-ZarrGroup-clone}{\code{ZarrGroup$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrGroup-new"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrGroup-new}{}}}
\subsection{Method \code{new()}}{
Create a new ZarrGroup instance.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrGroup$new(
  store,
  path = NA,
  read_only = FALSE,
  chunk_store = NA,
  cache_attrs = TRUE,
  synchronizer = NA
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{store}}{(\link{Store})\cr
Group store, already initialized.}

\item{\code{path}}{(\code{character(1)})\cr
Storage path.}

\item{\code{read_only}}{(\code{logical(1)})\cr
Whether the group is read-only.}

\item{\code{chunk_store}}{(\link{Store} or \code{NA})\cr
Separate storage for chunks.}

\item{\code{cache_attrs}}{(\code{logical(1)})\cr
Whether to cache attributes.}

\item{\code{synchronizer}}{(\code{ANY} or \code{NA})\cr
Object used to synchronize write access.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A \code{ZarrGroup} instance.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrGroup-get_store"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrGroup-get_store}{}}}
\subsection{Method \code{get_store()}}{
Get group store
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrGroup$get_store()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
\link{Store}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrGroup-get_path"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrGroup-get_path}{}}}
\subsection{Method \code{get_path()}}{
Get group path
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrGroup$get_path()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
\code{character(1)}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrGroup-get_meta"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrGroup-get_meta}{}}}
\subsection{Method \code{get_meta()}}{
Get group metadata
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrGroup$get_meta()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
\code{list()} or \code{NULL}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrGroup-get_name"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrGroup-get_name}{}}}
\subsection{Method \code{get_name()}}{
Get group name
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrGroup$get_name()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
\code{character(1)}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrGroup-get_read_only"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrGroup-get_read_only}{}}}
\subsection{Method \code{get_read_only()}}{
Is store read only?
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrGroup$get_read_only()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
\code{logical(1)}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrGroup-get_chunk_store"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrGroup-get_chunk_store}{}}}
\subsection{Method \code{get_chunk_store()}}{
Get group chunk store
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrGroup$get_chunk_store()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
\link{Store}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrGroup-get_synchronizer"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrGroup-get_synchronizer}{}}}
\subsection{Method \code{get_synchronizer()}}{
Get group synchronizer
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrGroup$get_synchronizer()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Synchronizer object or \code{NA}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrGroup-get_attrs"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrGroup-get_attrs}{}}}
\subsection{Method \code{get_attrs()}}{
Get group attributes
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrGroup$get_attrs()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
\link{Attributes}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrGroup-contains_item"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrGroup-contains_item}{}}}
\subsection{Method \code{contains_item()}}{
Test for group membership.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrGroup$contains_item(item)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{item}}{Character item to test for.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{logical(1)}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrGroup-get_item"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrGroup-get_item}{}}}
\subsection{Method \code{get_item()}}{
Obtain a group member.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrGroup$get_item(item)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{item}}{Character item to retrieve.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\link{ZarrArray} or \link{ZarrGroup}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrGroup-create_group"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrGroup-create_group}{}}}
\subsection{Method \code{create_group()}}{
create a group
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrGroup$create_group(name, overwrite = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{name}}{Character group name.}

\item{\code{overwrite}}{Logical overwrite.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\link{ZarrGroup}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrGroup-create_dataset"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrGroup-create_dataset}{}}}
\subsection{Method \code{create_dataset()}}{
Create a dataset (array) within this group.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrGroup$create_dataset(name, data = NA, ...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{name}}{Character dataset name.}

\item{\code{data}}{Data to add to group.}

\item{\code{...}}{Extra arguments to pass to \code{zarr_create()} or \code{array()}.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\link{ZarrArray}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrGroup-print"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrGroup-print}{}}}
\subsection{Method \code{print()}}{
Print a human-readable summary of the group.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrGroup$print(...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{...}}{Ignored.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{self} (invisibly).
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrGroup-tree"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrGroup-tree}{}}}
\subsection{Method \code{tree()}}{
Print an ASCII tree showing the group hierarchy.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrGroup$tree(level = NA)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{level}}{Maximum depth to display. \code{NA} (default) shows all levels.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{self} (invisibly).
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ZarrGroup-clone"></a>}}
\if{latex}{\out{\hypertarget{method-ZarrGroup-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ZarrGroup$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zarrs-bridge.R
\name{is_zarrs_available}
\alias{is_zarrs_available}
\title{Check whether the zarrs Rust backend is available}
\usage{
is_zarrs_available()
}
\value{
Logical scalar. \code{TRUE} when the package was installed with the
compiled zarrs backend (r-universe tier), \code{FALSE} on the CRAN tier
(pure R).
}
\description{
Check whether the zarrs Rust backend is available
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{chunk_fill}
\alias{chunk_fill}
\title{Fill in an R array with a single scalar value.}
\usage{
chunk_fill(chunk, value)
}
\arguments{
\item{chunk}{The R array to fill.}

\item{value}{The scalar value (after is_scalar() check).}
}
\description{
Fill in an R array with a single scalar value.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/creation.R
\name{zarr_open_group}
\alias{zarr_open_group}
\title{Open a group using file-mode-like semantics.}
\usage{
zarr_open_group(
  store = NA,
  mode = NA,
  cache_attrs = TRUE,
  synchronizer = NA,
  path = NA,
  chunk_store = NA,
  storage_options = NA,
  zarr_format = 2L
)
}
\arguments{
\item{store}{: Store
A mapping that supports string keys and bytes-like values.}

\item{mode}{: \code{'r', 'r+', 'a', 'w', 'w-'}, optional
Persistence mode: 'r' means read only (must exist); 'r+' means
read/write (must exist); 'a' means read/write (create if doesn't
exist); 'w' means create (overwrite if exists); 'w-' means create
(fail if exists).}

\item{cache_attrs}{: bool, optional
If True (default), user attributes will be cached for attribute read
operations. If False, user attributes are reloaded from the store prior
to all attribute read operations.}

\item{synchronizer}{: object, optional
Array synchronizer.}

\item{path}{: string, bytes, optional
Path under which array is stored.}

\item{chunk_store}{: Store, optional
Separate storage for chunks. If not provided, \code{store} will be used
for storage of both chunks and metadata.}

\item{storage_options}{: dict
If using an fsspec URL to create the store, these will be passed to
the backend implementation. Ignored otherwise.}

\item{zarr_format}{: int, optional
Zarr format version. Use \code{2L} (default) for Zarr V2 or \code{3L}
for Zarr V3.}
}
\value{
ZarrGroup
}
\description{
Open a group using file-mode-like semantics.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/meta.R
\name{create_v3_array_meta}
\alias{create_v3_array_meta}
\title{Create a V3 array metadata list.
Parallel to create_zarray_meta() for V2. Builds the zarr.json metadata
for a V3 array node. All scalar values are wrapped with jsonlite::unbox()
so the result can be encoded with toJSON(auto_unbox = FALSE).}
\usage{
create_v3_array_meta(
  shape,
  chunks,
  dtype,
  compressor,
  fill_value,
  order,
  filters,
  chunk_key_encoding = NULL,
  attributes = NULL,
  dimension_names = NULL
)
}
\arguments{
\item{shape}{Integer vector. Array shape.}

\item{chunks}{Integer vector. Chunk shape.}

\item{dtype}{Character. V2-style dtype string (e.g., "<f8", "|b1").}

\item{compressor}{Codec object or NA.}

\item{fill_value}{Fill value (numeric, logical, or NA).}

\item{order}{Character. "C" or "F" (unused in V3 codec pipeline but stored).}

\item{filters}{List of Codec objects or NA.}

\item{chunk_key_encoding}{List or NULL. If NULL, defaults to default "/" encoding.}

\item{attributes}{List or NULL. Embedded attributes. Default empty.}

\item{dimension_names}{Character vector or NULL. Omitted from zarr.json if NULL.}
}
\value{
A list ready for jsonlite::toJSON encoding.
}
\description{
Create a V3 array metadata list.
Parallel to create_zarray_meta() for V2. Builds the zarr.json metadata
for a V3 array node. All scalar values are wrapped with jsonlite::unbox()
so the result can be encoded with toJSON(auto_unbox = FALSE).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/meta.R
\name{Metadata2}
\alias{Metadata2}
\title{Zarr V2 Metadata Codec}
\format{
\link[R6:R6Class]{R6::R6Class} object.
}
\description{
Handles encoding and decoding of Zarr V2 metadata. Provides methods
to parse raw JSON bytes into R lists and to serialize metadata back
to JSON bytes, for both array and group nodes. Array metadata includes
shape, chunks, dtype, compressor, fill_value, order, and filters.
Group metadata contains only the zarr_format field.
}
\seealso{
Other Metadata classes: 
\code{\link{Metadata3}}
}
\concept{Metadata classes}
\keyword{internal}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-Metadata2-decode_metadata}{\code{Metadata2$decode_metadata()}}
\item \href{#method-Metadata2-encode_metadata}{\code{Metadata2$encode_metadata()}}
\item \href{#method-Metadata2-decode_array_metadata}{\code{Metadata2$decode_array_metadata()}}
\item \href{#method-Metadata2-decode_group_metadata}{\code{Metadata2$decode_group_metadata()}}
\item \href{#method-Metadata2-encode_array_metadata}{\code{Metadata2$encode_array_metadata()}}
\item \href{#method-Metadata2-encode_group_metadata}{\code{Metadata2$encode_group_metadata()}}
\item \href{#method-Metadata2-clone}{\code{Metadata2$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Metadata2-decode_metadata"></a>}}
\if{latex}{\out{\hypertarget{method-Metadata2-decode_metadata}{}}}
\subsection{Method \code{decode_metadata()}}{
Parse metadata from raw bytes or return as-is if already a list.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Metadata2$decode_metadata(s, auto_unbox = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{s}}{(\code{raw()} | \code{list()} | \code{NULL})\cr
Raw JSON bytes, a pre-parsed list, or \code{NULL}.}

\item{\code{auto_unbox}}{(\code{logical(1)})\cr
Passed to \code{\link[jsonlite:fromJSON]{jsonlite::toJSON()}}. Default \code{FALSE}.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{list()} or \code{NULL}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Metadata2-encode_metadata"></a>}}
\if{latex}{\out{\hypertarget{method-Metadata2-encode_metadata}{}}}
\subsection{Method \code{encode_metadata()}}{
Encode metadata list to raw JSON bytes.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Metadata2$encode_metadata(meta, auto_unbox = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{meta}}{(\code{list()})\cr
Metadata to encode.}

\item{\code{auto_unbox}}{(\code{logical(1)})\cr
Passed to \code{\link[jsonlite:fromJSON]{jsonlite::toJSON()}}. Default \code{FALSE}.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{raw()}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Metadata2-decode_array_metadata"></a>}}
\if{latex}{\out{\hypertarget{method-Metadata2-decode_array_metadata}{}}}
\subsection{Method \code{decode_array_metadata()}}{
Decode and validate V2 array metadata.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Metadata2$decode_array_metadata(s)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{s}}{(\code{raw()} | \code{list()} | \code{NULL})\cr
Raw JSON bytes or pre-parsed list.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{list()} or \code{NULL}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Metadata2-decode_group_metadata"></a>}}
\if{latex}{\out{\hypertarget{method-Metadata2-decode_group_metadata}{}}}
\subsection{Method \code{decode_group_metadata()}}{
Decode and validate V2 group metadata.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Metadata2$decode_group_metadata(s)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{s}}{(\code{raw()} | \code{list()} | \code{NULL})\cr
Raw JSON bytes or pre-parsed list.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{list()} or \code{NULL}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Metadata2-encode_array_metadata"></a>}}
\if{latex}{\out{\hypertarget{method-Metadata2-encode_array_metadata}{}}}
\subsection{Method \code{encode_array_metadata()}}{
Encode array metadata to raw JSON bytes, setting zarr_format to 2.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Metadata2$encode_array_metadata(meta)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{meta}}{(\code{list()})\cr
Array metadata list.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{raw()}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Metadata2-encode_group_metadata"></a>}}
\if{latex}{\out{\hypertarget{method-Metadata2-encode_group_metadata}{}}}
\subsection{Method \code{encode_group_metadata()}}{
Encode group metadata to raw JSON bytes, setting zarr_format to 2.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Metadata2$encode_group_metadata(meta = NA)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{meta}}{(\code{list()} | \code{NA})\cr
Ignored; a fresh group metadata list is created.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{raw()}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Metadata2-clone"></a>}}
\if{latex}{\out{\hypertarget{method-Metadata2-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Metadata2$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/slicing.R
\name{zb_slice}
\alias{zb_slice}
\title{Convenience function for the internal Slice R6 class constructor
with zero-based indexing and exclusive stop index.}
\usage{
zb_slice(start, stop = NA, step = NA)
}
\arguments{
\item{start}{The start index.}

\item{stop}{The stop index.}

\item{step}{The step size. Negative values reverse the direction of the
slice, matching Python/NumPy semantics (e.g., \code{step = -1} iterates
backward from \code{start} toward \code{stop}).}
}
\value{
A Slice instance with the specified parameters.
}
\description{
Convenience function for the internal Slice R6 class constructor
with zero-based indexing and exclusive stop index.
}
\examples{
# Equivalent to Python's arr[5:0:-1]
zb_slice(5, 0, -1)

# Equivalent to Python's arr[-1::-1] (full reverse)
zb_slice(-1, NA, -1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/meta.R
\name{create_zarray_meta}
\alias{create_zarray_meta}
\title{Create a list of zarray metadata.}
\usage{
create_zarray_meta(
  shape = NA,
  chunks = NA,
  dtype = NA,
  compressor = NA,
  fill_value = NA,
  order = NA,
  filters = NA,
  dimension_separator = NA
)
}
\arguments{
\item{shape}{: int or tuple of ints
Array shape.}

\item{chunks}{: int or tuple of ints, optional
Chunk shape. If True, will be guessed from \code{shape} and \code{dtype}. If
False, will be set to \code{shape}, i.e., single chunk for the whole array.
If an int, the chunk size in each dimension will be given by the value
of \code{chunks}. Default is True.}

\item{dtype}{: string or dtype, optional
NumPy dtype.}

\item{compressor}{: Codec, optional
Primary compressor.}

\item{fill_value}{: object
Default value to use for uninitialized portions of the array.}

\item{order}{: \code{'C', 'F'}, optional
Memory layout to be used within each chunk.}

\item{filters}{: sequence, optional
Sequence of filters to use to encode chunk data prior to compression.}

\item{dimension_separator}{: \code{'.', '/'}, optional
Separator placed between the dimensions of a chunk.}
}
\value{
A list.
}
\description{
Create a list of zarray metadata.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/creation.R
\name{init_group}
\alias{init_group}
\title{Initialize a group store. Note that this is a low-level function and there should be no
need to call this directly from user code.}
\usage{
init_group(
  store,
  overwrite = FALSE,
  path = NA,
  chunk_store = NA,
  zarr_format = 2L
)
}
\arguments{
\item{store}{: Store
A mapping that supports string keys and bytes-like values.}

\item{overwrite}{: bool, optional
If True, erase all data in \code{store} prior to initialisation.}

\item{path}{: string, bytes, optional
Path under which array is stored.}

\item{chunk_store}{: Store, optional
Separate storage for chunks. If not provided, \code{store} will be used
for storage of both chunks and metadata.}
}
\description{
Initialize a group store. Note that this is a low-level function and there should be no
need to call this directly from user code.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{is_int64_dtype}
\alias{is_int64_dtype}
\title{Check if a dtype represents a 64-bit integer type}
\usage{
is_int64_dtype(dtype_obj)
}
\arguments{
\item{dtype_obj}{A Dtype instance.}
}
\value{
Logical.
}
\description{
Check if a dtype represents a 64-bit integer type
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zarrs-dispatch.R
\name{can_use_zarrs_create}
\alias{can_use_zarrs_create}
\title{Check whether the zarrs create path can handle this array}
\usage{
can_use_zarrs_create(store, dtype)
}
\arguments{
\item{store}{A Store object.}

\item{dtype}{A Dtype object.}
}
\value{
Logical scalar.
}
\description{
Returns TRUE when zarrs is available, the store is writable and
has a store identifier, and the dtype is numeric/logical (not
object or string).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/indexing-types.R
\name{ChunkProjection}
\alias{ChunkProjection}
\title{Chunk Projection}
\format{
\link[R6:R6Class]{R6::R6Class} object.
}
\description{
A mapping of items from chunk to output array. Can be used to extract
items from the chunk array for loading into an output array. Can also be
used to extract items from a value array for setting/updating in a chunk
array.
}
\seealso{
Other Indexing classes: 
\code{\link{ChunkDimProjection}},
\code{\link{DimIndexer}},
\code{\link{Indexer}}
}
\concept{Indexing classes}
\keyword{internal}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{chunk_coords}}{(\code{integer()})\cr
Indices of chunk.}

\item{\code{chunk_sel}}{(\code{list()})\cr
Selection of items from chunk array.}

\item{\code{out_sel}}{(\code{list()})\cr
Selection of items in target (output) array.}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-ChunkProjection-new}{\code{ChunkProjection$new()}}
\item \href{#method-ChunkProjection-clone}{\code{ChunkProjection$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ChunkProjection-new"></a>}}
\if{latex}{\out{\hypertarget{method-ChunkProjection-new}{}}}
\subsection{Method \code{new()}}{
Creates a new instance of this \link[R6:R6Class]{R6} class.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ChunkProjection$new(chunk_coords, chunk_sel, out_sel)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{chunk_coords}}{(\code{integer()})\cr
Indices of chunk.}

\item{\code{chunk_sel}}{(\code{list()})\cr
Selection of items from chunk array.}

\item{\code{out_sel}}{(\code{list()})\cr
Selection of items in target (output) array.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ChunkProjection-clone"></a>}}
\if{latex}{\out{\hypertarget{method-ChunkProjection-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ChunkProjection$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/creation.R
\name{init_array}
\alias{init_array}
\title{Initialize an array store with the given configuration. Note that this is a low-level
function and there should be no need to call this directly from user code.}
\usage{
init_array(
  store,
  shape,
  chunks = TRUE,
  dtype = NA,
  compressor = NA,
  fill_value = NA,
  order = NA,
  overwrite = FALSE,
  path = NA,
  chunk_store = NA,
  filters = NA,
  object_codec = NA,
  dimension_separator = NA,
  zarr_format = 2L,
  dimension_names = NULL
)
}
\arguments{
\item{store}{: Store
A mapping that supports string keys and bytes-like values.}

\item{shape}{: int or tuple of ints
Array shape.}

\item{chunks}{: bool, int or tuple of ints, optional
Chunk shape. If True, will be guessed from \code{shape} and \code{dtype}. If
False, will be set to \code{shape}, i.e., single chunk for the whole array.}

\item{dtype}{: string or dtype, optional
NumPy dtype.}

\item{compressor}{: Codec, optional
Primary compressor.}

\item{fill_value}{: object
Default value to use for uninitialized portions of the array.}

\item{order}{: \code{'C', 'F'}, optional
Memory layout to be used within each chunk.}

\item{overwrite}{: bool, optional
If True, erase all data in \code{store} prior to initialisation.}

\item{path}{: string, bytes, optional
Path under which array is stored.}

\item{chunk_store}{: Store, optional
Separate storage for chunks. If not provided, \code{store} will be used
for storage of both chunks and metadata.}

\item{filters}{: sequence, optional
Sequence of filters to use to encode chunk data prior to compression.}

\item{object_codec}{: Codec, optional
A codec to encode object arrays, only needed if dtype=object.}

\item{dimension_separator}{: \code{'.', '/'}, optional
Separator placed between the dimensions of a chunk.}
}
\description{
Initialize an array store with the given configuration. Note that this is a low-level
function and there should be no need to call this directly from user code.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/meta.R
\name{Metadata3}
\alias{Metadata3}
\title{Zarr V3 Metadata Codec}
\format{
\link[R6:R6Class]{R6::R6Class} object.
}
\description{
Handles decoding of Zarr V3 metadata. In V3, each node has a single
\code{zarr.json} file containing \code{zarr_format}, \code{node_type}, and all
metadata (shape, data_type, codecs, chunk_grid, chunk_key_encoding,
fill_value, attributes, etc.). This class is read-only; no encode
methods are provided.
}
\seealso{
Other Metadata classes: 
\code{\link{Metadata2}}
}
\concept{Metadata classes}
\keyword{internal}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-Metadata3-decode_metadata}{\code{Metadata3$decode_metadata()}}
\item \href{#method-Metadata3-decode_array_metadata}{\code{Metadata3$decode_array_metadata()}}
\item \href{#method-Metadata3-decode_group_metadata}{\code{Metadata3$decode_group_metadata()}}
\item \href{#method-Metadata3-encode_array_metadata}{\code{Metadata3$encode_array_metadata()}}
\item \href{#method-Metadata3-encode_group_metadata}{\code{Metadata3$encode_group_metadata()}}
\item \href{#method-Metadata3-clone}{\code{Metadata3$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Metadata3-decode_metadata"></a>}}
\if{latex}{\out{\hypertarget{method-Metadata3-decode_metadata}{}}}
\subsection{Method \code{decode_metadata()}}{
Parse metadata from raw bytes or return as-is if already a list.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Metadata3$decode_metadata(s)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{s}}{(\code{raw()} | \code{list()} | \code{NULL})\cr
Raw JSON bytes, a pre-parsed list, or \code{NULL}.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{list()} or \code{NULL}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Metadata3-decode_array_metadata"></a>}}
\if{latex}{\out{\hypertarget{method-Metadata3-decode_array_metadata}{}}}
\subsection{Method \code{decode_array_metadata()}}{
Decode and validate V3 array metadata from \code{zarr.json}.
V3 arrays have \code{node_type = "array"} and contain shape, data_type,
chunk_grid, chunk_key_encoding, codecs, fill_value, and attributes.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Metadata3$decode_array_metadata(s)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{s}}{(\code{raw()} | \code{list()} | \code{NULL})\cr
Raw JSON bytes or pre-parsed list.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{list()} or \code{NULL}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Metadata3-decode_group_metadata"></a>}}
\if{latex}{\out{\hypertarget{method-Metadata3-decode_group_metadata}{}}}
\subsection{Method \code{decode_group_metadata()}}{
Decode and validate V3 group metadata from \code{zarr.json}.
V3 groups have \code{node_type = "group"} and may contain attributes.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Metadata3$decode_group_metadata(s)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{s}}{(\code{raw()} | \code{list()} | \code{NULL})\cr
Raw JSON bytes or pre-parsed list.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{list()} or \code{NULL}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Metadata3-encode_array_metadata"></a>}}
\if{latex}{\out{\hypertarget{method-Metadata3-encode_array_metadata}{}}}
\subsection{Method \code{encode_array_metadata()}}{
Encode V3 array metadata to raw JSON bytes.
Expects a list built by \code{create_v3_array_meta()} with all scalars
already wrapped in \code{jsonlite::unbox()}.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Metadata3$encode_array_metadata(meta)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{meta}}{(\code{list()})\cr
V3 array metadata list.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{raw()}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Metadata3-encode_group_metadata"></a>}}
\if{latex}{\out{\hypertarget{method-Metadata3-encode_group_metadata}{}}}
\subsection{Method \code{encode_group_metadata()}}{
Encode V3 group metadata to raw JSON bytes.
Produces \verb{\{"zarr_format":3,"node_type":"group","attributes":\{\}\}}.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Metadata3$encode_group_metadata(meta = NA, attributes = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{meta}}{(\code{list()} | \code{NA})\cr
Ignored; a fresh group metadata list is created.}

\item{\code{attributes}}{(\code{list()})\cr
Attributes to embed in zarr.json. Default empty.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{raw()}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Metadata3-clone"></a>}}
\if{latex}{\out{\hypertarget{method-Metadata3-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Metadata3$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/config.R
\name{pizzarr_config}
\alias{pizzarr_config}
\title{Get or set pizzarr configuration}
\usage{
pizzarr_config(
  nthreads = NULL,
  concurrent_target = NULL,
  http_batch_range_requests = NULL
)
}
\arguments{
\item{nthreads}{Integer or NULL. Number of threads for the rayon thread
pool. NULL uses all CPUs (the default). The pool can only be initialised
once per R session; later changes require a restart. Use the
\code{PIZZARR_NTHREADS} environment variable for reliable session-level
control.}

\item{concurrent_target}{Integer or NULL. Codec concurrency level — how
many codec operations zarrs runs in parallel within a single read/write
call. NULL uses the zarrs default (CPU count). Can be changed at any time.}

\item{http_batch_range_requests}{Logical or NULL. Whether HTTP stores use
multipart range requests (default TRUE). Set to FALSE for servers with
incomplete multipart range support. Takes effect on the next
\code{zarr_open()} or \code{zarrs_get_subset()} call that opens a new
HTTP store; existing cached stores are not affected (use
\code{zarrs_close_store(url)} to force re-creation).}
}
\value{
When called with no arguments, a named list of current settings.
When called with arguments, the previous values (invisibly).
}
\description{
Controls parallelism and HTTP behaviour for the zarrs backend.
Called with no arguments, returns the current settings as a named list.
Called with arguments, sets the specified options and applies them to
the Rust backend immediately.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/int.R
\name{int}
\alias{int}
\title{Convenience function for the internal Int class constructor.}
\usage{
int(index, zero_based = FALSE)
}
\arguments{
\item{index}{The integer index.}

\item{zero_based}{The index of the dimension. By default, FALSE for R-like behavior.}
}
\value{
A Int instance with the specified parameters.
}
\description{
Convenience function for the internal Int class constructor.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/slicing.R
\name{is_slice}
\alias{is_slice}
\title{Check if a value is a Slice instance.}
\usage{
is_slice(s)
}
\arguments{
\item{s}{The value to check.}
}
\value{
TRUE if the value is a Slice instance, FALSE otherwise.
}
\description{
Check if a value is a Slice instance.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zarrs-bridge.R
\name{pizzarr_compiled_features}
\alias{pizzarr_compiled_features}
\title{List compiled zarrs features}
\usage{
pizzarr_compiled_features()
}
\value{
Character vector of feature names (e.g. \code{"filesystem"},
\code{"gzip"}).
}
\description{
Returns the feature flags compiled into the zarrs backend, or
\code{character(0)} with a message when the backend is absent.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/atomic.R
\name{is_integer_vec}
\alias{is_integer_vec}
\title{Check that a value is a vector of one or more integers and has not been
explicitly tagged as a scalar.}
\usage{
is_integer_vec(s)
}
\description{
Check that a value is a vector of one or more integers and has not been
explicitly tagged as a scalar.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/array-raw.R
\docType{class}
\name{RawArray}
\alias{RawArray}
\title{RawArray Class}
\description{
TODO: Add export annotation / remove keywords internal annotation once implemented.
}
\details{
The Zarr RawArray class.
}
\keyword{internal}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-RawArray-new}{\code{RawArray$new()}}
\item \href{#method-RawArray-clone}{\code{RawArray$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-RawArray-new"></a>}}
\if{latex}{\out{\hypertarget{method-RawArray-new}{}}}
\subsection{Method \code{new()}}{
Create a new RawArray instance.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{RawArray$new()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
A \code{RawArray} instance.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-RawArray-clone"></a>}}
\if{latex}{\out{\hypertarget{method-RawArray-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{RawArray$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/atomic.R
\name{is_scalar}
\alias{is_scalar}
\title{Check if a value is a scalar (i.e., a one-element vector that was converted with as_scalar).}
\usage{
is_scalar(s)
}
\arguments{
\item{s}{The value to check.}
}
\value{
TRUE if the value is a scalar, FALSE otherwise.
}
\description{
Check if a value is a scalar (i.e., a one-element vector that was converted with as_scalar).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/indexing.R
\docType{class}
\name{BasicIndexer}
\alias{BasicIndexer}
\title{BasicIndexer Class}
\description{
An indexer class to normalize a selection of an array and provide an iterator
of indexes over the dimensions of an array.
}
\details{
The Zarr BasicIndexer class.
}
\keyword{internal}
\section{Super class}{
\code{\link[pizzarr:Indexer]{pizzarr::Indexer}} -> \code{BasicIndexer}
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{dim_indexers}}{List of per-dimension indexers (IntDimIndexer or SliceDimIndexer).}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-BasicIndexer-new}{\code{BasicIndexer$new()}}
\item \href{#method-BasicIndexer-iter}{\code{BasicIndexer$iter()}}
\item \href{#method-BasicIndexer-clone}{\code{BasicIndexer$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-BasicIndexer-new"></a>}}
\if{latex}{\out{\hypertarget{method-BasicIndexer-new}{}}}
\subsection{Method \code{new()}}{
Create a new BasicIndexer instance.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{BasicIndexer$new(selection, array)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{selection}}{Selection as with ZarrArray: scalar, string, or Slice.}

\item{\code{array}}{\link{ZarrArray} object that will be indexed.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A \code{BasicIndexer} instance.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-BasicIndexer-iter"></a>}}
\if{latex}{\out{\hypertarget{method-BasicIndexer-iter}{}}}
\subsection{Method \code{iter()}}{
An iterator over the dimensions of an array
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{BasicIndexer$iter()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
A list of ChunkProjection objects
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-BasicIndexer-clone"></a>}}
\if{latex}{\out{\hypertarget{method-BasicIndexer-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{BasicIndexer$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/normalize.R
\name{str_to_vec}
\alias{str_to_vec}
\title{Convert a string into a character vector.}
\usage{
str_to_vec(s)
}
\arguments{
\item{s}{The string.}
}
\value{
A vector where each element is an individual character.
}
\description{
Convert a string into a character vector.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/array-nested.R
\name{as.array.NestedArray}
\alias{as.array.NestedArray}
\title{S3 method for as.array}
\usage{
\method{as.array}{NestedArray}(x, ...)
}
\arguments{
\item{x}{Object.}

\item{...}{Not used.}
}
\description{
S3 method for as.array
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/indexing.R
\docType{class}
\name{BoolArrayDimIndexer}
\alias{BoolArrayDimIndexer}
\title{BoolArrayDimIndexer Class}
\description{
Indexer for a boolean array selection along one dimension, selecting elements where the mask is TRUE.
}
\details{
The Zarr BoolArrayDimIndexer class.
}
\keyword{internal}
\section{Super class}{
\code{\link[pizzarr:DimIndexer]{pizzarr::DimIndexer}} -> \code{BoolArrayDimIndexer}
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{dim_sel}}{Selection on dimension.}

\item{\code{dim_len}}{Dimension length.}

\item{\code{dim_chunk_len}}{Dimension chunk length.}

\item{\code{num_chunks}}{Number of chunks.}

\item{\code{chunk_nitems}}{Number of items per chunk.}

\item{\code{chunk_nitems_cumsum}}{Offsets into the output array.}

\item{\code{dim_chunk_ixs}}{Chunks that should be visited.}

\item{\code{dim_out_sel}}{Output array indices for the selected elements.}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-BoolArrayDimIndexer-new}{\code{BoolArrayDimIndexer$new()}}
\item \href{#method-BoolArrayDimIndexer-iter}{\code{BoolArrayDimIndexer$iter()}}
\item \href{#method-BoolArrayDimIndexer-clone}{\code{BoolArrayDimIndexer$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-BoolArrayDimIndexer-new"></a>}}
\if{latex}{\out{\hypertarget{method-BoolArrayDimIndexer-new}{}}}
\subsection{Method \code{new()}}{
Create a new BoolArrayDimIndexer instance.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{BoolArrayDimIndexer$new(dim_sel, dim_len, dim_chunk_len)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{dim_sel}}{Integer dimension selection.}

\item{\code{dim_len}}{Integer dimension length.}

\item{\code{dim_chunk_len}}{Integer dimension chunk length.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A \code{BoolArrayDimIndexer} instance.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-BoolArrayDimIndexer-iter"></a>}}
\if{latex}{\out{\hypertarget{method-BoolArrayDimIndexer-iter}{}}}
\subsection{Method \code{iter()}}{
An iterator over the dimensions of an array
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{BoolArrayDimIndexer$iter()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
A list of ChunkProjection objects
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-BoolArrayDimIndexer-clone"></a>}}
\if{latex}{\out{\hypertarget{method-BoolArrayDimIndexer-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{BoolArrayDimIndexer$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/creation.R
\name{zarr_open}
\alias{zarr_open}
\title{Convenience function to open a group or array using file-mode-like semantics.}
\usage{
zarr_open(store = NA, mode = NA, path = NA, ...)
}
\arguments{
\item{store}{: Store
A mapping that supports string keys and bytes-like values.}

\item{mode}{: \code{'r', 'r+', 'a', 'w', 'w-'}, optional
Persistence mode: 'r' means read only (must exist); 'r+' means
read/write (must exist); 'a' means read/write (create if doesn't
exist); 'w' means create (overwrite if exists); 'w-' means create
(fail if exists).}

\item{path}{: string, bytes, optional
Path under which array is stored.}

\item{...}{Additional arguments to pass to zarr_open_array or zarr_open_group.}
}
\value{
ZarrArray or ZarrGroup
}
\description{
Convenience function to open a group or array using file-mode-like semantics.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/atomic.R
\name{ensure_list}
\alias{ensure_list}
\title{Ensure that scalars, single slices, and R integer vectors are converted
to a list containing either R integer vectors or slice instances as values.}
\usage{
ensure_list(selection)
}
\description{
Ensure that scalars, single slices, and R integer vectors are converted
to a list containing either R integer vectors or slice instances as values.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{integer64_to_raw}
\alias{integer64_to_raw}
\title{Convert integer64 vector to raw bytes}
\usage{
integer64_to_raw(x, endian = "little")
}
\arguments{
\item{x}{A bit64::integer64 vector.}

\item{endian}{"little" or "big".}
}
\value{
A raw vector.
}
\description{
Convert integer64 vector to raw bytes
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/creation.R
\name{zarr_create_array}
\alias{zarr_create_array}
\title{Create an array initialized with data.}
\usage{
zarr_create_array(data, ...)
}
\arguments{
\item{data}{A base R array() or pizzarr NestedArray instance.}

\item{...}{The params of zarr_create()}
}
\value{
ZarrArray
}
\description{
Create an array initialized with data.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{obj_list}
\alias{obj_list}
\title{Create an empty named list}
\usage{
obj_list(...)
}
\arguments{
\item{...}{A variable number of list entries.}
}
\value{
A named list.
}
\description{
A helper function to construct an empty list which converts to a JSON object rather than a JSON array.
}
\examples{
default_window <- obj_list(
  min = 0,
  max = 255
)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zarrs-dispatch.R
\name{can_use_zarrs_write}
\alias{can_use_zarrs_write}
\title{Check whether the zarrs write path can handle this selection}
\usage{
can_use_zarrs_write(indexer, store)
}
\arguments{
\item{indexer}{A BasicIndexer or OrthogonalIndexer object.}

\item{store}{A Store object.}
}
\value{
Logical scalar.
}
\description{
Same eligibility as read. Separate function for future extensibility
(e.g., read-only remote stores).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/indexing-types.R
\name{DimIndexer}
\alias{DimIndexer}
\title{Dimension Indexer}
\format{
\link[R6:R6Class]{R6::R6Class} object.
}
\description{
Abstract base class for per-dimension chunk indexers. A dimension
indexer maps a selection on a single array dimension to a sequence of
\link{ChunkDimProjection} objects. Concrete subclasses include
\link{IntDimIndexer}, \link{SliceDimIndexer}, \link{IntArrayDimIndexer}, and
\link{BoolArrayDimIndexer}.
}
\seealso{
Other Indexing classes: 
\code{\link{ChunkDimProjection}},
\code{\link{ChunkProjection}},
\code{\link{Indexer}}
}
\concept{Indexing classes}
\keyword{internal}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{num_items}}{(\code{integer(1)})\cr
Number of items in the selection.}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-DimIndexer-iter}{\code{DimIndexer$iter()}}
\item \href{#method-DimIndexer-clone}{\code{DimIndexer$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DimIndexer-iter"></a>}}
\if{latex}{\out{\hypertarget{method-DimIndexer-iter}{}}}
\subsection{Method \code{iter()}}{
Iterate over \link{ChunkDimProjection} objects for this dimension.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{DimIndexer$iter()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
\code{list()} of \link{ChunkDimProjection} objects.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DimIndexer-clone"></a>}}
\if{latex}{\out{\hypertarget{method-DimIndexer-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{DimIndexer$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zarrs-dispatch.R
\name{compressor_to_preset}
\alias{compressor_to_preset}
\title{Map an R compressor object to a zarrs codec preset string}
\usage{
compressor_to_preset(compressor, compressor_config)
}
\arguments{
\item{compressor}{A codec object or NA.}

\item{compressor_config}{A config list or NA.}
}
\value{
Character scalar.
}
\description{
Returns one of \code{"none"}, \code{"gzip"}, \code{"blosc"},
\code{"zstd"}, or \code{NA_character_} if the compressor is not
recognised (signalling fallback to R-native).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/creation.R
\name{zarr_open_array}
\alias{zarr_open_array}
\title{Open an array using file-mode-like semantics.}
\usage{
zarr_open_array(
  store = NA,
  storage_options = NA,
  mode = NA,
  shape = NA,
  chunks = TRUE,
  dtype = NA,
  compressor = NA,
  fill_value = NA,
  order = NA,
  synchronizer = NA,
  overwrite = FALSE,
  path = NA,
  chunk_store = NA,
  filters = NA,
  cache_metadata = TRUE,
  cache_attrs = TRUE,
  object_codec = NA,
  dimension_separator = NA,
  write_empty_chunks = TRUE,
  zarr_format = 2L
)
}
\arguments{
\item{store}{: Store
A mapping that supports string keys and bytes-like values.}

\item{storage_options}{: dict
If using an fsspec URL to create the store, these will be passed to
the backend implementation. Ignored otherwise.}

\item{mode}{: \code{'r', 'r+', 'a', 'w', 'w-'}, optional
Persistence mode: 'r' means read only (must exist); 'r+' means
read/write (must exist); 'a' means read/write (create if doesn't
exist); 'w' means create (overwrite if exists); 'w-' means create
(fail if exists).}

\item{shape}{: int or tuple of ints
Array shape.}

\item{chunks}{: bool, int or tuple of ints, optional
Chunk shape. If True, will be guessed from \code{shape} and \code{dtype}. If
False, will be set to \code{shape}, i.e., single chunk for the whole array.}

\item{dtype}{: string or dtype, optional
NumPy dtype.}

\item{compressor}{: Codec, optional
Primary compressor.}

\item{fill_value}{: object
Default value to use for uninitialized portions of the array.}

\item{order}{: \code{'C', 'F'}, optional
Memory layout to be used within each chunk.}

\item{synchronizer}{: object, optional
Array synchronizer.}

\item{overwrite}{: bool, optional
If True, erase all data in \code{store} prior to initialisation.}

\item{path}{: string, bytes, optional
Path under which array is stored.}

\item{chunk_store}{: Store, optional
Separate storage for chunks. If not provided, \code{store} will be used
for storage of both chunks and metadata.}

\item{filters}{: sequence, optional
Sequence of filters to use to encode chunk data prior to compression.}

\item{cache_metadata}{: bool, optional
If True, array configuration metadata will be cached for the
lifetime of the object. If False, array metadata will be reloaded
prior to all data access and modification operations (may incur
overhead depending on storage and data access pattern).}

\item{cache_attrs}{: bool, optional
If True (default), user attributes will be cached for attribute read
operations. If False, user attributes are reloaded from the store prior
to all attribute read operations.}

\item{object_codec}{: Codec, optional
A codec to encode object arrays, only needed if dtype=object.}

\item{dimension_separator}{: \code{'.', '/'}, optional
Separator placed between the dimensions of a chunk.}

\item{write_empty_chunks}{: bool, optional
If True (default), all chunks will be stored regardless of their
contents. If False, each chunk is compared to the array's fill value
prior to storing. If a chunk is uniformly equal to the fill value, then
that chunk is not be stored, and the store entry for that chunk's key
is deleted. This setting enables sparser storage, as only chunks with
non-fill-value data are stored, at the expense of overhead associated
with checking the data of each chunk.}

\item{zarr_format}{: int, optional
Zarr format version. Use \code{2L} (default) for Zarr V2 or \code{3L}
for Zarr V3.}
}
\value{
ZarrArray
}
\description{
Open an array using file-mode-like semantics.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/atomic.R
\name{ensure_integer_vec}
\alias{ensure_integer_vec}
\title{Ensure that scalars and lists of integers are converted to
an R vector of integers.}
\usage{
ensure_integer_vec(selection)
}
\description{
Ensure that scalars and lists of integers are converted to
an R vector of integers.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/creation.R
\name{zarr_save_array}
\alias{zarr_save_array}
\title{Convenience function to save a ZarrArray to the local file system.}
\usage{
zarr_save_array(store, arr, ...)
}
\arguments{
\item{store}{: Store
A mapping that supports string keys and bytes-like values.}

\item{arr}{: ZarrArray
The array with data to save.}

\item{...}{Additional arguments to pass to zarr_create_array().}
}
\description{
Convenience function to save a ZarrArray to the local file system.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{is_key_error}
\alias{is_key_error}
\title{Check if an error is a KeyError.}
\usage{
is_key_error(e)
}
\arguments{
\item{e}{The error to check.}
}
\value{
TRUE if the error is a KeyError, FALSE otherwise.
}
\description{
Check if an error is a KeyError.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/attributes.R
\docType{class}
\name{Attributes}
\alias{Attributes}
\title{Attributes Class}
\format{
\link[R6:R6Class]{R6::R6Class}
}
\description{
Class providing access to user attributes on an array or group.
}
\details{
The Zarr Attributes class.
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{store}}{Attributes store, already initialized.}

\item{\code{key}}{The key under which the attributes will be stored.}

\item{\code{read_only}}{If True, attributes cannot be modified.}

\item{\code{cache}}{If True (default), attributes will be cached locally.}

\item{\code{synchronizer}}{Only necessary if attributes may be modified from multiple threads or processes.}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-Attributes-new}{\code{Attributes$new()}}
\item \href{#method-Attributes-to_list}{\code{Attributes$to_list()}}
\item \href{#method-Attributes-refresh}{\code{Attributes$refresh()}}
\item \href{#method-Attributes-contains}{\code{Attributes$contains()}}
\item \href{#method-Attributes-get_item}{\code{Attributes$get_item()}}
\item \href{#method-Attributes-set_item}{\code{Attributes$set_item()}}
\item \href{#method-Attributes-del_item}{\code{Attributes$del_item()}}
\item \href{#method-Attributes-set_cached_v3_attrs}{\code{Attributes$set_cached_v3_attrs()}}
\item \href{#method-Attributes-clone}{\code{Attributes$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Attributes-new"></a>}}
\if{latex}{\out{\hypertarget{method-Attributes-new}{}}}
\subsection{Method \code{new()}}{
Create a new Attributes instance.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Attributes$new(
  store,
  key = NA,
  read_only = FALSE,
  cache = TRUE,
  synchronizer = NA,
  zarr_format = NULL
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{store}}{(\link{Store})\cr
Attributes store, already initialized.}

\item{\code{key}}{(\code{character(1)})\cr
Key to use for attributes (\code{.zattrs} is default).}

\item{\code{read_only}}{(\code{logical(1)})\cr
Whether the attributes are read-only.}

\item{\code{cache}}{(\code{logical(1)})\cr
Whether to cache attributes.}

\item{\code{synchronizer}}{(\code{ANY} or \code{NA})\cr
Synchronizer object.}

\item{\code{zarr_format}}{(\code{integer(1)} or \code{NULL})\cr
Zarr format version: \code{2L} for V2 (\code{.zattrs}), \code{3L} for V3 (zarr.json).}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
An \code{Attributes} instance.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Attributes-to_list"></a>}}
\if{latex}{\out{\hypertarget{method-Attributes-to_list}{}}}
\subsection{Method \code{to_list()}}{
convert attributes to list
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Attributes$to_list()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
\code{list()}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Attributes-refresh"></a>}}
\if{latex}{\out{\hypertarget{method-Attributes-refresh}{}}}
\subsection{Method \code{refresh()}}{
refresh attributes
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Attributes$refresh()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
\code{NULL} (called for side effects).
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Attributes-contains"></a>}}
\if{latex}{\out{\hypertarget{method-Attributes-contains}{}}}
\subsection{Method \code{contains()}}{
check if object contains item
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Attributes$contains(x)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{x}}{Object to test.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{logical(1)}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Attributes-get_item"></a>}}
\if{latex}{\out{\hypertarget{method-Attributes-get_item}{}}}
\subsection{Method \code{get_item()}}{
get attribute
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Attributes$get_item(item)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{item}}{Character attribute name.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
The attribute value.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Attributes-set_item"></a>}}
\if{latex}{\out{\hypertarget{method-Attributes-set_item}{}}}
\subsection{Method \code{set_item()}}{
set attribute
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Attributes$set_item(item, value)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{item}}{Character attribute name.}

\item{\code{value}}{Value to add or update.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{NULL} (called for side effects).
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Attributes-del_item"></a>}}
\if{latex}{\out{\hypertarget{method-Attributes-del_item}{}}}
\subsection{Method \code{del_item()}}{
delete attribute
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Attributes$del_item(item)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{item}}{Character attribute name.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{NULL} (called for side effects).
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Attributes-set_cached_v3_attrs"></a>}}
\if{latex}{\out{\hypertarget{method-Attributes-set_cached_v3_attrs}{}}}
\subsection{Method \code{set_cached_v3_attrs()}}{
Set cached attributes from V3 embedded metadata.
In V3, attributes are part of zarr.json rather than a separate .zattrs file.
This method pre-populates the cache so the normal .zattrs read path is skipped.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Attributes$set_cached_v3_attrs(attrs_list)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{attrs_list}}{A named list of attributes from V3 zarr.json.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{NULL} (modifies cache in place).
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Attributes-clone"></a>}}
\if{latex}{\out{\hypertarget{method-Attributes-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Attributes$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{get_list_product}
\alias{get_list_product}
\title{Generate a product of lists.}
\usage{
get_list_product(dim_indexer_iterables)
}
\arguments{
\item{dim_indexer_iterables}{A list of lists.}
}
\value{
A list of lists.
}
\description{
Generate a product of lists.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stores.R
\docType{class}
\name{MemoryStore}
\alias{MemoryStore}
\title{MemoryStore Class}
\format{
\link[R6:R6Class]{R6::R6Class} inheriting from \link{Store}.
}
\description{
Store class that uses a hierarchy of list objects,
thus all data will be held in main memory.
}
\details{
MemoryStore for Zarr
}
\seealso{
Other Store classes: 
\code{\link{DirectoryStore}},
\code{\link{GcsStore}},
\code{\link{HttpStore}},
\code{\link{S3Store}},
\code{\link{Store}}
}
\concept{Store classes}
\keyword{internal}
\section{Super class}{
\code{\link[pizzarr:Store]{pizzarr::Store}} -> \code{MemoryStore}
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{root}}{The root list for the store.}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-MemoryStore-new}{\code{MemoryStore$new()}}
\item \href{#method-MemoryStore-get_parent}{\code{MemoryStore$get_parent()}}
\item \href{#method-MemoryStore-get_item}{\code{MemoryStore$get_item()}}
\item \href{#method-MemoryStore-set_item}{\code{MemoryStore$set_item()}}
\item \href{#method-MemoryStore-contains_item}{\code{MemoryStore$contains_item()}}
\item \href{#method-MemoryStore-listdir}{\code{MemoryStore$listdir()}}
\item \href{#method-MemoryStore-delete_item}{\code{MemoryStore$delete_item()}}
\item \href{#method-MemoryStore-rmdir}{\code{MemoryStore$rmdir()}}
\item \href{#method-MemoryStore-print}{\code{MemoryStore$print()}}
\item \href{#method-MemoryStore-clone}{\code{MemoryStore$clone()}}
}
}
\if{html}{\out{
<details><summary>Inherited methods</summary>
<ul>
<li><span class="pkg-link" data-pkg="pizzarr" data-topic="Store" data-id="close"><a href='../../pizzarr/html/Store.html#method-Store-close'><code>pizzarr::Store$close()</code></a></span></li>
<li><span class="pkg-link" data-pkg="pizzarr" data-topic="Store" data-id="get_consolidated_metadata"><a href='../../pizzarr/html/Store.html#method-Store-get_consolidated_metadata'><code>pizzarr::Store$get_consolidated_metadata()</code></a></span></li>
<li><span class="pkg-link" data-pkg="pizzarr" data-topic="Store" data-id="get_store_identifier"><a href='../../pizzarr/html/Store.html#method-Store-get_store_identifier'><code>pizzarr::Store$get_store_identifier()</code></a></span></li>
<li><span class="pkg-link" data-pkg="pizzarr" data-topic="Store" data-id="is_erasable"><a href='../../pizzarr/html/Store.html#method-Store-is_erasable'><code>pizzarr::Store$is_erasable()</code></a></span></li>
<li><span class="pkg-link" data-pkg="pizzarr" data-topic="Store" data-id="is_listable"><a href='../../pizzarr/html/Store.html#method-Store-is_listable'><code>pizzarr::Store$is_listable()</code></a></span></li>
<li><span class="pkg-link" data-pkg="pizzarr" data-topic="Store" data-id="is_readable"><a href='../../pizzarr/html/Store.html#method-Store-is_readable'><code>pizzarr::Store$is_readable()</code></a></span></li>
<li><span class="pkg-link" data-pkg="pizzarr" data-topic="Store" data-id="is_writeable"><a href='../../pizzarr/html/Store.html#method-Store-is_writeable'><code>pizzarr::Store$is_writeable()</code></a></span></li>
<li><span class="pkg-link" data-pkg="pizzarr" data-topic="Store" data-id="rename"><a href='../../pizzarr/html/Store.html#method-Store-rename'><code>pizzarr::Store$rename()</code></a></span></li>
</ul>
</details>
}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MemoryStore-new"></a>}}
\if{latex}{\out{\hypertarget{method-MemoryStore-new}{}}}
\subsection{Method \code{new()}}{
Create a new memory store.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{MemoryStore$new()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
A new \code{MemoryStore} object.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MemoryStore-get_parent"></a>}}
\if{latex}{\out{\hypertarget{method-MemoryStore-get_parent}{}}}
\subsection{Method \code{get_parent()}}{
Get the parent of an item.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{MemoryStore$get_parent(item)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{item}}{The item key.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A list with the keys \code{parent} and \code{key}.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MemoryStore-get_item"></a>}}
\if{latex}{\out{\hypertarget{method-MemoryStore-get_item}{}}}
\subsection{Method \code{get_item()}}{
Get an item from the store.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{MemoryStore$get_item(item = NA)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{item}}{The item key.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
The item data in a vector of type raw.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MemoryStore-set_item"></a>}}
\if{latex}{\out{\hypertarget{method-MemoryStore-set_item}{}}}
\subsection{Method \code{set_item()}}{
Set an item in the store.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{MemoryStore$set_item(item, value)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{item}}{The item key.}

\item{\code{value}}{The item value as a vector of type raw.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{NULL} (called for side effects).
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MemoryStore-contains_item"></a>}}
\if{latex}{\out{\hypertarget{method-MemoryStore-contains_item}{}}}
\subsection{Method \code{contains_item()}}{
Determine whether the store contains an item.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{MemoryStore$contains_item(item)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{item}}{The item key.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A boolean value.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MemoryStore-listdir"></a>}}
\if{latex}{\out{\hypertarget{method-MemoryStore-listdir}{}}}
\subsection{Method \code{listdir()}}{
List the store directory.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{MemoryStore$listdir(path = NA)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{path}}{Character path.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{character()} vector of entries.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MemoryStore-delete_item"></a>}}
\if{latex}{\out{\hypertarget{method-MemoryStore-delete_item}{}}}
\subsection{Method \code{delete_item()}}{
Delete an item from the store.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{MemoryStore$delete_item(key)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{key}}{The item key.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{NULL} (called for side effects).
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MemoryStore-rmdir"></a>}}
\if{latex}{\out{\hypertarget{method-MemoryStore-rmdir}{}}}
\subsection{Method \code{rmdir()}}{
Remove a path within a Store.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{MemoryStore$rmdir(item)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{item}}{Character item.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{NULL} (called for side effects).
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MemoryStore-print"></a>}}
\if{latex}{\out{\hypertarget{method-MemoryStore-print}{}}}
\subsection{Method \code{print()}}{
Print a human-readable summary of the store.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{MemoryStore$print(...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{...}}{Ignored.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
\code{self} (invisibly).
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-MemoryStore-clone"></a>}}
\if{latex}{\out{\hypertarget{method-MemoryStore-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{MemoryStore$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/atomic.R
\name{is_integer}
\alias{is_integer}
\title{Check if a value is an integer R vector or scalar.}
\usage{
is_integer(s)
}
\description{
Check if a value is an integer R vector or scalar.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/creation.R
\name{zarr_create_zeros}
\alias{zarr_create_zeros}
\title{Create an array filled with zeros.}
\usage{
zarr_create_zeros(shape, ...)
}
\arguments{
\item{shape}{: int or tuple of ints
Array shape.}

\item{...}{The params of zarr_create()}
}
\value{
ZarrArray
}
\description{
Create an array filled with zeros.
}
